from xiv_xsf.nextra.phase10.client import XCLIClient
from xiv_xsf.nextra.phase10.client.exceptions import IncorrectCredentials
from xiv_xsf.nextra.phase10.client.xcli.transport.exceptions import UnableToConnect

from exceptions import AuthenticationError, ConnectionError

NETWORK_ADDRESS_AUTHORITY = '5'
IEEE_COMPANY_ID = '001738'

class Machine(object):
  """
  A simple class to gap between the XML reply differences of 10.0 and 10.1
  (Mainly helpful for config_get)
  """
  version = None
  config = None
  xcli_client = None
  
  def __init__(self, xcli_client):
    """
    Constructor. Accepts a phase10 XCLIClient instance
    The constructed instance will contain the "config"
    and "version" attributes
    """
    self.xcli_client = xcli_client
    self.version = self.__get_version()
    self.__cache_config()
    self.cached_host_list = self.xcli_client.host_list()

  @classmethod
  def connect_ssl(cls, *args, **kwargs):
    xcli_client = XCLIClient.connect_ssl(*args, **kwargs)
    machine = Machine(xcli_client)
    return machine

  def __get_version(self):
    """
    Get the version from the XCLIClient instance
    """
    version_info = self.xcli_client.version_get()
    machine_version = version_info.as_list[0].system_version
    return machine_version

  def __cache_config(self):
    """
    Cache the configuration of the machine from the XCLIClient
    the configuration is converted to a dict of ["option"]="value"
    If the machine version is not recognized or supported a
    NotImplementedError is raised with the "unsupported_version"
    attribute
    """
    original_machine_config = self.xcli_client.config_get()
    simplified_config = dict()
    if self.version.startswith("10.1.") or self.version == "10.1":
      config_dict = original_machine_config.as_dict("name")
      for config_name, config_data in config_dict.iteritems():
        simplified_config[config_name] = config_data.value
    elif self.version.startswith("10.0.") or self.version == "10.0":
      config_dict = original_machine_config.as_single_element
      simplified_config.update(config_dict)
    else:
      e = NotImplementedError("Machine version is not supported (%s)" % self.version)
      e.unsupported_version = version
      raise e
    self.config = simplified_config
    for option, value in simplified_config.iteritems():
      self.__dict__[option] = value
    
  def get_wwnn(self):
    hex_system_id = hex(int(self.system_id))[2:].rjust(5,"0")
    wwnn = NETWORK_ADDRESS_AUTHORITY + IEEE_COMPANY_ID + \
                 hex_system_id + "0000"
    return wwnn

  def get_mirrors_by_vol(self):
    mirrors_by_vol = dict()
    for mirror in self.xcli_client.mirror_list():
      try:
        vol_name = mirror.local_peer_name
      except AttributeError, e:
        vol_name = mirror.local_volume_name
      if not mirrors_by_vol.has_key(vol_name):
        mirrors_by_vol[vol_name] = list()
      mirrors_by_vol[vol_name].append(mirror)
    return mirrors_by_vol
  
  def maps_by_vol(self):
    maps_by_vol = dict()
    for host in self.cached_host_list:
      for map in self.xcli_client.mapping_list(host=host.name):
        map.host_id = host.id
        if not maps_by_vol.has_key(map.volume):
          maps_by_vol[map.volume] = list()
        maps_by_vol[map.volume].append(map)
    return maps_by_vol

