import logging

from lxml.etree import Element, SubElement

from srm.exceptions import SRMError
from xiv_xsf.nextra.phase10.client.xcli.response.exceptions import SoftExecutionFailedError


NETWORK_ADDRESS_AUTHORITY = '5'
IEEE_COMPANY_ID = '001738'

SRA_TESTFAILOVER_SNAPSHOT_PREFIX="srm_testfailover_"
SRA_TESTFAILOVER_VOL_PREFIX="srm_testfailover_"


def get_response_code_from_xcli_exception(xcli_exception):
    return xcli_exception.response_etree.xpath("command/code")[0].get("value")


def get_local_vol_name_from_xcli_mirror(mirror):
  """
  Accepts an 10.1 or 10.0 xcli mirror and returns the local volume name
  """
  try: local_vol_name = mirror.local_peer_name # 10.1
  except AttributeError, e: local_vol_name = mirror.local_volume_name # 10.0
  return local_vol_name


def get_remote_vol_name_from_xcli_mirror(mirror):
  """
  Accepts an 10.1 or 10.0 xcli mirror and returns the remote volume name
  """
  try: remote_vol_name = mirror.remote_peer_name # 10.1
  except AttributeError, e: remote_vol_name = mirror.remote_volume_name # 10.0
  return remote_vol_name


def is_xcli_mirror_master(mirror):
  """
  Accepts an 10.1 or 10.0 xcli mirror and returns True of the machine is the master
  otherwise, returns False
  """
  is_master = False
  try: # 10.0
    if mirror.local_is_master == "yes":
      is_master = True
  except AttributeError, e:
    if mirror.current_role == "Master":
      is_master = True
  return is_master

def is_xcli_mirror_active(mirror):
  if mirror.active == "yes":
    return True
  return False

def xcli_config_get_as_dict(xcli_config_get):
  """
  Accepts an XCLI 'config_get' response, returns a dict of [config_option] = value
  """
  config = dict()
  try:  # 10.1
    for option_name, option in xcli_config_get.as_dict("name").iteritems():
      config[option_name] = option.value
  except AttributeError, e:
    config = xcli_config_get.as_single_element
  return config


def get_wwnn_from_machine_id(system_id):
  """
  Calculate the WWNN of an XIV system according to its system_id
  """
  hex_system_id = hex(int(system_id))[2:].rjust(5,"0")
  wwnn = NETWORK_ADDRESS_AUTHORITY + IEEE_COMPANY_ID + \
               hex_system_id + "0000"
  return wwnn

class SRMLunPeer(object):
  """
  Represents a mirror connection to a remote replicated/replicating lun
  """
  def __init__(self, peer_vol_name, peer_system_uid, master=True, active=True):
    self.vol_name = peer_vol_name
    self.system_uid = peer_system_uid
    self.master = master
    self.active = active

  def as_xml(self):
    """
    Format as XML according to the SRM API specifications
    """
    peer_root = Element("Peer")
    SubElement(peer_root, "ArrayKey").text = self.system_uid
    SubElement(peer_root, "ReplicaLunKey").text = self.vol_name
    return peer_root

  @classmethod
  def from_xcli(cls, xcli_mirror, target_uid):
    remote_vol_name = get_remote_vol_name_from_xcli_mirror(xcli_mirror)
    is_master = is_xcli_mirror_master(xcli_mirror)
    peer_system_uid = target_uid
    is_active = is_xcli_mirror_active(xcli_mirror)
    return SRMLunPeer(remote_vol_name, peer_system_uid, master=is_master, active=is_active)


class SRMLunNumber(object):
  """
  Represents a volume-mapping
  """
  def __init__(self, initiator_group, lun):
    self.initiator_group = initiator_group
    self.lun = lun

  @property
  def initiator_group_id(self):
    return self.initiator_group.get_id()

  def is_visible_to_port(self, port_id):
    port_id = port_id.replace(":","").upper()
    return self.initiator_group.initiators_by_id.has_key(port_id)
    

  def as_xml(self):
    """
    Format as XML according to the SRM API specifications
    """
    lun_number_root = Element("Number", {"initiatorGroupId": self.initiator_group_id})
    lun_number_root.text = str(self.lun)
    return lun_number_root


class SRMLun(object):
  """
  Represents a volume
  """
  def __init__(self, vol, vol_numbers , vol_peers):
    vol_name = vol.name
    self.id = vol.name
    self.vol = vol
    if vol.cg_id != "0":
      self.consistency_group = vol.cg_id or None
    else:
      self.consistency_group = None
    self.peers = vol_peers
    self.numbers = vol_numbers
    self.vol_name = vol_name

  def is_visible_to_port(self, port_id):
    for lun in self.numbers:
      if lun.is_visible_to_port(port_id):
        return True
    return False

  def is_master(self):
    for peer in self.peers:
      if peer.master:
        return True
    return False
 
  def get_mapped_hosts(self):
      return [ srm_lun_number.initiator_group.host.name for srm_lun_number in self.numbers ]
  
  def as_xml(self):
    """
    Format as XML according to the SRM API specifications
    """
    lun_attr = {'id': self.id}
    if self.consistency_group:
      lun_attr['consistencyGroupId'] = self.consistency_group
    lun_root = Element("Lun", lun_attr)
    for lun_number in self.numbers:
      lun_root.append(lun_number.as_xml())
    for peer in self.peers:
      lun_root.append(peer.as_xml())
    return lun_root

  def as_replica_xml(self):
    """
    Format as replica XML according to the SRM API specifications
    """
    lun_attr = {'key': self.id}
    lun_root = Element("ReplicaLun", lun_attr)
    for lun_number in self.numbers:
      lun_root.append(lun_number.as_xml())
    return lun_root


class SRMInitiatorGroup(object):
  """
  Represents a host
  """
  def __init__(self, host, array_uid):
    self.host = host
    self.array_uid = array_uid
    self.initiators_by_id = dict()
    if host.fc_ports:
      for fc_wwpn in host.fc_ports.split(','):
        self.initiators_by_id[fc_wwpn] = "FC"
    if host.iscsi_ports:
      for iscsi_iqn in host.iscsi_ports.split(','):
        self.initiators_by_id[iscsi_iqn] = "ISCSI"

  def get_initiator_count(self):
    return len(self.initiators_by_id)

  def get_id(self):
    return "%s-%s" % (self.array_uid, self.host.id)

  def as_xml(self):
    """
    Format as XML according to the SRM API specifications
    """
    group_id = self.get_id()
    group_root = Element("InitiatorGroup", {'id': group_id})
    for initiator_id, initiator_type in self.initiators_by_id.iteritems():
      initiator_attrs = {'id': initiator_id, 'type': initiator_type}
      SubElement(group_root, "Initiator", initiator_attrs)
    return group_root


class SRMInitiatorGroupList(object):
  """
  Represents a host list
  """
  initiator_groups_by_id = None

  def __init__(self):
    self.initiator_groups_by_id = dict()

  @property
  def initiator_groups(self):
    return self.initiator_groups_by_id.values()

  @classmethod
  def from_xcli_data(cls, xcli_host_list, array_uid):
    self = SRMInitiatorGroupList()
    self.initiator_groups_by_id = dict()
    for host in xcli_host_list:
      initiator_group = SRMInitiatorGroup(host, array_uid)
      if initiator_group.get_initiator_count() > 0:
        self.initiator_groups_by_id[initiator_group.get_id()] = initiator_group
      else:
        logging.info("Initiator group for host '%s' has no initiator IDs" % initiator_group.host.name)
    return self

  def as_xml(self):
    """
    Format as XML according to the SRM API specifications
    """
    list_root = Element("InitiatorGroupList")
    for initiator_group in self.initiator_groups_by_id.values():
      list_root.append(initiator_group.as_xml())
    return list_root

  def append_to_xml(self, list_root):
    """
    Append children to an exsiting root element
    """
    for initiator_group in self.initiator_groups_by_id.values():
      list_root.append(initiator_group.as_xml())

  def has_group(self, initiator_group_id):
    return self.initiator_groups_by_id.has_key(initiator_group_id)

  def subset(self, subset_group_id_list, ignore_missing=True):
    """
    Returns an SRMInitiatorGroupList with only the groups with the specified IDs
    """
    subset_list = SRMInitiatorGroupList()
    for group_id in subset_group_id_list:
      try:
        subset_list.initiator_groups_by_id[group_id] = self.initiator_groups_by_id[group_id]
      except KeyError, e:
        if ignore_missing:
          continue
        return e
    return subset_list


class SRMStoragePort(object):
  """
  Represents a host_port
  """
  def __init__(self, id, type):
    self.id = id
    self.type = type
  
  def as_xml(self):
    """
    Format as XML according to the SRM API specifications
    """
    storage_port_attrs = {'id': self.id, 'type': self.type}
    storage_port_root = Element("StoragePort", storage_port_attrs)
    return storage_port_root 



class SRMStoragePortList(object):
  """
  Represents a fc_port_list + iscsi name
  """
  def __init__(self):
    self.storage_ports = list()
  
  @classmethod
  def from_xcli_client(cls, fc_port_list, iscsi_name):
    self = SRMStoragePortList()
    fc_ports = fc_port_list.as_list
    for fc_port in fc_ports:
      storage_port = SRMStoragePort(fc_port.wwpn, 'FC')
      self.storage_ports.append(storage_port)
    iscsi_port = SRMStoragePort(iscsi_name, 'ISCSI')
    self.storage_ports.append(iscsi_port)
    return self

  def as_xml(self):
    """
    Format as XML according to the SRM API specifications
    """
    storage_port_list_root = Element("StoragePortList")
    for storage_port in self.storage_ports:
      storage_port_list_root.append(storage_port.as_xml())
    return storage_port_list_root



class SRMArray(object):
  id = None
  product = None
  model = None
  xcli_client = None
  luns = None
  initiator_group_list = None
  lun_access_by_vol_id = None
  _use_lun_access = False

  # Cache
  use_cache = True
  _cache = None

  def __init__(self):
      self._cache = dict()

  @classmethod
  def from_xcli_client(cls, xcli_client):
    self = cls()
    self.xcli_client = xcli_client
    return self

  def _get_cache(self, cache_key):
      if self.use_cache is True:
          if self._cache.has_key(cache_key):
              return self._cache[cache_key]
      return None

  def _set_cache(self, cache_key, value):
      self._cache[cache_key] = value

  @property
  def luns(self):
      cache_key = "luns"
      if self._get_cache(cache_key) is not None:
          return self._get_cache(cache_key)
      logging.info("Retrieving lun list from XIV system")
      luns_by_id = dict()
      for vol_name in self.xcli_vols_by_name.keys():
          luns_by_id[vol_name] = self.get_srm_lun(vol_name)
      self._set_cache(cache_key, luns_by_id)
      return luns_by_id

  @property
  def wwnn(self):
    xiv_system_id = self.config['system_id']
    return get_wwnn_from_machine_id(xiv_system_id)

  @property
  def initiator_group_list(self):
      cache_key = "initiator_group_list"
      if self._get_cache(cache_key) is not None:
          return self._get_cache(cache_key)
      logging.info("Retrieving initiator groups from XIV system")
      host_list = self.xcli_client.host_list()
      array_uid = self.get_uid()
      initiator_group_list = SRMInitiatorGroupList.from_xcli_data(host_list, array_uid)
      self._set_cache(cache_key, initiator_group_list)
      return initiator_group_list

  @property
  def storage_port_list(self):
      cache_key = "storage_port_list"
      if self._get_cache(cache_key) is not None:
          return self._get_cache(cache_key)
      logging.info("Retrieving storage port list from XIV system")
      fc_port_list = self.xcli_client.fc_port_list()
      srm_storage_port_list = SRMStoragePortList.from_xcli_client(fc_port_list, self.config['iscsi_name'])
      self._set_cache(cache_key, srm_storage_port_list)
      return srm_storage_port_list

  @property
  def config(self):
      cache_key = "system_config"
      if self._get_cache(cache_key) is not None:
          return self._get_cache(cache_key)
      logging.info("Retrieving XIV system configuration")
      system_config = xcli_config_get_as_dict(self.xcli_client.config_get())
      self._set_cache(cache_key, system_config)
      return system_config


  def _get_remote_target_system_name(self, remote_target_name, retry=3):
      cache_key = "remote_system_name_by_target"
      system_name_by_target = dict()
      if self._get_cache(cache_key) is not None:
          system_name_by_target = self._get_cache(cache_key)
      else:
          self._set_cache(cache_key, system_name_by_target)
      if system_name_by_target.has_key(remote_target_name):
          return system_name_by_target[remote_target_name]
      logging.debug("Obtaining remote system name for target '%s'" % remote_target_name)
      remote_system_name = None
      try:
          remote_config_get = self.xcli_client.config_get(remote_target=remote_target_name)
          remote_config = xcli_config_get_as_dict(remote_config_get)
          remote_system_name = remote_config['system_name']
          system_name_by_target[remote_target_name] = remote_system_name
          logging.info("Remote system name for target '%s' is '%s'" % (remote_target_name, remote_system_name))
      except Exception as e:
          if (retry > 0):
              return self._get_remote_target_system_name(remote_target_name, retry-1)
          logging.error("Failed to fetch remote system_name for target '%s': %s" % (remote_target_name, str(e)))
          return None
      return remote_system_name

  
  def _get_remote_target_system_id(self, remote_target_name, retry=3):
      cache_key = "remote_system_id_by_target"
      system_id_by_target = dict()
      if self._get_cache(cache_key) is not None:
          system_id_by_target = self._get_cache(cache_key)
      else:
          self._set_cache(cache_key, system_id_by_target)
      if system_id_by_target.has_key(remote_target_name):
          return system_id_by_target[remote_target_name]
      logging.debug("Obtaining remote system id for target '%s'" % remote_target_name)
      remote_system_id = None
      try:
          remote_config_get = self.xcli_client.config_get(remote_target=remote_target_name)
          remote_config = xcli_config_get_as_dict(remote_config_get)
          remote_system_id = remote_config['system_id']
          system_id_by_target[remote_target_name] = remote_system_id
          logging.info("Remote system id for target '%s' is '%s'" % (remote_target_name, remote_system_id))
      except Exception as e:
          logging.error("Failed to fetch remote system_id for target '%s': %s" % (remote_target_name, str(e)))
          return None
      return remote_system_id

  def is_lun_accessible(self, vol_name, port_id_list):
      srm_lun = self.get_srm_lun(vol_name)
      for port_id in port_id_list:
          if srm_lun.is_visible_to_port(port_id):
              return True
      return False

  @property
  def xcli_vols_by_name(self):
      cache_key = "xcli_vols_by_name"
      if self._get_cache(cache_key) is not None:
          return self._get_cache(cache_key)
      vol_list_by_name = self.xcli_client.vol_list().as_dict("name")
      self._set_cache(cache_key, vol_list_by_name)
      return vol_list_by_name

  @property
  def xcli_mirrors_by_vol(self):
      cache_key = "xcli_mirrors_by_vol"
      if self._get_cache(cache_key) is not None:
          return self._get_cache(cache_key)
      logging.debug("Caching mirrors...")
      mirrors_by_vol = dict()
      for xcli_mirror in self.xcli_client.mirror_list():
          vol_name = get_local_vol_name_from_xcli_mirror(xcli_mirror)
          if not mirrors_by_vol.has_key(vol_name):
              mirrors_by_vol[vol_name] = list()
          mirrors_by_vol[vol_name].append(xcli_mirror)
      self._set_cache(cache_key, mirrors_by_vol)
      return mirrors_by_vol

  @property
  def xcli_target_list(self):
      cache_key = "xcli_target_list"
      if self._get_cache(cache_key) is not None:
          return self._get_cache(cache_key)
      xcli_target_list = self.xcli_client.target_list()
      self._set_cache(cache_key, xcli_target_list)
      return xcli_target_list

  @property
  def xcli_host_list(self):
      cache_key = "xcli_host_list"
      if self._get_cache(cache_key) is not None:
          return self._get_cache(cache_key)
      xcli_host_list = self.xcli_client.host_list()
      self._set_cache(cache_key, xcli_host_list)
      return xcli_host_list

  def _get_srm_lun_numbers_for_vol(self, vol_name):
      cache_key = "lun_numbers_by_vol"
      if self._get_cache(cache_key) is not None:
          srm_lun_numbers_by_vol = self._get_cache(cache_key)
      else:
          srm_lun_numbers_by_vol = self.__direct_get_srm_lun_numbers_by_vol()
          self._set_cache(cache_key, srm_lun_numbers_by_vol)
      if srm_lun_numbers_by_vol.has_key(vol_name):
          return srm_lun_numbers_by_vol[vol_name]
      elif self.xcli_vols_by_name.has_key(vol_name):
          return list()
      raise SRMError("Volume '%s' does not seem to exist" % (vol_name,))

  def _get_srm_lun_peers_for_vol(self, vol_name):
      if not self.xcli_mirrors_by_vol.has_key(vol_name):
          if self.xcli_vols_by_name.has_key(vol_name):
              return list() # no peers
          else:
              raise SRMError("Volume '%s' does not seem to exist" % (vol_name,)) # no vol :)
      srm_lun_peers = list()
      vol_mirrors = self.xcli_mirrors_by_vol[vol_name]
      for xcli_mirror in vol_mirrors:
        target_uid = self.get_remote_uid(xcli_mirror.target_name)
        srm_lun_peer = SRMLunPeer.from_xcli(xcli_mirror, target_uid)
        srm_lun_peers.append(srm_lun_peer)
      return srm_lun_peers

  def get_srm_lun(self, vol_name):
      srm_lun_peers = self._get_srm_lun_peers_for_vol(vol_name)
      srm_lun_numbers = self._get_srm_lun_numbers_for_vol(vol_name)
      xcli_vol = self.xcli_vols_by_name[vol_name]
      return SRMLun(xcli_vol, srm_lun_numbers, srm_lun_peers)

  def host_mapping(self, host_name):
      cache_key = "mappings_by_host"
      mappings_by_host = self._get_cache(cache_key)
      if mappings_by_host is None:
          mappings_by_host = dict()
          self._set_cache(cache_key, mappings_by_host)
      if mappings_by_host.has_key(host_name):
          return mappings_by_host[host_name]
      mapping_list = self.xcli_client.mapping_list(host=host_name)
      mappings_by_host[host_name] = mapping_list
      return mapping_list

  def __direct_get_srm_lun_numbers_by_vol(self):
    xcli_client = self.xcli_client
    srm_lun_numbers_by_vol = dict()
    for initiator_group in self.initiator_group_list.initiator_groups:
      host = initiator_group.host
      maps = self.host_mapping(host.name)
      for map in maps:
        if not srm_lun_numbers_by_vol.has_key(map.volume):
          srm_lun_numbers_by_vol[map.volume] = list()
        initiator_group_id = initiator_group.get_id()
        if not self.initiator_group_list.has_group(initiator_group_id):
          logging.error("Initiator group '%s' not found in list." % initiator_group_id)
          continue
        initiator_group = self.initiator_group_list.initiator_groups_by_id[initiator_group_id]
        srm_lun_number = SRMLunNumber(initiator_group, map.lun)
        srm_lun_numbers_by_vol[map.volume].append(srm_lun_number)
    return srm_lun_numbers_by_vol

  def get_used_initator_group_ids(self):
    """
    Return a list of srm initiator group ids referenced by the discovered srm luns of the array
    """
    used_initiator_group_ids = list()
    for srm_lun in self.luns.values():
      for srm_lun_num in srm_lun.numbers:
        used_initiator_group_ids.append(srm_lun_num.initiator_group_id)
    return used_initiator_group_ids

  def get_version_list(self):
    return self.xcli_client.version_get()[0].system_version.split('.')

  def reactivate_local_mirror_as_slave(self, vol_name):
    """
    Take a local mirror in any state and activate it as a slave
    """
    logging.debug("Mirror reactivate: Starting to activate mirror for '%s'" % (vol_name,))
    # Make sure local mirror is slave (And deactivated)
    mirror_list = self.xcli_client.mirror_list(vol=vol_name)
    if len(mirror_list) == 0:
      logging.warning("Mirror reactivate: There is no mirror for this volume (%s)" % (vol_name,))
      logging.error("Mirror reactivate: Aborted")
      raise SRMError("Mirror reactivate failed (%s)" % (vol_name,))
    local_mirror = mirror_list[0]
    if is_xcli_mirror_master(local_mirror):
      self.xcli_client.mirror_deactivate(vol=vol_name)
      self.xcli_client.mirror_change_role(vol=vol_name, new_role="Slave")
    # Make sure remote mirror is master (and activate it)
    remote_vol_name = get_remote_vol_name_from_xcli_mirror(local_mirror)
    target_name = local_mirror.target_name
    logging.debug("Mirror reactivate: Remote target: '%s', remote volume is '%s'" % \
                  (target_name,remote_vol_name,))
    logging.debug("Mirror reactivate: Remote target: getting current mirror role")
    remote_mirror_list = self.xcli_client.mirror_list(vol=remote_vol_name, remote_target=target_name)
    remote_mirror = remote_mirror_list[0]
    if is_xcli_mirror_master(remote_mirror):
       logging.debug("Mirror reactivate: Remote target: Already master")
    else:
       logging.debug("Mirror reactivate: Remote target: Need to change role to Master")
       self.xcli_client.mirror_change_role(remote_target=target_name, \
                                           vol=remote_vol_name, new_role="Master")
       logging.debug("Mirror reactivate: Remote target: Deactivating mirror")
       self.xcli_client.mirror_deactivate(remote_target=target_name, \
                                           vol=remote_vol_name)
       logging.debug("Mirror reactivate: Remote target: Reactivating mirror")
       self.xcli_client.mirror_activate(remote_target=target_name, \
                                           vol=remote_vol_name)
    logging.info("Mirror reactivate: Complete: '%s'" % (vol_name,))
    return

  def break_local_slave_mirror(self, vol_name):
    """
    Break a local mirror in a Slave state. The mirror will be moved to master state
    and deactivated
    """
    logging.debug("Mirror break: Starting to break mirror for '%s'" % (vol_name,))
    logging.debug("Mirror break: Obtaining mirror information...")
    mirror_list = self.xcli_client.mirror_list(vol=vol_name)
    if len(mirror_list) == 0:
      logging.warning("Mirror break: There is no mirror for this volume (%s)" % (vol_name,))
      logging.error("Mirror break: Aborted")
      raise SRMError("Mirror break failed (%s)" % (vol_name,))
    mirror = mirror_list[0]
    # Abort if mirror is in master state
    if is_xcli_mirror_master(mirror):
      logging.error("Mirror break: Mirror for " + vol_name + " was expected to be in Slave state," + \
                    " but is master. Will not break mirroring.") 
      raise SRMError("Mirror break failed (%s)" % (vol_name,))
    self.xcli_client.mirror_change_role(vol=vol_name, new_role="Master")
    self.xcli_client.mirror_deactivate(vol=vol_name)
    logging.info("Mirror break: Complete: '%s'" % (vol_name,))
    return

  def _get_used_luns_for_host(self, host_name):
      xcli_mapping_list = self.host_mapping(host_name)
      used_luns = [ int(mapping.lun) for mapping in xcli_mapping_list ]
      return used_luns

  def _unallocate_lun_from_host(self, host_name, lun_id):
      cache_key = "used_luns_by_host"
      used_luns_by_host = self._get_cache(cache_key)
      if used_luns_by_host is None:
          used_luns_by_host = dict()
          self._set_cache(cache_key, used_luns_by_host)
      if not used_luns_by_host.has_key(host_name):
          used_luns_by_host[host_name] = self._get_used_luns_for_host(host_name)
      used_luns_by_host[host_name].remove(lun_id)

  def _allocate_free_lun_for_host(self, host_name):
      cache_key = "used_luns_by_host"
      used_luns_by_host = self._get_cache(cache_key)
      if used_luns_by_host is None:
          used_luns_by_host = dict()
          self._set_cache(cache_key, used_luns_by_host)
      if not used_luns_by_host.has_key(host_name):
          used_luns_by_host[host_name] = self._get_used_luns_for_host(host_name)
      free_lun = 1
      while free_lun in used_luns_by_host[host_name]:
          free_lun += 1
      used_luns_by_host[host_name].append(free_lun)
      logging.debug("Allocated lun '%d' for '%s'" % (free_lun, host_name,))
      return free_lun

  def _map_volume_to_host(self, host_name, vol_name):
      allocated_lun = self._allocate_free_lun_for_host(host_name)
      try:
          self.xcli_client.map_vol(vol=vol_name, host=host_name, lun=allocated_lun)
          logging.info("Mapped '%s' to '%s' on lun '%d'" % (vol_name, host_name, allocated_lun))
      except SoftExecutionFailedError as e:
          response_code = get_response_code_from_xcli_exception(e)
          if response_code == "VOLUME_ALREADY_ASSIGNED":
              self._unallocate_lun_from_host(host_name, allocated_lun)
              logging.warning("Volume '%s' is already mapped to '%s'" % (vol_name,host_name))
          elif response_code == "LUN_ALREADY_IN_USE":
              logging.warning("Lun '%s' is already used on '%s'" % (allocated_lun, host_name))
              return self._map_volume_to_host(host_name, vol_name)
          else:
              logging.warning("Lun mapping failed for '%s': %s: %s" % (vol_name, str(e), response_code)) 
              raise(e)
      return allocated_lun

  def find_host_by_port(self, port_id):
      cache_key = "host_by_port_id"
      host_by_port_id = self._get_cache(cache_key)
      if host_by_port_id is None:
          host_by_port_id = dict()
          self._set_cache(cache_key, host_by_port_id)
      if host_by_port_id.has_key(port_id):
          return host_by_port_id[port_id]
      host = self._find_host_by_port(port_id)
      host_by_port_id[port_id] = host
      return host
          
  def _find_host_by_port(self, port_id):
      for xcli_host in self.xcli_host_list:
          fc_port = port_id.replace(":","").upper()
          if fc_port in xcli_host.fc_ports.upper().split(","):
              return xcli_host
          if port_id.lower() in xcli_host.iscsi_ports.lower().split(","):
              return xcli_host
      raise SRMError("Couldn't find host for initiator port '%s'" % (port_id,))                        

  def restrict_vol_to_initiator_ports(self, vol_name, port_id_list):
      logging.info("Going to restrict '%s' to specific initiators" % (vol_name,))
      srm_lun = self.get_srm_lun(vol_name)
      restricted_hosts = list()
      for port_id in port_id_list:
          port_host = self.find_host_by_port(port_id)
          restricted_hosts.append(port_host.name)
      mapped_hosts = srm_lun.get_mapped_hosts()
      for host_name in mapped_hosts:
          if host_name not in set(restricted_hosts):
              logging.info("Going to unmap '%s' from '%s'" % (vol_name, host_name))
              self.xcli_client.unmap_vol(host=host_name, vol=vol_name)
      for host_name in set(restricted_hosts):
          if host_name not in mapped_hosts:
              logging.info("Going to map '%s' to '%s'" % (vol_name, host_name))
              self._map_volume_to_host(host_name, vol_name)      
          
  def _unmap_volume_from_all_hosts(self, vol_name):
      srm_lun = self.get_srm_lun(vol_name)
      mapped_hosts = srm_lun.get_mapped_hosts()
      for host_name in mapped_hosts:
          self.xcli_client.unmap_vol(host=host_name, vol=vol_name)

  def reset_lun_cache(self):
      logging.info("Removing lun cache...")
      self._cache = dict()

  def create_testfailover_snapshot(self, vol_name, initiator_access_list=None):
    if not self.luns.has_key(vol_name):
        return
    logging.info("Going to create testFailover snapshot for '%s'" % vol_name)
    srm_lun = self.get_srm_lun(vol_name)
    # Check if device may be failed over
    hosts_to_map = srm_lun.get_mapped_hosts()
    if initiator_access_list is not None:
      for initiator_id in initiator_access_list:
          try:
            xcli_host = self.find_host_by_port(initiator_id)
            hosts_to_map.append(xcli_host.name)
          except SRMError, e:
            logging.error("Couldn't find a host for initiator '%s'" % (initiator_id,))
    if len(hosts_to_map) == 0:
        raise SRMError("Snapshot of '%s' will not be mapped anywhere. Aborting creation." % (vol_name,))
    logging.debug("Snapshot of '%s' will be mapped to %s" % (vol_name, str(hosts_to_map)))
    snapshot_name = self._create_snapshot_for_volume(vol_name)
    self._unlock_volume(snapshot_name)
    snapshot_srm_lun_numbers = list()
    for host_name in hosts_to_map:
        self._map_volume_to_host(host_name, snapshot_name)
    return (snapshot_name, vol_name)

  def get_testfailover_snapshot_srm_lun(self, vol_name):
      snapshot_name = SRA_TESTFAILOVER_SNAPSHOT_PREFIX + vol_name
      snapshot_srm_lun = self.get_srm_lun(snapshot_name)
      return snapshot_srm_lun

  def delete_testfailover_snapshot(self, vol_name):
      snapshot_name = SRA_TESTFAILOVER_SNAPSHOT_PREFIX + vol_name
      logging.debug("Test Failover: %s: unmapping snapshot from hosts" % (snapshot_name,))
      self._unmap_volume_from_all_hosts(snapshot_name)
      logging.debug("Test Failover: %s: unmapping snapshot from hosts" % (snapshot_name,))
      self.xcli_client.snapshot_delete(snapshot=snapshot_name)
      return (snapshot_name, vol_name)

  def _unlock_volume(self, vol_name):
    self.xcli_client.vol_unlock(vol=vol_name)

  def _create_snapshot_for_volume(self, vol_name):
    snapshot_name = SRA_TESTFAILOVER_SNAPSHOT_PREFIX + vol_name
    logging.debug("Creating snapshot '%s' for volume '%s'" % (snapshot_name, vol_name))
    try:
        self.xcli_client.snapshot_create(vol=vol_name, name=snapshot_name)
    except SoftExecutionFailedError as e:
        response_code = get_response_code_from_xcli_exception(e)
        if response_code == "VOLUME_EXISTS":
            raise SRMError("Failed to create snapshot '%s' for '%s': Already exists." % \
                    (snapshot_name, vol_name))
            #self._overwrite_snapshot_for_volume(vol_name, snapshot_name)
        else:
            logging.debug("Mirror creation error response: " + response_code)
            raise(SRMError(e))
    return snapshot_name

  def _overwrite_snapshot_for_volume(self, vol_name, snapshot_name):
    logging.debug("Overwriting snapshot '%s' for volume '%s'" % (snapshot_name, vol_name))
    self.xcli_client.snapshot_create(vol=vol_name, overwrite=snapshot_name)

  def get_uid(self):
      return self.config['system_name']

  def get_remote_uid(self, target_name):
      return self._get_remote_target_system_name(target_name)

