import logging

from lxml import etree
from lxml.etree import XMLSyntaxError
import xiv_xsf

from xiv_xsf.nextra.phase10.client import XCLIClient

from array import SRMArray
from exceptions import IncorrectCredentials, UnableToConnect, SRMCommandError, SRMError


class SRMConnectSpec(object):
  def __init__(self, name, username, password, address):
    self.name = name
    self.username = username
    self.password = password
    self.address = address
    self.as_tuple = (username, password, address)

  def __str__(self):
    return "(%s) %s@%s" % (self.name, self.username, self.address)

class SRMCommand(object):
  command_xml = None
  connect_specs = None
  name = None
  log_level = None
  output_filename = None
  array_id = None
 
  @classmethod
  def from_command(cls, srm_command):
    self = SRMCommand()
    try:
      for attr in ('command_xml', 'connect_specs', 'name', 'log_level', 'output_filename'):
        self.__dict__[attr] = srm_command.__dict__[attr]
    except KeyError, e:
      raise SRMCommandError(srm_command, "Not an SRMCommand or SRMCommand is missing an attribute: " + str(e))
    return self
 
  @classmethod
  def from_xml_string(cls, command_xml_text):
    self = SRMCommand()
    try:
      command_xml = etree.fromstring(command_xml_text)
    except XMLSyntaxError, e:
      raise SRMCommandError(command_xml_text, "Error parsing XML: " + str(e))
    try:
      self.command_xml = command_xml
      self.connect_specs = list()
      self.name = command_xml.find("Name").text
      self.log_level = command_xml.find("LogLevel").text
      self.output_filename = command_xml.find("OutputFile").text
      try:
          self.array_id = self.command_xml.find("ArrayId").text
      except AttributeError as e:
          pass
      for connect_spec_xml in command_xml.findall("ConnectSpec"):
        username = connect_spec_xml.find("Username").text
        password = connect_spec_xml.find("Password").text
        address = connect_spec_xml.find("Address").text
        title = connect_spec_xml.find("Name").text
        connect_spec = SRMConnectSpec(title, username, password, address)
        self.connect_specs.append(connect_spec)
    except AttributeError, e:
      raise SRMCommandError(command_xml_text, "Error parsing XML: missing command data")
    return self

  def get_single_srm_array(self):
    srm_array = None
    if len(self.connect_specs) > 1:
      raise SRMCommandError(self, "SRM command has multiple machines")
    try:
      srm_array = self.get_srm_arrays()[0]
    except IndexError, e:
      raise SRMCommandError(self, "SRM command has no machines")
    return srm_array 

  def get_srm_arrays(self):
    srm_arrays = list()
    for connect_spec in self.connect_specs:
      username, password, address = connect_spec.as_tuple
      try:
          xcli_client = XCLIClient.connect_ssl(username, password, address)
      except xiv_xsf.nextra.phase10.client.exceptions.CantInstantiateError as e:
          if e.message.find("not authenticated") >= 0:
              raise IncorrectCredentials(e.message)
      srm_array = SRMArray.from_xcli_client(xcli_client)
      if self.array_id is not None:
          if srm_array.get_uid() != self.array_id:
              raise SRMError(self, \
                       "Array ID in command (%s) does not match array ID in target machine (%s)" % \
                      (self.array_id, srm_array.get_uid()))
      srm_arrays.append(srm_array)
    return tuple(srm_arrays)

  def __str__(self):
    return etree.tostring(self.command_xml, pretty_print=True)

class SRMFailoverCommand(SRMCommand):
  array_id = None
  replica_lun_keys = None
  lun_access_by_id = None
  use_lun_access = False
  
  @classmethod
  def from_command(self, srm_command):
    self = SRMTestFailoverCommand()
    try:
      for attr in ('command_xml', 'connect_specs', 'name', 'log_level', 'output_filename'):
        self.__dict__[attr] = srm_command.__dict__[attr]
    except KeyError, e:
      raise SRMCommandError(srm_command, "Not an SRMCommand or SRMCommand is missing an attribute: " + str(e))
    try:
      self.array_id = self.command_xml.find("ArrayId").text
      self.split_replication = False
      try:
        if self.command_xml.find("SplitReplication").text == "true":
          self.split_replication = True
      except AttributeError, e:
        pass
      self.replica_lun_keys = list()
      for replica_lun_xml in self.command_xml.findall("ReplicaLunKeyList/ReplicaLunKey"):
        self.replica_lun_keys.append(replica_lun_xml.text)
      logging.info("Request contains %d luns." % (len(self.replica_lun_keys),))
      self.__generate_lun_access()
    except AttributeError, e:
      raise SRMCommandError(srm_command, "Error parsing XML: missing command data")
    return self

  def __generate_lun_access(self):
    lun_access_by_id = dict()
    lun_access_list_xml = self.command_xml.find("LunAccess")

    # Index the initiator ports for each initiator group
    initiators_by_group_id = dict()
    try:
      initiator_group_list_xml = self.command_xml.find("InitiatorGroupList")
      initiator_group_list_xml.findall("InitiatorGroup")
    except AttributeError, e:
      logging.info("No initiator access specified")
      return
    logging.info("Indexing supplied initiator groups")
    for initiator_group_xml in initiator_group_list_xml.findall("InitiatorGroup"):
      group_id = initiator_group_xml.get("id")
      logging.debug("Processing group '%s'" % (group_id,))
      initiators_by_group_id[group_id] = list()
      for initiator_xml in initiator_group_xml.findall("Initiator"):
        port_id = initiator_xml.get("id")
        logging.debug("Group '%s': Adding port: '%s'" % (group_id, port_id))
        initiators_by_group_id[group_id].append(port_id)
    
    # Pair the port ids to the luns with the relevant access(initiator) group
    if lun_access_list_xml is not None:
      self.use_lun_access = True
      logging.info("Processing lun access list")
      self.lun_access = dict()
      if lun_access_list_xml.find("ReplicaLun") is not None:
        for lun_access_xml in lun_access_list_xml.findall("ReplicaLun"):
          lun_id = lun_access_xml.get("key")
          lun_access_by_id[lun_id] = list()
          if lun_access_xml.find("InitiatorGroup") is not None:
            for lun_access_group_xml in lun_access_xml.findall("InitiatorGroup"):
              access_group_id = lun_access_group_xml.get("id")
              logging.debug("\tLun %s should have access from initiator group %s" % \
                              (lun_id, access_group_id))
              lun_access_by_id[lun_id] = initiators_by_group_id[access_group_id]
    self.lun_access_by_id = lun_access_by_id

  @classmethod
  def from_xml_string(cls, command_xml_text):
    srm_command = SRMCommand.from_xml_string(command_xml_text)
    self = SRMFailoverCommand.from_command(srm_command)
    return self


class SRMTestFailoverCommand(SRMFailoverCommand):
  action = None

  @classmethod
  def from_command(self, srm_command):
    self = SRMTestFailoverCommand()
    failover_command = SRMFailoverCommand.from_command(srm_command)
    self.__dict__.update(failover_command.__dict__)
    self.action = self.command_xml.find("Action").text
    return self

  @classmethod
  def from_xml_string(cls, command_xml_text):
    srm_command = SRMCommand.from_xml_string(command_xml_text)
    self = SRMTestFailoverCommand.from_command(srm_command)
    return self
