import logging
from time import sleep

from lxml import etree
from lxml.etree import Element, SubElement

from exceptions import SRMError, SRMCommandError

RETURN_SUCCESS = "0"
RETURN_COMMAND_ERROR = "1"
RETURN_CONNECTION_ERROR = "2"
RETURN_AUTHENTICATION_ERROR = "3"
RETURN_UNKNOWN_ERROR = "4"

# SRM Array Id is XIV System id
# SRM Lun Id is XIV Volumen name
# SRM Initiator group Id is <array_id>-<host_id>)


def generate_basic_response():
  response_root = Element("Response")
  return_code = SubElement(response_root, "ReturnCode")
  return_code.text = RETURN_UNKNOWN_ERROR
  return (response_root, return_code)

class SRMResponse(object):
  def __init__(self, srm_command):
    self.srm_command = srm_command

class SRMDiscoverArrays(SRMResponse):

  def __xml_from_srm_array(self, srm_array):
    array_attrs = {'id': srm_array.get_uid() , 'wwnn': srm_array.wwnn}
    array = Element("Array", array_attrs)
    SubElement(array, "Product").text = srm_array.config['machine_type']
    SubElement(array, "Model").text = srm_array.config['machine_model']
    storage_port_list = srm_array.storage_port_list
    array.append(storage_port_list.as_xml())
    return array

  def as_xml(self):
    srm_arrays = self.srm_command.get_srm_arrays()
    response_root, returncode = generate_basic_response()
    array_list = SubElement(response_root, "ArrayList")
    for srm_array in srm_arrays:
      array_xml_root = self.__xml_from_srm_array(srm_array)
      array_list.append(array_xml_root)
    returncode.text = RETURN_SUCCESS
    return response_root
      

class SRMDiscoverLuns(SRMResponse):
  def as_xml(self):
    response_root, returncode = generate_basic_response()
    srm_arrays = self.srm_command.get_srm_arrays()
    initiator_group_list = SubElement(response_root, "InitiatorGroupList")
    for srm_array in srm_arrays:
      lun_list = SubElement(response_root, "LunList", {'arrayId': srm_array.get_uid()})
      srm_array.initiator_group_list.append_to_xml(initiator_group_list)
      for lun in srm_array.luns.values():
        if lun.is_master():
          lun_list.append(lun.as_xml())
    returncode.text = RETURN_SUCCESS
    return response_root

class SRMBasicFailover(SRMResponse):

  def __init__(self, srm_failover_command):
    SRMResponse.__init__(self, srm_failover_command)
    self.status = RETURN_SUCCESS
    self.failover_command = srm_failover_command
    self.failover_srm_array = None
    self.failedover_vol_names = list()

  def failover(self):
      raise NotImplementedError

  def get_lun_ids_to_process(self):
      return list(self.srm_command.replica_lun_keys)
      if not self.failover_command.use_lun_access:
          return list(self.srm_command.replica_lun_keys)
      vols_to_process = list()
      lun_access_by_id = self.failover_command.lun_access_by_id
      for vol_name in self.srm_command.replica_lun_keys:
          if lun_access_by_id.has_key(vol_name):
              try:
                  if self.failover_srm_array.is_lun_accessible(vol_name, lun_access_by_id[vol_name]):
                      vols_to_process.append(vol_name)
              except SRMError as e:
                  logging.warning("Failed to look up look access for '%s': %s" % (vol_name, str(e)))
      return vols_to_process
    
  def as_xml(self):
    failover_srm_array = self.srm_command.get_single_srm_array()
    response_root, returncode = generate_basic_response()
    initiator_group_list_root = SubElement(response_root, "InitiatorGroupList")
    replica_lun_list = SubElement(response_root, "ReplicaLunList")
    used_initator_group_ids = list()
    replica_luns = self.failedover_vol_names
    for lun in replica_luns:
        replica_lun_list.append(lun.as_replica_xml())
        for lun_number in lun.numbers:
            used_initator_group_ids.append(lun_number.initiator_group_id)
    used_initiator_group_list = failover_srm_array.initiator_group_list.subset(used_initator_group_ids)
    used_initiator_group_list.append_to_xml(initiator_group_list_root)
    returncode.text = self.status
    return response_root
  
class SRMFailover(SRMBasicFailover):

  def failover(self):
    failover_srm_array = self.srm_command.get_single_srm_array()
    self.failover_srm_array = failover_srm_array
    # Sanity check: The ID string in the command should match the ID string in the target f/o machine
    vols_to_process = self.get_lun_ids_to_process()
    vols_successfully_processed = list()
    lun_access_by_id = self.failover_command.lun_access_by_id
    for vol_name in vols_to_process:
      try:        
        try:
          failover_srm_array.break_local_slave_mirror(vol_name)
          if self.failover_command.use_lun_access:
              if lun_access_by_id.has_key(vol_name):
                  lun_port_ids = lun_access_by_id[vol_name]
                  failover_srm_array.restrict_vol_to_initiator_ports(vol_name, lun_port_ids)          
            # create a new host + map it
        except SRMError, e:
          logging.error("Volume '%s' failover failed, but includeing anyway" % (vol_name,))        
        vols_successfully_processed.append(vol_name)
      except SRMError, e:
        logging.error("Not listing volume in failover list: " + str(e))
      self.failover_srm_array.reset_lun_cache()
      failover_luns = list()
      for vol_name in vols_successfully_processed:
        try:
          srm_lun = failover_srm_array.get_srm_lun(vol_name)
          failover_luns.append(srm_lun)
        except SRMError, e:
          logging.error("Not listing volume in failover list: " + str(e))
      if len(failover_luns) == 0:
          self.status = RETURN_UNKNOWN_ERROR
      self.failedover_vol_names = failover_luns

class SRMTestFailover(SRMBasicFailover):
  failover_machine = None

  def __init__(self, srm_test_failover_command):
    SRMBasicFailover.__init__(self, srm_test_failover_command)
    try:
      x = srm_test_failover_command.action
    except AttributeError, e:
      raise SRMCommandError(srm_test_failover_command, "SRM Test Failover command has no 'action'")

  def __start_test_failover(self, vol_name):
    self.failover_srm_array.break_local_slave_mirror(vol_name)

  def __stop_test_failover(self, vol_name):
    self.failover_srm_array.reactivate_local_mirror_as_slave(vol_name)

  def __start_failover_for_vols(self, vol_name_list):
    failedover_snapshot_name_by_vol = dict()
    for vol_name in vol_name_list:
        try:
            srm_lun = self.failover_srm_array.get_srm_lun(vol_name)
            if srm_lun.is_master():
                raise SRMError("Volume '%s' is Master. it may be in failover mode. Skipping." % (vol_name,))
            vol_lun_access = None
            if self.failover_command.use_lun_access:
              vol_lun_access = list()
              lun_access_by_id = self.failover_command.lun_access_by_id
              if lun_access_by_id.has_key(vol_name):
                vol_lun_access = lun_access_by_id[vol_name]
            snapshot_name, vol_name = self.failover_srm_array.create_testfailover_snapshot(vol_name, vol_lun_access)
            failedover_snapshot_name_by_vol[vol_name] = snapshot_name
        except SRMError, e:
            logging.error("Error performing test failover start command for '%s': %s" % (vol_name, str(e)))
    failedover_srm_luns = list()
    self.failover_srm_array.reset_lun_cache()
    for vol_name, snapshot_name in failedover_snapshot_name_by_vol.iteritems():
        try:
            srm_original_replica_lun = self.failover_srm_array.get_srm_lun(vol_name)
            srm_snapshot_replica_lun = self.failover_srm_array.get_srm_lun(snapshot_name)
            srm_original_replica_lun.numbers = srm_snapshot_replica_lun.numbers
            failedover_srm_luns.append(srm_original_replica_lun)
        except SRMError as e:
            logging.error("Failed to process volume: " + vol_name + ": " + str(e))
    if len(failedover_srm_luns) == 0:
        self.status = RETURN_UNKNOWN_ERROR
    return failedover_srm_luns

  def __stop_failover_for_vols(self, vol_name_list):
      failedover_srm_luns = list()
      for vol_name in vol_name_list:
         try:
              srm_original_replica_lun = self.failover_srm_array.get_srm_lun(vol_name)
              srm_snapshot_replica_lun = self.failover_srm_array.get_testfailover_snapshot_srm_lun(vol_name)
              snapshot_name, vol_name = self.failover_srm_array.delete_testfailover_snapshot(vol_name)
              srm_original_replica_lun.numbers = srm_snapshot_replica_lun.numbers
              failedover_srm_luns.append(srm_original_replica_lun)
         except SRMError as e:
              logging.error("Error performing test failover stop command for '%s': %s" % (vol_name, str(e)))
      return failedover_srm_luns

  def __process_all_failover_srm_luns(self, vol_name_list):
    action = self.srm_command.action
    if action == "start":
        return self.__start_failover_for_vols(vol_name_list)
    if action == "stop":
        return self.__stop_failover_for_vols(vol_name_list)

  def __process_failover_lun(self, vol_name, action):
    if action == "start":
        snapshot_name, vol_name = self.failover_srm_array.create_testfailover_snapshot(vol_name)
    elif action == "stop":
        try:
            snapshot_name, vol_name = self.failover_srm_array.delete_testfailover_snapshot(vol_name)
        except SRMError as e:
            logging.error(e)
            logging.warning("Reporting '%s' eitherway" % (vol_name,))
            return (None, vol_name)
    logging.debug("Successfully processed '%s'" % (snapshot_name,))
    return (snapshot_name, vol_name)

  def failover(self):
    failover_srm_array = self.srm_command.get_single_srm_array()
    self.failover_srm_array = failover_srm_array
    # Sanity check: The ID string in the command should match the ID string in the target f/o machine
    vols_to_process = self.get_lun_ids_to_process()
    vols_successfully_processed = self.__process_all_failover_srm_luns(vols_to_process)
    self.failedover_vol_names = vols_successfully_processed
    if len(vols_successfully_processed) == 0:
      self.status = RETURN_UNKNOWN_ERROR
