#!/usr/bin/env python

from logging import Formatter

color_names = [
    'dark_gray',
    'red',
    'green',
    'yellow',
    'blue',
    'magenta',
    'cyan',
    'white',
]

COLORS = dict(
    (name, dict(color=color_num, light=True))
    for color_num, name in enumerate(color_names)
)

for orig in COLORS.keys():
    COLORS['dark_' + orig] = dict(COLORS[orig], light=False)

for orig, dark in [('dark_gray', 'black'),
                    ('white', 'gray'),]:
    COLORS[dark] = dict(COLORS[orig], light=False)

def _stylify(text, color, background, light=False, underline=False):    
    if not underline and not light:
        prefix_str = "\033[0"
    elif not underline or not light:
        prefix_str = "\033[%d" % (light and 1 or 4)
    else:
        prefix_str = "\033[4m\033[1"
    return "%s;%d;%dm%s%s" % (prefix_str,
                              background + 40,
                              color + 30,
                              text,
                              "\033[m")

def stylify(text, color='white', background='black', underline=False):    
    background = COLORS[background]['color']
    params = dict(COLORS[color], background=background, underline=underline)
    return _stylify(text, **params)

class StylingFormatter(Formatter):
    SEVERITY_COLOR_THRESHOLDS = (
        (0, 'dark_gray'),
        (10, 'white'),
        (20, 'green'),
        (30, 'yellow'),
        (40, 'red'),
        (50, 'magenta'),
    )
    def format(self, record):
        if record.msg[0] == '*' and record.msg[-1] == '*':
            record.msg = record.msg[1:-1]
            return stylify(Formatter.format(self, record), color='blue')
        plain_result = Formatter.format(self, record)
        for severity_threshold, severity_color in self.SEVERITY_COLOR_THRESHOLDS:
            if record.levelno >= severity_threshold:
                color = severity_color
            else:
                break
        return stylify(plain_result, color=color)

def test():
    r"""
    >>> for line in test():
    ...   print repr(line)
    '\x1b[4m\x1b[1;44;34mcolor=blue background=blue, underline=True\x1b[m'
    '\x1b[1;44;34mcolor=blue background=blue, underline=False\x1b[m'
    '\x1b[4m\x1b[1;47;34mcolor=blue background=gray, underline=True\x1b[m'
    '\x1b[1;47;34mcolor=blue background=gray, underline=False\x1b[m'
    '\x1b[4m\x1b[1;42;34mcolor=blue background=dark_green, underline=True\x1b[m'
    '\x1b[1;42;34mcolor=blue background=dark_green, underline=False\x1b[m'
    '\x1b[4m\x1b[1;44;34mcolor=blue background=dark_blue, underline=True\x1b[m'
    '\x1b[1;44;34mcolor=blue background=dark_blue, underline=False\x1b[m'
    '\x1b[4m\x1b[1;45;34mcolor=blue background=dark_magenta, underline=True\x1b[m'
    '\x1b[1;45;34mcolor=blue background=dark_magenta, underline=False\x1b[m'
    '\x1b[4m\x1b[1;40;34mcolor=blue background=dark_gray, underline=True\x1b[m'
    '\x1b[1;40;34mcolor=blue background=dark_gray, underline=False\x1b[m'
    '\x1b[4m\x1b[1;43;34mcolor=blue background=dark_yellow, underline=True\x1b[m'
    '\x1b[1;43;34mcolor=blue background=dark_yellow, underline=False\x1b[m'
    '\x1b[4m\x1b[1;43;34mcolor=blue background=yellow, underline=True\x1b[m'
    '\x1b[1;43;34mcolor=blue background=yellow, underline=False\x1b[m'
    '\x1b[4m\x1b[1;40;34mcolor=blue background=dark_dark_gray, underline=True\x1b[m'
    '\x1b[1;40;34mcolor=blue background=dark_dark_gray, underline=False\x1b[m'
    '\x1b[4m\x1b[1;41;34mcolor=blue background=dark_red, underline=True\x1b[m'
    '\x1b[1;41;34mcolor=blue background=dark_red, underline=False\x1b[m'
    '\x1b[4m\x1b[1;47;34mcolor=blue background=dark_white, underline=True\x1b[m'
    '\x1b[1;47;34mcolor=blue background=dark_white, underline=False\x1b[m'
    '\x1b[4m\x1b[1;42;34mcolor=blue background=green, underline=True\x1b[m'
    '\x1b[1;42;34mcolor=blue background=green, underline=False\x1b[m'
    '\x1b[4m\x1b[1;46;34mcolor=blue background=cyan, underline=True\x1b[m'
    '\x1b[1;46;34mcolor=blue background=cyan, underline=False\x1b[m'
    '\x1b[4m\x1b[1;45;34mcolor=blue background=magenta, underline=True\x1b[m'
    '\x1b[1;45;34mcolor=blue background=magenta, underline=False\x1b[m'
    '\x1b[4m\x1b[1;46;34mcolor=blue background=dark_cyan, underline=True\x1b[m'
    '\x1b[1;46;34mcolor=blue background=dark_cyan, underline=False\x1b[m'
    '\x1b[4m\x1b[1;47;34mcolor=blue background=white, underline=True\x1b[m'
    '\x1b[1;47;34mcolor=blue background=white, underline=False\x1b[m'
    '\x1b[4m\x1b[1;41;34mcolor=blue background=red, underline=True\x1b[m'
    '\x1b[1;41;34mcolor=blue background=red, underline=False\x1b[m'
    '\x1b[4m\x1b[1;40;34mcolor=blue background=black, underline=True\x1b[m'
    '\x1b[1;40;34mcolor=blue background=black, underline=False\x1b[m'
    '\x1b[4;44;37mcolor=gray background=blue, underline=True\x1b[m'
    '\x1b[0;44;37mcolor=gray background=blue, underline=False\x1b[m'
    '\x1b[4;47;37mcolor=gray background=gray, underline=True\x1b[m'
    '\x1b[0;47;37mcolor=gray background=gray, underline=False\x1b[m'
    '\x1b[4;42;37mcolor=gray background=dark_green, underline=True\x1b[m'
    '\x1b[0;42;37mcolor=gray background=dark_green, underline=False\x1b[m'
    '\x1b[4;44;37mcolor=gray background=dark_blue, underline=True\x1b[m'
    '\x1b[0;44;37mcolor=gray background=dark_blue, underline=False\x1b[m'
    '\x1b[4;45;37mcolor=gray background=dark_magenta, underline=True\x1b[m'
    '\x1b[0;45;37mcolor=gray background=dark_magenta, underline=False\x1b[m'
    '\x1b[4;40;37mcolor=gray background=dark_gray, underline=True\x1b[m'
    '\x1b[0;40;37mcolor=gray background=dark_gray, underline=False\x1b[m'
    '\x1b[4;43;37mcolor=gray background=dark_yellow, underline=True\x1b[m'
    '\x1b[0;43;37mcolor=gray background=dark_yellow, underline=False\x1b[m'
    '\x1b[4;43;37mcolor=gray background=yellow, underline=True\x1b[m'
    '\x1b[0;43;37mcolor=gray background=yellow, underline=False\x1b[m'
    '\x1b[4;40;37mcolor=gray background=dark_dark_gray, underline=True\x1b[m'
    '\x1b[0;40;37mcolor=gray background=dark_dark_gray, underline=False\x1b[m'
    '\x1b[4;41;37mcolor=gray background=dark_red, underline=True\x1b[m'
    '\x1b[0;41;37mcolor=gray background=dark_red, underline=False\x1b[m'
    '\x1b[4;47;37mcolor=gray background=dark_white, underline=True\x1b[m'
    '\x1b[0;47;37mcolor=gray background=dark_white, underline=False\x1b[m'
    '\x1b[4;42;37mcolor=gray background=green, underline=True\x1b[m'
    '\x1b[0;42;37mcolor=gray background=green, underline=False\x1b[m'
    '\x1b[4;46;37mcolor=gray background=cyan, underline=True\x1b[m'
    '\x1b[0;46;37mcolor=gray background=cyan, underline=False\x1b[m'
    '\x1b[4;45;37mcolor=gray background=magenta, underline=True\x1b[m'
    '\x1b[0;45;37mcolor=gray background=magenta, underline=False\x1b[m'
    '\x1b[4;46;37mcolor=gray background=dark_cyan, underline=True\x1b[m'
    '\x1b[0;46;37mcolor=gray background=dark_cyan, underline=False\x1b[m'
    '\x1b[4;47;37mcolor=gray background=white, underline=True\x1b[m'
    '\x1b[0;47;37mcolor=gray background=white, underline=False\x1b[m'
    '\x1b[4;41;37mcolor=gray background=red, underline=True\x1b[m'
    '\x1b[0;41;37mcolor=gray background=red, underline=False\x1b[m'
    '\x1b[4;40;37mcolor=gray background=black, underline=True\x1b[m'
    '\x1b[0;40;37mcolor=gray background=black, underline=False\x1b[m'
    '\x1b[4;44;32mcolor=dark_green background=blue, underline=True\x1b[m'
    '\x1b[0;44;32mcolor=dark_green background=blue, underline=False\x1b[m'
    '\x1b[4;47;32mcolor=dark_green background=gray, underline=True\x1b[m'
    '\x1b[0;47;32mcolor=dark_green background=gray, underline=False\x1b[m'
    '\x1b[4;42;32mcolor=dark_green background=dark_green, underline=True\x1b[m'
    '\x1b[0;42;32mcolor=dark_green background=dark_green, underline=False\x1b[m'
    '\x1b[4;44;32mcolor=dark_green background=dark_blue, underline=True\x1b[m'
    '\x1b[0;44;32mcolor=dark_green background=dark_blue, underline=False\x1b[m'
    '\x1b[4;45;32mcolor=dark_green background=dark_magenta, underline=True\x1b[m'
    '\x1b[0;45;32mcolor=dark_green background=dark_magenta, underline=False\x1b[m'
    '\x1b[4;40;32mcolor=dark_green background=dark_gray, underline=True\x1b[m'
    '\x1b[0;40;32mcolor=dark_green background=dark_gray, underline=False\x1b[m'
    '\x1b[4;43;32mcolor=dark_green background=dark_yellow, underline=True\x1b[m'
    '\x1b[0;43;32mcolor=dark_green background=dark_yellow, underline=False\x1b[m'
    '\x1b[4;43;32mcolor=dark_green background=yellow, underline=True\x1b[m'
    '\x1b[0;43;32mcolor=dark_green background=yellow, underline=False\x1b[m'
    '\x1b[4;40;32mcolor=dark_green background=dark_dark_gray, underline=True\x1b[m'
    '\x1b[0;40;32mcolor=dark_green background=dark_dark_gray, underline=False\x1b[m'
    '\x1b[4;41;32mcolor=dark_green background=dark_red, underline=True\x1b[m'
    '\x1b[0;41;32mcolor=dark_green background=dark_red, underline=False\x1b[m'
    '\x1b[4;47;32mcolor=dark_green background=dark_white, underline=True\x1b[m'
    '\x1b[0;47;32mcolor=dark_green background=dark_white, underline=False\x1b[m'
    '\x1b[4;42;32mcolor=dark_green background=green, underline=True\x1b[m'
    '\x1b[0;42;32mcolor=dark_green background=green, underline=False\x1b[m'
    '\x1b[4;46;32mcolor=dark_green background=cyan, underline=True\x1b[m'
    '\x1b[0;46;32mcolor=dark_green background=cyan, underline=False\x1b[m'
    '\x1b[4;45;32mcolor=dark_green background=magenta, underline=True\x1b[m'
    '\x1b[0;45;32mcolor=dark_green background=magenta, underline=False\x1b[m'
    '\x1b[4;46;32mcolor=dark_green background=dark_cyan, underline=True\x1b[m'
    '\x1b[0;46;32mcolor=dark_green background=dark_cyan, underline=False\x1b[m'
    '\x1b[4;47;32mcolor=dark_green background=white, underline=True\x1b[m'
    '\x1b[0;47;32mcolor=dark_green background=white, underline=False\x1b[m'
    '\x1b[4;41;32mcolor=dark_green background=red, underline=True\x1b[m'
    '\x1b[0;41;32mcolor=dark_green background=red, underline=False\x1b[m'
    '\x1b[4;40;32mcolor=dark_green background=black, underline=True\x1b[m'
    '\x1b[0;40;32mcolor=dark_green background=black, underline=False\x1b[m'
    '\x1b[4;44;34mcolor=dark_blue background=blue, underline=True\x1b[m'
    '\x1b[0;44;34mcolor=dark_blue background=blue, underline=False\x1b[m'
    '\x1b[4;47;34mcolor=dark_blue background=gray, underline=True\x1b[m'
    '\x1b[0;47;34mcolor=dark_blue background=gray, underline=False\x1b[m'
    '\x1b[4;42;34mcolor=dark_blue background=dark_green, underline=True\x1b[m'
    '\x1b[0;42;34mcolor=dark_blue background=dark_green, underline=False\x1b[m'
    '\x1b[4;44;34mcolor=dark_blue background=dark_blue, underline=True\x1b[m'
    '\x1b[0;44;34mcolor=dark_blue background=dark_blue, underline=False\x1b[m'
    '\x1b[4;45;34mcolor=dark_blue background=dark_magenta, underline=True\x1b[m'
    '\x1b[0;45;34mcolor=dark_blue background=dark_magenta, underline=False\x1b[m'
    '\x1b[4;40;34mcolor=dark_blue background=dark_gray, underline=True\x1b[m'
    '\x1b[0;40;34mcolor=dark_blue background=dark_gray, underline=False\x1b[m'
    '\x1b[4;43;34mcolor=dark_blue background=dark_yellow, underline=True\x1b[m'
    '\x1b[0;43;34mcolor=dark_blue background=dark_yellow, underline=False\x1b[m'
    '\x1b[4;43;34mcolor=dark_blue background=yellow, underline=True\x1b[m'
    '\x1b[0;43;34mcolor=dark_blue background=yellow, underline=False\x1b[m'
    '\x1b[4;40;34mcolor=dark_blue background=dark_dark_gray, underline=True\x1b[m'
    '\x1b[0;40;34mcolor=dark_blue background=dark_dark_gray, underline=False\x1b[m'
    '\x1b[4;41;34mcolor=dark_blue background=dark_red, underline=True\x1b[m'
    '\x1b[0;41;34mcolor=dark_blue background=dark_red, underline=False\x1b[m'
    '\x1b[4;47;34mcolor=dark_blue background=dark_white, underline=True\x1b[m'
    '\x1b[0;47;34mcolor=dark_blue background=dark_white, underline=False\x1b[m'
    '\x1b[4;42;34mcolor=dark_blue background=green, underline=True\x1b[m'
    '\x1b[0;42;34mcolor=dark_blue background=green, underline=False\x1b[m'
    '\x1b[4;46;34mcolor=dark_blue background=cyan, underline=True\x1b[m'
    '\x1b[0;46;34mcolor=dark_blue background=cyan, underline=False\x1b[m'
    '\x1b[4;45;34mcolor=dark_blue background=magenta, underline=True\x1b[m'
    '\x1b[0;45;34mcolor=dark_blue background=magenta, underline=False\x1b[m'
    '\x1b[4;46;34mcolor=dark_blue background=dark_cyan, underline=True\x1b[m'
    '\x1b[0;46;34mcolor=dark_blue background=dark_cyan, underline=False\x1b[m'
    '\x1b[4;47;34mcolor=dark_blue background=white, underline=True\x1b[m'
    '\x1b[0;47;34mcolor=dark_blue background=white, underline=False\x1b[m'
    '\x1b[4;41;34mcolor=dark_blue background=red, underline=True\x1b[m'
    '\x1b[0;41;34mcolor=dark_blue background=red, underline=False\x1b[m'
    '\x1b[4;40;34mcolor=dark_blue background=black, underline=True\x1b[m'
    '\x1b[0;40;34mcolor=dark_blue background=black, underline=False\x1b[m'
    '\x1b[4;44;35mcolor=dark_magenta background=blue, underline=True\x1b[m'
    '\x1b[0;44;35mcolor=dark_magenta background=blue, underline=False\x1b[m'
    '\x1b[4;47;35mcolor=dark_magenta background=gray, underline=True\x1b[m'
    '\x1b[0;47;35mcolor=dark_magenta background=gray, underline=False\x1b[m'
    '\x1b[4;42;35mcolor=dark_magenta background=dark_green, underline=True\x1b[m'
    '\x1b[0;42;35mcolor=dark_magenta background=dark_green, underline=False\x1b[m'
    '\x1b[4;44;35mcolor=dark_magenta background=dark_blue, underline=True\x1b[m'
    '\x1b[0;44;35mcolor=dark_magenta background=dark_blue, underline=False\x1b[m'
    '\x1b[4;45;35mcolor=dark_magenta background=dark_magenta, underline=True\x1b[m'
    '\x1b[0;45;35mcolor=dark_magenta background=dark_magenta, underline=False\x1b[m'
    '\x1b[4;40;35mcolor=dark_magenta background=dark_gray, underline=True\x1b[m'
    '\x1b[0;40;35mcolor=dark_magenta background=dark_gray, underline=False\x1b[m'
    '\x1b[4;43;35mcolor=dark_magenta background=dark_yellow, underline=True\x1b[m'
    '\x1b[0;43;35mcolor=dark_magenta background=dark_yellow, underline=False\x1b[m'
    '\x1b[4;43;35mcolor=dark_magenta background=yellow, underline=True\x1b[m'
    '\x1b[0;43;35mcolor=dark_magenta background=yellow, underline=False\x1b[m'
    '\x1b[4;40;35mcolor=dark_magenta background=dark_dark_gray, underline=True\x1b[m'
    '\x1b[0;40;35mcolor=dark_magenta background=dark_dark_gray, underline=False\x1b[m'
    '\x1b[4;41;35mcolor=dark_magenta background=dark_red, underline=True\x1b[m'
    '\x1b[0;41;35mcolor=dark_magenta background=dark_red, underline=False\x1b[m'
    '\x1b[4;47;35mcolor=dark_magenta background=dark_white, underline=True\x1b[m'
    '\x1b[0;47;35mcolor=dark_magenta background=dark_white, underline=False\x1b[m'
    '\x1b[4;42;35mcolor=dark_magenta background=green, underline=True\x1b[m'
    '\x1b[0;42;35mcolor=dark_magenta background=green, underline=False\x1b[m'
    '\x1b[4;46;35mcolor=dark_magenta background=cyan, underline=True\x1b[m'
    '\x1b[0;46;35mcolor=dark_magenta background=cyan, underline=False\x1b[m'
    '\x1b[4;45;35mcolor=dark_magenta background=magenta, underline=True\x1b[m'
    '\x1b[0;45;35mcolor=dark_magenta background=magenta, underline=False\x1b[m'
    '\x1b[4;46;35mcolor=dark_magenta background=dark_cyan, underline=True\x1b[m'
    '\x1b[0;46;35mcolor=dark_magenta background=dark_cyan, underline=False\x1b[m'
    '\x1b[4;47;35mcolor=dark_magenta background=white, underline=True\x1b[m'
    '\x1b[0;47;35mcolor=dark_magenta background=white, underline=False\x1b[m'
    '\x1b[4;41;35mcolor=dark_magenta background=red, underline=True\x1b[m'
    '\x1b[0;41;35mcolor=dark_magenta background=red, underline=False\x1b[m'
    '\x1b[4;40;35mcolor=dark_magenta background=black, underline=True\x1b[m'
    '\x1b[0;40;35mcolor=dark_magenta background=black, underline=False\x1b[m'
    '\x1b[4m\x1b[1;44;30mcolor=dark_gray background=blue, underline=True\x1b[m'
    '\x1b[1;44;30mcolor=dark_gray background=blue, underline=False\x1b[m'
    '\x1b[4m\x1b[1;47;30mcolor=dark_gray background=gray, underline=True\x1b[m'
    '\x1b[1;47;30mcolor=dark_gray background=gray, underline=False\x1b[m'
    '\x1b[4m\x1b[1;42;30mcolor=dark_gray background=dark_green, underline=True\x1b[m'
    '\x1b[1;42;30mcolor=dark_gray background=dark_green, underline=False\x1b[m'
    '\x1b[4m\x1b[1;44;30mcolor=dark_gray background=dark_blue, underline=True\x1b[m'
    '\x1b[1;44;30mcolor=dark_gray background=dark_blue, underline=False\x1b[m'
    '\x1b[4m\x1b[1;45;30mcolor=dark_gray background=dark_magenta, underline=True\x1b[m'
    '\x1b[1;45;30mcolor=dark_gray background=dark_magenta, underline=False\x1b[m'
    '\x1b[4m\x1b[1;40;30mcolor=dark_gray background=dark_gray, underline=True\x1b[m'
    '\x1b[1;40;30mcolor=dark_gray background=dark_gray, underline=False\x1b[m'
    '\x1b[4m\x1b[1;43;30mcolor=dark_gray background=dark_yellow, underline=True\x1b[m'
    '\x1b[1;43;30mcolor=dark_gray background=dark_yellow, underline=False\x1b[m'
    '\x1b[4m\x1b[1;43;30mcolor=dark_gray background=yellow, underline=True\x1b[m'
    '\x1b[1;43;30mcolor=dark_gray background=yellow, underline=False\x1b[m'
    '\x1b[4m\x1b[1;40;30mcolor=dark_gray background=dark_dark_gray, underline=True\x1b[m'
    '\x1b[1;40;30mcolor=dark_gray background=dark_dark_gray, underline=False\x1b[m'
    '\x1b[4m\x1b[1;41;30mcolor=dark_gray background=dark_red, underline=True\x1b[m'
    '\x1b[1;41;30mcolor=dark_gray background=dark_red, underline=False\x1b[m'
    '\x1b[4m\x1b[1;47;30mcolor=dark_gray background=dark_white, underline=True\x1b[m'
    '\x1b[1;47;30mcolor=dark_gray background=dark_white, underline=False\x1b[m'
    '\x1b[4m\x1b[1;42;30mcolor=dark_gray background=green, underline=True\x1b[m'
    '\x1b[1;42;30mcolor=dark_gray background=green, underline=False\x1b[m'
    '\x1b[4m\x1b[1;46;30mcolor=dark_gray background=cyan, underline=True\x1b[m'
    '\x1b[1;46;30mcolor=dark_gray background=cyan, underline=False\x1b[m'
    '\x1b[4m\x1b[1;45;30mcolor=dark_gray background=magenta, underline=True\x1b[m'
    '\x1b[1;45;30mcolor=dark_gray background=magenta, underline=False\x1b[m'
    '\x1b[4m\x1b[1;46;30mcolor=dark_gray background=dark_cyan, underline=True\x1b[m'
    '\x1b[1;46;30mcolor=dark_gray background=dark_cyan, underline=False\x1b[m'
    '\x1b[4m\x1b[1;47;30mcolor=dark_gray background=white, underline=True\x1b[m'
    '\x1b[1;47;30mcolor=dark_gray background=white, underline=False\x1b[m'
    '\x1b[4m\x1b[1;41;30mcolor=dark_gray background=red, underline=True\x1b[m'
    '\x1b[1;41;30mcolor=dark_gray background=red, underline=False\x1b[m'
    '\x1b[4m\x1b[1;40;30mcolor=dark_gray background=black, underline=True\x1b[m'
    '\x1b[1;40;30mcolor=dark_gray background=black, underline=False\x1b[m'
    '\x1b[4;44;33mcolor=dark_yellow background=blue, underline=True\x1b[m'
    '\x1b[0;44;33mcolor=dark_yellow background=blue, underline=False\x1b[m'
    '\x1b[4;47;33mcolor=dark_yellow background=gray, underline=True\x1b[m'
    '\x1b[0;47;33mcolor=dark_yellow background=gray, underline=False\x1b[m'
    '\x1b[4;42;33mcolor=dark_yellow background=dark_green, underline=True\x1b[m'
    '\x1b[0;42;33mcolor=dark_yellow background=dark_green, underline=False\x1b[m'
    '\x1b[4;44;33mcolor=dark_yellow background=dark_blue, underline=True\x1b[m'
    '\x1b[0;44;33mcolor=dark_yellow background=dark_blue, underline=False\x1b[m'
    '\x1b[4;45;33mcolor=dark_yellow background=dark_magenta, underline=True\x1b[m'
    '\x1b[0;45;33mcolor=dark_yellow background=dark_magenta, underline=False\x1b[m'
    '\x1b[4;40;33mcolor=dark_yellow background=dark_gray, underline=True\x1b[m'
    '\x1b[0;40;33mcolor=dark_yellow background=dark_gray, underline=False\x1b[m'
    '\x1b[4;43;33mcolor=dark_yellow background=dark_yellow, underline=True\x1b[m'
    '\x1b[0;43;33mcolor=dark_yellow background=dark_yellow, underline=False\x1b[m'
    '\x1b[4;43;33mcolor=dark_yellow background=yellow, underline=True\x1b[m'
    '\x1b[0;43;33mcolor=dark_yellow background=yellow, underline=False\x1b[m'
    '\x1b[4;40;33mcolor=dark_yellow background=dark_dark_gray, underline=True\x1b[m'
    '\x1b[0;40;33mcolor=dark_yellow background=dark_dark_gray, underline=False\x1b[m'
    '\x1b[4;41;33mcolor=dark_yellow background=dark_red, underline=True\x1b[m'
    '\x1b[0;41;33mcolor=dark_yellow background=dark_red, underline=False\x1b[m'
    '\x1b[4;47;33mcolor=dark_yellow background=dark_white, underline=True\x1b[m'
    '\x1b[0;47;33mcolor=dark_yellow background=dark_white, underline=False\x1b[m'
    '\x1b[4;42;33mcolor=dark_yellow background=green, underline=True\x1b[m'
    '\x1b[0;42;33mcolor=dark_yellow background=green, underline=False\x1b[m'
    '\x1b[4;46;33mcolor=dark_yellow background=cyan, underline=True\x1b[m'
    '\x1b[0;46;33mcolor=dark_yellow background=cyan, underline=False\x1b[m'
    '\x1b[4;45;33mcolor=dark_yellow background=magenta, underline=True\x1b[m'
    '\x1b[0;45;33mcolor=dark_yellow background=magenta, underline=False\x1b[m'
    '\x1b[4;46;33mcolor=dark_yellow background=dark_cyan, underline=True\x1b[m'
    '\x1b[0;46;33mcolor=dark_yellow background=dark_cyan, underline=False\x1b[m'
    '\x1b[4;47;33mcolor=dark_yellow background=white, underline=True\x1b[m'
    '\x1b[0;47;33mcolor=dark_yellow background=white, underline=False\x1b[m'
    '\x1b[4;41;33mcolor=dark_yellow background=red, underline=True\x1b[m'
    '\x1b[0;41;33mcolor=dark_yellow background=red, underline=False\x1b[m'
    '\x1b[4;40;33mcolor=dark_yellow background=black, underline=True\x1b[m'
    '\x1b[0;40;33mcolor=dark_yellow background=black, underline=False\x1b[m'
    '\x1b[4m\x1b[1;44;33mcolor=yellow background=blue, underline=True\x1b[m'
    '\x1b[1;44;33mcolor=yellow background=blue, underline=False\x1b[m'
    '\x1b[4m\x1b[1;47;33mcolor=yellow background=gray, underline=True\x1b[m'
    '\x1b[1;47;33mcolor=yellow background=gray, underline=False\x1b[m'
    '\x1b[4m\x1b[1;42;33mcolor=yellow background=dark_green, underline=True\x1b[m'
    '\x1b[1;42;33mcolor=yellow background=dark_green, underline=False\x1b[m'
    '\x1b[4m\x1b[1;44;33mcolor=yellow background=dark_blue, underline=True\x1b[m'
    '\x1b[1;44;33mcolor=yellow background=dark_blue, underline=False\x1b[m'
    '\x1b[4m\x1b[1;45;33mcolor=yellow background=dark_magenta, underline=True\x1b[m'
    '\x1b[1;45;33mcolor=yellow background=dark_magenta, underline=False\x1b[m'
    '\x1b[4m\x1b[1;40;33mcolor=yellow background=dark_gray, underline=True\x1b[m'
    '\x1b[1;40;33mcolor=yellow background=dark_gray, underline=False\x1b[m'
    '\x1b[4m\x1b[1;43;33mcolor=yellow background=dark_yellow, underline=True\x1b[m'
    '\x1b[1;43;33mcolor=yellow background=dark_yellow, underline=False\x1b[m'
    '\x1b[4m\x1b[1;43;33mcolor=yellow background=yellow, underline=True\x1b[m'
    '\x1b[1;43;33mcolor=yellow background=yellow, underline=False\x1b[m'
    '\x1b[4m\x1b[1;40;33mcolor=yellow background=dark_dark_gray, underline=True\x1b[m'
    '\x1b[1;40;33mcolor=yellow background=dark_dark_gray, underline=False\x1b[m'
    '\x1b[4m\x1b[1;41;33mcolor=yellow background=dark_red, underline=True\x1b[m'
    '\x1b[1;41;33mcolor=yellow background=dark_red, underline=False\x1b[m'
    '\x1b[4m\x1b[1;47;33mcolor=yellow background=dark_white, underline=True\x1b[m'
    '\x1b[1;47;33mcolor=yellow background=dark_white, underline=False\x1b[m'
    '\x1b[4m\x1b[1;42;33mcolor=yellow background=green, underline=True\x1b[m'
    '\x1b[1;42;33mcolor=yellow background=green, underline=False\x1b[m'
    '\x1b[4m\x1b[1;46;33mcolor=yellow background=cyan, underline=True\x1b[m'
    '\x1b[1;46;33mcolor=yellow background=cyan, underline=False\x1b[m'
    '\x1b[4m\x1b[1;45;33mcolor=yellow background=magenta, underline=True\x1b[m'
    '\x1b[1;45;33mcolor=yellow background=magenta, underline=False\x1b[m'
    '\x1b[4m\x1b[1;46;33mcolor=yellow background=dark_cyan, underline=True\x1b[m'
    '\x1b[1;46;33mcolor=yellow background=dark_cyan, underline=False\x1b[m'
    '\x1b[4m\x1b[1;47;33mcolor=yellow background=white, underline=True\x1b[m'
    '\x1b[1;47;33mcolor=yellow background=white, underline=False\x1b[m'
    '\x1b[4m\x1b[1;41;33mcolor=yellow background=red, underline=True\x1b[m'
    '\x1b[1;41;33mcolor=yellow background=red, underline=False\x1b[m'
    '\x1b[4m\x1b[1;40;33mcolor=yellow background=black, underline=True\x1b[m'
    '\x1b[1;40;33mcolor=yellow background=black, underline=False\x1b[m'
    '\x1b[4;44;30mcolor=dark_dark_gray background=blue, underline=True\x1b[m'
    '\x1b[0;44;30mcolor=dark_dark_gray background=blue, underline=False\x1b[m'
    '\x1b[4;47;30mcolor=dark_dark_gray background=gray, underline=True\x1b[m'
    '\x1b[0;47;30mcolor=dark_dark_gray background=gray, underline=False\x1b[m'
    '\x1b[4;42;30mcolor=dark_dark_gray background=dark_green, underline=True\x1b[m'
    '\x1b[0;42;30mcolor=dark_dark_gray background=dark_green, underline=False\x1b[m'
    '\x1b[4;44;30mcolor=dark_dark_gray background=dark_blue, underline=True\x1b[m'
    '\x1b[0;44;30mcolor=dark_dark_gray background=dark_blue, underline=False\x1b[m'
    '\x1b[4;45;30mcolor=dark_dark_gray background=dark_magenta, underline=True\x1b[m'
    '\x1b[0;45;30mcolor=dark_dark_gray background=dark_magenta, underline=False\x1b[m'
    '\x1b[4;40;30mcolor=dark_dark_gray background=dark_gray, underline=True\x1b[m'
    '\x1b[0;40;30mcolor=dark_dark_gray background=dark_gray, underline=False\x1b[m'
    '\x1b[4;43;30mcolor=dark_dark_gray background=dark_yellow, underline=True\x1b[m'
    '\x1b[0;43;30mcolor=dark_dark_gray background=dark_yellow, underline=False\x1b[m'
    '\x1b[4;43;30mcolor=dark_dark_gray background=yellow, underline=True\x1b[m'
    '\x1b[0;43;30mcolor=dark_dark_gray background=yellow, underline=False\x1b[m'
    '\x1b[4;40;30mcolor=dark_dark_gray background=dark_dark_gray, underline=True\x1b[m'
    '\x1b[0;40;30mcolor=dark_dark_gray background=dark_dark_gray, underline=False\x1b[m'
    '\x1b[4;41;30mcolor=dark_dark_gray background=dark_red, underline=True\x1b[m'
    '\x1b[0;41;30mcolor=dark_dark_gray background=dark_red, underline=False\x1b[m'
    '\x1b[4;47;30mcolor=dark_dark_gray background=dark_white, underline=True\x1b[m'
    '\x1b[0;47;30mcolor=dark_dark_gray background=dark_white, underline=False\x1b[m'
    '\x1b[4;42;30mcolor=dark_dark_gray background=green, underline=True\x1b[m'
    '\x1b[0;42;30mcolor=dark_dark_gray background=green, underline=False\x1b[m'
    '\x1b[4;46;30mcolor=dark_dark_gray background=cyan, underline=True\x1b[m'
    '\x1b[0;46;30mcolor=dark_dark_gray background=cyan, underline=False\x1b[m'
    '\x1b[4;45;30mcolor=dark_dark_gray background=magenta, underline=True\x1b[m'
    '\x1b[0;45;30mcolor=dark_dark_gray background=magenta, underline=False\x1b[m'
    '\x1b[4;46;30mcolor=dark_dark_gray background=dark_cyan, underline=True\x1b[m'
    '\x1b[0;46;30mcolor=dark_dark_gray background=dark_cyan, underline=False\x1b[m'
    '\x1b[4;47;30mcolor=dark_dark_gray background=white, underline=True\x1b[m'
    '\x1b[0;47;30mcolor=dark_dark_gray background=white, underline=False\x1b[m'
    '\x1b[4;41;30mcolor=dark_dark_gray background=red, underline=True\x1b[m'
    '\x1b[0;41;30mcolor=dark_dark_gray background=red, underline=False\x1b[m'
    '\x1b[4;40;30mcolor=dark_dark_gray background=black, underline=True\x1b[m'
    '\x1b[0;40;30mcolor=dark_dark_gray background=black, underline=False\x1b[m'
    '\x1b[4;44;31mcolor=dark_red background=blue, underline=True\x1b[m'
    '\x1b[0;44;31mcolor=dark_red background=blue, underline=False\x1b[m'
    '\x1b[4;47;31mcolor=dark_red background=gray, underline=True\x1b[m'
    '\x1b[0;47;31mcolor=dark_red background=gray, underline=False\x1b[m'
    '\x1b[4;42;31mcolor=dark_red background=dark_green, underline=True\x1b[m'
    '\x1b[0;42;31mcolor=dark_red background=dark_green, underline=False\x1b[m'
    '\x1b[4;44;31mcolor=dark_red background=dark_blue, underline=True\x1b[m'
    '\x1b[0;44;31mcolor=dark_red background=dark_blue, underline=False\x1b[m'
    '\x1b[4;45;31mcolor=dark_red background=dark_magenta, underline=True\x1b[m'
    '\x1b[0;45;31mcolor=dark_red background=dark_magenta, underline=False\x1b[m'
    '\x1b[4;40;31mcolor=dark_red background=dark_gray, underline=True\x1b[m'
    '\x1b[0;40;31mcolor=dark_red background=dark_gray, underline=False\x1b[m'
    '\x1b[4;43;31mcolor=dark_red background=dark_yellow, underline=True\x1b[m'
    '\x1b[0;43;31mcolor=dark_red background=dark_yellow, underline=False\x1b[m'
    '\x1b[4;43;31mcolor=dark_red background=yellow, underline=True\x1b[m'
    '\x1b[0;43;31mcolor=dark_red background=yellow, underline=False\x1b[m'
    '\x1b[4;40;31mcolor=dark_red background=dark_dark_gray, underline=True\x1b[m'
    '\x1b[0;40;31mcolor=dark_red background=dark_dark_gray, underline=False\x1b[m'
    '\x1b[4;41;31mcolor=dark_red background=dark_red, underline=True\x1b[m'
    '\x1b[0;41;31mcolor=dark_red background=dark_red, underline=False\x1b[m'
    '\x1b[4;47;31mcolor=dark_red background=dark_white, underline=True\x1b[m'
    '\x1b[0;47;31mcolor=dark_red background=dark_white, underline=False\x1b[m'
    '\x1b[4;42;31mcolor=dark_red background=green, underline=True\x1b[m'
    '\x1b[0;42;31mcolor=dark_red background=green, underline=False\x1b[m'
    '\x1b[4;46;31mcolor=dark_red background=cyan, underline=True\x1b[m'
    '\x1b[0;46;31mcolor=dark_red background=cyan, underline=False\x1b[m'
    '\x1b[4;45;31mcolor=dark_red background=magenta, underline=True\x1b[m'
    '\x1b[0;45;31mcolor=dark_red background=magenta, underline=False\x1b[m'
    '\x1b[4;46;31mcolor=dark_red background=dark_cyan, underline=True\x1b[m'
    '\x1b[0;46;31mcolor=dark_red background=dark_cyan, underline=False\x1b[m'
    '\x1b[4;47;31mcolor=dark_red background=white, underline=True\x1b[m'
    '\x1b[0;47;31mcolor=dark_red background=white, underline=False\x1b[m'
    '\x1b[4;41;31mcolor=dark_red background=red, underline=True\x1b[m'
    '\x1b[0;41;31mcolor=dark_red background=red, underline=False\x1b[m'
    '\x1b[4;40;31mcolor=dark_red background=black, underline=True\x1b[m'
    '\x1b[0;40;31mcolor=dark_red background=black, underline=False\x1b[m'
    '\x1b[4;44;37mcolor=dark_white background=blue, underline=True\x1b[m'
    '\x1b[0;44;37mcolor=dark_white background=blue, underline=False\x1b[m'
    '\x1b[4;47;37mcolor=dark_white background=gray, underline=True\x1b[m'
    '\x1b[0;47;37mcolor=dark_white background=gray, underline=False\x1b[m'
    '\x1b[4;42;37mcolor=dark_white background=dark_green, underline=True\x1b[m'
    '\x1b[0;42;37mcolor=dark_white background=dark_green, underline=False\x1b[m'
    '\x1b[4;44;37mcolor=dark_white background=dark_blue, underline=True\x1b[m'
    '\x1b[0;44;37mcolor=dark_white background=dark_blue, underline=False\x1b[m'
    '\x1b[4;45;37mcolor=dark_white background=dark_magenta, underline=True\x1b[m'
    '\x1b[0;45;37mcolor=dark_white background=dark_magenta, underline=False\x1b[m'
    '\x1b[4;40;37mcolor=dark_white background=dark_gray, underline=True\x1b[m'
    '\x1b[0;40;37mcolor=dark_white background=dark_gray, underline=False\x1b[m'
    '\x1b[4;43;37mcolor=dark_white background=dark_yellow, underline=True\x1b[m'
    '\x1b[0;43;37mcolor=dark_white background=dark_yellow, underline=False\x1b[m'
    '\x1b[4;43;37mcolor=dark_white background=yellow, underline=True\x1b[m'
    '\x1b[0;43;37mcolor=dark_white background=yellow, underline=False\x1b[m'
    '\x1b[4;40;37mcolor=dark_white background=dark_dark_gray, underline=True\x1b[m'
    '\x1b[0;40;37mcolor=dark_white background=dark_dark_gray, underline=False\x1b[m'
    '\x1b[4;41;37mcolor=dark_white background=dark_red, underline=True\x1b[m'
    '\x1b[0;41;37mcolor=dark_white background=dark_red, underline=False\x1b[m'
    '\x1b[4;47;37mcolor=dark_white background=dark_white, underline=True\x1b[m'
    '\x1b[0;47;37mcolor=dark_white background=dark_white, underline=False\x1b[m'
    '\x1b[4;42;37mcolor=dark_white background=green, underline=True\x1b[m'
    '\x1b[0;42;37mcolor=dark_white background=green, underline=False\x1b[m'
    '\x1b[4;46;37mcolor=dark_white background=cyan, underline=True\x1b[m'
    '\x1b[0;46;37mcolor=dark_white background=cyan, underline=False\x1b[m'
    '\x1b[4;45;37mcolor=dark_white background=magenta, underline=True\x1b[m'
    '\x1b[0;45;37mcolor=dark_white background=magenta, underline=False\x1b[m'
    '\x1b[4;46;37mcolor=dark_white background=dark_cyan, underline=True\x1b[m'
    '\x1b[0;46;37mcolor=dark_white background=dark_cyan, underline=False\x1b[m'
    '\x1b[4;47;37mcolor=dark_white background=white, underline=True\x1b[m'
    '\x1b[0;47;37mcolor=dark_white background=white, underline=False\x1b[m'
    '\x1b[4;41;37mcolor=dark_white background=red, underline=True\x1b[m'
    '\x1b[0;41;37mcolor=dark_white background=red, underline=False\x1b[m'
    '\x1b[4;40;37mcolor=dark_white background=black, underline=True\x1b[m'
    '\x1b[0;40;37mcolor=dark_white background=black, underline=False\x1b[m'
    '\x1b[4m\x1b[1;44;32mcolor=green background=blue, underline=True\x1b[m'
    '\x1b[1;44;32mcolor=green background=blue, underline=False\x1b[m'
    '\x1b[4m\x1b[1;47;32mcolor=green background=gray, underline=True\x1b[m'
    '\x1b[1;47;32mcolor=green background=gray, underline=False\x1b[m'
    '\x1b[4m\x1b[1;42;32mcolor=green background=dark_green, underline=True\x1b[m'
    '\x1b[1;42;32mcolor=green background=dark_green, underline=False\x1b[m'
    '\x1b[4m\x1b[1;44;32mcolor=green background=dark_blue, underline=True\x1b[m'
    '\x1b[1;44;32mcolor=green background=dark_blue, underline=False\x1b[m'
    '\x1b[4m\x1b[1;45;32mcolor=green background=dark_magenta, underline=True\x1b[m'
    '\x1b[1;45;32mcolor=green background=dark_magenta, underline=False\x1b[m'
    '\x1b[4m\x1b[1;40;32mcolor=green background=dark_gray, underline=True\x1b[m'
    '\x1b[1;40;32mcolor=green background=dark_gray, underline=False\x1b[m'
    '\x1b[4m\x1b[1;43;32mcolor=green background=dark_yellow, underline=True\x1b[m'
    '\x1b[1;43;32mcolor=green background=dark_yellow, underline=False\x1b[m'
    '\x1b[4m\x1b[1;43;32mcolor=green background=yellow, underline=True\x1b[m'
    '\x1b[1;43;32mcolor=green background=yellow, underline=False\x1b[m'
    '\x1b[4m\x1b[1;40;32mcolor=green background=dark_dark_gray, underline=True\x1b[m'
    '\x1b[1;40;32mcolor=green background=dark_dark_gray, underline=False\x1b[m'
    '\x1b[4m\x1b[1;41;32mcolor=green background=dark_red, underline=True\x1b[m'
    '\x1b[1;41;32mcolor=green background=dark_red, underline=False\x1b[m'
    '\x1b[4m\x1b[1;47;32mcolor=green background=dark_white, underline=True\x1b[m'
    '\x1b[1;47;32mcolor=green background=dark_white, underline=False\x1b[m'
    '\x1b[4m\x1b[1;42;32mcolor=green background=green, underline=True\x1b[m'
    '\x1b[1;42;32mcolor=green background=green, underline=False\x1b[m'
    '\x1b[4m\x1b[1;46;32mcolor=green background=cyan, underline=True\x1b[m'
    '\x1b[1;46;32mcolor=green background=cyan, underline=False\x1b[m'
    '\x1b[4m\x1b[1;45;32mcolor=green background=magenta, underline=True\x1b[m'
    '\x1b[1;45;32mcolor=green background=magenta, underline=False\x1b[m'
    '\x1b[4m\x1b[1;46;32mcolor=green background=dark_cyan, underline=True\x1b[m'
    '\x1b[1;46;32mcolor=green background=dark_cyan, underline=False\x1b[m'
    '\x1b[4m\x1b[1;47;32mcolor=green background=white, underline=True\x1b[m'
    '\x1b[1;47;32mcolor=green background=white, underline=False\x1b[m'
    '\x1b[4m\x1b[1;41;32mcolor=green background=red, underline=True\x1b[m'
    '\x1b[1;41;32mcolor=green background=red, underline=False\x1b[m'
    '\x1b[4m\x1b[1;40;32mcolor=green background=black, underline=True\x1b[m'
    '\x1b[1;40;32mcolor=green background=black, underline=False\x1b[m'
    '\x1b[4m\x1b[1;44;36mcolor=cyan background=blue, underline=True\x1b[m'
    '\x1b[1;44;36mcolor=cyan background=blue, underline=False\x1b[m'
    '\x1b[4m\x1b[1;47;36mcolor=cyan background=gray, underline=True\x1b[m'
    '\x1b[1;47;36mcolor=cyan background=gray, underline=False\x1b[m'
    '\x1b[4m\x1b[1;42;36mcolor=cyan background=dark_green, underline=True\x1b[m'
    '\x1b[1;42;36mcolor=cyan background=dark_green, underline=False\x1b[m'
    '\x1b[4m\x1b[1;44;36mcolor=cyan background=dark_blue, underline=True\x1b[m'
    '\x1b[1;44;36mcolor=cyan background=dark_blue, underline=False\x1b[m'
    '\x1b[4m\x1b[1;45;36mcolor=cyan background=dark_magenta, underline=True\x1b[m'
    '\x1b[1;45;36mcolor=cyan background=dark_magenta, underline=False\x1b[m'
    '\x1b[4m\x1b[1;40;36mcolor=cyan background=dark_gray, underline=True\x1b[m'
    '\x1b[1;40;36mcolor=cyan background=dark_gray, underline=False\x1b[m'
    '\x1b[4m\x1b[1;43;36mcolor=cyan background=dark_yellow, underline=True\x1b[m'
    '\x1b[1;43;36mcolor=cyan background=dark_yellow, underline=False\x1b[m'
    '\x1b[4m\x1b[1;43;36mcolor=cyan background=yellow, underline=True\x1b[m'
    '\x1b[1;43;36mcolor=cyan background=yellow, underline=False\x1b[m'
    '\x1b[4m\x1b[1;40;36mcolor=cyan background=dark_dark_gray, underline=True\x1b[m'
    '\x1b[1;40;36mcolor=cyan background=dark_dark_gray, underline=False\x1b[m'
    '\x1b[4m\x1b[1;41;36mcolor=cyan background=dark_red, underline=True\x1b[m'
    '\x1b[1;41;36mcolor=cyan background=dark_red, underline=False\x1b[m'
    '\x1b[4m\x1b[1;47;36mcolor=cyan background=dark_white, underline=True\x1b[m'
    '\x1b[1;47;36mcolor=cyan background=dark_white, underline=False\x1b[m'
    '\x1b[4m\x1b[1;42;36mcolor=cyan background=green, underline=True\x1b[m'
    '\x1b[1;42;36mcolor=cyan background=green, underline=False\x1b[m'
    '\x1b[4m\x1b[1;46;36mcolor=cyan background=cyan, underline=True\x1b[m'
    '\x1b[1;46;36mcolor=cyan background=cyan, underline=False\x1b[m'
    '\x1b[4m\x1b[1;45;36mcolor=cyan background=magenta, underline=True\x1b[m'
    '\x1b[1;45;36mcolor=cyan background=magenta, underline=False\x1b[m'
    '\x1b[4m\x1b[1;46;36mcolor=cyan background=dark_cyan, underline=True\x1b[m'
    '\x1b[1;46;36mcolor=cyan background=dark_cyan, underline=False\x1b[m'
    '\x1b[4m\x1b[1;47;36mcolor=cyan background=white, underline=True\x1b[m'
    '\x1b[1;47;36mcolor=cyan background=white, underline=False\x1b[m'
    '\x1b[4m\x1b[1;41;36mcolor=cyan background=red, underline=True\x1b[m'
    '\x1b[1;41;36mcolor=cyan background=red, underline=False\x1b[m'
    '\x1b[4m\x1b[1;40;36mcolor=cyan background=black, underline=True\x1b[m'
    '\x1b[1;40;36mcolor=cyan background=black, underline=False\x1b[m'
    '\x1b[4m\x1b[1;44;35mcolor=magenta background=blue, underline=True\x1b[m'
    '\x1b[1;44;35mcolor=magenta background=blue, underline=False\x1b[m'
    '\x1b[4m\x1b[1;47;35mcolor=magenta background=gray, underline=True\x1b[m'
    '\x1b[1;47;35mcolor=magenta background=gray, underline=False\x1b[m'
    '\x1b[4m\x1b[1;42;35mcolor=magenta background=dark_green, underline=True\x1b[m'
    '\x1b[1;42;35mcolor=magenta background=dark_green, underline=False\x1b[m'
    '\x1b[4m\x1b[1;44;35mcolor=magenta background=dark_blue, underline=True\x1b[m'
    '\x1b[1;44;35mcolor=magenta background=dark_blue, underline=False\x1b[m'
    '\x1b[4m\x1b[1;45;35mcolor=magenta background=dark_magenta, underline=True\x1b[m'
    '\x1b[1;45;35mcolor=magenta background=dark_magenta, underline=False\x1b[m'
    '\x1b[4m\x1b[1;40;35mcolor=magenta background=dark_gray, underline=True\x1b[m'
    '\x1b[1;40;35mcolor=magenta background=dark_gray, underline=False\x1b[m'
    '\x1b[4m\x1b[1;43;35mcolor=magenta background=dark_yellow, underline=True\x1b[m'
    '\x1b[1;43;35mcolor=magenta background=dark_yellow, underline=False\x1b[m'
    '\x1b[4m\x1b[1;43;35mcolor=magenta background=yellow, underline=True\x1b[m'
    '\x1b[1;43;35mcolor=magenta background=yellow, underline=False\x1b[m'
    '\x1b[4m\x1b[1;40;35mcolor=magenta background=dark_dark_gray, underline=True\x1b[m'
    '\x1b[1;40;35mcolor=magenta background=dark_dark_gray, underline=False\x1b[m'
    '\x1b[4m\x1b[1;41;35mcolor=magenta background=dark_red, underline=True\x1b[m'
    '\x1b[1;41;35mcolor=magenta background=dark_red, underline=False\x1b[m'
    '\x1b[4m\x1b[1;47;35mcolor=magenta background=dark_white, underline=True\x1b[m'
    '\x1b[1;47;35mcolor=magenta background=dark_white, underline=False\x1b[m'
    '\x1b[4m\x1b[1;42;35mcolor=magenta background=green, underline=True\x1b[m'
    '\x1b[1;42;35mcolor=magenta background=green, underline=False\x1b[m'
    '\x1b[4m\x1b[1;46;35mcolor=magenta background=cyan, underline=True\x1b[m'
    '\x1b[1;46;35mcolor=magenta background=cyan, underline=False\x1b[m'
    '\x1b[4m\x1b[1;45;35mcolor=magenta background=magenta, underline=True\x1b[m'
    '\x1b[1;45;35mcolor=magenta background=magenta, underline=False\x1b[m'
    '\x1b[4m\x1b[1;46;35mcolor=magenta background=dark_cyan, underline=True\x1b[m'
    '\x1b[1;46;35mcolor=magenta background=dark_cyan, underline=False\x1b[m'
    '\x1b[4m\x1b[1;47;35mcolor=magenta background=white, underline=True\x1b[m'
    '\x1b[1;47;35mcolor=magenta background=white, underline=False\x1b[m'
    '\x1b[4m\x1b[1;41;35mcolor=magenta background=red, underline=True\x1b[m'
    '\x1b[1;41;35mcolor=magenta background=red, underline=False\x1b[m'
    '\x1b[4m\x1b[1;40;35mcolor=magenta background=black, underline=True\x1b[m'
    '\x1b[1;40;35mcolor=magenta background=black, underline=False\x1b[m'
    '\x1b[4;44;36mcolor=dark_cyan background=blue, underline=True\x1b[m'
    '\x1b[0;44;36mcolor=dark_cyan background=blue, underline=False\x1b[m'
    '\x1b[4;47;36mcolor=dark_cyan background=gray, underline=True\x1b[m'
    '\x1b[0;47;36mcolor=dark_cyan background=gray, underline=False\x1b[m'
    '\x1b[4;42;36mcolor=dark_cyan background=dark_green, underline=True\x1b[m'
    '\x1b[0;42;36mcolor=dark_cyan background=dark_green, underline=False\x1b[m'
    '\x1b[4;44;36mcolor=dark_cyan background=dark_blue, underline=True\x1b[m'
    '\x1b[0;44;36mcolor=dark_cyan background=dark_blue, underline=False\x1b[m'
    '\x1b[4;45;36mcolor=dark_cyan background=dark_magenta, underline=True\x1b[m'
    '\x1b[0;45;36mcolor=dark_cyan background=dark_magenta, underline=False\x1b[m'
    '\x1b[4;40;36mcolor=dark_cyan background=dark_gray, underline=True\x1b[m'
    '\x1b[0;40;36mcolor=dark_cyan background=dark_gray, underline=False\x1b[m'
    '\x1b[4;43;36mcolor=dark_cyan background=dark_yellow, underline=True\x1b[m'
    '\x1b[0;43;36mcolor=dark_cyan background=dark_yellow, underline=False\x1b[m'
    '\x1b[4;43;36mcolor=dark_cyan background=yellow, underline=True\x1b[m'
    '\x1b[0;43;36mcolor=dark_cyan background=yellow, underline=False\x1b[m'
    '\x1b[4;40;36mcolor=dark_cyan background=dark_dark_gray, underline=True\x1b[m'
    '\x1b[0;40;36mcolor=dark_cyan background=dark_dark_gray, underline=False\x1b[m'
    '\x1b[4;41;36mcolor=dark_cyan background=dark_red, underline=True\x1b[m'
    '\x1b[0;41;36mcolor=dark_cyan background=dark_red, underline=False\x1b[m'
    '\x1b[4;47;36mcolor=dark_cyan background=dark_white, underline=True\x1b[m'
    '\x1b[0;47;36mcolor=dark_cyan background=dark_white, underline=False\x1b[m'
    '\x1b[4;42;36mcolor=dark_cyan background=green, underline=True\x1b[m'
    '\x1b[0;42;36mcolor=dark_cyan background=green, underline=False\x1b[m'
    '\x1b[4;46;36mcolor=dark_cyan background=cyan, underline=True\x1b[m'
    '\x1b[0;46;36mcolor=dark_cyan background=cyan, underline=False\x1b[m'
    '\x1b[4;45;36mcolor=dark_cyan background=magenta, underline=True\x1b[m'
    '\x1b[0;45;36mcolor=dark_cyan background=magenta, underline=False\x1b[m'
    '\x1b[4;46;36mcolor=dark_cyan background=dark_cyan, underline=True\x1b[m'
    '\x1b[0;46;36mcolor=dark_cyan background=dark_cyan, underline=False\x1b[m'
    '\x1b[4;47;36mcolor=dark_cyan background=white, underline=True\x1b[m'
    '\x1b[0;47;36mcolor=dark_cyan background=white, underline=False\x1b[m'
    '\x1b[4;41;36mcolor=dark_cyan background=red, underline=True\x1b[m'
    '\x1b[0;41;36mcolor=dark_cyan background=red, underline=False\x1b[m'
    '\x1b[4;40;36mcolor=dark_cyan background=black, underline=True\x1b[m'
    '\x1b[0;40;36mcolor=dark_cyan background=black, underline=False\x1b[m'
    '\x1b[4m\x1b[1;44;37mcolor=white background=blue, underline=True\x1b[m'
    '\x1b[1;44;37mcolor=white background=blue, underline=False\x1b[m'
    '\x1b[4m\x1b[1;47;37mcolor=white background=gray, underline=True\x1b[m'
    '\x1b[1;47;37mcolor=white background=gray, underline=False\x1b[m'
    '\x1b[4m\x1b[1;42;37mcolor=white background=dark_green, underline=True\x1b[m'
    '\x1b[1;42;37mcolor=white background=dark_green, underline=False\x1b[m'
    '\x1b[4m\x1b[1;44;37mcolor=white background=dark_blue, underline=True\x1b[m'
    '\x1b[1;44;37mcolor=white background=dark_blue, underline=False\x1b[m'
    '\x1b[4m\x1b[1;45;37mcolor=white background=dark_magenta, underline=True\x1b[m'
    '\x1b[1;45;37mcolor=white background=dark_magenta, underline=False\x1b[m'
    '\x1b[4m\x1b[1;40;37mcolor=white background=dark_gray, underline=True\x1b[m'
    '\x1b[1;40;37mcolor=white background=dark_gray, underline=False\x1b[m'
    '\x1b[4m\x1b[1;43;37mcolor=white background=dark_yellow, underline=True\x1b[m'
    '\x1b[1;43;37mcolor=white background=dark_yellow, underline=False\x1b[m'
    '\x1b[4m\x1b[1;43;37mcolor=white background=yellow, underline=True\x1b[m'
    '\x1b[1;43;37mcolor=white background=yellow, underline=False\x1b[m'
    '\x1b[4m\x1b[1;40;37mcolor=white background=dark_dark_gray, underline=True\x1b[m'
    '\x1b[1;40;37mcolor=white background=dark_dark_gray, underline=False\x1b[m'
    '\x1b[4m\x1b[1;41;37mcolor=white background=dark_red, underline=True\x1b[m'
    '\x1b[1;41;37mcolor=white background=dark_red, underline=False\x1b[m'
    '\x1b[4m\x1b[1;47;37mcolor=white background=dark_white, underline=True\x1b[m'
    '\x1b[1;47;37mcolor=white background=dark_white, underline=False\x1b[m'
    '\x1b[4m\x1b[1;42;37mcolor=white background=green, underline=True\x1b[m'
    '\x1b[1;42;37mcolor=white background=green, underline=False\x1b[m'
    '\x1b[4m\x1b[1;46;37mcolor=white background=cyan, underline=True\x1b[m'
    '\x1b[1;46;37mcolor=white background=cyan, underline=False\x1b[m'
    '\x1b[4m\x1b[1;45;37mcolor=white background=magenta, underline=True\x1b[m'
    '\x1b[1;45;37mcolor=white background=magenta, underline=False\x1b[m'
    '\x1b[4m\x1b[1;46;37mcolor=white background=dark_cyan, underline=True\x1b[m'
    '\x1b[1;46;37mcolor=white background=dark_cyan, underline=False\x1b[m'
    '\x1b[4m\x1b[1;47;37mcolor=white background=white, underline=True\x1b[m'
    '\x1b[1;47;37mcolor=white background=white, underline=False\x1b[m'
    '\x1b[4m\x1b[1;41;37mcolor=white background=red, underline=True\x1b[m'
    '\x1b[1;41;37mcolor=white background=red, underline=False\x1b[m'
    '\x1b[4m\x1b[1;40;37mcolor=white background=black, underline=True\x1b[m'
    '\x1b[1;40;37mcolor=white background=black, underline=False\x1b[m'
    '\x1b[4m\x1b[1;44;31mcolor=red background=blue, underline=True\x1b[m'
    '\x1b[1;44;31mcolor=red background=blue, underline=False\x1b[m'
    '\x1b[4m\x1b[1;47;31mcolor=red background=gray, underline=True\x1b[m'
    '\x1b[1;47;31mcolor=red background=gray, underline=False\x1b[m'
    '\x1b[4m\x1b[1;42;31mcolor=red background=dark_green, underline=True\x1b[m'
    '\x1b[1;42;31mcolor=red background=dark_green, underline=False\x1b[m'
    '\x1b[4m\x1b[1;44;31mcolor=red background=dark_blue, underline=True\x1b[m'
    '\x1b[1;44;31mcolor=red background=dark_blue, underline=False\x1b[m'
    '\x1b[4m\x1b[1;45;31mcolor=red background=dark_magenta, underline=True\x1b[m'
    '\x1b[1;45;31mcolor=red background=dark_magenta, underline=False\x1b[m'
    '\x1b[4m\x1b[1;40;31mcolor=red background=dark_gray, underline=True\x1b[m'
    '\x1b[1;40;31mcolor=red background=dark_gray, underline=False\x1b[m'
    '\x1b[4m\x1b[1;43;31mcolor=red background=dark_yellow, underline=True\x1b[m'
    '\x1b[1;43;31mcolor=red background=dark_yellow, underline=False\x1b[m'
    '\x1b[4m\x1b[1;43;31mcolor=red background=yellow, underline=True\x1b[m'
    '\x1b[1;43;31mcolor=red background=yellow, underline=False\x1b[m'
    '\x1b[4m\x1b[1;40;31mcolor=red background=dark_dark_gray, underline=True\x1b[m'
    '\x1b[1;40;31mcolor=red background=dark_dark_gray, underline=False\x1b[m'
    '\x1b[4m\x1b[1;41;31mcolor=red background=dark_red, underline=True\x1b[m'
    '\x1b[1;41;31mcolor=red background=dark_red, underline=False\x1b[m'
    '\x1b[4m\x1b[1;47;31mcolor=red background=dark_white, underline=True\x1b[m'
    '\x1b[1;47;31mcolor=red background=dark_white, underline=False\x1b[m'
    '\x1b[4m\x1b[1;42;31mcolor=red background=green, underline=True\x1b[m'
    '\x1b[1;42;31mcolor=red background=green, underline=False\x1b[m'
    '\x1b[4m\x1b[1;46;31mcolor=red background=cyan, underline=True\x1b[m'
    '\x1b[1;46;31mcolor=red background=cyan, underline=False\x1b[m'
    '\x1b[4m\x1b[1;45;31mcolor=red background=magenta, underline=True\x1b[m'
    '\x1b[1;45;31mcolor=red background=magenta, underline=False\x1b[m'
    '\x1b[4m\x1b[1;46;31mcolor=red background=dark_cyan, underline=True\x1b[m'
    '\x1b[1;46;31mcolor=red background=dark_cyan, underline=False\x1b[m'
    '\x1b[4m\x1b[1;47;31mcolor=red background=white, underline=True\x1b[m'
    '\x1b[1;47;31mcolor=red background=white, underline=False\x1b[m'
    '\x1b[4m\x1b[1;41;31mcolor=red background=red, underline=True\x1b[m'
    '\x1b[1;41;31mcolor=red background=red, underline=False\x1b[m'
    '\x1b[4m\x1b[1;40;31mcolor=red background=black, underline=True\x1b[m'
    '\x1b[1;40;31mcolor=red background=black, underline=False\x1b[m'
    '\x1b[4;44;30mcolor=black background=blue, underline=True\x1b[m'
    '\x1b[0;44;30mcolor=black background=blue, underline=False\x1b[m'
    '\x1b[4;47;30mcolor=black background=gray, underline=True\x1b[m'
    '\x1b[0;47;30mcolor=black background=gray, underline=False\x1b[m'
    '\x1b[4;42;30mcolor=black background=dark_green, underline=True\x1b[m'
    '\x1b[0;42;30mcolor=black background=dark_green, underline=False\x1b[m'
    '\x1b[4;44;30mcolor=black background=dark_blue, underline=True\x1b[m'
    '\x1b[0;44;30mcolor=black background=dark_blue, underline=False\x1b[m'
    '\x1b[4;45;30mcolor=black background=dark_magenta, underline=True\x1b[m'
    '\x1b[0;45;30mcolor=black background=dark_magenta, underline=False\x1b[m'
    '\x1b[4;40;30mcolor=black background=dark_gray, underline=True\x1b[m'
    '\x1b[0;40;30mcolor=black background=dark_gray, underline=False\x1b[m'
    '\x1b[4;43;30mcolor=black background=dark_yellow, underline=True\x1b[m'
    '\x1b[0;43;30mcolor=black background=dark_yellow, underline=False\x1b[m'
    '\x1b[4;43;30mcolor=black background=yellow, underline=True\x1b[m'
    '\x1b[0;43;30mcolor=black background=yellow, underline=False\x1b[m'
    '\x1b[4;40;30mcolor=black background=dark_dark_gray, underline=True\x1b[m'
    '\x1b[0;40;30mcolor=black background=dark_dark_gray, underline=False\x1b[m'
    '\x1b[4;41;30mcolor=black background=dark_red, underline=True\x1b[m'
    '\x1b[0;41;30mcolor=black background=dark_red, underline=False\x1b[m'
    '\x1b[4;47;30mcolor=black background=dark_white, underline=True\x1b[m'
    '\x1b[0;47;30mcolor=black background=dark_white, underline=False\x1b[m'
    '\x1b[4;42;30mcolor=black background=green, underline=True\x1b[m'
    '\x1b[0;42;30mcolor=black background=green, underline=False\x1b[m'
    '\x1b[4;46;30mcolor=black background=cyan, underline=True\x1b[m'
    '\x1b[0;46;30mcolor=black background=cyan, underline=False\x1b[m'
    '\x1b[4;45;30mcolor=black background=magenta, underline=True\x1b[m'
    '\x1b[0;45;30mcolor=black background=magenta, underline=False\x1b[m'
    '\x1b[4;46;30mcolor=black background=dark_cyan, underline=True\x1b[m'
    '\x1b[0;46;30mcolor=black background=dark_cyan, underline=False\x1b[m'
    '\x1b[4;47;30mcolor=black background=white, underline=True\x1b[m'
    '\x1b[0;47;30mcolor=black background=white, underline=False\x1b[m'
    '\x1b[4;41;30mcolor=black background=red, underline=True\x1b[m'
    '\x1b[0;41;30mcolor=black background=red, underline=False\x1b[m'
    '\x1b[4;40;30mcolor=black background=black, underline=True\x1b[m'
    '\x1b[0;40;30mcolor=black background=black, underline=False\x1b[m'
    """
    for color in COLORS.iterkeys():
        for bg in COLORS.iterkeys():            
            for ul in (True, False):
                yield stylify('color=%s background=%s, underline=%s' % (color, bg, ul), color, bg, ul)

if __name__ == "__main__":
    for line in test():
        print line
    import logging

    h = logging.StreamHandler()
    f = StylingFormatter(datefmt='%Y%m%d-%H%M%S', fmt='%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
    l = logging.getLogger()

    h.setLevel(0)
    l.setLevel(0)

    h.setFormatter(f)
    l.addHandler(h)
    l.log(5, 'level 5')
    l.debug('debug')
    l.info('info')
    l.warning('warning')
    l.log(35, 'level 35')
    l.error('error')
    l.critical('critical')
