"""
Common utility functions for console based UI.
"""
import os
import subprocess
from xiv.dtypes import Bunch

DIALOG_EXIT_CODES = Bunch(
        ERROR = -1,
        OK_PRESSED = 0,
        CANCEL_PRESSED = 1,
        HELP_PRESSED  =  2,
        EXTRA_PRESSED = 3,
        )

ESCAPE = Bunch(
               black='0',
               red='1',
               green='2',
               yellow='3',
               blue='4',
               magenta='5',
               cyan='6',
               white='7',
               bold='b',
               unbold='B',
               reverse='r',
               unreverse='R',
               underline='u',
               no_underline='U',
               reset='n',
              )
ESCAPE=Bunch((name, '\Z'+code) for name, code in ESCAPE.items())

def _prepr(obj):
    "'patched' repr: returns the representation of an object without escaping backslashes"
    return repr(obj).replace('\\\\', '\\')

def verify_dialog_exists():
    """
    This module relies on the 'dialog' binary, this function verifes it can be found in the path.
    """
    try:
        for subpath in os.environ['PATH'].split(':'):
            if os.access(os.path.join(subpath, 'dialog'), os.X_OK):
                return True
    except KeyError:
        pass
    raise EnvironmentError("'dialog' binary not found in PATH")

if 'XIV_MAKE_SUER_DIALOG_IS_IN_PATH' in os.environ:
    try:
        verify_dialog_exists()
    except EnvironmentError , e:
        raise ImportError(str(e))


def ask(question, default=True):
    """
    Asks a question and awaits user input, either true or false. This function /does not/ rely on 'dialog'.
    >>> ui.ask('Do you like finding screws?')
    Do you like finding screws? [Y/n] n
    False
    >>> 
    """
    if not question.endswith("?"):
        question = question + "?"
    if default:
        options = "[Y/n]"
    else:
        options = "[y/N]"
    answer = raw_input("%s %s " % (question, options))
    if not answer.strip():
        return default
    else:
        return answer.lower() in ("y", "yes")
    
class DialogCanceled(Exception):
    """
    Raised when the user chooses to cancel a displayed dialogue.
    """

class DialogExtraButtonPressed(Exception):
    """
    Raised when the user chooses the 'extra' button in  a displayed dialogue.
    """

class DialogError(Exception):
    """
    Raised when dialog internal error occurs

    """

def _dialog(command, command_options,*args, **kwargs):
    """
    Internal function for calling the dialog binary.
    """
    clear = kwargs.pop('clear', True)
    if kwargs:
        raise TypeError('_dialog() got an unexpected keyword argument %s' % kwargs.keys()[0])
    if not command_options:
        command_options = {}
    p = subprocess.Popen(
        "dialog --colors %s --%s %s" % (" ".join(map("%s %s".__mod__, command_options.iteritems())), command, " ".join(map(str, args)), ),
         shell=True,
         close_fds=True, 
         stderr=subprocess.PIPE
    )
    p.wait()
    if clear:
        clear_screen()
    return p.returncode, p.stderr.read()

class progress_bar(object):
    """
    General purpose class for displaying a dialog box.
    >>> def progress():
    ...     P = ui.progress_bar('Looking for screws...', total_steps=10)
    ...     for i in range(0,10):
    ...          P.next_step()
    ...          time.sleep(0.5)
    """
    def __init__(self, prompt, percentage=0, total_steps=None):
        self.total_steps = total_steps
        self.step = 0
        self.percentage = percentage
        self.p = subprocess.Popen("dialog --gauge %r 6 60 %d" % (prompt, percentage), shell=True, close_fds=True, stdin=subprocess.PIPE)
        self.closed = False
    def close(self):
        if self.closed:
            return
        self.p.stdin.close()
        self.p.wait()
        self.closed = True
        os.system("clear")
    def set_prompt(self, text):
        self.p.stdin.write("XXX\n%s\nXXX\n" % text)
    def next_step(self):
        assert self.total_steps is not None
        self.step += 1
        self.set_percent((float(self.step) / self.total_steps) * 100)
    def set_percent(self, percentage):
        self.percentage = percentage
        self.p.stdin.write("%d\n" % percentage)
        
def choice(options, title, selected=None):
    """
    Present the user with multiple options.
    >>> ui.choice(['nail', 'nail', 'nail', 'screw', 'nail'], "Can you find the screw, Jacob?")
    """
    args = []
    for index, option in enumerate(options):
        args.extend((option, '""', index == selected and "on" or "off"))    
    return_code, output = _dialog("radiolist", {}, _prepr(title), 0, 0, 0, *args)
    if return_code == DIALOG_EXIT_CODES.CANCEL_PRESSED :
        raise DialogCanceled()
    return output

def menu(options, title="title", header="menu",common_options=None):
    """
    Presents the user with a menu of items; the menu can include a title, a header (a description of what should be 
     chosen in the menu) and a set of key/description pairs for the items.
    >>> ui.menu(
    ...     options={'phillips': 'a screw with a plus-like head', 'flat': 'a simple screw'},
    ...     title='Screw Type',
    ...     header='What kind of screws are you trying to find? Pick the one that will match best...',
    ... )
    """

    if type(options) in (list, tuple):
        options = dict((str(i), k) for i, k in enumerate(options))
    return_code, output = _dialog("menu" , {"--title":_prepr(title)}, "'%s'" % header, 0, 0, 0, *map("'%s' '%s'".__mod__, sorted(options.iteritems())))
    if return_code == DIALOG_EXIT_CODES.CANCEL_PRESSED :
        raise DialogCanceled()
    elif return_code == DIALOG_EXIT_CODES.EXTRA_PRESSED :
        raise DialogExtraButtonPressed()
    elif return_code == DIALOG_EXIT_CODES.ERROR:
        raise DialogError()
    try:
        return int(output.strip())
    except ValueError:
        return output.strip()
    
def checklist(options, header = "header", title="title"):
    """
    Displays a list of checkboxes with default values and description, returns an identical list with user values.
    'options' is a list of (description, boolean), which is used to initialize the checklist.
    >>> ui.checklist(
    ...     options=[('jacob', True), ('yaniv', False), ('kariel', True)], 
    ...     header='Who are the screw finders?', 
    ...     title='Screw Finder List'
    )
    """
    return_code, output = _dialog("checklist", {"--title":_prepr(title)}, _prepr(header), 0, 0, 0, *["%d %s %s" % (i, k, v and "on" or "off") for i, (k, v) in enumerate(options)])
    if return_code == DIALOG_EXIT_CODES.CANCEL_PRESSED :
        raise DialogCanceled()
    on = map(int, output.replace('"', "").split())
    returned = [(options[index][0], index in on) for index , unused in enumerate(options)]
    return returned
        
def messagebox(title, message, height=0, width=0):
    """
    Displayes a message in a messagebox, contrary to a simple ask() (see above)
    >>> messagebox("Call For Arms", "Screw finders of the world, unite!")
    """
    _dialog("msgbox", {"--title":_prepr(title)}, _prepr(message), height, width)

def infobox(title, message, height=0, width=0):
    """
    Displayes a message without waiting for user input (terminates immediately, doesn't clear screen).
    >>> infobox("Call For Barns", "Foo bar baz qux, quux!")
    """
    _dialog("infobox", {"--title":_prepr(title)}, _prepr(message), height, width, clear=False)

def input(message, init = ""):
    """
    >>> input("What are you?", init='Screw finder!!!')
    """
    return_code, error = _dialog("inputbox", None, _prepr(message), 0, 0, init)
    if return_code == DIALOG_EXIT_CODES.CANCEL_PRESSED :
        raise DialogCanceled()
    return error

def yesno(message, title="", height=0, width=0):
    """
    Displays a yes/no dialog, returns boolean.
    >>> yesno("are you a screw finder?", title="Verify compliance:)
    """
    return_code, error = _dialog("yesno", {"--title":_prepr(title)}, _prepr(message), height, width)
    return return_code == 0

def clear_screen():
    os.system("clear")

def screen_dimensions():
    crap, err = subprocess.Popen(('dialog', '--print-maxsize'), stderr=subprocess.PIPE).communicate()
    crap, height, width = err.strip().replace(',', '').split()
    return int(height), int(width)
