import itertools
import keyword

class Bunch (dict): 
    """
    Generic container for attributes

    >>> a = Bunch(a = 2, s = "test_string")
    >>> a.a
    2
    >>> a.s == "test_string"
    True
    >>> len([x for x in a.keys()]) == 2
    True
    >>> a = Bunch(a=2, b=3)
    >>> b = Bunch(a=2, b=3)
    >>> a == b
    True
    >>> a.j = 33
    >>> hasattr(a, "j")
    True
    >>> a.nonexistent_attr #doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
       ...
    AttributeError: ...
    >>> a["nonexistent_attr"] #doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
       ...
    KeyError: ...
    >>> a == b
    False
    >>> b.j = 6
    >>> del b.j
    >>> hasattr(b, "j")
    False
    >>> del b.nonexistent_attr #doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
       ...
    AttributeError: ...
    >>> import copy
    >>> import pickle
    >>> b = copy.deepcopy(a) # Bunchs should be copyable
    >>> b == a
    True
    >>> import pickle #Bunchs should be picklable
    >>> temp = pickle.dumps(a)
    >>> a2 = pickle.loads(temp)
    >>> a2 == a
    True
    >>> a2 != a # != should work as well!
    False
    
    after unpickling, Bunch behavior should still
    be in effect:
    
    >>> a2.update({'new_attr' : 666})
    >>> hasattr(a2, 'new_attr')
    True
    >>> a2 is a
    False
    """

    def __init__(self, *attr_dicts, **attrs):
        for attr_dict in attr_dicts:
            self.update(attr_dict)
        self.update(attrs)
    
    def __getstate__(self):
        return self
        
    def __setstate__(self, d):
        self.update(d)
        
    def __repr__ (self):
	'''
	repr for bunch
	if all keys are valid variable names,
	show in keyword style,
	otherwise in dict style
	'''
	for key in self:
	    if not isinstance(key, str) or not ''.join(key.split('_')).isalnum() or not (
		key[0]=='_' or key[0].isalpha) or keyword.iskeyword(key):
		return type(self).__name__ + dict.__repr__(self)
	return type(self).__name__ + '('+', '.join('%s=%r'%it for it in self.iteritems())+')'
    
    def __setitem__(self, attr, val):
        self.__dict__[attr] = val
        return dict.__setitem__(self, attr, val)
    
    def __getattr__(self, attr):
        try:
            return self.__getitem__(attr)
        except KeyError:
            raise AttributeError("Bunch has no attribute %r" % (attr, ))

    def __setattr__(self, attr, value):
        return self.__setitem__(attr, value)
    
    def __delattr__(self, attr):
        super(Bunch, self).__delattr__(attr)
        #assumption - no attributes exist which are not keys...
        assert attr in self
        return self.__delitem__(attr)
    
    def update(self, src = None, **kwargs):
        if src is None:
            src = ()
        if isinstance(src, dict):
            src = src.iteritems()
        for key, val in itertools.chain(src, kwargs.iteritems()):
            self[key] = val
if __name__ == "__main__":
    import doctest
    doctest.testmod()
