class Symdict(dict):
    """Symdict is a dictionary in which keys point to values,
    and the values point back to the keys.
    This is done by keeping another dictionary to represent the
    reversed assignment.
    Symdict behaves just like an oridnary dictionary -- after
    assigning values to keys, the value can be obtained using
    the original keys. To get the reversed mapping, a seperate
    method named valueToKey should be used (to be convention
    friendly, value_to_key is also supported).
    Note that Symdict assures that it is consistent, meaning
    that in case a user tries to insert a new key with an
    existing value, ValueError will be raised.
    Also note that Symdict implicitly requires that your values
    will be immutable.

    >>> sd = Symdict()
    >>> sd[1] = 'Sunday'
    >>> sd[2] = 'Monday'
    >>> sd[3] = 'Tuesday'
    >>> sd
    Symdict({1: 'Sunday', 2: 'Monday', 3: 'Tuesday'})
    >>> sd.valueToKey('Sunday')
    1
    >>> sd[2]
    'Monday'
    >>> import pickle
    >>> p = pickle.dumps(sd)
    >>> pickle.loads(p)
    Symdict({1: 'Sunday', 2: 'Monday', 3: 'Tuesday'})

    >>> # Try adding an existing value
    >>> sd[4] = 'Tuesday'
    Traceback (most recent call last):
    ...
    ValueError: Cannot update key 4: will make value 'Tuesday' duplicate

    >>> # Make sure that a simply overload works
    >>> sd[3] = 'Tuesday'
    >>> sd[2] = 'Wednsday'

    >>> # Now make sure that we can't fetch 'Monday'
    >>> sd.valueToKey('Monday')
    Traceback (most recent call last):
    ...
    KeyError: 'Monday'
    """
    def __init__(self, *dicts, **attrs):
        self.reversed_dict = {}
        for newdict in dicts:
            self.update(newdict)
        self.update(attrs)

    def __repr__(self):
        return '%s(%s)' % (type(self).__name__, repr(dict(self)))

    def __setitem__(self, item, value):
        if (value in self.reversed_dict) and (item != self.reversed_dict[value]):
            raise ValueError("Cannot update key %r: will make value %r duplicate" % (item, value))
        try:
            oldvalue = self[item]
            del self.reversed_dict[oldvalue]
        except KeyError:
            pass
        dict.__setitem__(self, item, value)
        self.reversed_dict[value] = item

    def __delitem__(self, item):
        del self.reversed_dict[self[item]]
        dict.__delitem__(self, item)

    def valueToKey(self, value):
        return self.reversed_dict[value]

    value_to_key = valueToKey
