from xiv.pyutils import predicates
from xiv.dtypes.primitives import NOTHING

class ValueRange(predicates.Predicate):
    """
    >>> ValueRange(gt=2, lt=4)
    <(2 < X < 4)?>
    >>> ValueRange(ge=5)
    <(5 <= X)?>
    >>> ValueRange(ge=5, le=6)
    <(5 <= X <= 6)?>
    >>> x = ValueRange(ge=5, le=6)
    >>> x == 4
    False
    >>> x == 5
    True
    >>> x == 6.5
    False
    """
    def __init__(self, gt=NOTHING, lt=NOTHING, ge=NOTHING, le=NOTHING):
        super(ValueRange, self).__init__()
        assert (gt is NOTHING or ge is NOTHING)
        assert (lt is NOTHING or le is NOTHING)
        self.minInclusive = self.maxInclusive = False
        self.min = self.max = None
        if lt is not NOTHING:
            self.max = lt
        if le is not NOTHING:
            self.max = le
            self.maxInclusive = True
        if gt is not NOTHING:
            self.min = gt
        if ge is not NOTHING:
            self.min = ge
            self.minInclusive = True
    def __and__(self, other):
        """overrides predicates.Predicates.__and__"""
        if not isinstance(other, ValueRange):
            assert self.contains(other)
            return other
        raise NotImplementedError()

    def pickValue(self):
        from xiv.dtypes import units
        assert set([type(self.min), type(self.max)]) <= set([type(None), units.DataSize, int, long])
        if self.min is not None:
            returned = self.min * 2
            if self.max is not None:                
                returned = min(self.max, returned)
        else:            
            returned = self.max / 2
            if self.min is not None:
                returned = max(self.min, returned)
        if not self.contains(returned):
            raise NotImplementedError("pickValue is currently a very simple implementation")
        return returned
      
    
    def contains(self, value):
        if isinstance(value, ValueRange):
            raise NotImplementedError()
        if isinstance(value, predicates.Equality):
            value = value.value
        if self.min is not None:
            if value < self.min:
                return False
            if not self.minInclusive and value == self.min:
                return False
        if self.max is not None:
            if value > self.max:
                return False
            if not self.maxInclusive and value == self.max:
                return False
        return True
    applies = contains
    def _describe(self, variable):
        returned = "("
        if self.min is not None:
            if self.minInclusive:
                operator = "<="
            else:
                operator = "<"
            returned += "%s %s " % (self.min, operator)
        returned += variable
        if self.max is not None:
            if self.maxInclusive:
                operator = "<="
            else:
                operator = "<"
            returned += " %s %s" % (operator, self.max)
        returned += ")"
        return returned

