from xiv.dtypes import Bunch, ValueRange
import types
import copy

class Version(object):
    """
    A class representing the version of a Nextra system.
    """
    VERSION_MAGNITUDE_ELEMENTS = (
        ('major', types.IntType, None),
        ('minor', types.IntType, None),
        ('micro', types.IntType, None),
        ('revision', types.StringType, lambda(x): x.isalpha()),
    )
    VERSION_MAGNITUDE_NAMES = zip(*VERSION_MAGNITUDE_ELEMENTS)[0]
    def __init__(self, version, allow_major_zero=False):
        super(Version, self).__init__()
        self.allow_major_zero = allow_major_zero
        if isinstance(version, Version):
            self._version = copy.deepcopy(version._version)
        else:
            self.version = version

    def set_version(self, version_element):
        _version = Bunch()
        version_parts = str(version_element).split('.')
        version_parts.reverse()
        for part_name, part_type, filter in Version.VERSION_MAGNITUDE_ELEMENTS:
            try:
                current_part = version_parts.pop()
                current_part = part_type(current_part)
                if filter is not None:
                    if not filter(current_part):
                        raise ValueError
                _version[part_name] = current_part
            except IndexError:
                _version[part_name] = part_type('0')
            except ValueError:
                _version[part_name] = part_type('0')
                version_parts.append(current_part)
        # invalidate things like '9.a.a' or '91a'
        if len(version_parts) != 0:
            raise ValueError('encountered unparseable component: %s' % version_parts.pop())
        if _version['major'] == 0 and not self._allow_major_zero:
            raise ValueError('the major component of a version must not be zero')
        self._version = _version
    def get_version(self):
        return self._version
    version = property(get_version, set_version)

    def __getattr__(self, value):
        if value in Version.VERSION_MAGNITUDE_NAMES:
            # value is one of the known magnitudes
            return self._version[value]
        elif hasattr(self, value):
            return self.__dict__[value]
        else:
            raise AttributeError("%s has no attribute %s" % (self.__class__, value))

    def __str__(self):
        return '%(major)s.%(minor)s.%(micro)s.%(revision)s' % self._version

    def __repr__(self):
        return '(version ' + self.__str__() + ')'

    def __cmp__(self, other):
        other = Version(other)
        for value in Version.VERSION_MAGNITUDE_NAMES:
            relation = cmp(self._version[value], other._version[value])
            if relation != 0:
                return relation
        return 0
