"""
Human-readable representation of data sizes. Also allows
easy translation to different measurement units.

#-- *iB units -----------------------------------#
>>> bit, 8*bit, 16*bit, 48*bit, 49*bit, 128*bit
(bit, byte, 2*bytes, 6*bytes, 49*bits, 16*bytes)
>>> 1024*byte, 2048*byte
(KiB, 2*KiB)
>>> 1024*KiB, 2048*KiB
(MiB, 2*MiB)
>>> 1024*MiB, 2048*MiB
(GiB, 2*GiB)
>>> 1024*GiB, 2048*GiB
(TiB, 2*TiB)
>>> 1024*TiB, 2048*TiB
(PiB, 2*PiB)
>>> 1024*PiB, 2048*PiB
(XiB, 2*XiB)
>>> 1024*XiB, 4096*XiB
(1024*XiB, 4096*XiB)

#-- Decimal units -------------------------------#
>>> 1000*decimal.KB
1000000*bytes
>>> 1024*decimal.KB
1000*KiB
>>> 4000*decimal.GB
3906250000*KiB
>>> 4096*decimal.TB
3906250000*MiB
>>> 4000*decimal.TB
3906250000000*KiB
"""

from xiv.dtypes import Bunch

import itertools
import random

class DataSize(object):
    """
    >>> DataSize(1024, 'bits')
    128*bytes
    >>> _ * 5000
    625*KiB
    >>> _ / 5
    125*KiB
    >>> _ * 3000000
    375000000*KiB
    >>> _ // 1024
    375000000*bytes
    >>> _ * 1024 * 3000000
    1098632812500*MiB
    >>> _ // KiB
    1125000000000000
    >>> _ * TiB
    1098632812500*PiB
    >>> _._bits == (1125000000000000 * (2 ** 40) * 8)
    True
    """
    def __init__(self, num_bits, sanity):
        #don't let anyone use this constructor by mistake
        assert sanity == 'bits'
        self._bits = num_bits
    def __hash__(self):
        return hash(('data_size', self._bits))
    def _typecheck(self, other, action, relation='to', self_at_left=True):
        if self_at_left:
            a, b = self, other
        else:
            a, b = other, self
        if 0 != other and not isinstance(other, DataSize):
            raise ArithmeticError, 'units mismatch, %s %r %s %r' % (action, a, relation, b)
    def __cmp__(self, other):
        if 0 == other:
            return cmp(self._bits, 0)
        elif isinstance(other, DataSize):
            return cmp(self._bits, other._bits)
        else:
            return cmp((type(self), self._bits), other)
    # MATH OPERATIONS
    def __nonzero__(self):
        return self != 0
    def __abs__(self):
        return DataSize(abs(self._bits), 'bits')
    def __add__(self, other):
        self._typecheck(other, 'adding')
        if 0 == other:
            return self
        return DataSize(self._bits+other._bits, 'bits')
    __radd__ = __add__
    def __sub__(self, other):
        self._typecheck(other, 'substructing')
        if 0 == other:
            return self
        return DataSize(self._bits-other._bits, 'bits')
    def __rsub__(self, other):
        return other + (-self)
    def __mul__(self, mul):
        if isinstance(mul, DataSize):
            return TypeError, 'multiplying %s by %s' % (self, mul)
        return DataSize(int(self._bits*mul), 'bits')
    __rmul__ = __mul__
    def __floordiv__(self, div):
        if isinstance(div, DataSize):
            return self._bits//div._bits
        else:
            return DataSize(self._bits//div, 'bits')
    def __div__(self, div):
        if isinstance(div, DataSize):
            res = self._bits*1./div._bits
            if 0 == res % 1:
                return int(res)
            else:
                return res
        else:
            return DataSize(self._bits*1./div, 'bits')
    def __mod__(self, div):
        self._typecheck(div, 'performing modulu of', 'by')
        returned = DataSize(self._bits%div._bits, 'bits')
        if returned._bits == 0:
            return 0
        return returned
    def __rmod__(self, number):
        if number != 0:
            raise ArithmeticError("Cannot perform modulu of %s by %s" % (number, self))
        return 0
    def __rdiv__(self, size):
        if 0 == size:
            return 0
        self._typecheck(size, 'dividing', 'by', self_at_left=False)
        # with Rpyc left size of division can be remote datasize and it will call __rdiv__
        return size.__div__(self)
    __rfloordiv__ = __rdiv__
    def __neg__(self):
        return -1*self
    # ROUNDING FUNCTIONS
    def rounddown(self, unit):
        return (self // unit) * unit
    def roundup(self, unit):
        if 0 == self % unit:
            return self
        else:
            return (self // unit + 1) * unit
    def round(self, unit):
        if self % unit > unit / 2:
            add = unit
        else:
            add = 0
        return self.rounddown(unit) + add
    def ceildiv(self, other):
        return self.roundup(other) / other
    # REPRESENTATIONS
    def __str__(self):
        if 0 == self:
            return '0'
        if self in std_units_names:
            return '1 %s' % (std_units_names[self],)
        for unit in std_units_from_largest:
            name = std_units_names[unit]
            if name.islower():
                many = name+'s'
            else:
                many = name
            if 0 == self%unit:
                return '%d %s' % (self/unit, many)
            if self >= unit or (self >= unit/10 and self*10%unit == 0):
                return '%.1f %s' % (self/unit, many)
        return '%r bits' % self._bits
    def __repr__(self):
        if self in std_units_names:
            return std_units_names[self]
        for unit in std_units_from_largest:
            name = std_units_names[unit]
            if name.islower():
                many = name+'s'
            else:
                many = name
            if 0 == self%unit:
                return '%d*%s' % (self/unit, many)
        return '%r bits' % self._bits
    # USEFUL FUNCTIONS
    def randrange(self, divisor, start=None):
        'returns a random size below self dividing by divisor'
        if start is None:
            start = 0 * byte
        start = start.ceildiv(divisor)
        stop  = self.ceildiv(divisor)
        return random.randrange(start=start, stop=stop) * divisor
    

std_units_by_name = Bunch()
bit = std_units_by_name.bit = DataSize(1, 'bits')
for i, unit_name in enumerate(['byte', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'XiB']):
    std_units_by_name[unit_name] = DataSize(8 * (1024 ** i), 'bits')

std_units_names = dict((value, key) for key, value in std_units_by_name.iteritems())
std_units_from_largest = sorted(std_units_by_name.values(), reverse=True)
globals().update(std_units_by_name)

LogicalBlockInBytes = 512
LogicalBlock = LogicalBlockInBytes * byte

decimal = Bunch((unit_name, 10**((i+1)*3)*byte) for i, unit_name in enumerate(
    ['KB', 'MB', 'GB', 'TB', 'PB', 'XB']))

std_units_by_char = Bunch((key[:1], value) for key, value in std_units_by_name.iteritems())
std_units_by_char.b = std_units_by_name.bit
std_units_by_char.B = std_units_by_name.byte

def fromstr(text):
    unit_text, num_text = [''.join(func(str.isdigit, text))
                           for func in [itertools.dropwhile, itertools.takewhile]]
    try:
        if 1 == len(unit_text):
            unit = std_units_by_char[unit_text]
        else:
            unit = std_units_by_name[unit_text]
    except KeyError:
        raise ValueError, '%r not parsable as size - bad unit %r' % (text, unit_text)
    if '' == num_text:
        return unit
    num = int(num_text)
    return num * unit

def data_size_str(size):
    '''
    return a representation for data size.
    uses storage industry binary units. (KiB = 1024)
    '''
    
    for unit, name in [(2**60, 'XiB'), (2**50, 'PiB'), (2**40, 'TiB'),
                       (2**30, 'GiB'), (2**20, 'MiB'), (2**10, 'KiB'), (1, 'Bytes')]:
        if size >= unit:
            s = '%.1f' % (size*1./unit, )
        elif size >= unit/10:
            s = '%.2f' % (size*1./unit, )
        else:
            continue
        s = s.rstrip('0').rstrip('.')
        return '%s %s' % (s, name)
    assert size == 0
    return '0'
