"""
Command runner objects take care of the gory details for running processes,
obtaining their output and result, and managing pipes for interacting
with them.

Each command runner is parameterized with a ResultVerifier object,
which abstracts the verification process (from checking result==0 to
complex output checks)

Commanr runner factories are Abstract Factories for producing command
runner objects (decoupling their creation and parameterization).

>>> c = LocalCommandRunner("/bin/echo", SimpleResultVerifier())
>>> result, output, stderr = c.execute("hello there!")
>>> result
0
>>> output.startswith('hello there!')
True
>>> result, output, stderr = c.BLA("-t", 50, "-f", 60)
>>> result
0
>>> output.startswith("BLA -t 50 -f 60")
True
>>> c = LocalCommandRunner("/bin/test", DoNothingVerifier())
>>> result, output, stderr = c.execute('-f nonexistent_filename')
>>> result != 0
True
>>> c = LocalCommandRunner("/bin/test", SimpleResultVerifier(result))
>>> result, output, stderr = c.execute('-f nonexistent_filename')
>>> result != 0
True
>>> c = LocalCommandRunner("/bin/test", SimpleResultVerifier())
>>> result, output, stderr = c.execute('-f nonexistent_filename') #doctest: +IGNORE_EXCEPTION_DETAIL
Traceback (most recent call last):
    ...
CommandFailureException: ...
>>> c = LocalCommandRunner("/bin/echo", SimpleResultVerifier())
>>> result, output, stderr = c.executeStr("hello there")
>>> result == 0
True
>>> output
'hello there\\n'
>>> c = LocalCommandRunner("/bin/bash")
>>> result, output, stderr = c.execute('-c', "echo hello")
>>> output.startswith("hello")
True
>>> c = LocalCommandRunner()
>>> result, output, stderr = c.executeStr("grep ttt", "first\\nsecond\\nthirdttt\\nfourth")
>>> output
'thirdttt\\n'
>>> result, output, stderr = c.grep("ttt", stdin="first\\nsecond\\nthirdttt\\nfourth")
>>> output
'thirdttt\\n'
>>> a = c.executeStr("ls -la /", async=True)
>>> time.sleep(10)
>>> result, output, stderr = a.poll()
>>> result == 0 and len(output) > 0
True
>>> c.executeStr("sleep 120", timeout=10) #doctest: +IGNORE_EXCEPTION_DETAIL
Traceback (most recent call last):
  ...
CommandTimeout: ...
"""
import subprocess
import select
import time
import socket
import os
import cStringIO as StringIO

from xiv import logger

def _quote(s):
    """
    quotes a string with double quotes ("), while escaping
    internal double quotes with a backslash

    >>> print _quote('this is a " string')
    "this is a \\" string"
    >>> print _quote('this is a $shell variable')
    "this is a \\$shell variable"
    >>> print _quote(r'an escaped \\$')
    "an escaped \\\\\\$"
    """
    return '"' + s.replace("\\", "\\\\").replace('"', '\\"').replace("$", "\\$") + '"'

def _BACKGROUND_JOB(cmd_str):
    """
    translates a 'normal' command string, to a form which will
    run itself in the background, printing its pid.

    >>> _BACKGROUND_JOB('echo "hello world"')
    'FN=`mktemp /tmp/r.XXXXXX` && echo "echo \\\\"hello world\\\\" &" >> $FN && echo echo $\\\\! >> $FN && bash $FN && rm $FN'
    """
    return 'FN=`mktemp /tmp/r.XXXXXX` && echo %s >> $FN && echo echo $\! >> $FN && bash $FN && rm $FN' % (_quote(cmd_str + " &"),)


def REMOTE_BACKGROUND_PID(runner, cmd):
    """
    runs a comand through runner.executePopen() in the
    background, and returns its pid.
    """
    cmd = _BACKGROUND_JOB(cmd)
    logger.debug("Running background job. cmd=%r", cmd)
                    
    p = runner.executePopen(cmd)
    logger.debug("reading one line from popen...")
    #reat pid
    pid = int(p.stdout.readline().strip())
    logger.debug("pid is %d", pid)
    return pid

#----------------------------#
# Errors and Verifiers       #
#----------------------------#

class CommandRunnerException(Exception):
    pass

class CommandTimeout(CommandRunnerException):    
    def __init__(self, runner, command, timeout):
        CommandRunnerException.__init__(self, runner, command, timeout)
        self.runner = runner
        self.command = command
        self.timeout = timeout
    def __str__(self):
        return "Command %r with %s timed out (>%ds)" % (self.command, self.runner, self.timeout)

class CommandFailureException(CommandRunnerException):
    """
    Raised when a verification of the result/output
    failed
    """
    def __init__(self, command, result, output, stderr, host):
        CommandRunnerException.__init__(self, command, result, output, stderr)
        self.host = host
        self.command = command
        self.result = result
        self.output = output
        self.stderr = stderr
    def __str__(self):
        if self.host not in (None, 'localhost'):
            host_str = " on " + self.host
        else:
            host_str = ""
        return "Command failed%s.\ncmdline: %r\nresult: %d\noutput: %s\nstderr: %s\n" % (host_str, self.command, self.result, self.output, self.stderr)


class ResultVerifier(object):
    """
    result verifier for the command runner.
    """
    def verify(self, host, command, status, output, stderr):
        """
        the verify method is called for a given status and output from a program
        @raise  CommandFailureException: in a case where the verification
            resulted in a false result
        """
        raise NotImplementedError("verify is not implemented for this instance")

class DoNothingVerifier(ResultVerifier):
    """
    null object for verifying result.
    does not verify anything
    """
    def verify(self, *args, **kwargs):
        pass

class SimpleResultVerifier(ResultVerifier):
    """
    verifies that the status equals a given value
    """
    def __init__(self, expectedValue = 0):
        self.expected = expectedValue
    def verify(self, host, command, status, output, stderr):
        if status != self.expected:
            raise CommandFailureException(command,
                                          status,
                                          output,
                                          stderr,
                                          host)

class AsyncResult(object):
    def __init__(self, runner, fullString, popener, resultVerifier, numRetries, verify):
        super(AsyncResult, self).__init__()
        self.popener = popener
        self.popen = popener()
        self.runner = runner
        self.fullString = fullString
        self.resultVerifier = resultVerifier
        self.numRetries = numRetries
        self.performVerification = verify
        
    def verify(self, result, output, stderr):
        if not self.performVerification:
            return True
        try:
            self.resultVerifier.verify(self.runner.host, self.fullString, result, output, stderr)
        except CommandRunnerException, e:
            logger.debug(
                "Verifier threw exception %r while verifying %s. output=RED<<%r>>, stderr=RED<<%r>>",
                e, self.fullString, output, stderr)
            self.numRetries -= 1
            if 0 >= self.numRetries:
                raise
            self.popen = self.popener()
            return False
        return True
        
    def poll(self):
        result = self.popen.poll()
        if result is None:
            return
        output, stderr = self.popen.stdout.read(), self.popen.stderr.read()
        if self.verify(result, output, stderr):
            return result, output, stderr

    def wait(self, timeout = None):
        while True:
            output, stderr = self.communicate_with_timeout(timeout)
            result = self.popen.wait()
            if self.verify(result, output, stderr):
                return result, output, stderr

    def communicate_with_timeout(self, timeout):
        
        read_set = [self.popen.stdout, self.popen.stderr]
        stdout = StringIO.StringIO()
        stderr = StringIO.StringIO()
        stdin = self.popen.stdin
        self.popen.stdin = None
        if stdin:
            stdin.flush()
            stdin.close()

        if timeout is not None:
            end_time = time.time() + timeout
        while read_set:
            if timeout is not None and time.time() > end_time:
                raise CommandTimeout(self.runner, self.fullString, timeout)
            if timeout is None:
                select_timeout = None
            else:
                select_timeout = end_time - time.time()
            try:
                rlist, wlist, unused = select.select(read_set, [], [], select_timeout)
            except KeyboardInterrupt:
                print "KBD interrupt, with stdout", stdout.getvalue(), "stderr", stderr.getvalue()
                raise
            for (handle, buffer) in [(self.popen.stdout, stdout), 
                                        (self.popen.stderr, stderr)]:
                if handle in rlist:
                    data = os.read(handle.fileno(), 1024)
                    if not data:
                        handle.close()
                        read_set.remove(handle)
                    buffer.write(data)
        self.popen.wait()
        return (stdout.getvalue(), stderr.getvalue())

_DEFAULT_TIMEOUT = 5 * 60 #five minutes

#----------------------------#
# Command Runners            #
#----------------------------#

class CommandRunner(object):
    """
    this class is useful for running commands through command line,
    and capturing output.

    All methods of running commands through this object return a tuple of
    (result, output).

    This class can be used in several ways:
        1. calling the execute(*args) method, which runs a specified command
            line and returning output. the command line is obtained through
            concatenating the args given to the function while applying str()
            to them (thus, execute("hello", "there", 20) will result in
            the command line string "hello there 20"). Note that arguments
            which contain spaces go through quoting first.
        2. calling executeStr(s), which runs a string for a command line
            explicitly.
        2. accessing an attribute with a certain name, and calling it with
            *args. this will insert the attribute name as a first argument

    @see: module doctest example for more detatils            

    @ivar _basicRunString: the basic string to run for the given command.    
    """
    def __init__(self, basicRunString="", verifier=SimpleResultVerifier()):
        """
        @param basicRunString: basic string to which future arguments are to be
            concatenated
        @param verifier: an instance of L{ResultVerifier} or its derived class,
            which will be responsible for verifying results from this command
            runner.
        """
        super(CommandRunner, self).__init__()
        self._basicRunString = basicRunString
        self.verifier = verifier
        self.host = "localhost"

    def __getattr__(self, attr):
        if attr.startswith("_"):
            #assume this is not a running request, but a failed attribute request
            raise AttributeError("%s object has no attribute %s" % (type(self).__name__, attr))
        def execute(*args, **kwargs):
            return self.execute(attr, *args, **kwargs)
        return execute

    def getOutput(self, *args, **kwargs):
        if len(args) != 1:
            method = self.execute
        else:
            method = self.executeStr
        return method(*args, **kwargs)[1]

    def executeStr(self, s, stdin = None, async = False, timeout=_DEFAULT_TIMEOUT, numRetries = 1, verify=True):
        """
        similar to execute, only receives a single string of parameters, instead
        of seperate arguments.
        @param s: string to execute
        """
        fullString =  self._basicRunString
        #prevent an unnecessary space in case of an empty
        #basic string..
        if fullString:
            fullString += " "
        fullString += s
        
        def runPopen():
            p = self._runPopen(fullString)
            if stdin is not None:
                p.stdin.write(stdin)
            return p

        result = AsyncResult(self, fullString, runPopen, self.verifier, numRetries, verify)
        
        if async:
            return result
        else:
            return result.wait(timeout)

    def executePopen(self, cmd):
        """
        a low-level api which returns a subprocess.popen object...
        """
        return self._runPopen(self._basicRunString+cmd)
    
    def execute(self, *args, **kwargs):
        """
        executes the given arguments, with addition to the basic string

        @return: (result, output) tuple        
        """
            
        args = list(args)
        for index, arg in enumerate(args):
            if " " in str(arg):
                args[index] = _quote(arg)
        return self.executeStr(" ".join(map(str, args)), **kwargs)
        

    def _runPopen(self, commandString, stdin = None):
        raise NotImplementedError("_runPopen not implemented")



class LocalCommandRunner(CommandRunner):
    """
    A command runner for running commands on the local host
    """
    def __init__(self,
                 basicRunString = "",
                 verifier=SimpleResultVerifier()):
        super(LocalCommandRunner, self).__init__(basicRunString, verifier)
        self.isLocal = True
        

    def _runPopen(self, commandString):
        logger.debug("executing %r", commandString)
        return subprocess.Popen(commandString,
                                preexec_fn=os.setsid,
                                shell=True,
                                close_fds = True,
                                stdin = subprocess.PIPE,
                                stdout = subprocess.PIPE,
                                stderr = subprocess.PIPE)
    
        
    def __repr__(self):
        return "<Local command runner>"

class RemoteCommandRunner(CommandRunner):
    """
    >>> r = RemoteCommandRunner("pro1", "", SimpleResultVerifier())
    >>> r.echo("hello")[:2]
    (0, 'hello\\n')
    """
    def __init__(self,
                    host,
                    basicRunString="",
                    verifier=SimpleResultVerifier(),
                    baseCommandRunnerFactory = None,
                    baseCommandRunner = None,
                    user   = None,
                    sshkey = None):
        """
        initializes a remote command runner.
        @param host: ip or hostname for the host on which the command will be
            run.
        @param baseCommandRunnerFactory: command runner factory from which this
            command runner shall get its base command runner (i.e - host on
            which to perform the ssh command). If this is omitted, it defaults
            to L{LocalCommandRunnerFactory}
        @see: L{CommandRunner.__init__}
        """
        super(RemoteCommandRunner, self).__init__(basicRunString, verifier)
        if baseCommandRunner is None:       
            if baseCommandRunnerFactory is None:
                baseCommandRunnerFactory = LocalCommandRunnerFactory()
            baseCommandRunner = baseCommandRunnerFactory.getCommandRunner(verifier=verifier)
        self.command = baseCommandRunner
        self.host = host
        self.user = user
        self.sshkey = sshkey
    @property
    def _sshOptions(self):
        returned = "-q -o StrictHostKeyChecking=no -o PasswordAuthentication=no -o CheckHostIP=no"
        if self.sshkey:
            returned += " -i " + self.sshkey
        return returned
    @property
    def _destination(self):
        if self.user:
            return self.user + '@' + self.host
        else:
            return self.host
    def _runPopen(self, cmd):
        command_string = 'ssh %s %s %s' % (self._sshOptions, self._destination, _quote(cmd))
        return self.command._runPopen(command_string)
    def __repr__(self):
        return "<Remote command runner to %r>" % self._destination
    def uploadFile(self, localFileName, remoteFileName):
        self.command.executeStr("scp %s %s %s:%s" % (self._sshOptions, localFileName, self._destination, remoteFileName))
    def downloadFile(self, remoteFileName, localFileName):
        self.command.executeStr("scp %s %s:%s %s" % (self._sshOptions, self._destination, remoteFileName, localFileName))

#----------------------------#
# Command Runner Factories   #
#----------------------------#

class CommandRunnerFactory(object):
    """
    Represents an Abstract Factory for producing command runners.
    Inherit from this interface to customize command factory creation
    """
    def __init__(self):
        self._runStringPrefix = ''
    def getHost(self):
        raise NotImplementedError()
    def getCommandRunner(self, basicRunString="", verifier=SimpleResultVerifier()):
        runner = self._getCommandRunner(self._runStringPrefix+basicRunString, verifier)
        assert isinstance(runner, CommandRunner) # for wing
        return runner
    def _getCommandRunner(self, basicRunString, verifier):
        return CommandRunner(basicRunString, verifier)
    def appendToRunStringPrefix(self, runString):
        self._runStringPrefix += runString

class LocalCommandRunnerFactory(CommandRunnerFactory):
    """
    A CommandRunnerFactory which produces LocalCommandProxies.
    """
    def getHost(self):
        return 'localhost'
    def _getCommandRunner(self, basicRunString, verifier):
        return LocalCommandRunner(basicRunString, verifier)
    def __repr__(self):
        return "<Local command runner factory>"

class RemoteCommandRunnerFactory(CommandRunnerFactory):
    """
    Produces RemoteCommandRunner objects
    """
    def __init__(self, host, baseCommandRunnerFactory=None, user=None, sshkey=None):
        super(RemoteCommandRunnerFactory, self).__init__()
        self.baseCommandRunnerFactory = baseCommandRunnerFactory
        self.host = host
        self.user = user
        self.sshkey = sshkey
    def getHost(self):
        return self.host
    def _getCommandRunner(self, basicRunString, verifier):
        return RemoteCommandRunner(self.host, basicRunString, verifier,
                                   self.baseCommandRunnerFactory,
                                   user=self.user, sshkey=self.sshkey)
    def __repr__(self):
        return "<Remote command runner factory to %r>" % self.host

#----------------------------#
# MR. Chain                  #
#----------------------------#

class CommandRunnerChain(CommandRunnerFactory):
    """
    Chains several command runners, starting from
    a local one, to run a command on a remote machine

    >>> c = CommandRunnerChain('pro206', 'pro1')
    >>> r = c.getCommandRunner()
    >>> r.host == 'pro1'
    True
    >>> for src, srcf, dest, destf in c.traverseLinks():
    ...     print src, dest
    localhost pro206
    pro206 pro1
    >>> c.push('blabla')
    >>> c.pop()[0] == 'blabla'
    True
    """
    def __init__(self, *hosts):
        """
        @param hosts: list of addresses to hosts. Does not include
            a string representing the local host.
        """
        super(CommandRunnerChain, self).__init__()

        self.hosts = []
        self.factories = []
        self.root = LocalCommandRunnerFactory()
        last = self.root
        for host in hosts:
            if type(host) in (tuple, list):
                user, host, sshkey = host
            else:
                user, host, sshkey = None, host, None
            new_factory = RemoteCommandRunnerFactory(host, last, user=user, sshkey=sshkey)
            last = new_factory
            self.factories.append(new_factory)
            self.hosts.append(host)

    def getHost(self):
        if not self.hosts:
            return 'localhost'
        else:
            return self.hosts[-1]

    def __repr__(self):
        return "<Command runner chain (%s)>" % (" => ".join(["localhost"] + self.hosts),)

    def _getCommandRunner(self, basicRunString, verifier):
        return self._last().getCommandRunner(basicRunString, verifier)

    def clone(self):
        return type(self)(*self.hosts)

    def traverseLinks(self):
        """
        a generator yielding a tuple of the form:
        (source_hostname, source_factory, dest_hostname, dest_factory)
        for each link defined.
        """
        hosts_factories = zip(self.hosts, self.factories)
        last = ("localhost", self.root)
        for host, factory in hosts_factories:
            yield last + (host, factory)
            last = (host, factory)

    def _last(self):
        if not self.factories:
            return self.root
        else:
            return self.factories[-1]
        
    def push(self, host):
        """
        @param host: hostname or ip
        """
        self.hosts.append(host)
        self.factories.append(RemoteCommandRunnerFactory(host, self._last()))
    def pop(self):
        host = self.hosts.pop()
        factory = self.factories.pop()
        return host, factory

    def getTargetAddress(self):
        return self.hosts[-1]

#----------------------------#
# Ready-made local runner    #
#----------------------------#

def concurrentExecute(string, *runners, **kwargs):
    kwargs['async'] = True
    
    results = [None for runner in runners]
    asyncs = [runner.executeStr(string, **kwargs) for runner in runners]
    for index, async in enumerate(asyncs):
        results[index] = async.wait()
    return results


runner     = LocalCommandRunner()
getOutput  = runner.getOutput
execute    = runner.execute
executeStr = runner.executeStr

if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose=True)
