#!/usr/bin/python

import telnetlib
import re

class TelnetError(Exception):
    pass

class LogicalError(Exception):
    pass

class MDS(object):
    def __init__(self, name):
        super(MDS, self).__init__()
        self.name = name
        self.link = self.create_admin_link()
        self.execute('terminal length 511')
        self.zones = {}
        self.get_active_zones()
        
    def create_admin_link(self):
        link = telnetlib.Telnet(self.name)
        link.read_until("login: ")
        link.write("admin\n")
        link.read_until("Password: ")
        link.write("mds-admin\n")
        link.read_until('#')
        return link

    def get_output(self, cmd):
        if not cmd.endswith('\n'):
            cmd = cmd + '\n'
        self.link.write(cmd)
        result = [line.strip() for line in self.link.read_until('#').split('\n')]
        return result

    def execute(self, cmd, *expected_errors):
        result = self.get_output(cmd)

        # always except no errors
        if expected_errors == ():
            expected_errors = ([],)
        
        if ((len(result) < 2) or (result[0] != cmd.rstrip()) or
            (not re.match("^%s(\(.*\))?#"%(self.name), result[-1])) or
            (not result[1:-1] in expected_errors)):
            raise TelnetError(result)

    def __getitem__(self, item):
        if not isinstance(item, str):
            raise TypeError('Non-string names are not supported')
        if item not in self.zones:
            self.zones[item] = Zone(self, item)
        return self.zones[item]

    def get_active_zones(self):
        active_zones = []
        current_zone = None
        zone_name_re = re.compile("zone name (?P<zone_name>[^ ]+) vsan [0-9]+")
        wwpn_re = re.compile(".*\[?pwwn (?P<wwpn>[:0-9a-fA-F]+)\]?")
        self.link.write('show zone active\n')
        while True:
            idx, match, text = self.link.expect(["\n", "#"])
            if idx == 1:
                break
            match = zone_name_re.match(text)
            if match:
                name = match.group("zone_name")
                active_zones.append(name)
                self.zones[name] = Zone(self, name)
                current_zone = name
                continue
            match = wwpn_re.match(text)
            if match:
                self.zones[current_zone].add_member(match.group("wwpn"))
                
        return active_zones
                               
class Zone(object):
    active_zoneset = "XIV"
    vsan = "1"

    def __init__(self, mds, name):
        super(Zone, self).__init__()
        self.members=[]
        self.mds = mds
        self.name = name

    def normalize_wwpn(self, wwpn):
        if re.match("([0-9a-fA-F]{2}:){7}[0-9a-fA-F]{2}", wwpn):
            return wwpn
        if re.match("[0-9a-fA-F]{16}", wwpn):
            _wwpn = ""
            for idx in xrange(0, 16, 2):
                _wwpn = _wwpn + wwpn[idx:idx+2] + ':' 
            _wwpn = _wwpn[:-1]
            return _wwpn
        raise ValueError("invalid wwpn (%s)" % (wwpn,))            
        
    def add_member(self, wwpn):
        _wwpn = self.normalize_wwpn(wwpn)
        if (_wwpn in self.members):
            return
        self.members.append(_wwpn)

    def del_member(self, wwpn):
        if (wwpn not in self.members):
            raise LogicalError("wwpn %s not in zone '%s'"%(wwpn, self.name))
        self.members.remove(wwpn)
        if (self.members == []):
            self.delete()
        self.mds.execute("config t\n", ['Enter configuration commands, one per line.  End with CNTL/Z.',])
        self.mds.execute("zone name %s vsan %s\n"%(self.name, self.vsan))        
        self.mds.execute("no member pwwn %s"%(wwpn,))
        self.mds.execute("exit\n")
        self.mds.execute("exit\n")
        self.commit_changes()        
        
    def commit_changes(self):
        self.mds.execute("config t\n", ['Enter configuration commands, one per line.  End with CNTL/Z.',])
        self.mds.execute("zoneset activate name %s vsan %s\n"%(self.active_zoneset, self.vsan),
                         ['Zoneset activation initiated. check zone status',],
                         ['WARNING: Specified zoneset already active and unchanged',])
        self.mds.execute("exit\n")
        
    def create(self):
        self.mds.execute("config t\n", ['Enter configuration commands, one per line.  End with CNTL/Z.',])
        self.mds.execute("zone name %s vsan %s\n"%(self.name, self.vsan))
        for wwpn in self.members:
            self.mds.execute("member pwwn %s\n"%(wwpn,), ['Duplicate member'])
        self.mds.execute("exit\n")
        self.mds.execute("zoneset name %s vsan %s\n"%(self.active_zoneset, self.vsan))
        self.mds.execute("member %s\n"%(self.name,))
        self.mds.execute("exit\n")
        self.mds.execute("exit\n")
        self.commit_changes()

    def delete(self):
        self.mds.execute("config t\n", ['Enter configuration commands, one per line.  End with CNTL/Z.',])
        self.mds.execute("no zone name %s vsan %s\n"%(self.name, self.vsan),
                         [],
                         ['Zone not present',])
        self.mds.execute("exit\n")
        self.commit_changes()
        self.members = []
        
if __name__ == "__main__":
    mds = MDS('mds')
    zone = mds['tlib_auto_zone_10_117']
    print "members of '%s' are"%(zone.name,), zone.members
    ##zone.delete()
    zone.add_member("1234567890abcdef")
    zone.add_member("1234567890abcde0")
    zone.create()
    
