from __future__ import with_statement

import md5
import os
import time

from xiv import logger

def chainCopyFile(runner_chain, local_filename, remote_filename):
    """
    Copies a file over a command runner chain.

    >>> import tempfile
    >>> from .command_runner import CommandRunnerChain
    >>> 
    >>> # This checks that chainCopyFile can copy files locally:
    >>> local_filename = tempfile.mktemp()
    >>> with file(local_filename, 'wb') as local_file:
    ...     local_file.write('X' * 1024)
    ... 
    >>> chain = CommandRunnerChain('localhost')
    >>> remote_filename = tempfile.mktemp()
    >>> chainCopyFile(chain, local_filename, remote_filename)
    >>> file(local_filename).read() == file(remote_filename).read()
    True
    >>> os.unlink(remote_filename)
    >>>
    >>> # This checks that files can be copied over a real chain
    >>> import socket
    >>> chain = CommandRunnerChain('bleys', socket.gethostname())
    >>> chainCopyFile(chain, local_filename, remote_filename)
    >>> file(local_filename).read() == file(remote_filename).read()
    True
    >>> os.unlink(remote_filename)
    >>> os.unlink(local_filename)
    """
    source_file_name = None

    logger.debug("chainCopyFile over %s (%r)", runner_chain, list(runner_chain.traverseLinks()))
    temp_filename = "/tmp/__temporary_transfer_%s_%s" % (md5.md5(local_filename).hexdigest(),
                                                    time.time())
    
    for src, src_factory, dest, dest_factory in runner_chain.traverseLinks():
        assert None not in (src, dest)

        #choose temporary file on remote machine
        src_runner = src_factory.getCommandRunner()
        
        if source_file_name is None:
            #first link...
            src_runner.scp(local_filename, "%s:%s" % (dest, temp_filename))
            #don't remove source file...
        else:
            #not first link...
            src_runner.scp(source_file_name, "%s:%s" % (dest, temp_filename))
            #remove file
            src_runner.rm(source_file_name)
        source_file_name = temp_filename
    #move file on last machine (supports also single-link chains)
    if source_file_name is None:
        runner_chain.getCommandRunner().cp(local_filename, remote_filename)
    else:
        runner_chain.getCommandRunner().mv(source_file_name, remote_filename)

def chainWriteFile(runner_chain, data, remote_filename):
    logger.debug("chainWriteFile over %s (%r)", runner_chain, list(runner_chain.traverseLinks()))
    filename = "/tmp/__temporary_transfer_%s_%s" % (md5.md5(data).hexdigest(),
                                               time.time())
    f = file(filename, "wb")
    f.write(data)
    f.close()
    chainCopyFile(runner_chain, filename, remote_filename)
    try:
        os.unlink(filename)
    except OSError, e:
        pass
    
class TemporaryFile(object):
    def __init__(self, runner_chain, filename, data = None):
        super(TemporaryFile, self).__init__()
        self.filename = filename
        self.chain = runner_chain
        if data is not None:
            chainWriteFile(runner_chain, data, filename)
    def __del__(self):
        self.chain.getCommandRunner().rm("-f", self.filename)
