import socket
import os
import time
import pdb
from Shared import *
from PacketTypes import *
from xiv.dtypes.Bunch import Bunch
from xiv import logger

class Server(Session):
    """This class implements a tftp server object."""

    def __init__(self, tftproot='/tmp', listenip="0.0.0.0", listenport=DEF_TFTP_PORT,):
        """Class constructor. It takes a single optional argument, which is
        the path to the tftproot directory to serve files from and/or write
        them to."""
        self.listenip = listenip
        self.sock = None
        self.root = tftproot
        self.remote = None
        self.is_buffer = False
        self.buffer = None
        # A dict of handlers, where each session is keyed by a string like
        # ip:tid for the remote end.
        self.handlers = {}

        self._validate_root()
        self._init_socket(listenip, listenport)
        
        
    def _validate_root(self):
        if not os.path.exists(self.root):
            raise TftpException, "The tftproot '%s' does not exist." %(self.root,)
        if not os.path.isdir(self.root):
            raise TftpException, "The tftproot must be a directory."
        if not os.access(self.root, os.R_OK):
            raise TftpException, "The tftproot must be readable"
        if not os.access(self.root, os.W_OK):
            logger.warn("The tftproot %s is not writable" % (self.root,))

    def _init_socket(self, listenip, listenport):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        logger.info("Server requested on ip %s, port %s" % (listenip, listenport))
        self.sock.bind((listenip, listenport)) # Will raise an exception if already used

    def get_sockname(self):
        return self.sock.getsockname()
    

    def serve(self,  timeout=SOCK_TIMEOUT, incoming_buffer = Bunch(name=''),
              outgoing_buffer = Bunch(name=''), is_single_transfer = True):
        """Start a server listening on the supplied interface and port. This
        defaults to INADDR_ANY (all interfaces) and UDP port 69. You can also
        supply a different socket timeout value, if desired. The server also supports handling of a single download/upload
        transaction given through the incoming_buffer and outgoing_buffer parameters that should contain a dictionary with
        a 'name' field to specify the name of the file to handle and 'buffer' field with the buffer data"""
        import select
        
        tftp_factory = PacketFactory()

        self.incoming_buffer = incoming_buffer
        self.outgoing_buffer = outgoing_buffer

        
        logger.info("Starting receive loop...")
        self.handlers = {self.sock: self}
        
        while True:
            self.validate_timeouts()
            readyinput, readyoutput, readyspecial = select.select(self.handlers.keys(), [], [], SOCK_TIMEOUT)
            if not readyinput:
                continue

            readysock = readyinput[0]

            buffer, (raddress, rport) = readysock.recvfrom(MAX_BLKSIZE)
            recvpkt = tftp_factory.parse(buffer)
            #logger.debug("received packet %r" % (recvpkt, ))

            try:
                self.handlers[readysock].handle(recvpkt, raddress, rport)
            except TftpException, err:
                ret = None                
                is_buffer =  self.handlers[readysock].is_buffer
                if is_buffer:
                    logger.debug("exception returned from a buffer handler")
                    ret = self.handlers[readysock].buffer                    
                if self.handlers[readysock].is_finished():
                    logger.info("Successful transfer.")
                logger.debug("Deleting handler %r is_buffer %r" % (self.handlers[readysock].remote,is_buffer))
                del self.handlers[readysock]
                logger.debug("value of 'is_single_transfer or is_buffer' is %r" % (is_single_transfer or is_buffer, ) )
                if is_single_transfer or is_buffer:
                    return ret

    def validate_timeouts(self):
        now = time.time()
        for sock, handler in self.handlers.iteritems():
            try:
                handler.check_timeout(now)
            except TftpException, err:
                logger.error("Fatal exception thrown from handler: %s" % (str(err),))
                del self.handlers[sock]
                return self.validate_timeouts() # to stop the current iteration, but find more to remove.


    def handle(self, recvpkt, raddress, rport):
        client = (raddress, rport)
        if not self._is_client_new(client):
            logger.error("Received RRW or WRW %r for existing session %r. aborting" % (recvpkt, client))
            return
        
        if isinstance(recvpkt, PacketRRQ):
            try:
                logger.debug("New download request, client = %r" % (client, ))
                new_handler = ServerDownloadHandler(client, recvpkt, self.root, self.listenip, self.outgoing_buffer)
                self.handlers[new_handler.get_socket()] = new_handler
            except TftpException, err:
                logger.error("Fatal exception thrown from handler: %s" % (str(err),))
        elif isinstance(recvpkt, PacketWRQ):
            try:
                logger.debug("New upload request, client = %r" % (client, ))
                new_handler = ServerUploadHandler(client, recvpkt, self.root, self.listenip, self.incoming_buffer)
                self.handlers[new_handler.get_socket()] = new_handler
            except TftpException, err:
                logger.error("Fatal exception thrown from handler: %s" % (str(err),))                    
        else:
            # FIXME - this will have to change if we do symmetric UDP
            logger.error("Should only receive RRQ or WRQ packets on main listen port. Received %s" % recvpkt)
            self.senderror(TftpErrors.IllegalTftpOp, remote = client)
 
    def check_timeout(self, now):
        return True
    
    def _is_client_new(self, client):
        for handler in self.handlers.values():
            if client == handler.remote:
                return False
        return True


class ServerHandler(Session):
    """ Parent class for the upload and download sessions """
    def __init__(self, client, root, listenip):
        Session.__init__(self)
        self.remote = client
        self.listenip = listenip
        self.root = root
        self.filename = None
        self.options = {}
        self.timesent = 0
        self.timeouts = 0
        self.buffer = None
        self.sock = self._init_new_socket(listenip)        

    def _init_new_socket(self, listenip):
        """This method generates a new UDP socket, whose listening port must
        be randomly generated, and not conflict with any already in use. For
        now, let the OS do this."""
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.bind((listenip, 0))
        logger.debug("Bound socket to port %r" % (sock.getsockname()[1],))
        return sock

    def get_socket(self):
        return self.sock

    def _set_transfer_mode(self, packet):
        if packet.mode in ['octet']:
            self.is_ascii = False
        elif packet.mode in ['netascii', 'ascii', 'Ascii']:
            self.is_ascii = True
        else:
            self.senderror(TftpErrors.IllegalTftpOp)
            raise TftpException, "Unsupported mode: %s" % recvpkt.mode

                
    def _start_session_with_options(self, packet):
        if not packet.options:
            logger.debug("Client %r requested no options."  % (self.remote,))
            return self.start_work()
        try:
            blksize = int(packet.options['blksize'])
            if blksize < MIN_BLKSIZE or MAX_BLKSIZE < blksize:
                logger.warn("Client requested invalid blocksize %d." % (blksize, ))
                blksize = DEF_BLKSIZE
            self.options['blksize'] = blksize
            return self.send_oack()

        except KeyError: #no blksize
            logger.warn("Client %r requested unsupported options: %s" % (self.remote, packet.options))
            self.senderror(TftpErrors.FailedNegotiation)
            raise TftpException, "Failed option negotiation"

    def send_oack(self):
        """This method sends an OACK packet based on current params."""
        logger.debug("Composing and sending OACK packet")
        oack = PacketOACK()
        oack.options = self.options
        self.sock.sendto(oack.encode().buffer, self.remote)
        self.timesent = time.time()
        self.state = 'oack'

    def check_timeout(self, now):
        """This method checks to see if we've timed-out waiting for traffic
        from the client."""
        if self.timesent:
            if now - self.timesent > SOCK_TIMEOUT:
                self.timeouts += 1
                if self.timeouts > TIMEOUT_RETRIES:
                    raise TftpException, "Hit max retries, giving up"
                self.handle_timeout()

        
    def handle(self):
        raise NotImplementedError, "Abstract method"


class ServerDownloadHandler(ServerHandler):
    """This class implements a handler for a given server download session, handling
    the work for one download."""

    def __init__(self, client, packet, root, listenip, outgoing_buffer = {}):
        ServerHandler.__init__(self, client, root, listenip)
        logger.info("Starting new download handler. Key %r." % (client,))

        self.fetcher = None
        self.is_buffer = False
        self.buffer = None
        self._init_read_request(packet, outgoing_buffer)

    
    def _assure_requested_file_is_valid(self, packet, outgoing_buffer):
        if packet.filename == outgoing_buffer.name:
           self.filename = packet.filename
           self.is_buffer = True
           self.buffer = outgoing_buffer.buffer
           return
        
        self.filename = os.path.join(self.root, packet.filename)
        logger.debug("The path to the desired file is %s" % (self.filename,))
        self.filename = os.path.abspath(self.filename)
        
        if not self.filename.startswith(os.path.abspath(self.root)):
            logger.error("Insecure path: filename is %s root is %s" % (self.filename, self.root))
            self.senderror(TftpErrors.AccessViolation)
            raise TftpException, "Insecure path: %s" % (self.filename,)

        if not os.path.exists(self.filename):
            logger.error("Requested file %s does not exist." % (self.filename,))
            self.senderror(TftpErrors.FileNotFound)
            raise TftpException, "Requested file not found: %s" % (self.filename,)

            
    def _init_read_request(self, packet, outgoing_buffer):
        logger.debug("Requested file is %s, mode is %s packet is %r" % (packet.filename, packet.mode, packet))

        self._set_transfer_mode(packet)
        self._assure_requested_file_is_valid(packet, outgoing_buffer)
        self._start_session_with_options(packet)
        

    def handle_timeout(self):
        """This method handles a timeout condition."""
        # FIXME - still need to handle Sorceror's Apprentice problem
        if self.state == 'dat':
            logger.debug("Timing out on DAT. Need to resend.")
            self.send_dat(resend=True)
        elif self.state == 'oack':
            logger.debug("Timing out on OACK. Need to resend.")
            self.send_oack()
        else:
            tftpassert(False,
                       "Timing out in unsupported state %s" %
                       self.state)


    def handle_ack_packet(self, recvpkt):
        if recvpkt.blocknumber == 0 and self.state == 'oack':
            logger.debug("Received ACK with 0 blocknumber, starting download")
            return self.start_work()

        if self.state == 'dat':
            if recvpkt.blocknumber != self.fetcher.get_block_counter():
                logger.warn("Received an ACK for invalid block number %d. needed : %d" % (recvpkt.blocknumber, self.fetcher.get_block_counter()))

            if self.is_finished():
                logger.debug("Received last ACK for session %r" % (self.remote,))
                raise TftpException, "Finished downloading"
            self.send_dat()
        else:
            logger.error("Received ACK with block number %d while in state %s" % (recvpkt.blocknumber, self.state))

    def handle(self, recvpkt, raddress, rport):
        """This method informs a handler instance that it has data waiting on
        its socket that it must read and process."""
        if isinstance(recvpkt, PacketACK):
            self.handle_ack_packet(recvpkt)
        elif isinstance(recvpkt, PacketERR):
            logger.error("Received error packet from client: %s" % recvpkt)
            self.state = 'err'
            raise TftpException, "Received error from client"
        else:
            logger.error("Received packet %s while handling a download" % recvpkt)
            self.senderror(TftpErrors.IllegalTftpOp)
            raise TftpException, "Invalid packet received during download"

    def is_finished(self):
        return self.fetcher.is_finished()
    
    def start_work(self):
        """This method opens self.filename, stores the resulting file object
        in self.fileobj, and calls send_dat()."""
        self.state = 'dat'
        if self.is_buffer:
            logger.debug("Using the BufferFetcher")
            self.fetcher = BufferFetcher(self.options.get('blksize', DEF_BLKSIZE), self.buffer, is_ascii = self.is_ascii)
        else:
            self.fetcher = BlockFetcher(self.filename, self.options.get('blksize', DEF_BLKSIZE), is_ascii = self.is_ascii)
        self.send_dat()

    def send_dat(self, resend=False):
        """This method reads sends a DAT packet based on what is in self.buffer."""
        if not resend:
            self.fetcher.process_next_block()
        else:
            logger.warn("Resending block number %d" % (self.fetcher.get_block_counter(),))

        block, number = self.fetcher.get_current_block()
        
        dat = PacketDAT(data = block, block_number = number)
        self.sock.sendto(dat.encode().buffer, self.remote)
        self.timesent = time.time()
        #logger.debug("Sent block number %d with length of %d to remote %r" % (number, len(block), self.remote))


class ServerUploadHandler(ServerHandler):
    """ This class implements a handler for a given server upload session,
    handling the work for a single session """

    def __init__(self, client, packet, root, listenip, incoming_buffer = {}):
        ServerHandler.__init__(self, client, root, listenip)
        logger.info("Starting new upload handler. client is %r" % (client, ))
        self.dumper = None
        self.counter = 0
        self.is_eof = False
        self._init_write_request(packet, incoming_buffer)

    def _init_write_request(self, packet, incoming_buffer):
        buffer_name = incoming_buffer.name
        logger.debug("Requested written file is '%s' buff_name is '%s'. mode is %s. packet is %r" %
                     (packet.filename ,buffer_name, packet.mode, packet))
        self._set_transfer_mode(packet)
        if packet.filename == buffer_name:
            self.is_buffer = True
            self.buffer = ''
        else:
            self.is_buffer = False
            self.fileobj = open(os.path.join(self.root, packet.filename), "w") # If it fails -- nothing bad happens.
        self._start_session_with_options(packet)

    def start_work(self):
        self.state = 'dat'
        self.send_ack()

    def send_ack(self):
        logger.debug("Sending ack with counter %d" % (self.counter, ))
        ack = PacketACK(blocknumber = self.counter)
        self.sock.sendto(ack.encode().buffer, self.remote)
        self.timesent = time.time()

    def handle_next_data_block(self, packet):
        if packet.blocknumber == self.counter:
            return self.send_ack()

        if packet.blocknumber != (self.counter + 1):
            logger.error("Receivef packet out of order. Received %d needed %d" % (packet.blocknumber, self.counter +1))
            self.senderror(TftpErrors.IllegalTftpOp)
            raise TftpException, "Received invalid block number from client"
        # Send ack right away so other side will send next packet
        self.counter += 1
        self.send_ack()

        buff = packet.data
        if self.is_ascii:
            buff = buff.replace('\r', '')

        if self.is_buffer:
            self.buffer = self.buffer + buff
        else:
            self.fileobj.write(buff)
            self.fileobj.flush()


        if len(packet.data) <  self.options.get('blksize', DEF_BLKSIZE):
            self.is_eof = True
            raise TftpException, "Successful transfer"

    
    def handle(self, recvpkt, raddress, rport):
        if isinstance(recvpkt, PacketDAT):
            self.handle_next_data_block(recvpkt)

        elif isinstance(recvpkt, PacketERR):
            logger.error("Received error packet from client: %s" % recvpkt)
            self.state = 'err'
            raise TftpException, "Received error from client"
        else:
            logger.error("Received packet %s while handling a download" % recvpkt)
            self.senderror(TftpErrors.IllegalTftpOp)
            raise TftpException, "Invalid packet received during download"
        

    def is_finished(self):
        return self.is_eof
    
    def handle_timeout(self):
        """ We timed out on data reception. We should resend the ack, and hope for good."""
        self.send_ack()
