from xiv import logger

MIN_BLKSIZE = 8
DEF_BLKSIZE = 512
MAX_BLKSIZE = 65536
SOCK_TIMEOUT = 5
MAX_DUPS = 20
TIMEOUT_RETRIES = 5
DEF_TFTP_PORT = 69


def tftpassert(condition, msg):
    """This function is a simple utility that will check the condition
    passed for a false state. If it finds one, it throws a TftpException
    with the message passed. This just makes the code throughout cleaner
    by refactoring."""
    if not condition:
        raise TftpException, msg

    
class TftpErrors(object):
    """This class is a convenience for defining the common tftp error codes, and making
    them more readable in the code."""
    NotDefined = 0
    FileNotFound = 1
    AccessViolation = 2
    DiskFull = 3
    IllegalTftpOp = 4
    UnknownTID = 5
    FileAlreadyExists = 6
    NoSuchUser = 7
    FailedNegotiation = 8

class TftpException(Exception):
    """This class is the parent class of all exceptions regarding the handling
    of the TFTP protocol."""
    pass


class Session(object):
    """This class is the base class for the tftp client and server. Any shared
    code should be in this class."""

    def __init__(self):
        """Class constructor. Note that the state property must be a State
        object."""
        self.options = None
        self.dups = 0
        self.errors = 0
        
        
    def senderror(self, errorcode, sock = None,  remote = None):
        """This method uses the socket passed, and uses the errorcode, address and port to
        compose and send an error packet."""
        
        if not sock:
            sock = self.sock

        if not remote:
            remote = self.remote
            
        logger.debug("In senderror, being asked to send error %d to %r" % (errorcode, remote))
        from PacketTypes import PacketERR
        errpkt = PacketERR()
        errpkt.errorcode = errorcode
        self.sock.sendto(errpkt.encode().buffer, remote)


class GenericFetcher(object):
    def __init__(self, block_size):
        self.block_counter = 0
        self.is_eof = False
        self.block_size = block_size
        self.buffer = ""

        
    def _inc_block_counter(self):
        self.block_counter += 1
        if 65535 < self.block_counter:
            logger.debug("Block counter rolled over to 0")
            self.block_counter = 0
    
    def get_current_block(self):
        return self.buffer, self.block_counter

    def get_block_counter(self):
        return self.block_counter

    def is_finished(self):
        return self.is_eof

class BufferFetcher(GenericFetcher):
    """ This class maintains the position in the given buffer, and knows how to read in the 'netascii' mode
    of the protocol. """
    def __init__(self, block_size, given_buffer, is_ascii = False):
        GenericFetcher.__init__(self, block_size)
        if is_ascii:
            self.full_buffer = given_buffer.replace('\n', '\r\n')
        else :
            self.full_buffer = given_buffer
        logger.debug("Working with block size %d and given buffer of size %d" % (self.block_size, len(self.full_buffer)))


    def process_next_block(self):
        self._inc_block_counter()

        self.buffer = self.full_buffer[:self.block_size]
        self.full_buffer = self.full_buffer[self.block_size:]
        
        if self.full_buffer == "":
            logger.debug("Finished processing all of the buffer. counter is %d" % (self.block_counter, ))
            self.is_eof = True

class BlockFetcher(GenericFetcher):
    """ This class maintains the position in the current file, and knows how to read in the 'netascii' mode
    of the tftp protocol. It keeps the location of the file, and knows whether it is the end"""
    def __init__(self, filename, block_size, is_ascii = False):
        GenericFetcher.__init__(self, block_size)

        self.filename = filename
        self.fileobj = open(self.filename, "r")
        self.is_ascii = is_ascii
        self.left_over = ""
        if is_ascii:
            logger.warn("Setting block fetcher to ASCII mode")

    def _fix_ascii_buffer(self):
        self.buffer = self.left_over + self.buffer.replace('\n', '\r\n')
        self.left_over = self.buffer[self.block_size:]
        self.buffer = self.buffer[:self.block_size]

        if self.is_eof and self.left_over:
            self.is_eof = False
        
    def process_next_block(self):
        self._inc_block_counter()
        
        self.buffer = self.fileobj.read(self.block_size)    
        
        if len(self.buffer) < self.block_size:
            logger.debug("Reached EOF file file %s" % (self.filename, ))
            self.is_eof = True

        if self.is_ascii:
            self._fix_ascii_buffer()

