import platform
import tempfile
import os
import re

from xiv.dtypes import Bunch
from xiv.network.command_runner import LocalCommandRunner, DoNothingVerifier, CommandFailureException


class EZtmpfsException(Exception):
    pass
class EZtmpfsCommandFailed(EZtmpfsException):
    pass
class EZtmpfsNotSupported(EZtmpfsException):
    pass
class EZtmpfsInvalidOperationForDeploymentState(EZtmpfsException):
    pass
class EZtmpfsSizeException(EZtmpfsException):
    pass

def reraise_as_eztmpfs_exception_with_function_name(*catch_exceptions):
    def decor(func):
        def callable(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except catch_exceptions, e:
                raise EZtmpfsCommandFailed("%s failed: %s" % (func.__name__.replace('_', ' '), e))
        callable.__name__ = func.__name__
        callable.__doc__ = func.__doc__
        return callable
    return decor

class EZtmpfs(object):
    RESERVE_MEMORY_PERCENT = 10 # don't resize tmpfs so that less than this % of free memory (ram + swap) remains
    RESERVE_TMPFS_SIZE = 2**13 # don't resize the tmpfs thus that less than this size remains free on it
    def __init__(self, name=None, requested_fs_size=None, auto_load_tmpfs_module=False, auto_deploy=True):
        super(EZtmpfs, self).__init__()
        if auto_load_tmpfs_module:
            LocalCommandRunner(verifier=DoNothingVerifier()).modprobe('tmpfs')
        self.name = name
        self.verify_platform_support()
        self.mountpoint = None
        if auto_deploy:
            self.auto_deploy(requested_fs_size)

    def __str__(self):
        if not self.mountpoint:
            raise EZtmpfsInvalidOperationForDeploymentState("can't get string representation of undeployed filesystem")
        return self.mountpoint

    def __repr__(self):
        if self.mountpoint:
            status = 'deployed on %s' % self.mountpoint
        else:
            status = 'undeployed'
        return '<%s (%s)>' % (self.__class__.__name__, status)

    @classmethod
    def verify_platform_support(cls):
        if platform.system() != 'Linux':
            raise EZtmpfsNotSupported("EZtmpfs is supported only on Linux (you have %s)" % platform.system())
        try:
            if 'tmpfs' not in file('/proc/filesystems').read():
                raise EZtmpfsNotSupported("no tmpfs support found in /proc/filesystem")
        except IOError, e:
            raise EZtmpfsNotSupported("reading /proc/filesystem: " + e.strerror)

    def deploy(self):
        if self.mountpoint:
            raise EZtmpfsInvalidOperationForDeploymentState("can't deploy already deployed filesystem")
        self.create_mountpoint()
        self.mount()

    @reraise_as_eztmpfs_exception_with_function_name(IOError, OSError)
    def create_mountpoint(self):
        suffix = '-EZtmpfs'
        if self.name is not None:
            suffix += self.name
        self.mountpoint = tempfile.mkdtemp(suffix)

    @reraise_as_eztmpfs_exception_with_function_name(CommandFailureException)
    def mount(self):
        LocalCommandRunner().mount('-t', 'tmpfs', '/dev/null', self.mountpoint)

    @reraise_as_eztmpfs_exception_with_function_name(CommandFailureException)
    def resize(self, size):
        size = normalize_size(size)
        self.validate_size(size)
        LocalCommandRunner().mount('-o', 'remount,size=%s' % size, '-t', 'tmpfs', '/dev/null', self.mountpoint)

    def auto_deploy(self, size):
        self.deploy()
        if size is not None:
            size = normalize_size(size)
            self.validate_size(size)
            try:
                self.resize(size)
            except CommandFailureException, e:
                self.destroy()
                raise

    def destroy(self, dont_ignore_exceptions=False):
        if not self.mountpoint:
            return
        try:
            self.umount()
            self.remove_mountpoint()
            self.mountpoint = None
        except EZtmpfsException, error:
            if dont_ignore_exceptions:
                raise

    @reraise_as_eztmpfs_exception_with_function_name(CommandFailureException)
    def umount(self):
        LocalCommandRunner().umount(self.mountpoint)

    @reraise_as_eztmpfs_exception_with_function_name(OSError)
    def remove_mountpoint(self):
        os.rmdir(self.mountpoint)

    @reraise_as_eztmpfs_exception_with_function_name(CommandFailureException)
    def executing_df(self):
        if not self.mountpoint:
            raise EZtmpfsInvalidOperationForDeploymentState("must be deployed to see df parameters")
        line = LocalCommandRunner().df('-k', self.mountpoint)[1].splitlines()[1]
        return Bunch(zip(('device', 'capacity', 'used', 'available', 'use_percent', 'mountpoint'), line.split()))

    @property
    def used_space(self):
        return normalize_size(str(self.executing_df().used) + 'k')

    @property
    def available_space(self):
        return normalize_size(str(self.executing_df().available) + 'k')

    def validate_size(self, size):
        size = normalize_size(size)
        if size > total_memory() - total_memory() * (self.RESERVE_MEMORY_PERCENT/100):
            raise EZtmpfsSizeException("requested size %s exceeds total memory (%d), minus %s percent reserve" %
                (size, total_memory(), self.RESERVE_MEMORY_PERCENT))
        if size > free_memory() - total_memory() * (self.RESERVE_MEMORY_PERCENT/100):
            raise EZtmpfsSizeException("requested size %s exceeds free memory (%d), minus %s percent reserve of total" %
                (size, free_memory(), self.RESERVE_MEMORY_PERCENT))
        if self.mountpoint and size < self.used_space + self.RESERVE_TMPFS_SIZE:
            raise EZtmpfsSizeException("requested %s bytes is less than used space (%d and %d bytes reserve)" %
                (size, self.used_space, self.RESERVE_TMPFS_SIZE))

    def calculate_maximum_filesystem_size(self):
        return int(free_memory() - total_memory() * (self.RESERVE_MEMORY_PERCENT/100))

def normalize_size(size):
    size = str(size)
    quantifiers = { '': 2**0, 'k': 2**10, 'm': 2**20, 'g': 2**30}
    match = re.compile('(^[0-9]+)([kmg]?)$').match(size)
    if not match:
        raise ValueError("%s isn't a valid maximum size" % size)
    scalar = int(match.groups()[0])
    quantifier = quantifiers[match.groups()[1]]
    return scalar * quantifier

def meminfo():
    lines = file('/proc/meminfo').readlines()[3:]
    meminfo = Bunch()
    for line in lines:
        name, value, trash = line.split()
        meminfo[name.replace(':', '')] = value
    return meminfo

def free_memory():
    return normalize_size('%sk' % meminfo().MemFree) + normalize_size('%sk' % meminfo().SwapFree)

def total_memory():
    return normalize_size('%sk' % meminfo().MemTotal) + normalize_size('%sk' % meminfo().SwapTotal)

