"""
pyutils -- A package for messing around with python internals.
"""
import weakref as weakref_module
import struct
import sys
import dis
import gc
import time
import traceback
import itertools
import logging
import os
import types

def interact(locals=None):
    try:
        from IPython.Shell import IPShellEmbed
        interact_func = lambda locals: IPShellEmbed(argv=())(local_ns=locals)
    except ImportError:
        interact_func = lambda locals: code.interact(local=locals)
    interact_func(locals)

def unhandled_exception(exc_info, logger=None, message=None):
    if logger is None:
        logger = logging.getLogger('')
    if os.getenv('XIVDEBUG') is not None:
        list = traceback.format_exception(*exc_info)
        for line in list:
            logger.fatal(line.strip())
    else:
        logger.fatal('unhandled exception: %s' % exc_info[1])
    if message is not None:
        logger.fatal(message)

def wait(*args, **kwargs):
    """
    Waits for one of the args to become True for as long as 'timeout' seconds or 'iterations' iterations (one is required). 
    Raises 'exception' on timeout, calls 'callback' on each iteration, calls 'error_callback' before raising the exception and polls the args every 'interval' seconds. 
    The args may include function predicates or lists of function predicates (or both).
    """
    timeout = kwargs.pop('timeout', None)
    iterations = kwargs.pop('iterations', None)
    if timeout is None and iterations is None:
        raise TypeError("neither 'timeout' nor 'iterations' provided, at least one required")
    exception = kwargs.pop('exception', None)
    error_callback = kwargs.pop('error_callback', None)
    callback = kwargs.pop('callback', None)
    interval = kwargs.pop('interval', 1)
    if kwargs:
        raise ValueError("Unknown arguments: %s" % (kwargs.keys(),))

    callables = []
    for arg in args:
        try:
            callables.extend(arg)
        except:
            callables.append(arg)

    start_time = time.time() 
    if (iterations is None):
        iteration_counter = itertools.count()
    else:
        iteration_counter = range(iterations)

    for iteration in iteration_counter:
        if timeout is not None and time.time() - start_time > timeout:
            if error_callback is not None:
                error_return_object = error_callback()
                        
            if exception is not None:
                raise exception

        for callable in callables:
            if callable():
                return
        if callback is not None:
            callback()
        time.sleep(interval)
    if error_callback is not None:
        
        error_callback()
    if exception is not None:
        raise exception

def safe_forever_loop(timeout, on_timeout_exception=None):
    start = time.time()
    while time.time() - start < timeout:
        yield None
    if on_timeout_exception is not None:
        raise on_timeout_exception

def get_type_stats():
    """Returns a dictionary containing keys of type objects and
    a value representing how many objects of that type exist.
    """
    stats = {}
    for obj in gc.get_objects():
        stats[type(obj)] = stats.get(type(obj), 0) + 1
    return stats

def disassemble(codestring, from_position=0):
    """
    for a code string, yields tuples of (position, position_end, opcode, oparg)
    """
    position = 0
    n = len(codestring)
    extended_arg = 0
    while position < n:
        current_position = position
        op = ord(codestring[position])
        position += 1
        if op >= dis.HAVE_ARGUMENT:
            oparg = ord(codestring[position]) + ord(codestring[position+1])*256 + extended_arg
            extended_arg = 0
            position += 2
            if op == dis.EXTENDED_ARG:
                extended_arg = oparg*65536L
        else:
            oparg = None
        opname = dis.opname[op]        
        if current_position < from_position and position > from_position:
            raise ValueError("Cannot disassemble from %s", from_position)
        if current_position >= from_position:
            yield (current_position, position, opname, oparg)
def get_retval_recipient(level=1):
    """
    gets the name of the variable to which the return
    value of the calling function is to be returned.

    >>> def testfunc():
    ...     i = 2
    ...     j = 3
    ...     bla = otherfunc()
    ...     k = 666
    >>> def otherfunc():
    ...     print get_retval_recipient()
    ...     return (1,2,3)
    >>> testfunc()
    bla
    >>> global_variable=otherfunc()
    global_variable

    Also handles multiple recipients:
    >>> def testfunc():
    ...     a, b, c = otherfunc()
    >>> testfunc()
    ('a', 'b', 'c')
    >>> green, red, blue = otherfunc()
    ('green', 'red', 'blue')
    
    >>> a = b = c = otherfunc()
    a
    """
    frame = sys._getframe(level + 1)
    try:        
        code = frame.f_code.co_code
        disassembler = disassemble(code, from_position=frame.f_lasti)
        #move over function call
        disassembler.next()

        for unused, unused, instruction, arg in disassembler:
            if instruction != 'DUP_TOP':
                break
                
        if instruction == 'STORE_FAST':
            return frame.f_code.co_varnames[arg]
        elif instruction == 'STORE_NAME':
            return frame.f_code.co_names[arg]
        elif instruction == 'UNPACK_SEQUENCE':
            returned = []
            for i, (unused, unused, instruction, arg)  in zip(xrange(arg), disassembler):
                if instruction == 'STORE_FAST':
                    returned.append(frame.f_code.co_varnames[arg])
                elif instruction == 'STORE_NAME':
                    returned.append(frame.f_code.co_names[arg])
                else:
                    # unknown situation
                    return None
            return tuple(returned)
        else:
            return None
    finally:
        #frames might cause cyclic references...
        del frame  

def import_module(full_name):
    """Given a full module name, this function imports that module"""
    module = __import__(full_name)
    for module_name in full_name.split(".")[1:]:
        module = getattr(module, module_name)
    return module

def import_name(full_name):
    """Given a full function name, it imports and returns the function
    from within the module containing the function.
    Beware that this looks like a 'cool' feature, but is it used for
    cases in which the module or function name is dynamic and composed
    of input the program has only at runtime. PLEASE do not overuse.

    >>> unpack = import_name('struct.unpack')
    >>> unpack('LL', 'ABCDEFGHIJKLMNOP')
    (5208208757389214273, 5786930140093827657)
    """
    module_name, symbol = full_name.rsplit('.', 1)
    module = __import__(module_name, {}, {}, [''])
    try:
        return getattr(module, symbol)
    except AttributeError:
        raise ImportError(full_name)

def weakref(obj):
    if obj is None:
        return None
    if type(obj) is weakref_module.ProxyType:
        return obj
    return weakref_module.proxy(obj)

def safe_property(fget=None, fset=None, fdel=None, doc=None):
    """
    A pythonic property which raises a RuntimeError when an attribute error is raised within it. This is
    in order to avoid calls to __getattr__ when properties fail

    >>> def i_raise_an_exception():
    ...     raise AttributeError("blap")
    >>> class Test(object):
    ...     def some_prop(self):
    ...         return i_raise_an_exception()
    ...     def __getattr__(self, attr):
    ...         assert False
    ...     prop = property(some_prop)
    ...     safe_prop = safe_property(some_prop)
    >>> t = Test()
    >>> t.prop #doctest: +IGNORE_EXCEPTION_DETAIL +ELLIPSIS
    Traceback (most recent call last):
     ...
    AssertionError
    >>> t.safe_prop #doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
     ...
    RuntimeError: Attribute error within a property:
    Traceback (...):
     ...
    AttributeError: blap
    """
    if fget is not None:
        def callable(*args, **kwargs):
            try:
                return fget(*args, **kwargs)
            except AttributeError, e:
                raise RuntimeError("Attribute error within a property:\n%s" % (traceback.format_exc()))
        return property(callable, fset, fdel, doc)
    else:
        return property(fget, fset, fdel, doc)


def _get_entire_dict(class_obj):
    returned = {}
    for b in reversed(class_obj.__bases__):
        returned.update(_get_entire_dict(b))
    returned.update(class_obj.__dict__)
    return returned

class PreprocessedClass(type):
    """
    >>> class TestClass(object):
    ...     __metaclass__ = PreprocessedClass
    ...     SOME_INT = 2
    ...     SOME_STRING = 'bla'
    ...     def __process_class__(classname, classdict, bases):
    ...         print "*** Processing %s" % classname
    ...     def __process_attr__(classname, name, value, classdict, bases):
    ...         if name.upper() == name:
    ...             print "*** %s has an attribute %r=%r" % (classname, name, value)
    ...         return value
    ...     def __attr_delete_predicate__(classname, name, value, classdict, bases):
    ...         return name == 'SOME_STRING'
    *** Processing TestClass
    *** TestClass has an attribute 'SOME_INT'=2
    *** TestClass has an attribute 'SOME_STRING'='bla'
    >>> hasattr(TestClass, 'SOME_INT')
    True
    >>> hasattr(TestClass, 'SOME_STRING')
    False
    >>> class Derived(TestClass):
    ...     SOME_VALUE = 'value'
    *** Processing Derived
    *** Derived has an attribute 'SOME_VALUE'='value'
    """
    def __new__(cls, classname, bases, classdict):
        entire_dict = dict()
        
        for base in reversed(bases):
            entire_dict.update(_get_entire_dict(base))
        entire_dict.update(classdict)
        class_processor = entire_dict.get("__process_class__", None)
        attr_processor = entire_dict.get('__process_attr__', None)
        attr_delete_predicate = entire_dict.get('__attr_delete_predicate__', None)
        
        if class_processor:
            class_processor(classname, classdict, bases)
        for attr, value in classdict.items():
            if attr_processor:
                value = attr_processor(classname, attr, value, classdict, bases)
            if attr_delete_predicate and attr_delete_predicate(classname, attr, value, classdict, bases):
                classdict.pop(attr)
            else:
                classdict[attr] = value
        returned = type.__new__(cls, classname, bases, classdict)
        postprocessor = entire_dict.get('__postprocess_class__', None)
        if postprocessor is not None:
            postprocessor(returned, classdict)
        return returned

def renumerate(collection):
    for index in xrange(len(collection) - 1, -1, -1):
        yield (index, collection[index])

def flatten(seq, un_flattenable_types=(types.StringType,)):
    """Flattens any sequence, no matter it's type or depth.
    >>> list(flatten([1,2,'a']))
    [1, 2, 'a']
    >>> def a():
    ...     yield 7
    ...     yield 8
    >>> list(flatten([((1, ), ), [2], [[[[[3]]]]], [[4]], [[5], 6], a(), 9, 'a']))
    [1, 2, 3, 4, 5, 6, 7, 8, 9, 'a']
    """
    try:
        for un_flattenable_type in un_flattenable_types:
            if isinstance(seq, un_flattenable_types):
                raise TypeError()
        seq_iter = iter(seq)
    except TypeError:
        yield seq
    else:
        for i in seq:
            for j in flatten(i):
                yield j
