from __future__ import with_statement
import functools
import threading
import contextlib
import copy
import sys
import os

from xiv.dtypes import Bunch

class _ContextLocalInfo(threading.local):
    inside = False

class Context(object):
    """
    Handy context manager class

    >>> def enter():
    ...     print "ENTER"
    >>> def leave():
    ...     print "LEAVE"
    >>> with Context(reentrant=False, before=enter, after=leave) as c:
    ...     print "INSIDE"
    ENTER
    INSIDE
    LEAVE

    
    Reentrancy protection:
    >>> PASS = lambda:None
    >>> c = Context(reentrant=False, before=PASS, after=PASS)
    >>> with c: #doctest: +IGNORE_EXCEPTION_DETAIL
    ...     with c:
    ...         pass
    Traceback (most recent call last):
     ...
    RuntimeError: ...
    """
    def __init__(self, reentrant=False, before=None, after=None):
        super(Context, self).__init__()
        self._reentrant = reentrant
        self._local     = _ContextLocalInfo()
        self._before=before
        self._after=after

    def enter(self):
        if self._before:
            return self._before()
        raise NotImplementedError()
    def __call__(self, func):
        @functools.wraps(func)
        def new_func(*args, **kwargs):
            with self:
                return func(*args, **kwargs)
        return new_func
    def __add__(self, other):
        return nested(self, other)
    def __radd__(self, other):
        return nested(other, self)
    def __enter__(self):
        if self._local.inside and not self._reentrant:
            raise RuntimeError("Cannot re-enter context")
        returned = self.enter()
        self._local.inside = True
        return returned
    def exit(self, type, value, traceback):
        if self._after:
            return self._after()
        raise NotImplementedError()
    def __exit__(self, type, value, traceback):
        returned = self.exit(type, value, traceback)
        self._local.inside = False
        return returned
    
    
        

def nested(*contexts):
    """
    Flexible nested contexts

    >>> @contextlib.contextmanager
    ... def f(name):
    ...     print "->", name
    ...     yield name
    ...     print "<-", name
    
    >>> with nested(f("a"), f("b"), f("c")) as result:
    ...    print result
    -> a
    -> b
    -> c
    ['a', 'b', 'c']
    <- c
    <- b
    <- a
    >>> with nested(f(x) for x in "abc"):
    ...    print result
    -> a
    -> b
    -> c
    ['a', 'b', 'c']
    <- c
    <- b
    <- a
    >>> with nested([f(x) for x in "abc"]):
    ...    print result
    -> a
    -> b
    -> c
    ['a', 'b', 'c']
    <- c
    <- b
    <- a
    """
    from types import GeneratorType
    context_list = []
    for context in contexts:
        if type(context) in (tuple, list, GeneratorType):
            context_list.extend(context)
        else:
            context_list.append(context)
    return contextlib.nested(*context_list)

@contextlib.contextmanager
def ignore(*exceptions):
    """Ignore the given exceptions.
    >>> with ignore(TypeError, ValueError):
    ...   a, b = 1, 2, 3
    """
    try:
        yield None
    except exceptions:
        pass

@contextlib.contextmanager
def restore_path():
    """Context manager that preserves sys.path.
    >>> b4 = sys.path
    >>> b5 = os.getcwd()
    >>> with restore_path():
    ...   sys.path.append('/path/to/nowhere')
    ...   os.chdir('/')
    ... 
    >>> sys.path is b4
    >>> os.getcwd() == b5
    True
    """
    orig_path = sys.path
    orig_cwd  = os.getcwd()
    sys.path  = copy.copy(sys.path)
    try:
        yield None
    finally:
        sys.path = orig_path
        os.chdir(orig_cwd)

@contextlib.contextmanager
def restore_modules():
    """Context manager that preserves sys.modules.
    >>> b4 = sys.modules
    >>> with restore_modules():
    ...   sys.modules['blah'] = None
    ... 
    >>> sys.modules is b4
    True
    """
    orig_modules = sys.modules
    sys.modules  = copy.copy(sys.modules)
    try:
        yield None
    finally:
        sys.modules = orig_modules

def restore_env():
    """Nested context manager that restores sys.path and sys.modules."""
    return contextlib.nested(restore_path(), restore_modules())
