class Predicate(object):
    def applies(self, obj):
        raise NotImplementedError()
    def __eq__(self, obj):
        return self.applies(obj)
    def __ne__(self, obj):
        return not self == obj
    def __call__(self, obj):
        return self.applies(obj)
    def __str__(self):
        return self.describe()
    def __and__(self, other):
        return And(self, other)
    def __or__(self, other):
        return Or(self, other)
    def __repr__(self):
        return "<%s>" % self
    def describe(self, variable="X"):
        return self._describe(variable) + "?"
    def _describe(self, variable):
        raise NotImplementedError()
        

class FunctionPredicate(Predicate):
    def __init__(self, func):
        super(FunctionPredicate, self).__init__()
        self.func = func
    def applies(self, obj):
        if isinstance(obj, FunctionPredicate):
            return obj.func is self.func
        else:
            return self.func(obj)
    def _describe(self, variable):
        return "%s(%s)" % (self.func,variable)

class Equality(Predicate):
    def __init__(self, value):
        super(Equality, self).__init__()
        self.value = value
    def applies(self, obj):
        if isinstance(obj, Equality):
            return obj.value == self.value
        else:
            return obj == self.value
    def _describe(self, variable):
        return "%s==%s" % (variable, str(self.value))

class Or(Predicate):
    def __init__(self, *preds):
        super(Or, self).__init__()
        self.preds = map(MAKE_PREDICATE, preds)
    def applies(self, obj):
        return any(p == obj for p in self.preds)
    def _describe(self, variable):
        return " OR ".join("(%s)" % pred._describe(variable) for pred in self.preds)

class And(Predicate):
    def __init__(self, *preds):
        super(And, self).__init__()
        self.preds = map(MAKE_PREDICATE, preds)
    def applies(self, obj):
        return all(p == obj for p in self.preds)
    def _describe(self, variable):
        return " AND ".join("(%s)" % pred._describe(variable) for pred in self.preds)
class Not(Predicate):
    def __init__(self, pred):
        super(Not, self).__init__()
        self.pred = MAKE_PREDICATE(pred)
    def applies(self, obj):
        return not self.pred == obj
    def _describe(self, variable):
        return "NOT(%s)" % (self.pred._describe(variable),)
    
    
class _Dummy(Predicate):
    def __init__(self, retval, description = ""):
        self.retval = retval
        self.description = description
    def applies(self, other):
        return self.retval
    def _describe(self, variable):
        return self.description
IGNORE = _Dummy(True, "ANYTHING")
FAIL = _Dummy(False, "NOTHING")

class RecorderPredicate(Predicate):
    def __init__(self, recorder):
        import recorder as recorder_module
        assert type(recorder) is recorder_module.Recorder
        self.recorder = recorder
    def applies(self, obj):
        return bool(self.recorder._apply(obj))
    def _describe(self, variable):
        return str(self.recorder)
    
def MAKE_PREDICATE(expr):
    """
    common utility for making various expressions into predicates
    """
    from common import recorder
    if type(expr) is recorder.Recorder:
        return RecorderPredicate(expr)
    elif isinstance(expr, Predicate):
        return expr
    elif callable(expr):
        return FunctionPredicate(expr)    
    else:
        return Equality(expr)
P = MAKE_PREDICATE
