"""SSH wrapper module which lets the user run an SSH command on a remote host, implementng the SSH class.

The SSH class may raise LocalSSHError during execution of an SSH command, which means the command failed locally (ssh binary not found, illegal ssh options passed, etc). Remote command failures are relayed verbatim as results of the execution (status, stdout and stderr).

The SSH class also  has an instantiation test, which run_commands '/bin/true' by default on the remote host and refuses instantiation if that fails."""

from subprocess import PIPE
import subprocess
import logging
import os
import types

SSH_PATH = '/usr/bin/ssh'
SCP_PATH = '/usr/bin/scp'
DEFAULT_LOGLEVEL=logging.INFO

class BaseSSHException(Exception):
	pass

class LocalSSHError(BaseSSHException):
	"""Thrown when the ssh process returned exit status 255. This means no
	session was actually established successfully, and the remote command
	was not done."""
	pass
		
class CopyFailedError(BaseSSHException):
	"""Thrown when the scp process returned nonzero exit status 255."""

class CantInstantiateError(BaseSSHException):
	"""Thrown when an SSH instance with verified connectivity was requested,
	but a basic SSH command couldn't be run_commandd without a password."""

class SSH(object):
	def __init__(self, host, user=None, default_ssh_options={"StrictHostKeyChecking":"no","PasswordAuthentication":"no"}, \
		ssh_path=None, scp_path=None, verify_connectivity=True, logger=None, \
		verify_connectivity_command=None):
		"""Initializes ssh/scp binary path, username, host, options and logging
		object. Also, performs an optional connectivity sanity test by running
		'echo' command remotely."""
		self.ssh_path, self.scp_path = self.__detect_paths(ssh_path, scp_path)
		self.host = host
		self.user = self.__detect_user(user)
		self.options = default_ssh_options 
		self.logger = self.__detect_logger(logger)
		if verify_connectivity is True:
			self.__check_no_password_connectivity(verify_connectivity_command)

	def __detect_paths(self, ssh_paths, scp_paths):
		"""Placeholder function for currently unimplemented ability to detect
		the ssh/scp binaries in a list of paths."""
		if ssh_paths is None:
			ssh_path = SSH_PATH
		if scp_paths is None:
			scp_path = SCP_PATH
		return ssh_path, scp_path

	def __detect_user(self, user=None):
		"""If no user is provided as an argument, returns the currently running user."""
		if user is None:
			user = os.getlogin()
		return user

	def __detect_logger(self, logger):
		"""If no logger is provided as an arguments, returns a basically
		configured logger, for 'hassle-free' instantiation. Otherwise return the
		same logger as provided."""
		if logger is None:
			logging.basicConfig()
			logger = logging.getLogger('ssh')
			logger.setLevel(DEFAULT_LOGLEVEL)
			logger.debug('Using default logger with loglevel %s' \
				% DEFAULT_LOGLEVEL)
		return logger

	def __check_no_password_connectivity(self, command=None):
		"""Attempts to run some command, making sure no password authentication
		works. Note that following commands won't necessarily be run with
		the PasswordAuthentication='no' option. Also, this function doesn't
		honor default_ssh_options - should be fixed.""" # TODO
		self.logger.debug('checking passwordless connectivity')
		ssh_options = dict(PasswordAuthentication='no',StrictHostKeyChecking='no')
		if command is None:
			command = 'echo'
		try :
			self.run_command(command, options=ssh_options)
		except LocalSSHError , e :
			raise CantInstantiateError('failed key based authentication\n' + str(e))
	
	def run_critical_command(self,command,arguments=None,options=None):
		"""
		executes a remote command, raises and exception if execution retun code is non zero
		"""
		status,stdout,stderr = self.run_command(command,arguments,options)
		if status != 0:
			self.logger.debug('remote critical command failed')
			raise LocalSSHError('%s' % stderr)
		return status,stdout,stderr



	def run_command(self, command, arguments=None, options=None):
		"""executes a remote command using local ssh execution, and raises an
		exception if execution returns 255 (this is ssh's signal that the remote
		command was not run_commandd due to some local problem)."""
		self.logger.debug("running command:%s %s" % (command,arguments))
		results = self.__local_run_command(command, arguments, options)
		status, stdout, stderr = results
		if status == 255:
			self.logger.debug('remote command not run_commandd')
			raise LocalSSHError('%s' % stderr)
		self.logger.debug('succesfuly executed remote command:%s %s' % (command,arguments))
		return status, stdout, stderr

	def __local_run_command(self, command, arguments=None, options=None, ssh=True):
		"""Constructs the commandline and run_commands a command with arguments.
		Returns status, stdout, stderr."""
		if options is None:
			options = self.options
		if type(arguments) is types.StringType:
			arguments = [arguments]

		if ssh:	# ssh
			local_command = self.__make_ssh_invocation_list(options)
			remote_command = self.__make_remote_command_list(command, arguments)
			args = local_command + remote_command
		else:		# scp
			local_command = self.__make_scp_invocation_list(options)
			file_list = self.__make_scp_file_list(arguments)
			args = local_command + file_list

		self.logger.debug("running command:%s",args)
		p = subprocess.Popen(args=args, stdout=PIPE, stderr=PIPE)

		stdout, stderr = p.communicate()
		self.logger.debug('received status %d' % p.returncode)
		self.logger.debug('received output %r' % stdout)
		self.logger.debug('received error %r' % stderr)
		return p.returncode, stdout, stderr

	def __make_ssh_invocation_list(self, options):
		"""Returns a string with the path to ssh and ssh options."""
		invocation_list = [self.ssh_path] + self.__make_option_list(options) + \
			[self.__make_ssh_style_host_and_user()]
		self.logger.debug('local command %r' % invocation_list)
		return invocation_list

	def __make_scp_invocation_list(self, options):
		invocation_list = [self.scp_path] + self.__make_option_list(options)
		self.logger.debug('local command %r' % invocation_list)
		return invocation_list

	def __make_ssh_style_host_and_user(self):
		"""user@host"""
		return self.user + '@' + self.host

	def __make_remote_command_list(self, command, arguments):
		if arguments is None:
			results = [command]
		else:
			results = [command] + arguments
		self.logger.debug('remote command %r' % results)
		return results

	def __make_scp_file_list(self, arguments):
		self.logger.debug('copying %s to %s' % (arguments[0], arguments[1]))
		return [arguments[0], arguments[1]]

	def __make_option_list(self, options):
		"""Translates the options dictionary into a list of options, each
		matching SSH's format."""
		option_list = []
		for key in options:
			option_list.append('-o %s=%s' % (key, options[key]))
		return option_list

	def get_file(self, remote, local, options=None):
		"""Copies a file using scp from the remote host to the local host."""
		arguments = [self.__make_ssh_style_host_and_user()+':'+remote, local]
		self.__scp(arguments, options)

	def put_file(self, local, remote, options=None):
		"""Copies a file using scp from the local host to the remote host."""
		arguments = [local, self.__make_ssh_style_host_and_user()+':'+remote]
		self.__scp(arguments, options)

	def __scp(self, arguments, options):
		"""Copies files across computers using local scp execution, and raises an
		exception if execution returns non-zero (copy failed)."""
		command = None
		results = self.__local_run_command(command, arguments, options, ssh=False)
		status, stdout, stderr = results
		if status != 0:
			self.logger.error('file transfer failed')
			raise CopyFailedError('%s' % stderr)
		self.logger.info('file transfer ok')
		return
