from __future__ import with_statement

from cStringIO import StringIO
from contextlib import contextmanager
from contextlib import nested

class Formatter(object):
    INDENTATION_STR = "\t"
    def __init__(self, output):
        super(Formatter, self).__init__()
        self._output = output
        self._stack  = []
        self._indentation = ""
        self._newLine = True
        self._virgin  = True

    def _push(self):
        self._stack.append(self._output)
        self._output = StringIO()

    def _pop(self):
        returned = self._output.getvalue()
        self._output = self._stack.pop()
        return returned
        
    
    def write(self, data):
        if data:
            self._virgin = False
        new_line = data.endswith("\n")
        if new_line:
            data = data[:-1]
        data = data.split("\n")
        for index, line in enumerate(data):
            if self._newLine:
                self._output.write(self._indentation)
            self._output.write(line.replace("\n", ""))
            if index != len(data) - 1 or new_line:
                self._output.write("\n")
                self._newLine = True
            else:
                self._newLine = False       

    @contextmanager
    def _indented(self, length=1):
        self._indentation += self.INDENTATION_STR
        yield None
        self._indentation = self._indentation[:-len(self.INDENTATION_STR)]
    @contextmanager
    def _wrapped(self, begin, end):
        self.write(begin)
        yield None
        self.write(end)
        
class CFormatter(Formatter):
    """
    >>> output = StringIO()
    >>> f = CFormatter(output)
    >>> f.INDENTATION_STR = "->"
    >>> with f.block("if (0)"):
    ...     print >>f, "return 0;"
    >>> print output.getvalue() #doctest: +NORMALIZE_WHITESPACE
    if (0) {
    ->return 0;
    }
    """    
    def block(self, text=None):
        if text is not None:
            self.write(text)
            self.write(" ")
        return nested(self._wrapped("{\n", "}\n"), self._indented())
                      
class XMLFormatter(Formatter):
    """
    >>> output = StringIO()
    >>> f = XMLFormatter(output)
    >>> f.INDENTATION_STR = "->"
    >>> with f.tag("html"):
    ...     with f.tag("a", href="bla"):
    ...         with f.tag("empty"):
    ...             pass
    ...         print >>f, "hello"
    >>> print output.getvalue() #doctest: +NORMALIZE_WHITESPACE
    <html>
    -><a href="bla">
    ->-><empty/>
    ->->hello
    -></a>
    </html>
    
    """
    @contextmanager
    def tag(self, name, **attrs):
        self._push()
        yield None
        written = self._pop()
        self.write("<")
        self.write(name)
        for attr, value in attrs.iteritems():
            self.write(" %s=\"%s\"" % (attr, value))
        if written:
            self.write(">\n")
            with self._indented():
                self.write(written)
            self.write("</%s>" % name)
        else:
            self.write("/>")
        self.write("\n")
            
        
