import sys
import cx_Oracle
import os
import string
import subprocess
from xiv.pyutils import wait
import exceptions
from xiv_xsf.appcontrol.dbcontrol.oracle import StartupMode, ShutdownMode, OracleStatus

def safe_db_activity(exception_type, connection_mode = cx_Oracle.SYSDBA):
    def decorator(func):
        def callable(self, *args, **kwargs):
            try:
                os.environ["ORACLE_SID"] = self.sid
                connection = cx_Oracle.connect("/", mode = connection_mode)
                return func(self, connection, *args, **kwargs)
            except cx_Oracle.Error, e:
                newx = exception_type(str(e))
                newx.original_exception = e
                raise newx
        callable.__name__ = func.__name__
        return callable
    return decorator


class Instance(object):
    def __init__(self, sid):
        self.sid = sid

    def startup(self, startup_mode = StartupMode.OPEN):
        if hasattr(cx_Oracle, 'PRELIM_AUTH'):
            return self.oracle10_startup(startup_mode)
        else:
            return self.oracle9_startup(startup_mode)

    def shutdown(self, shutdown_mode = ShutdownMode.IMMEDIATE):
        if hasattr(cx_Oracle, 'PRELIM_AUTH'):
            return self.oracle10_shutdown(shutdown_mode)
        else:
            return self.oracle9_shutdown(shutdown_mode)

    def oracle10_startup(self, startup_mode = StartupMode.OPEN):
        try:
            os.environ["ORACLE_SID"] = self.sid
            # the connection must be in PRELIM_AUTH mode
            connection = cx_Oracle.connect("/", mode = cx_Oracle.SYSDBA | cx_Oracle.PRELIM_AUTH)
            connection.startup()
            # the following statements must be issued in normal SYSDBA mode
            connection = cx_Oracle.connect("/",
                    mode = cx_Oracle.SYSDBA)
            cursor = connection.cursor()
            if startup_mode == StartupMode.OPEN:
                cursor.execute("alter database mount")
                cursor.execute("alter database open")
            elif startup_mode == StartupMode.MOUNT:
                cursor.execute("alter database mount")
            elif startup_mode == StartupMode.NOMOUNT:
                pass
            else:
                raise exceptions.Startup("No such startup mode: %s" % (str(startup_mode)))
        except cx_Oracle.Error, e:
            raise exceptions.Startup(str(e))

    @safe_db_activity(exceptions.Shutdown)
    def oracle10_shutdown(self, connection, shutdown_mode = ShutdownMode.IMMEDIATE):
        # first shutdown() call must specify the mode
        if shutdown_mode == ShutdownMode.IMMEDIATE:
            connection.shutdown(mode = cx_Oracle.DBSHUTDOWN_IMMEDIATE)
            # now close and dismount the database
            cursor = connection.cursor()
            cursor.execute("alter database close normal")
            cursor.execute("alter database dismount")
            # perform the final shutdown call
            connection.shutdown(mode = cx_Oracle.DBSHUTDOWN_FINAL)
        elif shutdown_mode == ShutdownMode.ABORT:
            connection.shutdown(mode = cx_Oracle.DBSHUTDOWN_ABORT)
        else:
            raise exceptions.Shutdown("No such shutdown mode: %s" % (str(shutdown_mode)))

    def oracle9_startup(self, startup_mode = StartupMode.OPEN):
        if startup_mode == StartupMode.OPEN:
            self.run_in_sqlplus(['startup'])
        elif startup_mode == StartupMode.MOUNT:
            self.run_in_sqlplus(['startup mount'])
        elif startup_mode == StartupMode.NOMOUNT:
            self.run_in_sqlplus(['startup nomount'])
        else:
            raise exceptions.Startup("No such startup mode: %s" % (str(startup_mode)))

    def oracle9_shutdown(self, shutdown_mode = ShutdownMode.IMMEDIATE):
        if shutdown_mode == ShutdownMode.IMMEDIATE:
            self.run_in_sqlplus(['shutdown immediate'])
        elif shutdown_mode == ShutdownMode.ABORT:
            self.run_in_sqlplus(['shutdown abort'])
        else:
            raise exceptions.Shutdown("No such shutdown mode: %s" % (str(shutdown_mode)))


    @safe_db_activity(exceptions.BeginBackup)
    def begin_backup(self, connection, tablespaces = None):
        cursor = connection.cursor()
        # if there is no request for specific tablespaces, add all from database.
        if tablespaces is None:
            cursor.execute("select TABLESPACE_NAME \
                    from dba_tablespaces \
                    where CONTENTS<>'TEMPORARY' and STATUS <> 'READ ONLY'")
            tablespace_list = [record[0] for record in cursor]
        else:
            tablespace_list = tablespaces
            
        active_tablespaces = self.get_tablespaces_in_backup_mode()
        # if there are tablespaces in the list which are in backup mode, raise exception
        e = []
        for tablespace in tablespace_list:
            if tablespace in active_tablespaces:
                e.append(tablespace)
        if len(e) > 0:
            raise exceptions.TablespacesAlreadyInBackupMode(e)
        # begin backup mode for tablespaces in list
        for tablespace in tablespace_list:
            cursor.execute("alter tablespace %s begin backup" % (tablespace))

    @safe_db_activity(exceptions.EndBackup)
    def end_backup(self, connection, tablespaces = None):
        cursor = connection.cursor()
        # if there is no request for specific tablespaces, add all from database.
        if tablespaces is None:
            cursor.execute("select TABLESPACE_NAME \
                    from dba_tablespaces \
                    where CONTENTS<>'TEMPORARY' and STATUS <> 'READ ONLY'")
            tablespace_list = [record[0] for record in cursor]
        else:
            tablespace_list = tablespaces
                
        inactive_tablespaces = self.get_tablespaces_not_in_backup_mode()
        # if there are tablespaces in the list which are not in backup mode, raise exception
        e = []
        for tablespace in tablespace_list:
            if tablespace in inactive_tablespaces:
                e.append(tablespace)
        if len(e) > 0:
            raise exceptions.TablespacesNotInBackupModeException(e)
        # end backup mode for tablespaces in list
        for tablespace in tablespace_list:
            cursor.execute("alter tablespace %s end backup" % (tablespace))

    @safe_db_activity(exceptions.Query)
    def get_tablespaces_in_backup_mode(self, connection):
        cursor = connection.cursor()
        # get the list of tablespaces which are already in backup mode
        cursor.execute("select distinct D.TABLESPACE_NAME \
                from dba_data_files D, v$backup B \
                where B.FILE# = D.FILE_ID and B.STATUS = 'ACTIVE'")
        # if there are tablespaces in backup mode, raise exception
        active_tablespaces = [record[0] for record in cursor]
        return active_tablespaces

    @safe_db_activity(exceptions.Query)
    def get_tablespaces_not_in_backup_mode(self, connection):
        cursor = connection.cursor()
        # get the list of tablespaces which are already in backup mode
        cursor.execute("select distinct D.TABLESPACE_NAME \
                from dba_data_files D, v$backup B \
                where B.FILE# = D.FILE_ID and B.STATUS <> 'ACTIVE'")
        # if there are tablespaces in backup mode, raise exception
        inactive_tablespaces = [record[0] for record in cursor]
        return inactive_tablespaces

    # get the first scn # of the active redo log file
    @safe_db_activity(exceptions.Query)
    def get_current_redo_log_first_change_number(self, connection):
        cursor = connection.cursor()
        # get the first change# of the current log file
        cursor.execute("select first_change# from v$log where status = 'CURRENT'")
        scn = cursor.fetchone()
        return scn[0]

    # get the current scn #
    @safe_db_activity(exceptions.Query)
    def get_current_change_number(self, connection):
        cursor = connection.cursor()
        # get the first change# of the current log file
        cursor.execute("select checkpoint_change# from v$database")
        scn = cursor.fetchone()
        return scn[0]

    # get archive logs path names which contains the scn range (including)
    @safe_db_activity(exceptions.Query)
    def get_archivedlog_paths_between_scns(self, connection, start_scn, end_scn = None):
        cursor = connection.cursor()
        if end_scn is not None:
            cursor.execute("select name \
                    from v$archived_log \
                    where ( FIRST_CHANGE# <= %s and NEXT_CHANGE# > %s ) \
                    or ( FIRST_CHANGE# >= %s and FIRST_CHANGE# <= %s ) \
                    " % (start_scn, start_scn, start_scn, end_scn) )
        else:
            cursor.execute("select name \
                    from v$archived_log \
                    where ( FIRST_CHANGE# <= %s and NEXT_CHANGE# > %s ) \
                    or ( FIRST_CHANGE# >= %s ) \
                    " % (start_scn, start_scn, start_scn) )
        archived_logs = [record[0] for record in cursor]
        return archived_logs

    @safe_db_activity(exceptions.Query)
    def get_datafiles_paths(self, connection):
        cursor = connection.cursor()
        cursor.execute("select name from v$datafile")
        datafiles = [record[0] for record in cursor]
        return datafiles

    @safe_db_activity(exceptions.Query)
    def get_non_temporary_datafiles_paths(self, connection):
        cursor = connection.cursor()
        cursor.execute("select d.file_name \
                    from dba_data_files d, dba_tablespaces t \
                    where d.tablespace_name = t.tablespace_name and t.contents <> 'TEMPORARY'")
        datafiles = [record[0] for record in cursor]
        return datafiles

    @safe_db_activity(exceptions.Query)
    def get_redolog_paths(self, connection):
        cursor = connection.cursor()
        cursor.execute("select member from v$logfile")
        redologs = [record[0] for record in cursor]
        return redologs

    @safe_db_activity(exceptions.Query)
    def get_archivelog_paths(self, connection):
        cursor = connection.cursor()
        cursor.execute("select destination from v$archive_dest where status='VALID'")
        archivelogs = [record[0] for record in cursor]
        return archivelogs

    @safe_db_activity(exceptions.Operational)
    def switch_logfile(self, connection):
        cursor = connection.cursor()
        cursor.execute("alter system switch logfile")

    @safe_db_activity(exceptions.Operational)
    def system_checkpoint(self, connection):
        cursor = connection.cursor()
        cursor.execute("alter system checkpoint")

    @safe_db_activity(exceptions.Operational)
    def archivelog_current(self, connection):
        cursor = connection.cursor()
        cursor.execute("alter system archive log current")

    @safe_db_activity(exceptions.Operational)
    def backup_controlfile(self, connection, file_path, trace_backup = False):
        cursor = connection.cursor()
        if trace_backup:
            cursor.execute("alter database backup controlfile to trace as '%s' reuse" % (file_path) )
        else:
            cursor.execute("alter database backup controlfile to '%s'" % (file_path) )

    def execute_recovery_script(self, script):
        return self.run_in_sqlplus(script)

    def get_instance_status(self):
        try:
            os.environ["ORACLE_SID"] = self.sid
            connection = cx_Oracle.connect("/", mode = cx_Oracle.SYSDBA)
            cursor = connection.cursor()
            cursor.execute("select status from v$instance")
            status = cursor.fetchone()[0]
            if status == "STARTED":
                return OracleStatus.NOTMOUNTED
            elif status == "MOUNTED":
                return OracleStatus.MOUNTED
            elif status == "OPEN":
                return OracleStatus.OPEN
            else:
                return OracleStatus.UNKNOWN
        except cx_Oracle.DatabaseError:
            return OracleStatus.DOWN

    def run_in_sqlplus(self, sql_script_list):
        """
        running sql scripts in sqlplus command line
        sql_script_list - should be a list of sql commands in the right syntax. if ';' is expected at end of line it should be there.
        """
        try:
            SQLPLUS_TIMEOUT = 300
            SQLPLUS_INTERVAL = 5
            os.environ["ORACLE_SID"] = self.sid
            command = "sqlplus -s / as sysdba"
            proc = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE)
            # This is doing some troubles, for now, no error reporting :(            
            #proc.stdin.write('whenever sqlerror exit failure\n')
            for line in sql_script_list:
                proc.stdin.write(line + '\n')
            proc.stdin.write('quit' + '\n')
            wait(lambda: proc.poll() is not None, 
                timeout=SQLPLUS_TIMEOUT,
                interval=SQLPLUS_INTERVAL, 
                exception=exceptions.SqlPlusTimeOutError()
            )

            if proc.returncode != 0:
                raise exceptions.Operational(proc.stdout.read() + proc.stderr.read())

        except (OSError, ValueError, subprocess.CalledProcessError), e:
            raise exceptions.Operational(str(e))

        return proc.stdout.read()

