import inspect
from xiv.dtypes import Bunch
import pickle
import zlib
import os

WriterStatus = Bunch(
        DOWN = 1,
        UP = 2,
        PARTIALLY_UP = 3,
        UNKNOWN = 4)


class WriterException(Exception):
    pass

def notimplemented(func):
    def callable(*args, **kwargs):
        raise NotImplementedError
    callable.__name__ = func.__name__
    callable.notimplemented = True
    return callable


class Writer(object):
    def start(self, start_mode):
        pass
    def stop(self, stop_mode):
        pass
    def enter_hot_backup_mode(self):
        pass
    def finish_hot_backup_mode(self):
        pass
    def cleanup_faulted_backup_mode(self):
        pass
    def recover_hot_backup(self, hot_backup_meta_data):
        pass
    def check_state(self):
        pass
    def used_paths(self):
        pass



for name, method in inspect.getmembers(Writer, inspect.ismethod):
    setattr(Writer, name, notimplemented(method))


class HotBackupMetaData(object):
    def __init__(self, files_dict = None, strings_dict = None, alternate_paths_dict = None):
        self.files_dict = files_dict
        self.strings_dict = strings_dict
        self.alternate_paths_dict = alternate_paths_dict
        self.packed = False
        self.packed_obj = None

    def pack(self):
        try:
            if self.packed:
                raise WriterException("Error: Cannot pack already packed data")
            self.packed_obj = Bunch(
                    files_dict = Bunch(),
                    files_original_path = Bunch(),
                    strings_dict = self.strings_dict,
                    alternate_paths_dict = self.alternate_paths_dict,
            )
            if self.files_dict is not None:
                for file_key in self.files_dict:
                    self.packed_obj.files_original_path[file_key] = self.files_dict[file_key]
                    self.packed_obj.files_dict[file_key] = zlib.compress(file(self.files_dict[file_key], 'rb').read(), zlib.Z_BEST_SPEED)
            self.packed = True
        except OSError, e:
            print "Error: Unable to pack meta data: %s" % str(e)
            raise WriterException("Error: Unable to pack meta data: %s" % str(e))

    def deploy(self, file_handle = None):
        try:
            if file_handle is not None:
                self.load(file_handle)
            if not self.packed:
                raise WriterException("Error: Cannot deploy unpacked meta data without file handle parameter")
            self.files_dict = Bunch()
            for file_key in self.packed_obj.files_original_path:
                self.files_dict[file_key] = self.packed_obj.files_original_path[file_key]
                if not os.path.isdir(os.path.dirname(self.packed_obj.files_original_path[file_key])):
                    os.makedirs(os.path.dirname(self.packed_obj.files_original_path[file_key]))
                newfile = file(self.packed_obj.files_original_path[file_key], 'wb')
                newfile.write(zlib.decompress(self.packed_obj.files_dict[file_key]))
                newfile.close()
            self.alternate_paths_dict = self.packed_obj.alternate_paths_dict
            self.strings_dict = self.packed_obj.strings_dict
            self.packed_obj = None
            self.packed = False
        except OSError, e:
            print "Error: Unable to deploy meta data: %s" % str(e)
            raise WriterException("Error: Unable to deploy meta data: %s" % str(e))

    def dump(self, file_handle):
        try:
            if not self.packed:
                self.pack()             
            pickle.dump(self.packed_obj, file_handle)
        except pickle.PickleError, e:
            print "Error: Unable to dump meta data: %s" % str(e)
            raise WriterException("Error: Unable to dump meta data: %s" % str(e))

    def load(self, file_handle):
        try:
            self.packed_obj = pickle.load(file_handle)
            self.packed = True
        except pickle.PickleError, e:
            print "Error: Unable to load meta data: %s" % str(e)
            raise WriterException("Error: Unable to load meta data: %s" % str(e))

    def relocate_files_base_dir(self, old_base_path, new_base_path):
        if not self.packed:
            raise WriterException("Error: Cannot convert paths on non-packed object")
        for file_key in self.packed_obj.files_original_path:
            self.packed_obj.files_original_path[file_key] = self.packed_obj.files_original_path[file_key].replace(
                    old_base_path,
                    new_base_path,
                    1
            )

    def relocate_alternate_paths_base_dir(self, old_base_path, new_base_path):
        if not self.packed:
            raise WriterException("Error: Cannot convert paths on non-packed object")
        for file_key in self.packed_obj.alternate_paths_dict:
            self.packed_obj.alternate_paths_dict[file_key] = self.packed_obj.alternate_paths_dict[file_key].replace(
                    old_base_path,
                    new_base_path,
                    1
            )

def instanciate_metadata_from_packed_object(meta_data):
    new_meta_data = HotBackupMetaData()
    new_meta_data.packed_obj = meta_data.packed_obj
    new_meta_data.packed = True
    return new_meta_data

