import os
import time
import tempfile
import logging
from xiv_xsf.appcontrol.dbcontrol import oracle
from xiv_xsf import appwriters
from elementtree import ElementTree

class OracleWriterException(Exception):
    pass
class Operational(OracleWriterException):
    pass
class InvalidStatusForOperation(OracleWriterException):
    pass

class OracleWriter(appwriters.Writer):
    DEFAULT_LOGGER_NAME = "OracleWriter"
    DEFAULT_LOG_LEVEL = logging.DEBUG

    def __init__(self, appname, tmpdir = tempfile.gettempdir(), logger_name = None):
        super(OracleWriter,self).__init__()
        if logger_name is None:
            logger_name = self.DEFAULT_LOGGER_NAME
            logging.basicConfig()
        self.logger = logging.getLogger(logger_name)
        self.logger.setLevel(self.DEFAULT_LOG_LEVEL)
        
        self.sid = appname
        self.instance = oracle.operations.Instance(self.sid)
        self.start_scn = -1
        self.end_scn = -1
        self.temp_dir = os.path.join(tmpdir, time.strftime("%Y%m%d-%H%M%S", time.localtime()))
        try:
            # create temporary directory for files needed for recovery
            if not os.path.exists(self.temp_dir):
                os.makedirs(self.temp_dir)
            os.chmod(self.temp_dir, 0777)
        except OSError, e:
            self.logger.error( "Cannot create temp directory %s: %s" % (self.temp_dir, str(e)))
            raise Operational("Cannot create temp directory %s: %s" % (self.temp_dir, str(e)))
    
    def start(self, start_mode = oracle.StartupMode.OPEN):
        # Check database status before startup
        dbstatus = self.instance.get_instance_status()
        if dbstatus == oracle.OracleStatus.DOWN:
            try:
                self.instance.startup(startup_mode = start_mode)
            except oracle.exceptions.Startup, e:
                raise Operational("Cannot start instance %s" % (self.sid, str(e)))
        elif dbstatus == oracle.OracleStatus.UNKNOWN:
            raise Operational("Database is in Unknown Status")
        else:
            raise InvalidStatusForOperation("Cannot start instance %s since it is already up" % (self.sid))

    def stop(self, stop_mode = oracle.ShutdownMode.IMMEDIATE):
        # check database status before shutdown
        dbstatus = self.instance.get_instance_status()
        if dbstatus == oracle.OracleStatus.DOWN:
            ## should be replaced with logger
            self.logger.warning("Warning: Database %s is already down, ignoring request" % (self.sid))
        elif dbstatus == oracle.OracleStatus.UNKNOWN:
            raise Operational("Database is in Unknown Status")
        elif dbstatus != oracle.OracleStatus.OPEN and stop_mode != oracle.ShutdownMode.ABORT:
            raise InvalidStatusForOperation("Shutdown abort must be given for not opened instance %s" % (self.sid))
        else:
            try:
                self.instance.shutdown(shutdown_mode = stop_mode)
            except oracle.exceptions.Shutdown, e:
                raise Operational("Cannot stop instance %s" % (self.sid, str(e)))

    def enter_hot_backup_mode(self):
        try:
            # switch log file and get first change number, then begin backup mode
            self.instance.archivelog_current()
            self.instance.switch_logfile()
            self.start_scn = self.instance.get_current_redo_log_first_change_number()
            self.instance.begin_backup()
        except oracle.exceptions.Operational, e:
            self.logger.error("Error while trying to perform oracle activity on sid %s: %s" % (self.sid, str(e)))
            raise Operational("Error while trying to perform oracle activity on sid %s: %s" % (self.sid, str(e)))
        except oracle.exceptions.BeginBackup, e:
            self.logger.error("Error while trying to enter backup mode on sid %s: %s" % (self.sid, str(e)))
            raise Operational("Error while trying to enter backup mode on sid %s: %s" % (self.sid, str(e)))

    def finish_hot_backup_mode(self):
        try:
            # get out of backup mode
            self.instance.end_backup()
            # do post backup activities and return token for restore
            #self.end_scn = self.instance.get_current_change_number()
            self.instance.switch_logfile()
            self.instance.archivelog_current()
            self.end_scn = self.instance.get_current_redo_log_first_change_number() - 1
            # backup control file to trace
            trc_bck_ctl_file_name = os.path.join(self.temp_dir, "controlfile.trc")
            self.instance.backup_controlfile(trc_bck_ctl_file_name, True)
            # get archive logs between scns required for recovery
            archive_files = self.instance.get_archivedlog_paths_between_scns(self.start_scn, self.end_scn)
            # Build recovery script
            recovery_script = _create_recovery_script_to_scn_(trc_bck_ctl_file_name, self.end_scn, archive_files)
            # create dict of archive files required for restore
            files_dict = {}
            for arc in archive_files:
                files_dict[arc] = arc
            # keep the recovery script in dict of strings for restore
            strings_dict = { 'recovery_script': recovery_script }
            # keep other used paths in the alternate_paths_dict
            alternate_paths_dict = {}
            for redo in self.instance.get_redolog_paths():
                alternate_paths_dict[redo] = redo
            for data in self.instance.get_datafiles_paths():
                alternate_paths_dict[data] = data
            self.delete_temporary_directory()
            # return hot backup meta data
            return appwriters.HotBackupMetaData(files_dict, strings_dict, alternate_paths_dict)
        
        except oracle.exceptions.TablespacesNotInBackupMode, e:
            self.logger.warning("Warning: Tablespaces were not in backup mode: %s" % (str(e)))
            self.logger.warning("Warning: Clone might be inrecoverable")
            raise InvalidStatusForOperation("Tablespaces were not in backup mode")
        except (oracle.exceptions.Operational, oracle.exceptions.Query), e:
            self.logger.error("Error while trying to perform operation on sid %s: %s" % (self.sid, str(e)))
            raise Operational("Error while trying to perform operation on sid %s: %s" % (self.sid, str(e)))

    def delete_temporary_directory(self):
        try:
            # create temporary directory for files needed for recovery
            if os.path.exists(self.temp_dir):
                for tmpfile in os.listdir(self.temp_dir):
                    os.remove(os.path.join(self.temp_dir, tmpfile))
                os.rmdir(self.temp_dir)
        except OSError, e:
            self.logger.warning( "Cannot delete temp directory %s: %s" % (self.temp_dir, str(e)))

    def cleanup_faulted_backup_mode(self):
        try:
            tablespaces = self.instance.get_tablespaces_in_backup_mode()
            self.instance.end_backup(tablespaces)
        except oracle.exceptions.Query, e:
            self.logger.error("Error while trying to get a list of tablespaces in backup mode on sid %s: %s" % (self.sid, str(e)))
            self.logger.error("Warning: Database may still be in backup mode!!!")
        except oracle.exceptions.EndBackup, e:
            self.logger.error("Error while trying to end backup mode on sid %s, tablespaces %s: %s" % (self.sid, str(tablespaces), str(e)))
            self.logger.error("Warning: Database may still be in backup mode!!!")

    def check_state(self):
        status = self.instance.get_instance_status()
        if status == oracle.OracleStatus.DOWN:
            return appwriters.WriterStatus.DOWN
        elif status == oracle.OracleStatus.OPEN:
            return appwriters.WriterStatus.UP
        elif status == oracle.OracleStatus.NOTMOUNTED or status == oracle.OracleStatus.MOUNTED:
            return appwriters.WriterStatus.PARTIALLY_UP
        else:
            return appwriters.WriterStatus.UNKNOWN

    def paths_required_for_hot_backup(self):
        return self.instance.get_non_temporary_datafiles_paths()

    def paths_required_for_cold_backup(self):
        return self.instance.get_datafiles_paths() + self.instance.get_redolog_paths()

    def recover_hot_backup(self, hot_backup_meta_data):
        paths_dict = hot_backup_meta_data.alternate_paths_dict
        paths_dict.update(hot_backup_meta_data.files_dict)
        # replace original files paths with paths from the meta data in the recovery script
        recovery_script = _update_paths_and_sid_in_recovery_script_(hot_backup_meta_data.strings_dict['recovery_script'],
                                    paths_dict, self.sid)
        # run recovery for instance
        try:
#            self.instance.startup(oracle.StartupMode.NOMOUNT)
            self.instance.execute_recovery_script(recovery_script)
        except oracle.exceptions.Operational, e:
            self.logger.error("Error while trying to run recovery on sid %s: %s" % (self.sid, str(e)))
            raise Operational("Error while trying to run recovery on sid %s: %s" % (self.sid, str(e)))


# for internal use
# FIXME - should be replaced with parsing the trace file into an object some day
def _create_recovery_script_to_scn_(source_trace_file, scn, archive_files):
    SECTION_START = " Set #2. RESETLOGS case"
    srcfile = file(source_trace_file, "rb")
    lines = map(str.rstrip, srcfile.xreadlines())
    new_lines = []
    lines.reverse()
    for line in lines:
        if line.endswith(SECTION_START):
            break
        elif line.startswith("--"):
            continue
        elif line.startswith("#"):
            continue
        elif line.strip() == "":
            continue
#        elif line.startswith("STARTUP NOMOUNT"):
#            continue
        elif line.strip() == "RECOVER DATABASE USING BACKUP CONTROLFILE":
            new_lines.extend(("RECOVER DATABASE UNTIL CHANGE %d USING BACKUP CONTROLFILE" % scn,
                "SET AUTORECOVERY ON"))

            """ This should apply once support of relocation of archived logs is done
            for arc in archive_files:
                new_lines.append("ALTER DATABASE REGISTER LOGFILE '%s';" % arc )
            """

        else:
            new_lines.append(line)
    new_lines.reverse()
    return new_lines

# for internal use
# replace original files paths with paths from the meta data in the recovery script
def _update_paths_and_sid_in_recovery_script_(recovery_script, path_dict, new_sid):
    new_script = []
    for line in recovery_script:
        # TODO Next section is for support changing path of oracle data files and archived log path from original
        # It it causing bad performance if u have many lines in the script. should be re-written :(

        """
        for original_path in path_dict:
            new_line = line.replace(original_path, path_dict[original_path])
            if new_line != line:
                break
        """
        if line.startswith("CREATE CONTROLFILE"):
            new_script.append('CREATE CONTROLFILE REUSE SET DATABASE "%s" RESETLOGS ARCHIVELOG' % new_sid)
        else:
            # The following line is commented from the same reason as above lines. the line after is replacing it.
            # new_script.append(new_line)
            new_script.append(line)
    return new_script


"""
class BackupConfig(object):
    def __init__(self, sid, config_directory):
        self.sid = sid
        xml_configuration_file = "%s/xiv-ora-instances.xml" % (config_directory)
        instances = self.__process_configuration_elements__(xml_configuration_file)
        self.backup_directory = instances[sid]['backup_directory']
        self.auto_discovery = instances[sid]['auto_discovery']
        self.config_directory = config_directory

    def __process_configuration_elements__(self, xml_configuration_file):
        instances = {}
        raw_file = file(xml_configuration_file, 'r').read()
        elements = ElementTree.fromstring(raw_file)
        for ora_inst_element in elements.findall('instance'):
            inst_name = ora_inst_element.find('sid').attrib['value']
            instances[inst_name] = {}
            auto_discovery = ora_inst_element.find('auto_discovery').attrib['value']
            instances[inst_name]['auto_discovery'] = auto_discovery
            backup_directory = ora_inst_element.find('backup_directory').attrib['value']
            instances[inst_name]['backup_directory'] = backup_directory
            return instances
"""
