from errors import CanvasException
from widget import Widget

__all__ = ['ExactPlacement', 'Grid', 'WidgetGeometry']

class WidgetGeometry(object):

    def __init__(self, y, x, height, width):

        object.__init__(self)
        self.y = y
        self.x = x
        self.height = height
        self.width = width

    def show(self):

        print '(%(y)d, %(x)d, %(height)d, %(width)d)' % self.__dict__ 


class GridPosition(object):

    def __init__(self, row, column, rowspan, columnspan, pady, padx, anchor, sizing_mode):

        object.__init__(self)
        self.row = row
        self.column = column
        self.rowspan = rowspan
        self.columnspan = columnspan
        self.pady = pady
        self.padx = padx
        self.anchor = anchor
        self.sizing_mode = sizing_mode


class GeometryManager(object):

    def __init__(self):
        object.__init__(self)
        self.widgets = {} 

    def remove(self, widget):
        del self.widgets[widget]

    def get_geometry(self, widget):
        return self.widgets[widget]

    def set_canvas(self, canvas):
        self.canvas = canvas


class ExactPlacement(GeometryManager):

    def place(self, widget, geometry):
        self.widgets[widget] = geometry 



class Grid(GeometryManager):

    CENTER = 0
    LEFT = 1
    DOWN = 2
    UP = 4
    RIGHT = 8

    BEST_FIT = 10
    FILL = 11
    MANUAL = 12

    def __init__(self, rows, columns):
        GeometryManager.__init__(self)
        self.rows = rows
        self.columns = columns
        self.row_start_places = [float(i) / rows for i in range(rows)] + [1.]
        self.column_start_places = [float(i) / columns for i in range(columns)] + [1.]
        self.grid = [[False for i in range(self.columns)] for j in range(self.rows)]
        
    def place(self, widget, row, column, rowspan=1, columnspan=1, pady=1, padx=1, anchor=CENTER, sizing_mode=BEST_FIT):
        if row + rowspan > self.rows or column + columnspan > self.columns:
            raise CanvasException("Grid index out of bounds")

        for i in range(row, row + rowspan):
            for j in range(column, column + columnspan):
                if self.grid[i][j]:
                    raise CanvasException("Grid position already taken")
                self.grid[i][j] = True

        self.widgets[widget] = GridPosition(row, column, rowspan, columnspan, pady, padx, anchor, sizing_mode)
         

    def set_row_sizes(self, numbers):
        self.row_start_places = self._rel_cum_sum(numbers)

    def set_col_sizes(self, numbers):
        self.col_start_places = self._rel_cum_sum(numbers)

    def _rel_cum_sum(self, numbers):

        s = float(sum(numbers))
        res = [0.0]
        for i in numbers:
            res.append(res[-1] + i / s)
        return res

    def get_geometry(self, widget):

        screen_height, screen_width = self.canvas.get_screen_size()        
        grid_pos = self.widgets[widget]
        box_y = int(screen_height * self.row_start_places[grid_pos.row]) + grid_pos.pady
        box_x = int(screen_width * self.column_start_places[grid_pos.column]) + grid_pos.padx
        box_height = int(screen_height * (self.row_start_places[grid_pos.row + grid_pos.rowspan] - self.row_start_places[grid_pos.row])) - 2 * grid_pos.pady
        box_width = int(screen_width * (self.column_start_places[grid_pos.column + grid_pos.columnspan] - self.column_start_places[grid_pos.column])) - 2 * grid_pos.padx

        widget_height, widget_width = widget.best_fit(max_height=box_height, max_width=box_width)
        
        if grid_pos.sizing_mode == Grid.FILL:
            widget_height = box_height
            widget_width = box_width

        elif grid_pos.sizing_mode == Grid.MANUAL:
            if widget['height'] == Widget.AUTO or widget['width'] == Widget.AUTO:
                raise CanvasError("For manual sizing both height and width must be specified")
            widget_height = min(widget['height'], box_height)
            widget_width = min(widget['width'], box_width)
   
        assert widget_height <= box_height
        assert widget_width <= box_width
        
        y = int(box_y + (box_height - widget_height) / 2)
        x = int(box_x + (box_width - widget_width) / 2)

        if grid_pos.anchor & Grid.LEFT:
            x = box_x
        elif grid_pos.anchor & Grid.RIGHT:
            x = box_x + box_width - widget_width
        if grid_pos.anchor & Grid.UP:
            y = box_y
        elif grid_pos.anchor & Grid.DOWN:
            y = box_y + box_height - widget_height
            
        geometry = WidgetGeometry(y, x, widget_height, widget_width)
        #WidgetGeometry(box_y, box_x, box_height, box_width).show()

        #geometry.show()
        return geometry


    def remove(self, widget):
        grid_pos = self.widgets[widget]
        for i in range(grid_pos.row, grid_pos.row + grid_pos.rowspan):
            for j in range(grid_pos.column, grid_pos.column + grid_pos.columnspan):
                self.grid[i][j] = False

        GeometryManager.remove(self, widget)

