#!/usr/bin/env python

class NonXivScsiIdError(Exception):
    """This exception is raised when the SCSI Unit Serial Number doesn't
    belong to an XIV device"""
    def __init__(self, scsi_id):
        self.scsi_id = scsi_id

    def __str__(self):
        return "Cannot recognize the Unit Serial Number '%s'" % self.scsi_id


class XivScsiId(object):
    """XivScsiId class represents an XIV LUN, as seen from the host.
    It's mostly about the SCSI Unit Serial Number, which encapsulates some
    XIV internal data: Machine Serial Number, XIV Volume Serial Number

    Main properties:
    machine_serial_number
    xiv_volume_serial_number
    """

    """i.e. "MN0001A015F" returns "MN0001A" """
    machine_serial_number = None

    """i.e. "MN0001A015F" returns "015F" """
    xiv_volume_serial_number = None

    @property
    def machine_system_id(self):
        """The suffix of machine_serial_number, in decimal form.
        EXAMPLE: "MN0001A" returns 26
        GenII only.
        """
        if (self.nextra_gen == 1):
            return None
        elif (self.nextra_gen == 2):
            return int(self.machine_serial_number[2:], 16)

    @property
    def machine_manufacturing_prefix(self):
        """The manufactoring prefix of machine_serial_number.
        EXAMPLE: "MN0001A" returns "MN"
        GenII only.
        """
        if (self.nextra_gen == 1):
            return None
        elif (self.nextra_gen == 2):
            return self.machine_serial_number[0:2]

    @property
    def xiv_volume_serial_number_decimal(self):
        """Converts xiv_volume_serial_number to a regular decimal int,
        different behaviors for different nextra_generations
        """
        if (self.nextra_gen == 1):
            return int(self.xiv_volume_serial_number)
        elif (self.nextra_gen == 2):
            return int(self.xiv_volume_serial_number, 16)

    def detect_nextra_gen(self):
        """Analyzes self.scsi_unit_serial_number, and decides which
        XIV Nextra generation it is. Currently based on length only."""
        if (len(self.scsi_unit_serial_number) == 12):
            return 1
        elif (len(self.scsi_unit_serial_number) == 11):
            return 2
        else:
            return -1

    def __init__(self, scsi_unit_serial_number):
        """__init__ gets the raw scsi_unit_serial_number (the one in VPD page
        0x80 or 0x83), and updates the following properties accordingly:
        machine_serial_number, xiv_volume_serial_number properties
        """
        self.scsi_unit_serial_number = scsi_unit_serial_number

        self.nextra_gen = self.detect_nextra_gen()

        if (self.nextra_gen == 2):
            self.machine_serial_number = scsi_unit_serial_number[0:7]
            self.xiv_volume_serial_number = scsi_unit_serial_number[7:11]
        elif (self.nextra_gen == 1):
            self.machine_serial_number = scsi_unit_serial_number[0:6]
            self.xiv_volume_serial_number = scsi_unit_serial_number[6:12]
        else:
            raise NonXivScsiIdError(scsi_unit_serial_number)
        return

    def __str__(self):
        return "Machine Serial Number: " + self.machine_serial_number + ", " +\
            "XIV Volume Serial Number: " + self.xiv_volume_serial_number
