import platform
import os
import subprocess
import re


if platform.system().lower() != 'aix':
    raise ImportError('Platform %s is not supported' % platform.system().lower())

class ODMError(Exception):
    pass
class UnSupportedError(ODMError):
    pass
class ODMCommandTimeOut(ODMError):
    pass
class ODMCommandError(ODMError):
    pass

ODMGET_PATH = '/usr/bin/odmget'
COMMAND_TIMEOUT = 10
COMMAND_INTERVAL = 1

class ODM(object):
    def __init__(self):
        """
        This class supports aix ODM queries
        """
        if platform.system().lower() != 'aix':
            raise UnSupportedError('Platform %s is not supported' % platform.system().lower())
        if not os.path.isfile(ODMGET_PATH):
            raise UnSupportedError('Cannot find odmget command where expected: ' + ODMGET_PATH)

    def query(self, qriteria, object_class):
        """
        Query the odm
        qriteria example: name='oradata' and attribute='pv'
        class: CuAt / CuDv etc.
        return list of dicts build of output
        """

        try:
            command = '%s -q"%s" %s' % (ODMGET_PATH, qriteria, object_class)
            proc = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, err = proc.communicate()
            return self.parse_output(output=out, record_seperator=object_class)
        except (OSError, ValueError, subprocess.CalledProcessError), e:
            raise ODMCommandError(e)

    def parse_output(self, output, record_seperator):
        output_list = []
        for record in output.split(record_seperator):
            record_dict = {}
            for line in record.split('\n'):
                match = re.compile('(.*) = "(.*)"').match(line.strip())
                if match is not None:
                    record_dict[match.groups()[0]] = match.groups()[1]
            if len(record_dict) > 0:
                output_list.append(record_dict)

        return output_list

