#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#------------------------------------------------------------------------------------------------- 

import commands
import re
import os

from common import CommonDevices
from xiv_xsf.utils.system_command import SystemCommand

import fcntl
import array
import ctypes
import struct
from construct import Struct, ULInt32, UBInt32, Bytes, Padding, Container
from platform import processor

if 'i386' == processor():
  UPInt32 = ULInt32
else:
  UPInt32 = UBInt32
 
class OsDevices(CommonDevices):
    FORMAT_CMD = "/usr/sbin/format < /dev/null"
    RESCAN_CMD = [ "/usr/sbin/devfsadm" ]

    def rescan_disk_devices(self):
        system_command = SystemCommand(self.RESCAN_CMD)
        system_command.run()
        

    def get_list_of_scsi_disk_devices(self):
        out = commands.getoutput(self.FORMAT_CMD)

        out = out.split("\n")

        regexp_string = "^\s+\d+\.\s+(\S+)"
        regexp = re.compile(regexp_string)

        devices = []

        for line in out:
            results = regexp.search(line)
            if results is not None:
                if (line.find("<drive not available>") != -1):
                    continue

                device = "/dev/rdsk/" + results.groups()[0]
                devices.append(device)

        return devices

    def get_devfs_string(self, device_file):
        dest = os.readlink(device_file + "s2")
        regexp = re.compile("^../../devices(/scsi_vhci/\w+@[^:]+):\S+$")
        regexp_match = regexp.search(dest)
        if (regexp_match is None):
            raise Exception()

        return regexp_match.groups()[0]
        

    def get_multipath_info(self):
        mpxio_tool = MpxioTool()
        devices = self.get_list_of_scsi_disk_devices()
        for device in devices:
            try:
                devfs_path = self.get_devfs_string(device)
            except Exception, e:
                continue
            try:
                mp_info = mpxio_tool.inspect_multipath(devfs_path)
            except IOError, e:
                mp_info = None
            print device, mp_info

class MpxioTool(object):
    def __init__(self):
       self.set_defs()

    def set_defs(self):
        # DEFINE
        self.SV_DEV_CTL = "/devices/scsi_vhci:devctl"
        self.SCSI_VHCI_GET_CLIENT_MULTIPATH_INFO = 30721
        self.MAXPATHLEN = 1024
        self.MAXNAMELEN = 256

        self.sv_iocdata32 = Struct("iocdata", 
            UPInt32("client"),
            UPInt32("phci"),
            UPInt32("addr"),
            UPInt32("buf_elem"),
            UPInt32("ret_buf"),
            UPInt32("ret_elem"),
            )

        self.MDI_PATHINFO_STATE_INIT = 0
        self.MDI_PATHINFO_STATE_ONLINE = 1
        self.MDI_PATHINFO_STATE_STANDBY = 2
        self.MDI_PATHINFO_STATE_FAULT = 3
        self.MDI_PATHINFO_STATE_OFFLINE = 4

        self.sv_path_info = Struct("sv_path_info",
            Bytes("device", self.MAXPATHLEN),
            Bytes("ret_addr", self.MAXNAMELEN),
            UPInt32("ret_state"),
            UPInt32("ret_ext_state"),
            Padding(12),
            )

        self.path_info_len = self.sv_path_info.sizeof()


    def inspect_multipath(self, path):
        fd = file(self.SV_DEV_CTL)
        client = array.array("B", path) + \
            array.array("B", [0] * (1025-len(path)) )
        client_ptr, client_len = client.buffer_info()

        ret_elem = array.array("B", [0] * 4)
        ret_elem_ptr, ret_elem_len = ret_elem.buffer_info()

        ret_buf = array.array("B", [0] * 1300)
        ret_buf_ptr, ret_buf_len = ret_buf.buffer_info()

        data_buffer = array.array("B", [0] * 1300)
        data_buffer_ptr, data_buffer_len = data_buffer.buffer_info()

        number_of_elements = array.array("B", [0] * 4);
        number_of_elements_ptr, number_of_elements_len = number_of_elements.buffer_info()

        data = self.sv_iocdata32.build(Container(
            client = client_ptr,
            phci = 0,
            addr = 0,
            buf_elem = 0,
            ret_buf = ret_buf_ptr,
            ret_elem = ret_elem_ptr,
            ))

        resp = fcntl.ioctl(fd, self.SCSI_VHCI_GET_CLIENT_MULTIPATH_INFO, data)
        response = self.sv_iocdata32.parse(resp)
        np = struct.unpack("i", ret_elem.tostring())[0]
        if (np < 1):
            raise Exception("Not a multipath")


        # SECOND PHASE
        ret_elem = array.array("B", [0] * 4)
        ret_elem_ptr, ret_elem_len = ret_elem.buffer_info()

        client = array.array("B", path) + \
            array.array("B", [0] * (self.MAXPATHLEN-len(path)+1) )
        client_ptr, client_len = client.buffer_info()

        ret_buf = array.array("B", [0] * self.path_info_len * np)
        ret_buf_ptr, ret_buf_len = ret_buf.buffer_info()

        data = self.sv_iocdata32.build(Container(
            client = client_ptr,
            phci = 0,
            addr = 0,
            buf_elem = np,
            ret_buf = ret_buf_ptr,
            ret_elem = ret_elem_ptr,
            ))

        resp = fcntl.ioctl(fd, self.SCSI_VHCI_GET_CLIENT_MULTIPATH_INFO, data)

        count = 0
        for i in range(np):
            path_info = self.sv_path_info.parse(ret_buf)
            if (path_info.ret_state == self.MDI_PATHINFO_STATE_ONLINE):
                count += 1
            if (len(ret_buf) > self.path_info_len):
                ret_buf = ret_buf[self.path_info_len:]

        fd.close()
        return { "paths": np, "online": count }
