#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program  
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
# 
#-------------------------------------------------------------------------------------------------  

import re
import os

from xiv_xsf.utils.system_command import SystemCommand

class MountEntryNotFoundError(Exception):
    """Mount point search error"""
    def __init__(self, mount_point):
        self.mount_point = mount_point

    def __str__(self):
        return "Mount point %s wasn't found" % \
            (self.mount_point)

class FsUtils(object):
    """Common filesystem related functions"""
    def get_mount_point_of_path(self, path):
        """Returns the mount point of the given path"""
        mount_point = os.path.realpath(path)
        while not os.path.ismount(mount_point):
            mount_point = os.path.dirname(mount_point)

        return mount_point

    def get_device_of_mount_point(self, mount_point):
        """Finds the device of mount_point"""
        raise NotImplementedError
        
    def parse_mounttab_format(self, mount_point, mount_tab):
        """Returns the matching device to the given mount point,
            by parsing mount_tab's contents.
        """
        device = None

        regexp_string = "^(\S+)\s+%s\s+" % (mount_point)
        mount_regexp = re.compile(regexp_string)

        for line in mount_tab:
            match = mount_regexp.search(line)
            if match is not None:
                device = match.groups()[0]

        if (device is None):
            raise MountEntryNotFoundError(mount_point)

        return device

    def get_mount_tab_data(self):
        """Returns a dict representing a mount-tab-like format,
        dict's key is the mount point"""
        
        raise NotImplementedError()

class MountTabParser(object):
    def __init__(self, read_file=None, data=None,
            regexp="^(\S+)\s+(\S+)\s+(\S+)\s+"):
        """Use ONE of the arguments:
        read_file (i.e. /etc/mnttab)
        or
        data (a list of lines) i.e. output of "df -k" command"""

        self.read_file = read_file
        self.regexp_string = regexp
        self.data = data

    def parse(self):
        """Reads the mount tab data and returns a nice dict that holds the
        data. Key is mount point (might be configurable in the future)"""
        system_mounts = {}

        if (self.read_file is not None):
            myfile = file(self.read_file, 'r')
            contents = myfile.readlines()
            myfile.close()
        elif (self.data is not None):
            contents = self.data
        else:
            raise Exception("Bad arguments")
            
        regexp_obj = re.compile(self.regexp_string)
        for line in contents:
            match = regexp_obj.search(line)
            if (match is not None):
                key = match.groups()[1]

                system_mounts[key] = {
                    "device": match.groups()[0], 
                    "mntpoint": match.groups()[1],
                    "fstype": match.groups()[2],
                }

        return system_mounts

