from xiv_xsf.host.storage.nextravolume import crossplatform
import subprocess
import re
import os
from xiv.dtypes import Bunch

class NextraVolume(crossplatform.CrossPlatformNextraVolume):
    pass
class HandlerError(Exception):
    pass
class FileSystemWithNoVolumeGroup(Exception):
    pass

DF_CMD = ['/usr/sbin/df', '-l']
VOL_LIST_CMD = '/opt/xiv/host_attach/bin/vol_list'
VXDG_LIST_CMD = ['/usr/sbin/vxdg', '-q', 'list']
VXDISK_PATH = '/usr/sbin/vxdisk'

FILESYSTEM_TYPES = Bunch(
    Unknown = 1,
    VXFS = 2,
)
VOLUME_GROUP_TYPES = Bunch(
    Unknown = 1,
    VXDG = 2,
)
    

class Filesystem(object):
    def __init__(self, filesystem, raw_device):
        self.mount_point = filesystem
        self.raw_device = raw_device
        self.type = self.get_type()
        self.volume_group = self.get_volume_group()

    def get_type(self):
        if len(self.raw_device.split('/')) == 6 and self.raw_device.split('/')[2] == 'vx':
            return FILESYSTEM_TYPES.VXFS
        else:
            return FILESYSTEM_TYPES.Unknown

    def get_volume_group(self):
        if self.type == FILESYSTEM_TYPES.VXFS:
            return self.raw_device.split('/')[4]

class FilesystemCacheHandler(object):
    def __init__(self):
        if hasattr(self.__class__, 'cached_filesystem_handler'):
            self.filesystem_handler = self.__class__.cached_filesystem_handler
        else:
            self.filesystem_handler = FilesystemHandler()
            self.__class__.cached_filesystem_handler = self.filesystem_handler

class FilesystemHandler(object):
    def __init__(self):
        self.filesystems = {}
        for filesystem, raw_device in self.get_filesystems():
            try:
                self.filesystems[filesystem] = Filesystem(filesystem, raw_device)
            except FileSystemWithNoVolumeGroup, e:
                pass

    def get_filesystems(self):
        try:
            proc = subprocess.Popen(DF_CMD, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, err = proc.communicate()
            fs_regexp = re.compile('^(/\S*)\s*\((\S*)\s*\).*$')
            for line in out.split('\n'):
                if fs_regexp.match(line) is not None:
                    yield fs_regexp.match(line).groups()
        except (OSError, ValueError, subprocess.CalledProcessError), e:
            raise HandlerError('Failed executing %s: %s' % (DF_CMD, e))

class VolumeGroup(object):
    def __init__(self, name, type, device_files):
        self.name = name
        self.type = type
        self.device_files = device_files

class VolumeGroupCacheHandler(object):
    def __init__(self):
        if hasattr(self.__class__, 'cached_volume_group_handler'):
            self.volume_group_handler = self.__class__.cached_volume_group_handler
        else:
            self.volume_group_handler = VolumeGroupHandler()
            self.__class__.cached_volume_group_handler = self.volume_group_handler

class VolumeGroupHandler(object):
    def __init__(self):
        if os.path.isfile(VXDISK_PATH):
            self.volume_groups = self._get_vx_volume_groups()
    def _get_vx_volume_group_list(self):
        volume_groups_list = []
        try:
            proc = subprocess.Popen(VXDG_LIST_CMD, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, err = proc.communicate()
            vxdg_list_regexp = re.compile('^(\S*).*$')
            for line in out.split('\n'):
                if vxdg_list_regexp.match(line) is not None:
                    volume_groups_list.append(vxdg_list_regexp.match(line).groups()[0])
            return volume_groups_list
        except (OSError, ValueError, subprocess.CalledProcessError), e:
            raise HandlerError('Failed executing %s: %s' % (VXDG_LIST_CMD, e))
    def _get_vx_volume_groups(self):
        volume_groups = {}
        for volume_group in self._get_vx_volume_group_list():
            try:
                vxdisk_cmd = [VXDISK_PATH, '-p', '-g', volume_group, 'list']
                proc = subprocess.Popen(vxdisk_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                out, err = proc.communicate()
                vxdisk_list_regexp = re.compile('^LUN_SERIAL_NO\s*:\s*(\S*)$')
                disk_list = []
                disk_count = 0
                for line in out.split('\n'):
                    if vxdisk_list_regexp.match(line) is not None:
                        disk_count += 1
                        disk_device_path = DeviceFileCacheHandler().device_file_handler.get_device_file_path_from_serial_number(vxdisk_list_regexp.match(line).groups()[0])
                        if disk_device_path is not None:
                            disk_list.append(disk_device_path)
                if disk_count > 0:
                    if len(disk_list) < disk_count:
                        raise HandlerError('Error: Volume Group %s has mixed XIV and non XIV volumes' % volume_group)
                    volume_groups[volume_group] = VolumeGroup(volume_group, VOLUME_GROUP_TYPES.VXDG, disk_list)
            except (OSError, ValueError, subprocess.CalledProcessError), e:
                raise HandlerError('Failed executing %s: %s' % (vxdisk_cmd, e))
        return volume_groups

class DeviceFile(object):
    def __init__(self, path, system_id, volume_id, volume_name, size):
        self.path = path
        self.system_id = system_id
        self.volume_id = volume_id
        self.volume_name = volume_name
        self.size = size

class DeviceFileCacheHandler(object):
    def __init__(self):
        if hasattr(self.__class__, 'cached_device_file_handler'):
            self.device_file_handler = self.__class__.cached_device_file_handler
        else:
            self.device_file_handler = DeviceFileHandler()
            self.__class__.cached_device_file_handler = self.device_file_handler

class DeviceFileHandler(object):
    def __init__(self):
        self.device_files = self.get_device_files()
    def get_device_files(self):
        device_files = {}
        try:
            proc = subprocess.Popen(VOL_LIST_CMD, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, err = proc.communicate()
            vol_list_regexp = re.compile('^(\S*)\s*(\S*)\s*(\S*)\s*\S*\s*(\S*)\s*(\S*)$')
            for line in out.split('\n'):
                if vol_list_regexp.match(line) is not None:
                    device_files[vol_list_regexp.match(line).groups()[0]] = DeviceFile(
                        path=vol_list_regexp.match(line).groups()[0],
                        system_id=vol_list_regexp.match(line).groups()[3],
                        volume_id=vol_list_regexp.match(line).groups()[4],
                        volume_name=vol_list_regexp.match(line).groups()[1],
                        size=vol_list_regexp.match(line).groups()[2]
                    )
            return device_files
        except (OSError, ValueError, subprocess.CalledProcessError), e:
            raise HandlerError('Failed executing %s: %s' % (VOL_LIST_CMD, e))
    def get_device_file_path_from_serial_number(self, serial_number):
        for device_file in self.device_files:
            if serial_number in device_file:
                return device_file

def instanciate_from_path(path, generation = 1):
    """
    return NextraVolume objects that are underlying the path entered, as itterable
    """
    try:
        device_files = DeviceFileCacheHandler().device_file_handler.device_files
        filesystems = FilesystemCacheHandler().filesystem_handler.filesystems
        volume_groups = VolumeGroupCacheHandler().volume_group_handler.volume_groups
    except HandlerError, e:
        raise crossplatform.InstanciateError('Error: Unable to load file systems and volumes informations: %s' % e)
    potential_filesystem = os.path.realpath(path)
    while potential_filesystem not in filesystems:
        potential_filesystem = os.path.dirname(potential_filesystem)
    if filesystems[potential_filesystem].type == FILESYSTEM_TYPES.VXFS:
        for device_file in volume_groups[filesystems[potential_filesystem].volume_group].device_files:
            yield NextraVolume(device_files[device_file].system_id, device_files[device_file].volume_id)
    else:
        raise crossplatform.InstanciateError('Filesystem %s is from an unsupported type' % potential_filesystem)

def get_filesystem_list():
    filesystem_handler = FilesystemCacheHandler().filesystem_handler
    return filesystem_handler.filesystems.keys()
