#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#------------------------------------------------------------------------------------------------- 

import re

class OsVolumeManagerTools(object):
    """General OS-specific volume management stuff"""
    def generate_device_datatype(self, device_list):
        """Returns a dict holding block + character devices (on many
        OSes they are the same device)"""
        return map(lambda x: {"char": x,"block": x}, device_list)

    def generate_device_datatype_rdsk_trick(self, device_list):
        """Returns a dict holding block device and character device"""
        char_block_list = []

        regexp_string = "^\/dev\/dsk\/"
        regexp = re.compile(regexp_string)
        for block_device in device_list:
            char_device = None # default

            if (regexp.search(block_device)):
                char_device = block_device.replace("/dev/dsk/", "/dev/rdsk/")

            char_block_list.append({
                "char": char_device, "block": block_device
            })
        return char_block_list

class VolumeManagerMapping(object):
    """Base class for volume manager implementation"""
    def __init__(self, os_vm_tools):
        self.os_vm_tools = os_vm_tools

    def get_physical_devices_of_mount_device(self, mount_device):
        return self.os_vm_tools.generate_device_datatype([mount_device])

class VolumeManagerGeneralError(Exception):
    pass

class VolumeManagerDeviceError(VolumeManagerGeneralError):
    def __init__(self, device):
        self.device = device
    def __str__(self):
        return "Bad device %s" % self.device

class VolumeManagerVolumeError(VolumeManagerGeneralError):
    def __init__(self, volume):
        self.volume = volume
    def __str__(self):
        return "Bad volume %s" % self.volume

class VolumeManagerDiskGroupError(VolumeManagerGeneralError):
    def __init__(self, diskgroup):
        self.diskgroup = diskgroup
    def __str__(self):
        return "Bad disk/volume group %s" % self.diskgroup


class VeritasVM(VolumeManagerMapping):
    """Handles veritas volume manager"""
    VXDISK = '/usr/sbin/vxdisk'

    def __init__(self, os_vm_tools, device_prefix):
        self.device_prefix = device_prefix
        VolumeManagerMapping.__init__(self, os_vm_tools)

    def get_dg_from_device_name(self, device):
        regexp = re.compile("^/dev/vx/dsk/([^/]+)")
        match = regexp.match(device)
        if (match is not None):
            return match.groups()[0]

        raise VolumeManagerDeviceError(device)
            
    def get_physical_devices_of_mount_device(self, mount_device):
        """Returns the physical disks of the VX volume"""

        # Import it here, because IronPython doesn't have subprocess module
        from xiv_xsf.utils.system_command import SystemCommand

        devices = []

        disk_group = self.get_dg_from_device_name(mount_device)

        cmdline = [ self.VXDISK, "-q", "-e", "-g", disk_group, "list" ]
        proc = SystemCommand(cmdline)
        proc.run()

        for line in proc.out.split("\n"):
            output_fields = line.split();
            if (len(output_fields) >= 5):
                if (output_fields[4] == "online"):
                    devices.append(self.device_prefix + "/" + output_fields[5])

        devices_datatype = self.os_vm_tools.generate_device_datatype(devices)
        return devices_datatype
