#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#------------------------------------------------------------------------------------------------- 

import re

from xiv_xsf.host.storage.volume_managers.common import VolumeManagerMapping, \
    VeritasVM, VolumeManagerDeviceError, OsVolumeManagerTools
from xiv_xsf.utils.system_command import SystemCommand
from xiv_xsf.utils.parse import ParseUtils

class HpuxVolumeManagerTools(OsVolumeManagerTools):
    """General OS-specific volume management stuff"""
    def generate_device_datatype(self, device_list):
        """Returns a dict holding block + character devices"""
        return self.generate_device_datatype_rdsk_trick(device_list)

class HpuxNative(VolumeManagerMapping):
    """Native device mode"""
    pass

class HpuxLVM(VolumeManagerMapping):
    """HPUX Logical Volume Manager support class"""

    VGDISPLAY="/usr/sbin/vgdisplay"
    LVDISPLAY="/usr/sbin/lvdisplay"

    def get_volume_group_of_device(self, device):
        """Queries the HP-UX LVM: who is the VG of this device?"""
        volume_group = None
        cmdline = [self.LVDISPLAY, device]
        proc = System.Command(cmdline)
        proc.run()

        regexp = re.compile("^VG Name\s+(/dev/\S+)")
        for line in proc.out.split("\n"):
            match = regexp.search(line)
            if (match is not None):
                volume_group = match.groups()[0]

        return volume_group
 
    def get_physical_devices_of_mount_device(self, mount_device):
        """Returns the LVM Physical Volumes of a given volume"""
        devices = []
        volume_group = None

        volume_group = self.get_volume_group_of_device(mount_device)

        if (volume_group is None):
            raise VolumeManagerInfoError(device, \
                "Volume Group for %s not found" % device)

        cmdline = [self.VGDISPLAY, "-v", volume_group]
        proc = SystemCommand(cmdline)
        proc.run()

        regexp = re.compile("^\s*PV Name\s+(/dev/\S+)")
        for line in proc.out.split("\n"):
            match = regexp.search(line)
            if (match is not None):
                devices.append(match.groups()[0])

        devices_datatype = self.os_vm_tools.generate_device_datatype(devices)
        return devices_datatype

class VolumeManagerDetector(object):
    """Detect and inistantiate the matching Volume Manager object"""

    def __init__(self, device):
        self.device = device

    def detect_vm(self):
        """Checks volume type, by the way the device name looks"""
        vm_types = {
            "vxvm": ["^/dev/vx/"],
            "hpux_native": ["^/dev/dsk/"],
            "hpux_lvm": ["^/dev/"],
        }

        vm_type = ParseUtils().regexp_dictionary_match(vm_types, self.device)

        return vm_type

    def get_vm_object(self):
        """Returns a volume manager object, after detecting the correct
        volume manager (based on self.device)"""
        os_vm_tools = HpuxVolumeManagerTools()
        vm_type = self.detect_vm()

        if (vm_type == "hpux_lvm"):
            vm_obj = HpuxLVM(os_vm_tools)
        elif (vm_type == "vxvm"):
            vm_obj = VeritasVM(os_vm_tools, "/dev/dsk")
        elif (vm_type == "hpux_native"):
            vm_obj = HpuxNative(os_vm_tools)
        else:
            raise VolumeManagerDeviceError(self.device)

        return vm_obj
