#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#------------------------------------------------------------------------------------------------- 

import re

from xiv_xsf.host.storage.volume_managers.common import VolumeManagerMapping, \
    VeritasVM, VolumeManagerDeviceError, OsVolumeManagerTools
from xiv_xsf.utils.system_command import SystemCommand
from xiv_xsf.utils.parse import ParseUtils

class LinuxVolumeManagerTools(OsVolumeManagerTools):
    """General OS-specific volume management stuff"""
    pass

class LinuxLVM(VolumeManagerMapping):
    """Linux LVM support class"""
    LVS = "/usr/sbin/lvs"
    PVS = "/usr/sbin/pvs"
    
    def get_physical_devices_of_mount_device(self, mount_device):
        """Returns the LVM Physical Volumes of a given volume"""
        devices = []

        cmdline = [ self.LVS, "--noheadings", "--separator", ",", mount_device ]
        proc = SystemCommand(cmdline)
        proc.run()

        output_fields = proc.out.split(",")
        volume_group = output_fields[1]

        cmdline = [ self.PVS, "--noheadings", "--separator", "," ]
        proc = SystemCommand(cmdline)
        proc.run()

        for line in proc.out.split("\n"):
            output_fields = line.split(",")
            if (len(output_fields) > 2 and output_fields[1] == volume_group):
                devices.append(output_fields[0].replace(" ",""))

        devices_datatype = self.os_vm_tools.generate_device_datatype(devices)
        return devices_datatype

class LinuxNative(VolumeManagerMapping):
    """Native device (including multipath) mode"""
    pass

class VolumeManagerDetector(object):
    """Detect and inistantiate the matching Volume Manager object"""
    
    DMSETUP = "/sbin/dmsetup"

    def __init__(self, device):
        self.device = device

    def detect_vm(self):
        """Checks volume type, by the way the device name looks"""
        vm_types = {
            "mapper": ["^/dev/mapper/"],
            "linux_native": ["^/dev/sd\w{1,2}"],
            "vxvm": ["^/dev/vx/"],
        }

        vm_type = ParseUtils().regexp_dictionary_match(vm_types, self.device)

        # Handle the Device Mapper case
        if (vm_type == "mapper"):
            cmdline = [ self.DMSETUP, "info", "-c", "-o", "uuid", \
                "--noheading", self.device ]
            proc = SystemCommand(cmdline)
            proc.run()
            device_mapper_uuid = proc.out

            device_mapper_types = {
                "linux_lvm": ["^LVM-"],
                "linux_native": ["^mpath-"]
            }

            vm_type = ParseUtils().regexp_dictionary_match(\
                device_mapper_types, device_mapper_uuid)

        return vm_type

    def get_vm_object(self):
        """Returns a volume manager object, after detecting the correct
        volume manager (based on self.device)"""
        os_vm_tools = LinuxVolumeManagerTools()
        vm_type = self.detect_vm()

        if (vm_type == "linux_lvm"):
            vm_obj = LinuxLVM(os_vm_tools)
        elif (vm_type == "vxvm"):
            vm_obj = VeritasVM(os_vm_tools, "/dev")
        elif (vm_type == "linux_native"):
            vm_obj = LinuxNative(os_vm_tools)
        else:
            raise VolumeManagerDeviceError(self.device)

        return vm_obj
