"""
This file facilitates SSHTunneling on UNIX systems(with ssh installed).
"""

import socket
import string
import signal
import subprocess
import os
import signal
import types
import sys
from time import time, sleep

from xiv.pyutils import weakref, wait


class SSHTunnelFailed(Exception):
    pass

class SSHTunnel(object):
    """
    SSH Tunneling class, provides managed and safe creation of Local Port Forwarding over SSH Tunnel.
    The tunnel is created by ssh channel to a  proxy_host  which must have access to the target machine.

    Local Port is randomly chosen for the tunnel.

    Forwarding local port  to machine 'hermes' poirt 25 using proxy machine 10.0.0.10
    >>> smtp_tunnel = SSHTunnel(proxy_host="10.0.0.10",target_host="hermes",target_port="25",remote_username="root")
    """

    DEFAULT_TUNNEL_START_TIMEOUT = 3

    def __init__(self, proxy_host,target_host,target_port,remote_username=None,tunnel_start_timeout=None,verbose=False):
        super(SSHTunnel, self).__init__()

        if remote_username is None:
            remote_username = os.getlogin()

        self.remote_username = remote_username
        self.tunneling_process = None
        self.proxy_host = proxy_host
        self.target_host = target_host
        self.target_port = target_port
        self.establish_tunnel(tunnel_start_timeout,verbose=verbose)

    # once this function returns, the returned port number is free for anyone to grab;
    #  there's a race here by design: what if someone grabs the port before the caller of the function does?
    def pick_free_random_local_port_with_potential_race(self):
        test_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        test_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        # binding to port 0 makes the kernel pick a port
        test_socket.bind(('127.0.0.1', 0))
        result = test_socket.getsockname()[1]
        test_socket.close()
        return result

    def test_forwarding(self):
        if self.local_tunneling_port is None:
            raise SSHTunnelFailed("Tried testing null port")
        try:
            test_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            test_socket.connect(('127.0.0.1', self.local_tunneling_port))
        except socket.error:
            return False
        del(test_socket)
        return True

    def establish_tunnel(self,tunnel_start_timeout=None,verbose=False):

        if self.tunneling_process is not None:
            raise SSHTunnelFailed("Tried to establish tunnel when tunnel already established")

        if tunnel_start_timeout is None:
            tunnel_start_timeout = self.DEFAULT_TUNNEL_START_TIMEOUT
        self.local_tunneling_port = self.pick_free_random_local_port_with_potential_race()
        verbose_flag = ""
        if verbose:
                verbose_flag = "-v"
        ssh_parameters = dict(
                proxy_host =  self.proxy_host,
                source_port = self.local_tunneling_port,
                target_port = self.target_port,
                target_host = self.target_host,
                verbose_flag = verbose_flag,
                remote_username = self.remote_username
            )
        self.tunneling_process = subprocess.Popen(
                args = 'ssh -N %(proxy_host)s %(verbose_flag)s -o PasswordAuthentication=no -l %(remote_username)s -L %(source_port)s:%(target_host)s:%(target_port)s' % ssh_parameters,
                shell = True,
                stdout = subprocess.PIPE,
                stderr = subprocess.PIPE,
                )


        detailed_error_message = ""
        wait(
            self.test_forwarding,
            interval = 0.1,
            timeout = tunnel_start_timeout,
            error_callback = self.tunnel_error_analyzer,
            exception = SSHTunnelFailed('Failed to establish tunnel within the given timeout(%d)' % (tunnel_start_timeout))
        )

    def tunnel_error_analyzer(self):
        # Kill the tunnel
        tunnel_stdout,tunnel_stderr = self.tear_down_tunnel()
        tunnel_stderr = "\n".join(tunnel_stderr)
        tunnel_stdout = "\n".join(tunnel_stdout)
        # Return the proccess stderr output
        raise SSHTunnelFailed("Failed to establish tunnel within the given timeout\nError:%s" % tunnel_stderr )
        
    def tear_down_tunnel(self):
        if self.tunneling_process is None:
            raise SSHTunnelFailed("Tried to tear down non-established tunnel")
        os.kill(self.tunneling_process.pid, signal.SIGKILL)
        wait(
            lambda: self.tunneling_process.poll() is not None,
            timeout = 1,
            interval = 0.1,
            exception = SSHTunnelFailed("Tunnel process failed to die")
        )
        tunnel_stdout = self.tunneling_process.stdout.readlines()
        tunnel_stderr = self.tunneling_process.stderr.readlines()
        self.tunneling_process = None
        self.local_tunneling_port = None
        return((tunnel_stdout,tunnel_stderr))


