#!/usr/bin/env python

import sys, os, time
from optparse import OptionParser
import smtplib
from xiv_xsf.network.connectivity import SSHTunnel, SSHTunnelFailed


if sys.version[:3] == "2.4" :
    from email.Utils import COMMASPACE
    from email.MIMEText import MIMEText
    from email.MIMEMultipart import MIMEMultipart
    from email.MIMEBase import MIMEBase
    from email.MIMEAudio import MIMEAudio
    from email.MIMEImage import MIMEImage
    from email.Encoders import encode_base64
    
else :
    from email.utils import COMMASPACE
    from email.mime.text import MIMEText
    from email.mime.multipart import MIMEMultipart
    from email.mime.base import MIMEBase
    from email.mime.audio import MIMEAudio
    from email.mime.image import MIMEImage
    from email.mime.application import MIMEApplication
    from email.encoders import encode_base64 
import mimetypes

VERSION="1.0"

class SMTPClient(object):
    """
    Simple SMTPClient with SMTP proxy(tunnel) support  
    """
    
    DEFAULT_PROXY_SERVER_USERNAME = "root"
    
    def __init__(self, smtp_server, smtp_port=smtplib.SMTP_PORT, proxy_host=None, proxy_username=DEFAULT_PROXY_SERVER_USERNAME,
                 reuse_tunnel=True, verbose=False):
        super(SMTPClient, self).__init__()
        self.smtp_server = smtp_server
        self.proxy_host = proxy_host
        self.smtp_port = smtp_port
        self.proxy_username = proxy_username
        self.smtp_tunnel = None
        self.smtp_server_handle = None
        self.reuse_tunnel = reuse_tunnel
        self.verbose = verbose
        
    def raise_smtp_tunnel(self):
        """
        Raise smtp(SSH) tunnel through a proxy host
        """
        
        self.smtp_tunnel = SSHTunnel(self.proxy_host, self.smtp_server, self.smtp_port, self.proxy_username)
        return self.smtp_tunnel
    
    
    def connected(self):
        return  (self.smtp_server_handle is not None) 
    
    def connect(self):
        
        if self.smtp_server_handle is not None:
            raise smtplib.SMTPException("Attempt to connect while having a live SMTP session")
        
        if self.proxy_host is not None :
            if self.smtp_tunnel is None:
                try:
                    self.raise_smtp_tunnel()
                except SSHTunnelFailed, e:
                    raise SMTPClientException("Error setting SMTP proxy tunnel ", inner_exception=e)
                    
            smtp_host = "127.0.0.1" # We use tunnel
            smtp_port = self.smtp_tunnel.local_tunneling_port
        else :
            smtp_host = self.smtp_server
            smtp_port = self.smtp_port
                
        try:
            if self.verbose :
                smtplib.SMTP.debuglevel = 10        
            self.smtp_server_handle = smtplib.SMTP(host=smtp_host, port=smtp_port,
                                              local_hostname="127.0.0.1")
        except (smtplib.SMTPHeloError, smtplib.SMTPException), e:
            raise SMTPClientException("Error connecting to SMTP Server", inner_exception=e)
        
    def send_mail(self, recipients, message, subject="", sender_address=None, cc_recipients=None, bcc_recipients=None):
        """
        Low level email delivery mechanism
        
        message should be either str or MIMEMuiltipart object
        """
        if not isinstance(message, MIMEMultipart) and not isinstance(message, str):
            raise ValueError("Message should be either a string or MIMEMultipart")
        
        if isinstance(message, str):
            message = MIMEText(message)
            message['To'] = COMMASPACE.join(recipients)
            
        
        if isinstance(recipients, str):
            recipients = recipients.split(',')
        if isinstance(bcc_recipients, str):
            bcc_recipients = bcc_recipients.split(',')
        if isinstance(cc_recipients, str):
            cc_recipients = cc_recipients.split(',')
        if message.has_key('To'):
            message.replace_header('To', COMMASPACE.join(recipients))
        else:
            message['To'] = COMMASPACE.join(recipients)
            
        if not message.has_key('From'):
            if sender_address is None:
                raise ValueError("No Sender(From:) address specified")
            message['From'] = sender_address
       
        message['Subject'] = subject
        message['Date']  = time.ctime(time.time())
    
        if type(cc_recipients).__name__ == "list" :
            recipients.extend(cc_recipients)
        if type(bcc_recipients).__name__ == "list" :
            recipients.extend(bcc_recipients)
        
        # Send the mail
        self.connect()
        self.smtp_server_handle.sendmail(message['From'], recipients, message.as_string())
        self.disconnect()
    
    def disconnect(self):
        try:
            self.smtp_server_handle.close()
            self.smtp_server_handle = None
            if self.smtp_tunnel is not None and not self.reuse_tunnel:
                self.smtp_tunnel.tear_down_tunnel()
                self.smtp_tunnel = None
                
        except (smtplib.SMTPException), e:
            raise SMTPClientException("Error disconnecting from SMTP Server", inner_exception=e)
        
    def is_valid_smtp_server(self):
        
        try:
            self.connect()
            self.smtp_server_handle.helo()
            self.disconnect()
            return True
        except (smtplib.SMTPHeloError, smtplib.SMTPException, SMTPClientException), e:
            return False
        
    @staticmethod
    def attach_file(message, file_path):
        """
        Attach file to MIMEMultipart message
        """
        if not isinstance(message, MIMEMultipart):
            raise TypeError("Cannot add attachment to non MIMEMultipart message")
    
        try :
            file_handle = open(file_path, 'rb')
            file_data = file_handle.read()
            file_handle.close()
        except (OSError, IOError), e:
            raise(SMTPClientException("Unable to read attachment file:%s" % file_path, inner_exception=e)) 
        
        attachment=""
        
        # Guess the content type based on the file's extension.  Encoding
        # will be ignored, although we should check for simple things like
        # gzip'd or compressed files.
        ctype, encoding = mimetypes.guess_type(file_path)
        if ctype is None or encoding is not None:
            # No guess could be made, or the file is encoded (compressed), so
            # use a generic bag-of-bits type.
            ctype = 'application/octet-stream'
        maintype, subtype = ctype.split('/', 1)
        if maintype == 'text':
            # Note: we should handle calculating the charset
            attachment = MIMEText(file_data, _subtype=subtype)
        elif maintype == 'image':
            attachment = MIMEImage(file_data, _subtype=subtype)
        elif maintype == 'audio':
            attachment = MIMEAudio(file_data, _subtype=subtype)
        else:
            attachment = MIMEBase(maintype, subtype)
            attachment.set_payload(file_data)
            # Encode the payload using Base64
            encode_base64(attachment)
        
        # Set the filename parameter
        attachment.add_header('Content-Disposition', 'attachment', filename=os.path.basename(file_path))
        message.attach(attachment)

        

class SMTPClientException(Exception):
    """
    Basic Exception class for NetMail
    """
    def __init__(self, message, inner_exception=None):
        super(SMTPClientException, self).__init__()
        if inner_exception is not None:
            self.inner_exception = inner_exception




def main():
    
    message = None
    
    # CLI Argument Parsing
    cli_option_parser =  OptionParser(usage="%prog <options> [message text]\n Sends email messages via  specific SMTP server", version="%prog " + VERSION)
    cli_option_parser.add_option("-v", "--verbose", action="store_true", dest="verbose",
                                 default=False,
                                 help="Verbose")
    cli_option_parser.add_option("-i", "--interactive", action="store_true", dest="interactive_mode",
                                 default=False,
                                 help="Interactive Mode")
    cli_option_parser.add_option("-t", "--To", action="store", dest="recipients",
                                 help="List of recipients comma(',') delimited")
    cli_option_parser.add_option("-c", "--CC", action="store", dest="cc_recipients",
                                 help="List of CC recipients comma(',') delimited")
    cli_option_parser.add_option("-b", "--BCC", action="store", dest="bcc_recipients",
                                 help="List of recipients comma(',') delimited")
    cli_option_parser.add_option("-f", "--From", action="store", dest="sender",
                                 help="Sender Address (From:)")
    cli_option_parser.add_option("-s", "--subject", action="store", dest="subject",
                                 help="Email Subject")
    cli_option_parser.add_option("-S", "--smtp", action="store", dest="smtp_server",
                                 help="SMTP Server Address")
    cli_option_parser.add_option("-p", "--port", action="store", default=smtplib.SMTP_PORT, dest="smtp_port",
                                 help="SMTP Server Port ")
    cli_option_parser.add_option("-m", "--message_file", action="store", dest="message_file",
                                 help="File containing message text")
    
    cli_option_parser.add_option("-a", "--attachment", action="append", dest="attachments",
                                 help="Files to attach to Email")
    
    cli_option_parser.add_option("-P", "--proxy_host", action="store", dest="proxy_host",
                                 help="Send Email through SMTP Proxy host using ssh tunnel ")
    cli_option_parser.add_option("-u", "--proxy_user", action="store", dest="proxy_username",
                                 help="Username for ssh tunnel to SMTP Proxy")
    
    
    (parsed_options, unmatched_args) = cli_option_parser.parse_args()
    
    
    if parsed_options.interactive_mode :
        message = interactive_message_builder(parsed_options)
    
    # if message file supplied, read its contents into message buffer
    if parsed_options.message_file is not None:
        try:
            message = MIMEMultipart()
            message_text = open(parsed_options.message_file).readlines()
            message.attach(MIMEText(message_text))
        except OSError, e:
            print("Unable to open message file:%s\nError:%s" % (parsed_options.message_file, str(e)))
            cli_option_parser.print_help()
            
    
    # Try to read message text body from stdin
    #if len(unmatched_args) > 0 and message is None:
    #   message_text = read_message_text_from_stdin()
    #   message = MIMEMultipart()
    #   message.attach(MIMEText(message_text))
        
    # If we havent filled message text yet, try to get it from remaining cmd args
    if message is None:
        if len(unmatched_args) > 0 :
            message_text = " ".join(unmatched_args)
            message = MIMEMultipart()
            message.attach(MIMEText(message_text))
        else :
            print("Missing message text")
            cli_option_parser.print_help()
            sys.exit(2)
    
    
    # Attach files
    if hasattr(parsed_options.attachments, "__iter__"):
        for file_path in parsed_options.attachments :
            SMTPClient.attach_file(message, file_path)
    
    net_mail = SMTPClient(parsed_options.smtp_server, proxy_host=parsed_options.proxy_host,
                         proxy_username=parsed_options.proxy_username, smtp_port=parsed_options.smtp_port,
                         reuse_tunnel=False, verbose=parsed_options.verbose)
    
    net_mail.send_mail(parsed_options.recipients, message, parsed_options.subject,
                       parsed_options.sender,
                       cc_recipients=parsed_options.cc_recipients,
                       bcc_recipients=parsed_options.bcc_recipients)
    
    print("Mail sent")

def ask_question(question, default=True):
    """
    Asks a question and awaits user input, either true or false. 
    Do you like finding screws? [Y/n] 
    False
    >>> 
    """
    
    if not question.endswith("?"):
        question = question + "?"
    if default:
        options = "[Y/n]"
    else:
        options = "[y/N]"
    while True:
        try:
            print("\n")
            answer = raw_input("%s %s " % (question, options))
        except EOFError:
            continue
        if not answer.strip():
            return default
        else:
            return answer.lower() in ("y", "yes")

def interactive_message_builder(defined_options):
    """
    Gathers mail information interactively.
    """
    
    print("Entering Interactive Mode\n")
    # Delivery Fields
    if defined_options.smtp_server is None:
        defined_options.smtp_server = raw_input("SMTP Server: ")
    
    # Proxy Fields
    if defined_options.proxy_host is None and ask_question("Use proxy smtp server") == True:
        defined_options.proxy_host = raw_input("Proxy Host: ")
        defined_options.proxy_username = raw_input("Proxy (ssh) username: ")
    
    
    # Mail information fields
    if defined_options.recipients is None:
        defined_options.recipients = raw_input("To: ")
    if defined_options.recipients is None:
        defined_options.cc_recipients = raw_input("CC: ")
    if defined_options.recipients is None:
        defined_options.bcc_recipients = raw_input("BCC: ")
    
    if defined_options.subject is None:
        defined_options.subject = raw_input("Subject: ")
    
    if defined_options.sender is None:
        defined_options.sender = raw_input("From: ")
    
    message = MIMEMultipart()
    print("Please enter mail text. To finish input press (Unix: Ctl-D, Windows: Ctl-Z+Return)")
    message_text = read_message_text_from_stdin()
    message.attach(MIMEText(message_text))
    return message

def read_message_text_from_stdin(show_prompt=True):
    message_text = ""
    line_num = 0
    while True:
        try:
            line_num = line_num + 1
            if show_prompt:
                prompt="Mail Line %d: " % line_num
            line = raw_input(prompt) + "\n"
        except EOFError:
            print("\n")
            break
        if not line:
            print("\n")
            break
        message_text = message_text + line
    return message_text
    

        
        
if __name__ == '__main__':
    main()
        
