#!/usr/bin/env ipython

"""
This module supplies the NextraExplorer class, which permits the collection and analysis of Nextra's status (software and hardware). It is expected to be executed from within service scripts like the NextraSupportWizard or from a small utility program executed by technicians or other Nextra maintainers to manually collect all data and send it back home.
"""

__version__ = '1.0'
__author__ = 'yaniva@il.ibm.com'

import sys
import logging 
import os
import time
import cPickle
import subprocess
import socket
from BaseHTTPServer import HTTPServer

from xiv.network.command_runner import LocalCommandRunner, RemoteCommandRunner, SimpleResultVerifier, \
    CommandFailureException
from xiv.osutils.eztmpfs import EZtmpfs, EZtmpfsNotSupported
from xiv.pyutils import wait
from xiv_xsf.nextra.p9client.exceptions import NextraClientException
from xiv_xsf.nextra.p9constants import MANAFACE_NUMBERS, MANAFACE_NODES, CACHE_NODE_NUMBERS, \
    DISK_NUMBERS, MODULE_NUMBERS, UPS_NUMBERS, DIGI_NUMBER

from errors import NextraExplorerFailed
from collection_commands import full_collection
from decorators import log_function_name, requires_living_node, reraise_exceptions_as_explorer_failed_exception
from webserver import ExplorerHTTPRequestHandler

class NextraExplorer(object):
    def __init__(self, nextra, maintainer_name, maintainer_email, logger_name=None, password=None):
        if logger_name is None:
            logger_name = 'explorer'
        self.logger = logging.getLogger(logger_name)
        self.nextra = nextra
        self.system_id = self.nextra.system_id_as_octets
        if password is not None:
            os.environ['XIV_INTERNAL_PASSWORD'] = password
        self.maintainer_name = maintainer_name
        self.maintainer_email = maintainer_email
        self.initialization_time = time.localtime()
        self.initialize_eztmpfs()
        self.packing_process = None
        self.checking_sanity()
        self.logger.info('initialized Explorer; collection is performed by %s (%s)' % 
            (self.maintainer_name, self.maintainer_email))

    @reraise_exceptions_as_explorer_failed_exception(OSError)
    def initialize_eztmpfs(self):
        self.eztmpfs = EZtmpfs('-nextra-explorer')
        self.results_root = os.path.join(str(self.eztmpfs), 'NextraExplorer')
        os.mkdir(self.results_root)
        self.eztmpfs_handler = logging.FileHandler(os.path.join(self.results_root, 'NextraExplorer.log'))
        self.eztmpfs_handler.setFormatter(logging.Formatter(fmt='%(name)-12s %(levelname)-8s %(message)s'))
        logging.getLogger().addHandler(self.eztmpfs_handler)

    def dispatch_commands(self, command_sequence=None):
        if self.eztmpfs.mountpoint is None:
            raise NextraExplorerFailed("can't dispatch after results were packed")
        if command_sequence is None:
            command_sequence = full_collection
        for dispatcher, args, kwargs in command_sequence:
            dispatcher(self).dispatch(*args, **kwargs)

    def __repr__(self):
        return '<NextraExplorer (%s) for system %s>' % (__version__, self.system_id)

    @reraise_exceptions_as_explorer_failed_exception(AssertionError, NextraClientException, EZtmpfsNotSupported)
    @log_function_name(level=logging.DEBUG, num_args_to_log=None)
    def checking_sanity(self):
        external_managers = tuple(
            ip for ip in self.nextra.transport.aservers.values()
            if not ip.startswith('14.%s' % self.system_id)
        )
        if len(external_managers) != 0:
            self.logger.error('tried initialisation with external managers: %s' % external_managers)
            raise AssertionError('explorer must be used from within a nextra system ("14 net")')
        assert not self.is_in_rebuild(), "explorer mustn't be run when system is in rebuild"
        assert 'XIV_INTERNAL_PASSWORD' in os.environ, 'XIV_INTERNAL_PASSWORD not set in environment'
        EZtmpfs.verify_platform_support()

    def is_in_rebuild(self):
        return self.nextra.system_counters_get().as_single_element.system_status == 'Rebuilding'

    def auto_choose_filename(self, full_path=False):
        filename = 'NextraExplorer_%s_%s.tar.bz2' % \
            (time.strftime('%Y%m%d-%H%M', self.initialization_time), self.maintainer_email)
        if full_path:
            return os.path.join(str(self.eztmpfs), filename)
        else:
            return filename

    @reraise_exceptions_as_explorer_failed_exception(IOError)
    def pack_results_to_file(self, filename=None):
        if filename is None:
            filename = self.auto_choose_filename(full_path=True)
        self.start_packing_results(file(filename), 'w')
        wait(self.is_packing_complete, timeout=15*60, exception=NextraExplorerFailed('timeout packing results'))

    @reraise_exceptions_as_explorer_failed_exception(OSError)
    def start_packing_results(self, output=subprocess.PIPE):
        self.logger.info('packing results data')
        self.eztmpfs_handler.flush()
        self.total_size_to_pack = self.eztmpfs.used_space
        self.packing_process = subprocess.Popen(
            ('tar', '--bzip', '--remove-files', '--create', '--directory', str(self.eztmpfs), 'NextraExplorer'),
            stdout=output, stderr=subprocess.PIPE,
        )

    def packed_data_generator(self):
        assert self.packing_process.stdout is not None, 'tried reading a redirected result packer'
        while True:
            data = self.packing_process.stdout.read(4096)
            if data == '':
                break
            yield data

    def pack_completion_percent(self):
        return 100-int(self.eztmpfs.used_space/(float(self.total_size_to_pack)/100))

    def is_packing_complete(self):
        if self.packing_process.poll() is None:
            return False
        if self.packing_process.poll() != 0:
            raise NextraExplorerFailed("packing results failed: " + self.packing_process.stderr.read(256))
        return True

    def start_file_specific_webserver(self, filename=None):
        if filename is None:
            filename = self.auto_choose_filename()
        current_directory = os.getcwd()
        os.chdir(str(self.eztmpfs))
        ExplorerHTTPRequestHandler.fixed_file_name = filename
        try:
            server = HTTPServer(('0.0.0.0', 80), ExplorerHTTPRequestHandler)
        except socket.error, error:
            self.logger.warning('failed running server: %s' % error)
        self.logger.info('webserver ready, start download or abort with Ctrl-C')
        try:
            try:
                server.handle_request()
            except KeyboardInterrupt:
                self.logger.warning('server aborted by keyboard interruption')
        finally:
            del(server)
            os.chdir(current_directory)

    def deallocate_results(self):
        if self.eztmpfs_handler is not None:
            logging.getLogger().removeHandler(self.eztmpfs_handler)
            self.eztmpfs_handler.close()
            self.eztmpfs_handler = None
        if self.packing_process is not None and self.packing_process.poll() is None:
            os.kill(self.packing_process.pid, 9)
        if self.eztmpfs.mountpoint is not None:
            self.eztmpfs.destroy()

def main(argv):
    logging.basicConfig(level=10)
    os.environ['XIV_INTERNAL_PASSWORD'] = 'itRwo-@'
    print 'interactively testing nextra explorer; instantiating objects'
    try:
        global e
        global ex
        global n
        from xiv_xsf.nextra import p9client
        n = p9client.instantiators.internal_net('10.180')
        from xiv_xsf.nextra.hardware.explorer import NextraExplorer
        e = NextraExplorer(n, 'yaniv', 'yaniv@il.ibm.com')
    except Exception, ex:
        print 'caught %s during instantiation' % ex
        raise
    else:
        print 'you have an explorer object at variable "e"'

if __name__ == '__main__':
    main(sys.argv)
