import logging
import os
import subprocess
import cPickle
import time
import socket
from telnetlib import Telnet

from xiv_xsf.nextra.hardware.ipmi import IPMI, IPMIException
from xiv_xsf.nextra.p9constants import DIGI_NUMBER, MANAFACE_NUMBERS, CACHE_NODE_NUMBERS
from xiv.network.command_runner import LocalCommandRunner, RemoteCommandRunner, SimpleResultVerifier, \
    CommandFailureException
from xiv.pyutils import safe_forever_loop
from xiv.strutils import str_brief

from decorators import requires_living_node, reraise_exceptions_as_explorer_failed_exception, log_exceptions_and_resume
from errors import NextraExplorerFailed

def CMD(collector, *args, **kwargs):
    return (collector, args, kwargs)

class Collector(object):
    INVALID_PATH_CHARACTERS = ' :?*&";#'
    logger_name = 'base'
    def __init__(self, explorer):
        super(Collector, self).__init__()
        self.logger = logging.getLogger(self.logger_name)
        if not hasattr(self, 'filesystem_name'):
            self.filesystem_name = self.logger_name.upper()
        self.nextra = explorer.nextra
        self.results_root = str(explorer.results_root)
        self.explorer = explorer

    def __repr__(self):
        return '<%s for %s>' % (self.__class__.__name__, self.nextra.system_id_as_octets)

    def kwargs_to_filename(self, kwargs_dict):
        return '#' + '#'.join(
            '%s=%s' % (self.normalize_path_component(key), self.normalize_path_component(value))
            for key, value in kwargs_dict.items()
        )

    def args_to_filename(self, args):
        return '#' + '#'.join(str(arg) for arg in args)

    def format_args(self, args):
        if not args:
            return ''
        return '(' + ", ".join(str(arg) for arg in args) + ')'

    def format_kwargs(self, kwargs):
        if not kwargs:
            return ''
        return '{' + ', '.join('%s=%s' % (key, value) for key, value in kwargs.items()) + '}'

    def normalize_path_component(self, component):
        component = os.path.basename(str(component))
        for invalid_character in self.INVALID_PATH_CHARACTERS:
            component = component.replace(invalid_character, '_')
        return component

    @reraise_exceptions_as_explorer_failed_exception(CommandFailureException)
    def creating_results_subdirectory(self, *raw_path_components):
        normal_components = list(self.normalize_path_component(component) for component in raw_path_components)
        subdirectory = os.path.join(self.results_root, *normal_components)
        LocalCommandRunner().mkdir('-p', subdirectory)
        return subdirectory

    @reraise_exceptions_as_explorer_failed_exception(IOError)
    def storing_collector_results(self, filename, subdirectory, data):
        file(os.path.join(subdirectory, str(filename)), 'w').write(str(data))

class XCLICollector(Collector):
    logger_name = 'xcli'
    def dispatch(self, command, **kwargs):
        self.logger.info("collecting %s %s" % (command, self.format_kwargs(kwargs)))
        results = getattr(self.nextra, command)(self, command, **kwargs)
        subdirectory = self.creating_results_subdirectory(self.filesystem_name, command, 
                                                          self.kwargs_to_filename(kwargs))
        self.storing_collector_results("xml", subdirectory, results)

class ModuleDependantCollector(Collector):
    def is_ping_responsive(self, module_number):
        p = subprocess.Popen('ping -c 1 -w 1 ' + self.module_ip(module_number), shell=True, 
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return (p.wait() == 0)

    def module_ip(self, module_number):
        return '14.%s.%s' % (self.nextra.system_id_as_octets, module_number)

class ShellCollector(ModuleDependantCollector):
    MAX_DATA_PER_FETCH = 2**24 # 16m
    logger_name = 'shell'
    @requires_living_node
    def dispatch(self, module, command, *args):
        self.log_operation(module, command, args)
        runner = RemoteCommandRunner(self.module_ip(module), verifier=SimpleResultVerifier(), basicRunString=command)
        results = runner.execute(*args)[1][0:self.MAX_DATA_PER_FETCH]
        subdirectory = self.creating_results_subdirectory(module, self.filesystem_name)
        self.storing_collector_results(
            filename=self.normalize_path_component(command)+self.args_to_filename(args),
            subdirectory=subdirectory, data=results, 
        )

    def log_operation(self, module, command, args):
        self.logger.info("[%s] collecting %s %s" % (module, command, self.format_args(args)))

    def format_args(self, args):
        return " ".join(str(arg) for arg in args)

class OptionalShellCollector(ShellCollector):
    @log_exceptions_and_resume(CommandFailureException, filter=lambda(error): 'failed on %s' % error.host)
    def dispatch(self, module, command, *args):
        super(OptionalShellCollector, self).dispatch(module, command, *args)

class BaseDigiCollector(ModuleDependantCollector):
    def digi_port_from_module_number(self, module_number):
        if module_number in CACHE_NODE_NUMBERS:
            return 2000+module_number # 1 -> 2001, 2 -> 2002, ... 8 -> 2008
        elif module_number in MANAFACE_NUMBERS:
            return 2000+(module_number-90) # 101 -> 2011, 102 -> 2012, 103 -> 2013
        else:
            raise ValueError('unknown module number %d' % module_number)

class DigiCollector(BaseDigiCollector):
    MAX_DATA_PER_FETCH = 2**22 # 4m
    logger_name = 'digi'
    def dispatch(self, module):
        try:
            results = self.collect_through_manaface(self.nextra.active_manager, module)
            subdirectory = self.creating_results_subdirectory(module, self.filesystem_name)
            self.storing_collector_results('collected-from-%s' % self.nextra.active_manager, subdirectory, results)
        except CommandFailureException, error:
            self.logger.warning("failed collecting digi logs: %s" % error)

    @requires_living_node
    def collect_through_manaface(self, manaface, target_module):
        self.logger.info("[%s] collecting digi logs from manaface %s" % (target_module, manaface)) 
        runner = RemoteCommandRunner(self.module_ip(manaface), verifier=SimpleResultVerifier())
        potential_digi_files = self.find_digi_files_for_port(runner, self.digi_port_from_module_number(target_module))
        return self.collect_digi_data_from_digi_files(runner, potential_digi_files)

    def collect_digi_data_from_digi_files(self, runner, potential_digi_files):
        collected_bytes = ''
        desired_bytes = self.MAX_DATA_PER_FETCH
        while len(collected_bytes) < self.MAX_DATA_PER_FETCH and potential_digi_files:
            retval, out, err = runner.tail('-c', str(desired_bytes), potential_digi_files.pop())
            collected_bytes += out
        return collected_bytes
            
    def find_digi_files_for_port(self, runner, port):
        retval, out, err = runner.ls('-tr', os.path.sep + os.path.join('manager_logs', 'digi', 'port-%s-*' % port))
        return out.splitlines()

class IPMICollector(ModuleDependantCollector):
    MAX_DATA_PER_FETCH = 2**22 # 4m
    logger_name = 'ipmi'
    def dispatch(self, module, command):
        self.logger.info("[%s] collecting %r" % (module, command))
        answers_ping = self.is_ping_responsive(module)
        results = self.do_ipmi(module, command)
        subdirectory = self.creating_results_subdirectory(module, self.filesystem_name)
        self.storing_collector_results(self.normalize_path_component(command), subdirectory, results)
        if answers_ping and not self.is_ping_responsive(module):
            raise NextraExplorerFailed("running 'ipmi %s' killed module %s, don't use the Explorer and contact XIV" %
                (command, module))

    def do_ipmi(self, module, command):
        try:
            return IPMI(self.ipmi_ip(module), self.module_ip(module)).ipmitool(command)
        except IPMIException, error:
            self.logger.debug('%r failed on %s, %s' % (command, module, error))
            self.logger.warning("IPMI command %r failed on module %s, skipping" % (command, module))

    def ipmi_ip(self, module_number):
        return '2.%s.%s' % (self.nextra.system_id_as_octets, module_number)

class ProcTreeCollector(ModuleDependantCollector):
    MAX_DATA_PER_FETCH = 2**22 # 4m
    logger_name = 'proc'

    def dispatch(self, module, *paths):
        self.logger.info("[%s] collecting %s" % (module, ", ".join(paths)))
        results = self.fetch_proc_trees(module, paths)
        if results is None:
            return
        subdirectory = self.creating_results_subdirectory(module, 'FS')
        self.storing_collector_results(subdirectory, results)

    def normalize_proc_path_components(self, paths):
        for path in paths:
            if type(path) is not str:
                raise TypeError("path component %s isn't a string" % path)
            if path.startswith('/proc/'):
                yield repr(path)
            elif path.startswith('/'):
                yield repr('/proc' + path)
            else:
                yield repr('/proc/' + path)

    @log_exceptions_and_resume(CommandFailureException)
    @requires_living_node
    def fetch_proc_trees(self, module, paths):
        remote_script = """
import cPickle
import os
results = dict()
for path in (%(normalized_proc_paths)s):
    for dir, subdirs, filenames in os.walk(path):
        for filename in filenames: 
            results[os.path.join(dir, filename)] = file(os.path.join(dir, filename)).read()
print cPickle.dumps(results)
        """ % dict(normalized_proc_paths = ", ".join(self.normalize_proc_path_components(paths)))
        retval, out, err = RemoteCommandRunner(self.module_ip(module)).python('-c', remote_script)
        return cPickle.loads(out)

    @reraise_exceptions_as_explorer_failed_exception(IOError, CommandFailureException)
    def storing_collector_results(self, subdirectory, results):
        for relative_path, data in results.items():
            full_path = subdirectory+relative_path
            LocalCommandRunner().mkdir('-p', os.path.dirname(full_path))
            open(full_path, 'w').write(data)

class TreeCollector(ModuleDependantCollector):
    MAX_DATA_PER_FETCH = 2**26 # 64m
    LSI_DUMPS_TO_FETCH = 12
    logger_name = 'tree'
    @requires_living_node
    def dispatch(self, module, label, *include_list):
        self.logger.info("[%s] collecting %s" % (module, label))
        command = "tar -C / -c %s | head -c %s" % (" ".join(include_list), self.MAX_DATA_PER_FETCH)
        runner = RemoteCommandRunner(self.module_ip(module), verifier=SimpleResultVerifier(), basicRunString=command)
        results = runner.execute()[1]
        subdirectory = self.creating_results_subdirectory(module)
        self.storing_collector_results(subdirectory, results)

    @reraise_exceptions_as_explorer_failed_exception(OSError)
    def storing_collector_results(self, subdirectory, results):
        process = subprocess.Popen(("tar", "-x" "-C", subdirectory), stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
        out, err = process.communicate(results)
        if process.returncode != 0:
            self.logger.debug('extracting returned %s, %s' % (out, err))
            self.logger.warning('extraction failed, collected tree might be corrupt or incomplete')

    def format_args(self, args):
        return " ".join(str(arg) for arg in args)

    def collecting_lsi_dumps(self, module_number):
        command = "tar -c $(ls -tr /manager_logs/lsi_dumps | tail -%s) | tail -c %s" % (self.LSI_DUMPS_TO_FETCH,
                                                                                        self.MAX_DATA_PER_FETCH)
        return self.instantiate_command_runner(module_number, basicRunString=command).execute()[1]

class SataSmartCollector(ShellCollector):
    MAX_DATA_PER_FETCH = 2**24 # 16m
    logger_name = 'smart'
    @log_exceptions_and_resume(CommandFailureException, filter=lambda(error): '%s: %s' %
                               (error.host, error.output.strip()))
    def dispatch(self, module, sata_command):
        super(SataSmartCollector, self).dispatch(module, '/root_dir/stable/sata_smart', sata_command)

class UPSCredentialsError(Exception):
    pass
class UPSCollector(ModuleDependantCollector):
    FTP_CONNECTION_RETRIES = 2
    FTP_CONNECTION_TIMEOUT = 5
    MAX_DATA_PER_FETCH = 2**22 # 4m
    logger_name = 'ups'
    @requires_living_node
    def dispatch(self, ups_number):
        self.logger.info("[%s] collecting event log" % (ups_number))
        subdirectory = self.creating_results_subdirectory('UPS', ups_number)
        results = self.try_credentials(ups_number)
        self.storing_collector_results(filename = 'events.txt', subdirectory=subdirectory, data=results)

    @log_exceptions_and_resume(CommandFailureException)
    def try_credentials(self, ups_number):
        for username, password in (('root', 'rootroot'),
                                   ('root', 'root'),
                                   ('root', os.environ['XIV_INTERNAL_PASSWORD']),
                                   ('apc', 'apc')):
            try:
                return self.get_ups_log(ups_number, username, password)
            except UPSCredentialsError:
                pass
        raise NextraExplorerFailed("no known credentials work on UPS %s" % ups_number)
        
    def get_ups_log(self, ups_number, username, password):
        args = ('-c', '-u', username, '-r', self.FTP_CONNECTION_RETRIES, '-t', self.FTP_CONNECTION_TIMEOUT, '-p',
                password, self.module_ip(ups_number), 'event.txt')
        runner = RemoteCommandRunner(self.module_ip(self.nextra.active_manager))
        try:
            return runner.ncftpget(*args)[1][0:self.MAX_DATA_PER_FETCH]
        except CommandFailureException, error:
            if "password was not accepted for login" in error.stderr:
                raise UPSCredentialsError()
            raise

class LogFileCollector(ModuleDependantCollector):
    MAX_LINES_TO_FETCH = 100000
    logger_name = 'logfile'
    @requires_living_node
    def dispatch(self, module_number, directory, *log_file_names):
        self.logger.info('[%s] collecting %s log files' % (module_number, os.path.commonprefix(log_file_names)))
        subdirectory = self.creating_results_subdirectory(module_number, 'FS', *directory.split(os.path.sep))
        runner = RemoteCommandRunner(self.module_ip(module_number))
        total_fetched = 0
        for log_file_name in log_file_names:
            lines = self.fetch_lines(runner, os.path.join(directory, log_file_name))
            if not lines:
                break
            if len(lines) >= self.MAX_LINES_TO_FETCH - total_fetched:
                lines = lines[0:self.MAX_LINES_TO_FETCH - total_fetched]
            total_fetched += len(lines)
            self.storing_collector_results(filename=log_file_name, subdirectory=subdirectory, data="\n".join(lines))
            if total_fetched >= self.MAX_LINES_TO_FETCH:
                break

    def fetch_lines(self, runner, logfile):
        try:
            return runner.tail('-%s' % self.MAX_LINES_TO_FETCH, logfile)[1].splitlines()
        except CommandFailureException, error:
            self.logger.warning('fetching %s failed (%s), skipping rest' % (logfile, error.stderr.strip()))

class KDBCollectionFailed(Exception):
    pass
class KDBCollector(BaseDigiCollector):
    TELNET_SLEEP = 1
    MAXIMUM_COLLECTION_TIME = 30
    logger_name = 'kdb'
    @log_exceptions_and_resume(KDBCollectionFailed)
    def dispatch(self, module_number):
        try:
            self.open_console_handle(module_number)
        except socket.error:
            self.logger.error("can't connect to module %s's console: make sure digi is free and functional" %
                              (module_number))
            return
        if not self.is_in_kdb(module_number):
            self.logger.info("module %s (digi port %s) does not appear to be in kdb, skipping" %
                             (module_number, self.digi_port_from_module_number(module_number)))
            return
        self.logger.warning('module %s is in the kernel debugger, this output must reach XIV development!' %
                            module_number)
        subdirectory = self.creating_results_subdirectory(module_number, 'KDB')
        self.storing_collector_results("backtrace", subdirectory, self.collect_kdb_command('btc'))
        self.storing_collector_results("dmesg", subdirectory, self.collect_kdb_command('dmesg'))
        self.storing_collector_results("ps", subdirectory, self.collect_kdb_command('ps'))
        self.close_console_handle()

    def open_console_handle(self, module_number):
        self.console_handle = Telnet()
        self.console_handle.open(self.module_ip(DIGI_NUMBER), self.digi_port_from_module_number(module_number))
        time.sleep(self.TELNET_SLEEP)
        self.write_to_console(chr(21))
        self.read_from_console()

    def write_to_console(self, data):
        self.console_handle.write(data)
        time.sleep(self.TELNET_SLEEP)

    def read_from_console(self):
        return self.console_handle.read_very_eager()

    def close_console_handle(self):
        self.console_handle.close()

    def is_in_kdb(self, module_number):
        self.write_to_console('\r\n\r\n')
        response = self.read_from_console()
        self.logger.debug("module %s console returned %s" % (module_number, str_brief(response, lim=45)))
        if response.count('kdb') == 2:
            return True

    def collect_kdb_command(self, command):
        self.write_to_console(command + '\r\n')
        buffers = []
        while safe_forever_loop(self.MAXIMUM_COLLECTION_TIME):
            current_buffer = self.read_from_console()
            buffers.append(current_buffer)
            if 'more' not in current_buffer:
                break
            self.write_to_console('\r\n')
        if 'kdb' not in current_buffer:
            raise KDBCollectionFailed('kdb prompt lost during collection, skipping module')
        buffers.append('\n')
        return "".join(buffers)
