import logging
import time

from errors import NextraExplorerFailed

def log_function_name(level=logging.DEBUG, num_args_to_log=1):
    """Logs the name of the function it decorates.
    @param level: level to use for logging this message
    @num_args_to_log: number of arguments from *args to append to log, use None for none"""
    def decor(func):
        def callable(self, *args, **kwargs):
            message = func.__name__.replace('_', ' ')
            if num_args_to_log is not None:
                message = "%s: (%s)" % (message, "; ".join(str(obj) for obj in args[:num_args_to_log]))
            self.logger.log(level, message)
            return func(self, *args, **kwargs)
        callable.__name__ = func.__name__
        callable.__doc__ = func.__doc__
        return callable
    return decor

def reraise_exceptions_as_explorer_failed_exception(*exceptions):
    "Catches exceptions provided as *args, reraises them as a NextraExplorerFailed exception"
    def decor(func):
        def callable(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except exceptions, error:
                raise NextraExplorerFailed('%s: %s' % (func.__name__.replace('_', ' '), error))
        callable.__name__ = func.__name__
        callable.__doc__ = func.__doc__
        return callable
    return decor

def log_exceptions_and_resume(*exceptions, **logging_parms):
    """Catches exceptions provided as *args, logs them and proceeds.
    @param level: level to use for logging the exception
    @filter: filter the exception through this function prior to logging it"""
    def decor(func):
        def callable(self, *args, **kwargs):
            try:
                return func(self, *args, **kwargs)
            except exceptions, error:
                level = logging_parms.get('level', logging.WARNING)
                filter = logging_parms.get('filter', lambda(x): x)
                exception_message = filter(error)
                self.logger.log(level, '%s: %s' % (func.__name__, exception_message))
        callable.__name__ = func.__name__
        callable.__doc__ = func.__doc__
        return callable
    return decor

def dont_recollect(func):
    "Uses a caching mechanism to avoid running the decorated function twice"
    def callable(self, *args, **kwargs):
        key = (func.__name__, args, tuple((kwargs.items())))
        if key in self.results.executed_collectors:
            self.logger.debug('avoiding recollection of %s(%s, %s)' % (func.__name__, args, kwargs))
        else:
            retval = func(self, *args, **kwargs)
            self.results.executed_collectors[key] = retval
        return self.results.executed_collectors[key]
    callable.__name__ = func.__name__
    callable.__doc__ = func.__doc__
    return callable

def requires_living_node(func):
    """Tests that target node is alive (by ping) before executing the decorated function, does not execute and warns if the node is unreachable. Also verifies no harm came to the node as a result of the command and raises an exception if it died after the command.
The decorated function should receive an internal-network octet as an argument (1-8, 101-103, 150, 201-203, etc)"""
    def callable(self, module_number, *args, **kwargs):
        if not self.is_ping_responsive(module_number):
            self.logger.warning("can't collect from dead module %s, skipping" % (module_number))
            return 
        result = func(self, module_number, *args, **kwargs)
        if not self.is_ping_responsive(module_number):
            raise NextraExplorerFailed("%s lost ping after collection, stop using NextraExplorer and contact XIV" % \
                module_number)
        return result
    callable.__name__ = func.__name__
    callable.__doc__ = func.__doc__
    return callable

def store_results(attribute_name, num_args_to_append_to_attribute=1):
    def decor(func):
        def callable(self, *args, **kwargs):
            result = func(self, *args, **kwargs)
            if num_args_to_append_to_attribute is not None:
                formatted_args = [attribute_name]
                formatted_args.extend(str(arg).replace(' ', '_') for arg in args[0:num_args_to_append_to_attribute])
            setattr(self.results, '_'.join(formatted_args), result)
        callable.__name__ = func.__name__
        callable.__doc__ = func.__doc__
        return callable
    return decor
