import os.path
from subprocess import Popen, PIPE
from xiv.network.command_runner import RemoteCommandRunner, SimpleResultVerifier, CommandFailureException

class IPMIException(Exception):
    pass
class BinaryNotFound(IPMIException):
    pass
class UnknownIPMIPassword(IPMIException):
    pass
class IPMIReturnedNonZero(IPMIException):
    pass
class RemoteIPMIExecutionFailed(IPMIException):
    pass

class IPMI(object):
    IPMI_PATHS = ['/usr/bin/ipmitool']
    def __init__(self, ipmi_ip, internal_ip, root_password=None):
        self.ipmi_ip = ipmi_ip
        self.internal_ip = internal_ip
        self.ipmi_binary = self.detect_ipmi_binary()
        if 'XIV_INTERNAL_PASSWORD' in os.environ:
            root_password = os.environ['XIV_INTERNAL_PASSWORD']
        if root_password is not None:
            os.environ['IPMI_PASSWORD'] = root_password
        if 'IPMI_PASSWORD' not in os.environ:
            raise UnknownIPMIPassword('root password not provided and IPMI_PASSWORD not set in environment')
        self.safe_execute = RemoteCommandRunner(self.internal_ip, verifier=SimpleResultVerifier())

    def detect_ipmi_binary(self):
        for path in self.IPMI_PATHS:
            if os.path.isfile(path):
                return path
        raise BinaryNotFound("can't find ipmitool in %s." % ", ".join(path for path in self.IPMI_PATHS))

    def ipmitool(self, ipmi_args):
        execution_string = "%(ipmi_binary)s -U root -H %(host)s -E %(ipmi_args)s" % dict(
            ipmi_binary = self.ipmi_binary,
            host = self.ipmi_ip,
            ipmi_args = ipmi_args,
        )
        p = Popen(execution_string, shell=True, stdout=PIPE, stderr=PIPE)
        out, err = p.communicate()
        if p.returncode == 0:
            return out
        raise IPMIReturnedNonZero(str(err).replace('\n', '; '))

    def is_chassis_power_on(self):
        out = self.ipmitool('chassis power status')
        return "is on" in out

    def is_ipmi_accessible(self):
        try:
            self.ipmitool('mc info')
            return True
        except IPMIReturnedNonZero:
            return False

    def remote_ipmitool(self, *ipmi_args):
        try:
            self.safe_execute.ipmitool(*ipmi_args)
        except CommandFailureException, e:
            raise RemoteIPMIExecutionFailed(str(e))
 
    def set_ipmi_ip_on_module(self):
        self.remote_ipmitool('lan', 'set', '1', 'ipaddr', self.ipmi_ip, 'netmask', '0.0.0.0')
