#!/usr/bin/python

import logging
from xiv.network import command_runner
import re
import os

LOCAL_FC_UTILS_PATH = '/opt/xiv/python/pspackages/xiv_xsf/nextra/hardware/manaface'
MANAFACE_FC_UTILS_PATH = '/tmp'

class FcPortError(Exception):
    pass

class FcPortMismatchSettings(FcPortError):
    pass

class FCPort(object):
    """
    loads and setes FC port configuration
    """
    configuration_regexps = {
        'topology': re.compile("Link Topology:\s*(?P<link_topology>Auto)"),
        'active_firmware': re.compile("Current active firmware version is [0-9a-f]+ \((?P<active_firmware>[0-9.]+)\)"),
        'firmware': re.compile("Firmware image's version is (?P<hba_model>[A-Z0-9]+)-(?P<firmware_version>[0-9.]+).*"),
        'link_speed': re.compile("Link Speed:\s*(?P<link_speed>Auto|[0-9](?= +Gb))"),
        'initiator': re.compile("FCP Initiator protocol:\s*(?P<is_initiator>Disabled|Enabled)"),
        'target': re.compile("FCP Target protocol:\s*(?P<is_target>Disabled|Enabled)"),
        'lan': re.compile("LAN protocol:\s*(?P<is_lan>Disabled|Enabled)"),
        'bus_assignment': re.compile("Assignment of Bus/Target IDs:\s*(?P<sort_by>SortByWWN)"),
        'immediate_error_reply': re.compile("Immediate Error Reply:\s*(?P<immdiate_error_reply>Disabled|Enabled)"),
        'maintain_logins': re.compile("Maintain Logins:\s*(?P<maintain_logins>Disabled|Enabled)"),
        'al_pa': re.compile("Hard AL_PA:\s*(?P<hard_al_pa>[0-9a-fA-F]+)"),
        'device_timeout': re.compile("Initiator Device Timeout:\s*(?P<device_timeout>[0-9]+)"),
        'io_pending_timeout': re.compile("Initiator I/O Pending Timeout:\s*(?P<io_pending_timeout>[0-9]+)"),
        'multipathing': re.compile("Multi-pathing:\s*(?P<multi_pathing>Disabled|Enabled)"),
        'int_coalescing': re.compile("Interrupt Coalescing:\s*(?P<int_co_switch>Disabled|Enabled)(, timeout is (?P<int_co_timeout>[0-9]+) us, depth is (?P<int_co_depth>[0-9]+))?")
        }

    def __init__(self, manaface_ip, port_number, logger_name):
        self.logger = logging.getLogger(logger_name)
        self.manaface_ip = manaface_ip
        self.port_number = port_number

    def load_configuration(self):
        configuration_dict = {}
        unmatched = self.configuration_regexps.keys()[:]
        try:
            command = command_runner.RemoteCommandRunner(self.manaface_ip, "", command_runner.SimpleResultVerifier(),user="root")
            result, output, stderr = command.executeStr(s='lsiutil -p %s -i' % (self.port_number+1))
        except (command_runner.CommandFailureException, command_runner.CommandTimeout), e:
            raise FcPortError('Failed to get port %d configuration on %s: %s' % (self.port_number, self.manaface_ip, e))

        for line in output.split('\n'):
            for key in unmatched:
                if self.configuration_regexps[key].match(line) is not None:
                    configuration_dict.update(self.configuration_regexps[key].match(line).groupdict())
                    unmatched.remove(key)
                    break
        self.__dict__.update(configuration_dict)
        if ('firmware' in unmatched):
            self.hba_model, self.firmware_version = ('', '0')
            unmatched.remove('firmware')
        if configuration_dict['link_speed'] == 'Auto':
            self.link_speed = 0
        else:
            self.link_speed = int(configuration_dict['link_speed'])
        if len(unmatched) > 0:
            self.logger.warning('The following configuration attributes were not found by lsiutil: ' + ','.join(unmatched))

class FCPortsConfiguratorer(object):
    MODULE_NUMBER_TO_NODE_NUMBER = {
            101: 1,
            102: 2,
            103: 3
    }
    def __init__(self, connection_ip, system_id, module_id, logger_name):
        self.logger = logging.getLogger(logger_name)
        self.connection_ip = connection_ip
        self.system_id = system_id
        self.module_id = module_id
        self.fc_port_numbers = []
        self.ports_configuration = {}

    @property
    def system_id_for_wwnn(self):
        return '%s%s' % (int(self.system_id)/1000, int(self.system_id)%1000)

    def available_fc_ports(self):
        """
        could only be called after configure_wwnns or on active manaface (drivers should be up)
        """
        try:
            ioc_list = []
            ioc_regexp = re.compile("ioc(?P<ioc_idx>[0-9]+):.*")
            command = command_runner.RemoteCommandRunner(self.connection_ip, "", command_runner.SimpleResultVerifier(),user="root")
            result, output, stderr = command.executeStr(s='cat /proc/mpt/summary')
            for line in output.split('\n'):
                if ioc_regexp.match(line) is not None:
                    yield int(ioc_regexp.match(line).group('ioc_idx'))
        except (command_runner.CommandFailureException, command_runner.CommandTimeout), e:
            raise FcPortError('Failed to get available ports: %s' % e)

    def configure_wwnns(self):
        parameters_string = '-w %s %s' % (self.system_id_for_wwnn, self.MODULE_NUMBER_TO_NODE_NUMBER[self.module_id])
        try:
            command = command_runner.RemoteCommandRunner(self.connection_ip, "", command_runner.SimpleResultVerifier(),user="root")
            command.uploadFile(os.path.join(LOCAL_FC_UTILS_PATH,'config_system_fc.py'), os.path.join(MANAFACE_FC_UTILS_PATH,'config_system_fc.py'))
            command.uploadFile(os.path.join(LOCAL_FC_UTILS_PATH,'drivers.py'), os.path.join(MANAFACE_FC_UTILS_PATH,'drivers.py'))
            self.logger.debug('python %s %s' % (os.path.join(MANAFACE_FC_UTILS_PATH,'config_system_fc.py'), parameters_string))
            result, output, stderr = command.executeStr(s='python %s %s' % (os.path.join(MANAFACE_FC_UTILS_PATH,'config_system_fc.py'), parameters_string))
        except command_runner.CommandFailureException, e:
            try:
                self.logger.warning('config_system_fc.py returned invalid return code, trying again, could be false error')
                result, output, stderr = command.executeStr(s='python %s %s' % (os.path.join(MANAFACE_FC_UTILS_PATH,'config_system_fc.py'), parameters_string))
            except (command_runner.CommandFailureException, command_runner.CommandTimeout), e:
                raise FcPortError('Failed to configure ports on %s: %s' % (self.connection_ip, e))
        except command_runner.CommandTimeout, e:
            raise FcPortError('Failed to configure ports on %s: %s' % (self.connection_ip, e))

    def load_port_numbers(self):
        """
        could only be called after configure_wwnns or on active manaface (drivers should be up)
        """
        self.fc_port_numbers = list(self.available_fc_ports())

    def load_ports_configuration(self):
        for port in self.fc_port_numbers:
            self.ports_configuration[port] = FCPort(self.connection_ip, port, self.logger.name)
            self.ports_configuration[port].load_configuration()

    def configure_fc_ports_from_configuratorer(self, source_configuratorer):
        parameters_string = '-w %s %s' % (self.system_id_for_wwnn, self.MODULE_NUMBER_TO_NODE_NUMBER[self.module_id])
        for port in self.fc_port_numbers:
            if source_configuratorer.ports_configuration[port].is_initiator == 'Enabled' and source_configuratorer.ports_configuration[port].is_target == 'Enabled':
                parameters_string = parameters_string + ' -p %s b %s' % (port, source_configuratorer.ports_configuration[port].link_speed)
            elif source_configuratorer.ports_configuration[port].is_initiator == 'Enabled':
                parameters_string = parameters_string + ' -p %s i %s' % (port, source_configuratorer.ports_configuration[port].link_speed)
            elif source_configuratorer.ports_configuration[port].is_target == 'Enabled':
                parameters_string = parameters_string + ' -p %s t %s' % (port, source_configuratorer.ports_configuration[port].link_speed)
            else:
                raise FcPortError('port %s must be set as initiator, target or both' % port)

        try:
            command = command_runner.RemoteCommandRunner(self.connection_ip, "", command_runner.SimpleResultVerifier(),user="root")
            command.uploadFile(os.path.join(LOCAL_FC_UTILS_PATH,'config_system_fc.py'), os.path.join(MANAFACE_FC_UTILS_PATH,'config_system_fc.py'))
            command.uploadFile(os.path.join(LOCAL_FC_UTILS_PATH,'drivers.py'), os.path.join(MANAFACE_FC_UTILS_PATH,'drivers.py'))
            self.logger.debug('python %s %s' % (os.path.join(MANAFACE_FC_UTILS_PATH,'config_system_fc.py'), parameters_string))
            result, output, stderr = command.executeStr(s='python %s %s' % (os.path.join(MANAFACE_FC_UTILS_PATH,'config_system_fc.py'), parameters_string))
        except (command_runner.CommandFailureException, command_runner.CommandTimeout), e:
            raise FcPortError('Failed to configure ports on %s: %s' % (self.connection_ip, e))

def verify_configurators_fc_ports_settings_match(configurators_dict):
    first_one = True
    for module_number in configurators_dict:
        if first_one:
            ports_configuration = configurators_dict[module_number].ports_configuration
            first_one = False
        else:
            if len(ports_configuration.keys()) != len(configurators_dict[module_number].ports_configuration.keys()):
                raise FcPortMismatchSettings('Number of ports is different between current active manafaces')
            for port in ports_configuration:
                if ports_configuration[port].firmware_version != configurators_dict[module_number].ports_configuration[port].firmware_version:
                    raise FcPortMismatchSettings('Port %s firmware is different between current active manafaces' % port)
                if ports_configuration[port].link_speed != configurators_dict[module_number].ports_configuration[port].link_speed:
                    raise FcPortMismatchSettings('Port %s link speed is different between current active manafaces' % port)
                if ports_configuration[port].is_initiator != configurators_dict[module_number].ports_configuration[port].is_initiator:
                    raise FcPortMismatchSettings('Port %s initiator role is different between current active manafaces' % port)
                if ports_configuration[port].is_target != configurators_dict[module_number].ports_configuration[port].is_target:
                    raise FcPortMismatchSettings('Port %s target role is different between current active manafaces' % port)

