"""
This file facilitates querying port information from the switches found in P9 Nextras (Dell PowerConnect).
The file is mostly based on code ripped out of sysprobe (//leia/p9_2/leia/utils/sysprobe), a non-field safe tool which provides a lot more functionality.

The class P9NextraSwitchPortQuerier handles some ssh-tunneling-fu to enable the monitor (chubaka) to run this script, communicating with the switches over tunneled ssh connections. See P9NextraSwitchPortQuerier's constructor documentation for usage.
"""

import socket
import string
import signal
import subprocess
import os
import random
import signal
import types
import sys
from time import time, sleep

from xiv.pyutils import weakref, wait, safe_forever_loop

TELNET_PORT = 23
VALID_SWITCH_NUMBERS = (100, 200)
MINIMUM_TIME_BETWEEN_SWITCH_CONNECTIONS = 0.1
SWITCH_CONNECTION_TIMEOUT = 3*60

class SwitchOperationFailed(Exception):
    pass

class P9NextraSwitchPortQuerier(object):
    def __init__(self, nextra):
        super(P9NextraSwitchPortQuerier, self).__init__()
        if 'XIV_INTERNAL_PASSWORD' not in os.environ:
            raise AssertionError('XIV_INTERNAL_PASSWORD not set in environment')
        self.nextra = weakref(nextra)
        self.sock_commander = SwitchSockCommander('127.0.0.1', 
            username='root', password=os.environ['XIV_INTERNAL_PASSWORD'])
        self.tunneling_process = None
        self.tunneling_port = None
        self.connection_throttling_dictionary = {}

    def __repr__(self):
        return '<P9NextraSwitchPortQuerier>'

    def switch_ip(self, switch):
        if int(switch) not in VALID_SWITCH_NUMBERS:
            raise ValueError('invalid switch number %s' % switch)
        return '14.%s.%s' % (self.nextra.system_id_as_octets, switch)

    def pick_free_random_port(self):
        for attempts in range(0,10):
            port = self.pick_random_port()
            if port is not None:
                return port
        raise AssertionError('unable to find a free port to use for tunneling')

    def pick_random_port(self):
        port = random.randint(2**15, 2**16-1)
        test_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        test_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            test_socket.bind(('127.0.0.1', port))
            del(test_socket)
        except socket.error:
            return None
        return port

    def test_forwarding(self):
        assert self.tunneling_port is not None, 'tried testing null port'
        try:
            test_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            test_socket.connect(('127.0.0.1', self.tunneling_port))
        except socket.error:
            return False
        del(test_socket)
        return True

    def establish_tunnel(self, target_ip):
        assert self.tunneling_process is None, 'tried to establish tunnel when tunnel already established'
        # there's a race here; to be totally safe we should read ssh's stderr and verify tunneling succeeded;
        #  otherwise, we can pick a port, and then have someone else bind to that port before ssh takes it
        self.tunneling_port = self.pick_free_random_port()
        ssh_parameters = dict(
                manager_ip = self.nextra.get_manager_ip(self.nextra.active_manager),
                source_port = self.tunneling_port,
                target_port = TELNET_PORT,
                target_ip = target_ip,
            )
        self.tunneling_process = subprocess.Popen(
            args = 'ssh -N %(manager_ip)s -L %(source_port)s:%(target_ip)s:%(target_port)s' % ssh_parameters,
            shell = True,
            stdout = subprocess.PIPE,
            stderr = subprocess.PIPE,
        )
        wait(
            self.test_forwarding,
            interval = 0.1,
            timeout = 3,
            exception = AssertionError('port forwarding failed to forward connections after 3 seconds')
        )

    def tear_down_tunnel(self):
        assert self.tunneling_process is not None, 'tried to tear down non-established tunnel'
        os.kill(self.tunneling_process.pid, signal.SIGKILL)
        wait(
            lambda: self.tunneling_process.poll() is not None,
            timeout = 1,
            interval = 0.1,
            exception = AssertionError("tunnel process failed to die")
        )
        self.tunneling_process = None
        self.port = None

    # decorators don't need to receive self, if used only inside their class
    def switch_command_decorator(func):
        def callable(self, switch, *args, **kwargs):
            self.wait_if_necessary_to_throttle_connection_speed_to_switch(switch)
            self.establish_tunnel(self.switch_ip(switch))
            try:
                try:
                    self.sock_commander.connect(self.tunneling_port)
                    return func(self, switch, *args, **kwargs)
                    self.sock_commander.close()
                except socket.error, e:
                    raise SwitchOperationFailed('socket error while talking with switch')
            finally:
                self.tear_down_tunnel()
        callable.__name__ = func.__name__
        callable.__doc__ = func.__doc__
        return callable

    def wait_if_necessary_to_throttle_connection_speed_to_switch(self, switch):
            latest_connection_time = self.connection_throttling_dictionary.pop(switch, None)
            current_time = time()
            if latest_connection_time is not None:
                if current_time - latest_connection_time < MINIMUM_TIME_BETWEEN_SWITCH_CONNECTIONS:
                    sleep(MINIMUM_TIME_BETWEEN_SWITCH_CONNECTIONS - (current_time - latest_connection_time))
            self.connection_throttling_dictionary[switch] = time()

    @switch_command_decorator
    def query_port_info(self, switch, port):
        return self.query_port_over_open_session(port)

    @switch_command_decorator
    def query_ports_info(self, switch, ports):
        if type(ports) is types.IntType or type(ports) is types.StringType:
            ports = [int(ports)]
        if not hasattr(ports, '__iter__'):
            raise TypeError("provided port list isn't iterable")
        result = {}
        for port in ports:
            result[port] = self.query_port_over_open_session(port)
        return result

    def query_cache_node_info(self, module):
        if int(module) < 1 or int(module) > 8:
            raise ValueError('module %s is not a cache node number' % module)
        for switch in VALID_SWITCH_NUMBERS:
            yield self.query_port_info(switch, module)

    def query_port_over_open_session(self, port):
        s = tokenize(" ", self.sock_commander.command("show bridge address-table ethernet g" + str(port))[-1:][0])[1]
        if not s.startswith("-"):
            return s.lower()

def tokenize(token, string):
    return filter(lambda s: 0 < len(s), [s.strip() for s in string.split(token)])

class SockInteractor(object):
    def __init__(self, ip, port = None, newline = "\n", timeout = None):
        self.ip = str(ip)
        self.port = port
        self.newline = newline
        self.timeout = timeout
        self.buffer = []

    def __repr__(self):
        return '<SockInteractor>'

    def open(self, port=None):
        if port is not None:
            self.port = port
        assert self.port is not None, 'tried opening null port'
        self.so = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.so.connect((self.ip, self.port))
        if self.timeout is not None:
            self.so.settimeout(self.timeout)

    def close(self):
        self.so.close()

    def parse(self, string):
        for dummy in safe_forever_loop(SWITCH_CONNECTION_TIMEOUT, 
                                       SwitchOperationFailed('timeout expired for switch operation')):
            index = string.find("\x1b")
            if index < 0:
                break
            string = string[0:index] + string[index + 3:]
        return tokenize("\n", string.replace("\r", ""))

    def getch(self):
        if len(self.buffer):
            return self.buffer.pop()
        return self.so.recv(1)

    def ungetch(self, c):
        self.buffer.append(c)

    def readto(self, tokens):
        lens = [len(x) for x in tokens]
        idxs = [0 for x in tokens]
        argc = range(0, len(tokens))
        s = ""
        for dummy in safe_forever_loop(SWITCH_CONNECTION_TIMEOUT, 
                                       SwitchOperationFailed('timeout expired for switch operation')):
            c = self.getch()
            if 0 == len(c):
                return (s, None)
            s += c
            preempted = False
            for i in argc:
                if lens[i] <= idxs[i]:
                    if not preempted:
                        for j in xrange(idxs[i], lens[i] - 1, -1):
                            self.ungetch(s[-1])
                            s = s[:-1]
                        return (s[:-lens[i]], tokens[i])
                    else:
                        idxs[i] += 1
                elif c == tokens[i][idxs[i]]:
                    idxs[i] += 1
                    if idxs[i] == lens[i] and not preempted:
                        return (s[:-lens[i]], tokens[i])
                else:
                    idxs[i] = 0
                if 0 < idxs[i]:
                    preempted = True

    def write(self, string):
        s = str(string)
        self.so.send(s)
        return len(s)

    def writeln(self, string= ""):
        return self.write(str(string) + self.newline)

class SockCommander(object):
    def __init__(self, ip, username_prompt, username, password_prompt, password, prompt, newline = "\n",
      exit_command = None, repeat_prompt = None, repeat_command = None):
        self.username_prompt = username_prompt
        self.username = username
        self.password_prompt = password_prompt
        self.password = password
        self.prompt = prompt
        self.exit_command = exit_command
        self.repeat_command = repeat_command
        self.prompts = [self.prompt]
        if repeat_prompt is not None:
            self.prompts = [repeat_prompt] + self.prompts
        self.interactor = SockInteractor(ip, TELNET_PORT, newline, timeout=1)

    def __repr__(self):
        return 'SockCommander'

    def connect(self, port=None):
        self.interactor.open(port)
        self.interactor.readto([self.username_prompt])
        self.interactor.writeln(self.username)
        self.interactor.readto([self.password_prompt])
        self.interactor.writeln(self.password)
        return self.interactor.parse(self.interactor.readto([self.prompt])[0])

    def command(self, command, nonewline = False):
        if nonewline:
            index = self.interactor.write(str(command))
        else:
            index = self.interactor.writeln(str(command))
        result = []
        for dummy in safe_forever_loop(SWITCH_CONNECTION_TIMEOUT, 
                                       SwitchOperationFailed('timeout expired for switch operation')):
            rv = self.interactor.readto(self.prompts)
            result += self.interactor.parse(rv[0][index:])
            index = 0
            if self.prompt == rv[1]:
                return result
            self.interactor.write(self.repeat_command)

    def close(self):
        if self.exit_command is not None:
            self.interactor.write(self.exit_command)
        self.interactor.close()

class SwitchSockCommander(SockCommander):
    def __init__(self, ip, username, password):
        SockCommander.__init__(self, ip,
          username_prompt = "User Name:",
          username = username,
          password_prompt = "Password:",
          password = password,
          prompt = "console# ",
          newline = "\n",
          exit_command = "exit",
          repeat_prompt = "More: <space>,  Quit: q, One line: <return> ",
          repeat_command = " "
        )

    def __repr__(self):
        return '<SwitchSockCommander>'
