"""A module representing possible UPS access mediums. At the time of this writing, the only supported medium is telnet.
 * TODO:
    * Improve encapsulation of exceptions (TelnetSessionInterruptedError is poor)
    * Define 'pure-virtual' functions of the Medium class, to serve as an API
        for the module."""

import logging
import re
import telnetlib
import socket
from exceptions import UPSException

TELNET_TIMEOUT = 30

class MediumError(UPSException):
    """Base class for medium errors. Keeps the original exception object
    as a member."""
    def __init__(self, message=None, original_exception=None):
        if original_exception != None:
            self.original_exception = original_exception
        UPSException.__init__(self, message)

class TimeoutError(MediumError):
    """Thrown when the UPSs session doesn't respond within the timeout."""
    def __init__(self, expected, received):
        MediumError.__init__(self, 'timeout waiting for "%s", got "%s"' % (expected, received))
        self.expected = expected
        self.received = received

class TelnetSessionInterruptedError(MediumError):
    """Thrown when a read or write operation fails due to a telnetlib
    exception."""
    pass

class CantConnectError(MediumError):
    """Thrown when telnetlib is unable to establish a session with a UPS.
    The member self.exception contains the exception telnetlib has thrown."""
    pass

class InvalidParametersError(MediumError):
    """Thrown when a UPS is initialized with incorrect paramers (no username/
    password pair in the environment, malformed IP, etc)"""
    pass

class Medium(object):
    """Represents possible UPS Mediums."""
    def __init__(self, medium_parameter, logger_name):
        pass

class Telnet(Medium):
    """UPS-over-telnet/digi control class."""
    def __init__(self, logger_name, ip, port = 23):
        """Initializes a UPS object, without talking to the UPS at all."""
        super(Telnet, self).__init__(ip, logger_name)
        self.logger = logging.getLogger(logger_name + '.tn')
        self.set_ip(ip)
        self.set_port(port)
        self.timeout = TELNET_TIMEOUT
        self.telnet_session = None
        socket.setdefaulttimeout(3)

    def set_ip(self, ip):
        pattern = re.compile(r"""
            \b                   # beginning of the string
            ((25[0-5]|           # integer range 250-255 OR
            2[0-4][0-9]|         # integer range 200-249 OR
            [01]?[0-9][0-9]?)    # other combinations of 1-3 digits below 200
            \.){3}               # matches '.', repeats three octets
            (25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$       # repeat for last octet
            """, re.VERBOSE)
        if re.match(pattern, ip):
            self.ip = ip
        else:
            raise InvalidParametersError("invalid ip address (%s)" % ip)

    def set_port(self, port):
        if port >= 2014 and port <= 2016:
            self.port = port
        else:
            raise InvalidParametersError("invalid port number (%s)" % port)

    def open_connection(self):
        """Opens a connection to the UPS, throw a CantConnectError exception
        upon failure."""
        self.logger.log(5, 'opening connection')
        try:
            self.telnet_session = None
            self.telnet_session = telnetlib.Telnet(self.ip, self.port)
            if self.telnet_session == None:
                raise Exception("unknown error")
        except Exception, e:
            self.logger.log(6, 'connection failed')
            raise CantConnectError('opening telnet: %s' % e.__str__(), e)

    def close_connection(self):
        """Closes connection to the UPS."""
        self.logger.log(5, 'closing connection')
        if self.telnet_session == None:
            self.logger.log(6, 'closing telnet: no session to close')
            return
        self.telnet_session.close()

    def read_until(self, until_what, first_read=False, log_timeout=True):
        """A wrapper around telnetlib's read_until. This wrapper raises
        exceptions on read failure:
         * TimeoutError: On timeout
         * CantConnectError: very commonly, the session will open correctly but 
                the first read or write will trigger an exception. this is because 
                when a UPS already has another telnet session opened against it, 
                it accepts the TCP session open but immediately close the 
                connection silently.
         * TelnetSessionInterruptedError: For any other exception"""
        self.logger.log(5, 'reading until "%s"' % until_what.strip())
        try:
            ret = self.telnet_session.read_until(until_what, self.timeout)
            if until_what not in ret:
                raise TimeoutError(until_what, ret)
        except TimeoutError, e:
            if log_timeout:         
                self.logger.log(logging.DEBUG, str(e))
            raise
        except Exception, e:
            if first_read:
                error_string = 'first read: ' + str(e)
                self.logger.log(7, error_string)
                raise CantConnectError(error_string, e)
            else:
                error_string = 'read: ' + str(e)
                self.logger.log(logging.DEBUG, error_string)
                raise TelnetSessionInterruptedError(error_string, e)
        return ret

    def write(self, write_what):
        """Write a line to the UPS through an open telnet connection.
        Throws only TelnetSessionInterrupted for any failure."""
        self.logger.log(5, 'writing "%s"' % write_what.strip())
        try:
            ret = self.telnet_session.write(write_what)
        except Exception, e:
            raise TelnetSessionInterruptedError('writing telnet: %s' % e.__str__(), e)
        return ret

    def write_single_escape(self):
        self.write(chr(27))

