import time
import medium
from exceptions import UPSException, UPSManagerException, CommandFailedError, CredentialsIncorrectError
import logging
from xiv.dtypes import Bunch
from xiv_xsf.nextra import p9constants
from xiv_xsf.nextra.p9client import system_id_utils

DEFAULT_PROMPT = '> '
UPS_MAXIMUM_MENU_DEPTH = 4

def medium_session(operation):
    """This decorator opens and closes a session on the medium before executing operation."""
    def callable(self, *args, **kwargs):
        try:
            try:
                self.medium.open_connection()
                return operation(self, *args, **kwargs)
            except (medium.TimeoutError, medium.CantConnectError, medium.TelnetSessionInterruptedError, ValueError), e:
                self.logger.log(5, 'caught %s in %s' % (e, operation.__name__))
                exc = CommandFailedError('%s failed: %s' %(operation.__name__, e))
                exc.original_exception = e
                raise exc
        finally:
            self.medium.close_connection()
    return callable


def login_session(operation):
    @medium_session
    def callable(self, *args, **kwargs):
        try:
            self.login()
            retval = operation(self, *args, **kwargs)
        finally:
            self.logout()
        return retval
    return callable

class SessionManager(object):
    """Drives a UPS by means of a Medium object. Navigates through UPS screens
    to perform a specific action or find specific data.
    ip and port could be direct tcpip or through digi:
    sm = sessionmanager.DigiPortSessionManager('ran', '14.10.180.150', 2014, {'username': 'root', 'password': '123456'})
    sm.configure_tcpip(180, 201)
    """
    def __init__(self, logger_name, ip, port, credentials):
        manager_name = logger_name + '.mg'
        self.logger = logging.getLogger(manager_name)
        self.credentials = credentials
        self.medium = medium.Telnet(manager_name , ip, port)

    def login(self, credentials=None):
        if credentials is None:
            credentials = self.credentials
        self.logger.log(6, 'logging in')
        self.medium.write('\r\r')
        self.medium.read_until('User Name', True)
        self.medium.write(credentials['username']+'\n\r')
        self.medium.read_until('Password')
        self.medium.write(credentials['password']+'\n\r')
        try:
            self.medium.read_until(DEFAULT_PROMPT, log_timeout=False)
        except medium.TimeoutError, e:
            if 'User Name' in e.received:
                self.logger.debug('incorrect credentials')
                raise CredentialsIncorrectError()
            raise

    def logout(self):
        self.logger.log(6, 'logging out')
        self.escape_to_entry_screen()
        self.medium.write('4\n\r') # logout
        self.medium.read_until('Bye')

    @login_session
    def fetch(self):
        self.logger.log(7, 'fetching updated ups status')
        return Bunch(
            entry = self.entry_screen(),
            device = self.device_manager_screen(),
        )

    def check_credentials(self):
        try:
            self.logger.log(7, 'checking ability to login with the ups')
            login_session(lambda(dummy): True)(self)
        except (CommandFailedError, CredentialsIncorrectError), e:
            self.logger.debug('login to ups failed')
            return False
        return True


    def check_medium(self):
        try:
            self.logger.log(7, 'checking ability to open session with the ups')
            medium_session(lambda(dummy): True)(self)
        except CommandFailedError, e:
            self.logger.debug('opening test session failed')
            return False
        return True

    @medium_session
    def change_ups_credentials(self, old_credentials, new_credentials):
        try:
            self.logger.log(7, 'changing credentials')
            self.login(old_credentials)
            self.make_menu_selection(3, 'system menu')
            self.make_menu_selection(1, 'user manager')
            self.make_menu_selection(1, 'administrator', wait_for_prompt=False)
            self.medium.read_until('Enter current Administrator password', True)
            self.medium.write(old_credentials['password']+'\n\r')
            self.medium.read_until('Enter new Administrator user name', True)
            self.medium.read_until('to skip', True)
            self.medium.write(new_credentials['username']+'\n\r')
            self.medium.read_until('User name has been changed successfully', True)
            self.medium.read_until('to skip', True)
            self.medium.write(new_credentials['password']+'\n\r')
            self.medium.read_until('Confirm new Administrator password', True)
            self.medium.write(new_credentials['password']+'\n\r')
            self.medium.read_until('Password has been changed successfully', True)
            self.medium.read_until('to skip', True) # skip passphrase
            self.medium.write('\n\r')
            self.medium.write('\n\r')   # get back to user management menu
            self.medium.write('4\n\r')  # save changes
        finally:
            self.logout()

    @login_session
    def configure_tcpip(self, dotted_system_id, ups_id):
        """
        configure_tcpip(self, system_id, ups_id)
        example: configure_tcpip(10.140, 203)
        configure:
            ups hostname - ups-system_id-ups_id
            ip address - 14.system_id.ups_id
            subnet mask - 255.255.255.0
            default gateway - 14.system_id.100
            and more: port speed, manual boot
        """
        self.logger.log(7, 'configuring TCP/IP')
        system_id_utils.verify_dotted_system_id_string(dotted_system_id)
        self.verify_ups_id(ups_id)
        self.make_menu_selection(2, 'network menu')
        self.make_menu_selection(1, 'tcp/ip menu')
        self.make_menu_selection(5, 'advanced configuration')
        self.make_menu_selection(1, 'Hostname configuration', False)
        self.medium.read_until('Host Name', True)
        self.medium.write('ups-%s-%s\n\r' % (system_id_utils.system_id_string_to_low_octet(dotted_system_id), ups_id))
        self.wait_for_prompt()
        self.make_menu_selection(3, 'Port Speed configuration')
        self.make_menu_selection(5, 'selecting 100 Full')
        self.make_menu_selection(4, 'accepting changes')
        self.escape(1)
        self.make_menu_selection(4, 'bootmode menu')
        self.make_menu_selection(4, 'select Manual')
        self.make_menu_selection(1, 'IP configuration', False)
        self.medium.read_until('System IP', True)
        self.medium.write('14.%s.%s\n\r' % (dotted_system_id, ups_id))
        self.wait_for_prompt()
        self.make_menu_selection(2, 'Subnet Mask configuration', False)
        self.medium.read_until('Subnet Mask', True)
        self.medium.write('255.255.255.0\n\r')
        self.wait_for_prompt()
        self.make_menu_selection(3, 'Default Gateway configuration', False)
        self.medium.read_until('Default Gateway', True)
        self.medium.write('14.%s.100\n\r' % dotted_system_id)
        self.wait_for_prompt()

    @login_session
    def configure_syslog(self, dotted_system_id):
        """
        configure syslog parameters in the ups
        example: configure_syslog(140)
        """
        self.logger.log(7, 'configuring syslog')
        system_id_utils.verify_dotted_system_id_string(dotted_system_id)
        self.make_menu_selection(2, 'network menu')
        self.make_menu_selection(9, 'syslog menu')
        self.make_menu_selection(1, 'syslog settings menu')
        self.make_menu_selection(1, 'syslog option')
        self.make_menu_selection(2, 'enable syslog')
        self.make_menu_selection(3, 'accept changes')
        self.escape(1)
        self.make_menu_selection(2, 'server1 menu')
        self.make_menu_selection(1, 'server1 ip config', False)
        self.medium.read_until('Server IP', True)
        self.medium.write('14.%s.101\n\r' % dotted_system_id)
        self.wait_for_prompt()
        self.make_menu_selection(3, 'accept changes')
        self.escape(1)
        self.make_menu_selection(3, 'server2 menu')
        self.make_menu_selection(1, 'server2 ip config', False)
        self.medium.read_until('Server IP', True)
        self.medium.write('14.%s.102\n\r' % dotted_system_id)
        self.wait_for_prompt()
        self.make_menu_selection(3, 'accept changes')
        self.escape(1)
        self.make_menu_selection(4, 'server3 menu')
        self.make_menu_selection(1, 'server3 ip config', False)
        self.medium.read_until('Server IP', True)
        self.medium.write('14.%s.103\n\r' % dotted_system_id)
        self.wait_for_prompt()
        self.make_menu_selection(3, 'accept changes')

    @login_session
    def configure_snmp(self):
        self.logger.log(7, 'configuring snmp')
        self.make_menu_selection(2, 'network menu')
        self.make_menu_selection(8, 'snmp menu')
        self.make_menu_selection(1, 'snmp settings')
        self.make_menu_selection(1, 'snmpv1 access')
        self.make_menu_selection(2, 'enable snmp')
        self.make_menu_selection(3, 'accept changes')

    @login_session
    def configure_device_manager(self):
        self.logger.log(7, 'configuring device manager')
        self.make_menu_selection(1, 'device manager menu')
        self.make_menu_selection(1, 'Smart UPS')
        self.make_menu_selection(3, 'Configuration')
        self.make_menu_selection(4, 'General')
        self.make_menu_selection(2, 'UPS Position')
        self.make_menu_selection(2, 'RACK')
        self.make_menu_selection(6, 'accept changes')

    @login_session
    def configure_date(self):
        self.logger.log(7, 'configuring device manager')
        self.make_menu_selection(3, 'System menu')
        self.make_menu_selection(3, 'Date/Time')
        self.make_menu_selection(1, 'Manual')
        self.make_menu_selection(1, 'Set Date', False)
        self.medium.read_until('Date (mm/dd/yyyy)', True)
        self.medium.write('%s\n\r' % time.strftime('%m/%d/%Y', time.localtime()))
        self.wait_for_prompt()
        self.make_menu_selection(2, 'Set Time', False)
        self.medium.read_until('Time (hh:mm:ss)', True)
        self.medium.write('%s\n\r' % time.strftime('%H:%M:%S', time.localtime()))
        self.wait_for_prompt()
        self.make_menu_selection(3, 'accept changes')

    @login_session
    def entry_screen(self):
        self.logger.log(6, 'capturing entry screen')
        self.medium.write_single_escape()
        return self.wait_for_prompt()

    def device_manager_screen(self):
        self.logger.log(6, 'capturing device manager screen')
        self.medium.write('1\n\r')
        self.wait_for_prompt()
        self.medium.write('1\n\r')
        return self.wait_for_prompt()

    def escape(self, times):
        while times > 0:
            self.logger.log(4, 'escaping...')
            times -= 1
            self.medium.write_single_escape()
            self.wait_for_prompt()

    def wait_for_prompt(self, prompt=DEFAULT_PROMPT):
        self.logger.log(6, 'waiting for prompt (%s)' % prompt)
        return self.medium.read_until(prompt)

    def escape_to_entry_screen(self):
        self.logger.log(6, 'escape to entry')
        self.escape(UPS_MAXIMUM_MENU_DEPTH)

    def make_menu_selection(self, number, description, wait_for_prompt=True):
        self.logger.log(6, 'entering ' + description)
        self.medium.write('%d\n\r' % number)
        if wait_for_prompt:
            self.wait_for_prompt()

    def verify_ups_id(seld, ups_id):
        if ups_id not in p9constants.UPS_NUMBERS:
            raise ValueError('ups id %s is incorrect' % ups_id)

class DigiPortSessionManager(SessionManager):
    """SessionManager class for connecting through digi port"""
    def logout(self):
        self.logger.log(6, 'logging out')
        self.escape_to_entry_screen()
        self.medium.write('4\n\r') # logout
        self.medium.read_until('You are now in passthru mode.')

