from network import IPy
import subprocess
from xiv.network import command_runner
import tempfile
import os
import logging

class InterfaceConfigError(Exception):
    pass
class InterfaceConfigValueError(InterfaceConfigError):
    pass
class ConnectionIPNotReachableError(InterfaceConfigError):
    pass

NETWORK_DIRECTORY = '/etc/sysconfig/network-scripts'
INTERFACE_PREFIX = 'eth'
INTERFACE_FILE_PREFIX = 'ifcfg-'
ValidMtuSizes = (1500, 4500)

# HOSTNAME_CONF_PATH = '/etc/sysconfig/network'

class InterfaceConfig(object):
    """
    Configuring inteface IP address (for manaface)
    connection_ip is required for interacting with the node you wish to configure
    """
    DEFAULT_LOGGER_NAME = "ifacecfg"
    DEFAULT_LOG_LEVEL = logging.DEBUG

    def __init__(self, connection_ip, netmask = '255.255.255.0', logger_name = None):
        """
        connection_ip - ip address for interacting with the node
        """
        super(InterfaceConfig,self).__init__()
        if logger_name is None:
            logger_name = self.DEFAULT_LOGGER_NAME
            logging.basicConfig()
            self.logger = logging.getLogger(logger_name)
            self.logger.setLevel(self.DEFAULT_LOG_LEVEL)
        else:
            self.logger = logging.getLogger(logger_name)

        self.update_connection_ip(connection_ip, netmask)

    def check_connection_ip_connectivity(self):
        if not self.check_ip_connectivity(str(self.connection_ip)):
            raise ConnectionIPNotReachableError('no ping connection to ip %s' % self.connection_ip)

    def check_connection_ip_ssh_connectivity(self):
        if not self.check_ssh_connectivity(str(self.connection_ip)):
            raise ConnectionIPNotReachableError('unable to establish ssh connection to ip %s' % self.connection_ip)

    def update_connection_ip(self, connection_ip, netmask = '255.255.255.0'):
        try:
            self.connection_ip = IPy.IP(connection_ip)
            self.network = IPy.IP(connection_ip).make_net(netmask)
        except ValueError, e:
            raise InterfaceConfigValueError('Inappropriate values given for ip %s and/or netmask %s' % (connection_ip, netmask))

    def set_interface_ip(self, interface, ip, netmask = '255.255.255.0', mtu = 4500, apply_now = False):
        """
        - Sets the interface ip address to a new address
        - Example:
        - set_interface_ip(interface = 'eth0', ip = '10.10.180.103', netmask = '255.255.255.0', mtu = 4500)
        """
        self.logger.debug("validating request and checking connectivity")
        self.validate_ip_and_netmask(ip, netmask)
        self.validate_interface_name(interface)
        self.validate_mtu_size(mtu)
        self.check_connection_ip_ssh_connectivity()
        temporary_file = tempfile.NamedTemporaryFile()
        self.write_interface_config_file(temporary_file.file, interface, ip, netmask, mtu)
        self.logger.debug("Creating network configuration file %s on remote host" % os.path.join(NETWORK_DIRECTORY, '%s%s' % (INTERFACE_FILE_PREFIX, interface)))
        self.copy_file_to_node(temporary_file.name, os.path.join(NETWORK_DIRECTORY, '%s%s' % (INTERFACE_FILE_PREFIX, interface)))
        temporary_file.file.close()
        if apply_now:
            self.logger.debug("refreshing remote interface, this may cause disconnection from remote host")
            self.refresh_remote_interface(interface)
        else:
            self.logger.debug("configuration had changed but not applied, please refresh interface or reboot")

    def copy_file_to_node(self, local_path, destination_path):
        try:
            command = command_runner.RemoteCommandRunner(str(self.connection_ip), "", command_runner.SimpleResultVerifier(),user="root")
            command.uploadFile(local_path, destination_path)
        except command_runner.CommandFailureException, e:
            raise InterfaceConfigError('Failed to copy file %s to destination ip % path %s: %s' % (local_path, self.connection_ip, destination_path, e))

    def refresh_remote_interface(self, interface):
        try:
            command = command_runner.RemoteCommandRunner(str(self.connection_ip), "", command_runner.SimpleResultVerifier(),user="root")
            command.executeStr(
                s = ";". join((
                    '%s %s' % (os.path.join(NETWORK_DIRECTORY, 'ifdown'), interface),
                    '%s %s' % (os.path.join(NETWORK_DIRECTORY, 'ifup'), interface),
                )),
                timeout=30
            )
        except command_runner.CommandTimeout ,e:
            self.logger.info("remote host has disconnected because of interface reset (IP had changed)")

    @staticmethod
    def check_ip_connectivity(ip):
        p = subprocess.Popen('ping -c 1 -w 1 %s' % ip, shell=True, 
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return (p.wait() == 0)

    @staticmethod
    def check_ssh_connectivity(ip):
        if not InterfaceConfig.check_ip_connectivity(ip):
            return False
        try:
            command = command_runner.RemoteCommandRunner(ip, "", command_runner.SimpleResultVerifier(),user="root")
            command.echo() # Verify that remote execution works
            return True
        except command_runner.CommandFailureException ,e:
            return False

    @staticmethod
    def validate_ip_and_netmask(ip, netmask):
        try:
            interface_object = IPy.IP(ip)
            network_object = IPy.IP(ip).make_net(netmask)
        except ValueError, e:
            raise InterfaceConfigValueError('Inappropriate values given for ip %s and/or netmask %s' % (ip, netmask))

    @staticmethod
    def validate_interface_name(interface):
        if not (interface[:3] == INTERFACE_PREFIX and interface[3:].isdigit()):
            raise InterfaceConfigValueError('interface %s is not a valid interface name' % interface)

    @staticmethod
    def validate_mtu_size(mtu):
        if mtu not in ValidMtuSizes:
            raise InterfaceConfigValueError('MTU %s is not a valid MTU size' % mtu)

    @staticmethod
    def write_interface_config_file(open_file_handle, interface, ip, netmask, mtu):
        open_file_handle.write("\n".join((
            'DEVICE=%s' % interface,
            'BOOTPROTO=static',
            'IPADDR=%s' % ip,
            'MTU=%s' % mtu,
            'NETMASK=%s' % netmask,
            'ONBOOT=yes',
            ''
        )))
        open_file_handle.flush()

