"""
instantiators are static functions which aid in the instantiation of a Nextra or NextraState object. All instantiators receive the optional argument 'cls', which may be either the Nextra class or the NextraState class. The class instantiated will be of the type passed to cls. Read the individual instantiator's docstring for additional information.
"""

import logging
import guiconfig
from xcli import Nextra
import exceptions
from system_id_utils import _verify_octets, reraise_value_error_as_cant_instantiate_error, verify_dotted_system_id_string

INTERNAL_IP_ADDRESS_PREFIX = '14'
XIV_IP_ADDRESS_PREFIX = '10'
MANAGER_MODULE_NUMBERS = [101, 102, 103]
NEXTRA_SYSTEM_ID_PATH = '/etc/xiv/nextratab'
SYSTEM_ID_DETECTION_FUNCTIONS = []

_verify_octets = reraise_value_error_as_cant_instantiate_error(_verify_octets)
verify_dotted_system_id_string = reraise_value_error_as_cant_instantiate_error(verify_dotted_system_id_string)

def system_id_detector(func):
    SYSTEM_ID_DETECTION_FUNCTIONS.append(func)
    return func

def internal_net(system_id, logger_name=None, cls=Nextra, **kwargs):
    """
    Meant to instantiate an object on the internal net (14.x.y.z) of a Nextra machine. Receives the system id of the system we're in. This function is meant to be run from a Nextra module or a chubaka. It will instantiate managers 101-103.
    """
    verify_dotted_system_id_string(system_id)
    managers = _generate_module_parameters_on_internal_net(MANAGER_MODULE_NUMBERS, system_id)
    return cls(managers, logger_name, **kwargs)

def _generate_module_parameters_on_internal_net(module_numbers, system_id):
    modules = {}
    for module_number in module_numbers:
        modules[module_number] = '.'.join((INTERNAL_IP_ADDRESS_PREFIX, system_id, str(module_number)))
    return modules

def external_net(prefix, last_octet_offset=0, logger_name=None, cls=Nextra, **kwargs):
    """
    Meant to instantiate an object on an external net. Receives the IP prefix of the three manafaces (for example, '10.1.2') and the offset of the last octet (for example, '101'), and instantiates three manafaces on consecutive IP addresses. In the examples given here, the manafaces used would be: 101: 10.1.2.101, 102: 10.1.2.102, 103: 10.1.2.103.
    """
    _verify_octets(prefix, expected_octets = 3)
    managers = {}
    for module_number in MANAGER_MODULE_NUMBERS:
        managers[module_number] = '.'.join((prefix, str(module_number+last_octet_offset)))
    return cls(managers, logger_name, **kwargs)

def xiv_net(system_id, logger_name=None, cls=Nextra, **kwargs):
    """
    Meant to instantiate an object on XIV's Israeli network (10.x.y.z), of no use for customers. Receives the system id of the machine. For example, for xiv_net('10.140'), a Nextra with the following manafaces would be instantiated: 101: 10.10.140.101, 102: 10.10.140.102, 103: 10.10.140.103.
    """
    verify_dotted_system_id_string(system_id)
    return external_net('.'.join((XIV_IP_ADDRESS_PREFIX, system_id)), 
        logger_name=logger_name, cls=cls, **kwargs)

def gui_config(name=None, cls=Nextra, keep_exceptions=False, **kwargs):
    """
    Attempts to instantiates machines as defined in the GUI's XML configuration file (see the p9client.guiconfig module). Receives the name of a system and returns an instantiated object representing this system, or no parameters, instantiating every accessible machine defined in the GUI.
    """
    if name is not None:
        return cls(guiconfig.GUIConfig().nextras[name], logger_name=name, **kwargs)
    results = {}
    for nextra_name, nextra_ips in guiconfig.GUIConfig().nextras.items():
        try:
            results[nextra_name] = cls(nextra_ips, logger_name=nextra_name, **kwargs)
        except exceptions.CantInstantiateError, e:
            logging.getLogger(nextra_name).error("failed instantiating '%s', %s" % (nextra_name, e))
            if keep_exceptions:
                results[nextra_name] = e
    return results

@system_id_detector
def get_system_id_from_file(path=NEXTRA_SYSTEM_ID_PATH):
    try:
        system_id_lines = file(path).readlines()
    except IOError, e:
        raise exceptions.CantInstantiateError("system id file (%s): %s" % (NEXTRA_SYSTEM_ID_PATH, e.args[1]))
    if len(system_id_lines) == 0:
        return None
    if len(system_id_lines) > 1:
        raise NotImplementedError('multi-line %s not supported' % NEXTRA_SYSTEM_ID_PATH)
    system_id = system_id_lines[0].split(':')[0].strip()
    verify_dotted_system_id_string(system_id)
    return system_id


def chubaka(logger_name=None, cls=Nextra):
    """
    Attempts to instantiate the Nextra for which the chubaka on which the script is running is responsible.
    """
    system_id = None
    for detection_function in SYSTEM_ID_DETECTION_FUNCTIONS:
        system_id = detection_function()
        if system_id is not None:
            break
    if system_id is None:
        return None
    return internal_net(system_id, logger_name=logger_name, cls=cls)
