"""
High-level library for capturing a point-in-time state (list of volumes, hosts, mirrors, maps, ...) of a P9 Nextra. This library isn't entirely safe (as it does not use changes_get to update itself of changes which occur on the Nextra during the time Nextra information is being loaded), but it *should* either fail or succeed, without silent failures.
"""

from xiv.dtypes import ValueRange, Version, Bunch
from xiv.pyutils import wait
import configuration as conf
import time
import logging
from p9elements import *
from p9elements import typemap
from elementcontainer import ElementContainer
from xiv_xsf.nextra.p9client import Nextra, exceptions
from xiv_xsf.nextra.p9client.common import string_amount
from xiv.strutils import plural

class NextraState(object):
    """
    An object representing the point-in-time state of a Nextra system. Upon initialization it 
    """
    SUPPORTED_VERSIONS = ValueRange(ge=Version(9), lt=Version(10))
    def __init__(self, nextra_object, logger_name=None, verbosity=logging.INFO, element_types_to_load=None):
        """
        NextraState is initialized either with an XCLITransport object, much like the Nextra object or with an already initialized Nextra object. See p9client's main documentation and the documentation for the Nextra object.
        """
        super(NextraState, self).__init__()
        if isinstance(nextra_object, Nextra):
            self.system = nextra_object
        else:
            self.system = Nextra(nextra_object, logger_name, verbosity)
        if logger_name is None:
            logger_name = self.system.logger.name + '.' + 'ns'
            logging.basicConfig()
        self.logger = logging.getLogger(logger_name)
        self.logger.setLevel(verbosity)
        self._initialize(element_types_to_load)

    def _set_version(self, version):
        if not NextraState.SUPPORTED_VERSIONS.contains(version):
            raise exceptions.NotSupportedError("Nextra version %s not supported" % version)
        self._version = Version(version)
    def _get_version(self):
        return self._version
    version = property(_get_version, _set_version)

    def _initialize(self, element_types_to_load):
        self.logger.debug('getting system version')
        self.version = self.system.version_get().as_single_element.version
        self.logger.debug('getting system config')
        self.config = self.system.config_get().as_single_element

        if element_types_to_load is None:
            element_types_to_load = typemap.values()

        for element_type in element_types_to_load:
            self.load_element_type(element_type)

    def load_element_type(self, element_type):
        self.logger.debug('loading %s' % plural(element_type.description))
        setattr(self, plural(element_type.tag), ElementContainer(self, element_type))

    @property
    def element_containers(self):
        return [obj for name, obj in self.__dict__.iteritems() if type(obj) is ElementContainer]

    def is_loaded(self, element_type):
        return hasattr(self, plural(element_type.tag))

    def __str__(self):
        return str(self.system)

    def __repr__(self):
        return '<ns %s; %s>' % (self, string_amount(self.element_containers, "element type"))

    def resolve_element(self, type, name):
        return getattr(self, plural(type.tag))[name]
