import time
import types
from xiv.dtypes import Bunch
from xiv_xsf.nextra.p9client import exceptions
from elementcontainer import ElementContainer
from xiv.pyutils import weakref

def nextrastate_resolver(resolver):
    """
    A NextraState resolver is a method of a NextraStateElement descendant which resolves internal relationships 
    between NextraState elements during runtime. For example, a resolver could add a 'mapped_in_hosts' property to
    a volume. Resolvers raise ElementResolutionFailed when an expected element is not found within the internal
    state of a Nextra (i.e., tried to resolve the host which you expected to be linked to a map, but that host
    was not found).
    """
    def safe_resolver(self, *args, **kwargs):
        try:
            return resolver(self, *args, **kwargs)
        except exceptions.ElementResolutionFailed:
            e.seeking_on_behalf_of = self
            raise e
    safe_resolver.__name__ = resolver.__name__
    return safe_resolver

def nextrastate_resolver_property(resolver):
    return property(nextrastate_resolver(resolver))
 
class NextraStateElement(Bunch):
    """
    Base class for all Nextra Elements (volumes, maps, mirrors) a NextraState can hold. This class must be inherited
    to make sense.
    """
    # a human readable description of the object
    description = None
    # the tag used for this element in XCLI XMLs
    tag = None
    # the name of the command required to fetch this element from a nextra
    fetching_command = None
    # the attribute inside the element denoting this element's name
    name_attribute = 'name'
    # does this element has 'internal' elements as well as regular (customer) elements?
    has_internal_partition = False

    def __init__(self, nextrastate_parent, raw_data, internal):
        # initialize the Bunch parent with the data
        super(NextraStateElement, self).__init__(raw_data)
        self.instantiation_time = time.time()
        self.nextrastate_parent = weakref(nextrastate_parent)
        self.internal = internal

    @property
    def age(self):
        """
        Returns the element's age (time since instantiation from Nextra) in seconds.
        """
        return time.time() - self.instantiation_time

    def __str__(self):
        if hasattr(self, 'name_attribute'):
            return self.name
        raise LookupError('no name attribute defined')

    def __repr__(self):
        return "<ns.%s '%s'>" % (self.__class__.description, self.__str__())

    @nextrastate_resolver
    def expand_comma_separated_list(self, list_element_type, attribute_to_expand, internal, has_all=True):
        if type(list_element_type) is types.StringType:
            try:
                list_element_type = typemap[list_element_type]
            except KeyError:
                raise LookupError('unknown type ' + list_element_type)

        if has_all and getattr(self, attribute_to_expand) == 'all':
            return self.nextrastate_parent.get_container_for_type(list_element_type).as_list(internal)

        return list(
            self.nextrastate_parent.resolve_element(list_element_type, list_element_name)
            for list_element_name in getattr(self, attribute_to_expand).split(',')
        )

class NextraStateVolume(NextraStateElement):
    description = 'volume'
    tag = 'volume'
    fetching_command = 'vol_list'

    @nextrastate_resolver_property
    def mapped_in_maps(self):
        """
        Resolves and returns an iterable of maps the volume is mapped in.
        """
        for map in self.nextrastate_parent.maps:
            for lun in map.luns:
                if self.name == lun.vol_name:
                    yield map
                    break

    @nextrastate_resolver_property
    def mapped_for_hosts(self):
        """
        Resolves and returns the list of hosts the volume is mapped for via some map.
        """
        maps = list(self.mapped_in_maps)
        if maps:
            for host in self.nextrastate_parent.hosts:
                if host.map == "":
                    continue
                map = self.nextrastate_parent.maps[host.map]
                if map in maps:
                    yield host
                
    @nextrastate_resolver_property
    def mirror(self):
        """
        Resolves and returns the name of the remote target (as set in config_set).
        """
        for mirror in self.nextrastate_parent.mirrors:
            if mirror.local_volume_name == self.name:
                return mirror

class NextraStateHost(NextraStateElement):
    description = 'host'
    tag = 'host'
    fetching_command = 'host_list'
    
    @nextrastate_resolver_property
    def mapped_volumes(self):
        """
        Resolves and returns the list of volumes mapped to this host.
        """
        # FIXME: silently discards default without looking into the configuration
        if self.map != '' and self.map != 'default':
            for lun in self.nextrastate_parent.maps[self.map].luns:
                yield self.nextrastate_parent.volumes[lun.vol_name]

class NextraStateMap(NextraStateElement):
    description = 'map'
    tag = 'map'
    fetching_command = 'map_list'

    def __init__(self, nextrastate_parent, raw_data, internal):
        super(NextraStateMap, self).__init__(nextrastate_parent, raw_data, internal)
        self.luns = {}
        for lun in nextrastate_parent.system.map_list_luns(map=self.name):
            self.luns[int(lun.lun)] = lun

class NextraStateConsistencyGroup(NextraStateElement):
    description = 'consistency group'
    tag = 'consistency_group'
    fetching_command = 'cg_list'

    @nextrastate_resolver_property
    def volumes(self):
        for volume in self.nextrastate_parent.volumes:
            if volume.cg_name == self.name:
                yield volume

    @nextrastate_resolver_property
    def mapped_for_hosts(self):
        seen_hosts = set()
        for volume in self.volumes:
            for host in volume.mapped_for_hosts:
                if host.name in seen_hosts:
                    continue
                seen_hosts.add(host.name)
                yield host

    @nextrastate_resolver_property
    def mapped_in_maps(self):
        seen_maps = set()
        for volume in self.volumes.external:
            for map in volume.mapped_in_maps:
                if map.name in seen_maps:
                    continue
                seen_maps.add(map.name)
                yield map

class NextraStateMirror(NextraStateElement):
    description = 'mirror'
    tag = 'mirror'
    name_attribute = 'nextra_element_id'
    fetching_command = 'mirror_list'

    def __str__(self):
        return "[%s:%s] <-> [%s:%s]" % (self.nextrastate_parent, self.local_volume_name, 
            self.target_name, self.remote_volume_name)

class NextraStateRule(NextraStateElement):
    description = 'rule'
    tag = 'rule'
    fetching_command = 'rule_list'
    has_internal_partition = True

    @nextrastate_resolver_property
    def destinations_and_destination_groups(self):
        for name in self.dests.split(','):
            try:
                # if it's a plain destination
                yield self.nextrastate_parent.dests.get_element_partition(self.internal)[name]
            except exceptions.ElementResolutionFailed:
                # it wasn't found as a destination, so it's a destination group! aha!
                yield self.nextrastate_parent.destgroups.get_element_partition(self.internal)[name]

    @property
    def destinations_and_expanded_destination_groups(self):
        for obj in self.destinations_and_destination_groups:
            if type(obj) is NextraStateDestination:
                yield obj
            else:
                for destination in obj.destinations:
                    yield destination

class NextraStateDestination(NextraStateElement):
    description = 'destination'
    tag = 'dest'
    fetching_command = 'dest_list'
    has_internal_partition = True

    @property
    def smtpgws(self):
        try:
            return self.expand_comma_separated_list('smtpgw', 'gateways', self.internal)
        except Exception, e:
            print e

class NextraStateDestinationGroup(NextraStateElement):
    description = 'destination group'
    tag = 'destgroup'
    fetching_command = 'destgroup_list'
    has_internal_partition = True

class NextraStateSMTPGateway(NextraStateElement):
    description = 'SMTP gateway'
    tag = 'smtpgw'
    fetching_command = 'smtpgw_list'
    has_internal_partition = True

class NextraStateSMSGateway(NextraStateElement):
    description = 'SMS gateway'
    tag = 'smsgw'
    fetching_command = 'smsgw_list'
    has_internal_partition = True

typemap = dict(
    (cls.tag, cls)
    for cls in globals().values() if type(cls) is type
    and issubclass(cls, NextraStateElement)
    and cls is not NextraStateElement
)

__all__ = list(
    name for name, obj in globals().iteritems() 
    if isinstance(obj, type)
    and obj is not NextraStateElement
    and issubclass(obj, NextraStateElement)
)

