import logging
from transport import XCLITransport
import command
import response
from xiv_xsf.nextra.p9client import DEFAULT_OPTIONS
from xiv.pyutils import weakref
from xiv_xsf.nextra.p9client import exceptions
import os
from xiv_xsf.nextra.p9client.system_id_utils import system_id_string_to_high_octet, system_id_string_to_low_octet, \
    raw_system_id_string_to_dotted_octet_string

DEFAULT_VERBOSITY = logging.INFO

class Nextra(object):
    """
    Nextra instances provide a convenient wrapper to make Python method calls into XCLI calls (create XML, deliver to aserver, return response XML). Nextra objects require a transport parameter object to be initialized, see the class method from_object of XCLITransport for usage instruction regarding the creation of a transport parameter. Unless the 'initialize' parameter is set to False, a newly instantiated Nextra object will fetch its name from the Nextra, check the Nextra's current redundancy status and create stub methods for every command the Nextra supports as listed in its help() command.

    >>> from xiv_xsf.nextra import p9client ; n = p9client.instantiators.xiv_net('10.250')
    >>> c = n.config_get()
    >>> v = n.vol_list()
    >>> e = n.event_list()
    >>> c.as_single_element.machine_id
    '10250'
    >>> for V in v:
    ...     if 'bll' in V.name:
    ...             print V.name
    ... 
    bll_h1_1
    bll_h1_2
    >>> list(e['last_event_index'])
    ['689']
    >>> 
    """
    def __init__(self, transport_parameter, logger_name=None, verbosity=logging.INFO, options=DEFAULT_OPTIONS,
        initialize=True):
        super(Nextra, self).__init__()
        if 'XIVVERBOSE' in os.environ:
            verbosity = int(os.environ['XIVVERBOSE'])
        if logger_name is None:
            logger_name = 'nx'
            logging.basicConfig()
        self.logger = logging.getLogger(logger_name)
        self.logger.setLevel(verbosity)
        self.transport = XCLITransport.from_object(transport_parameter, logger_name + '.xp')
        self.options = options
        self.system_id = None
        if initialize:
            self._initialize()

    @property
    def active_manager(self):
        return int(self.system_counters_get().as_single_element.active_manager_id[1:])

    def get_manager_ip(self, manager):
        "Poorly designed function from the era where only TCP transport was supported. It's problematic when XCLI over SCSI is performed, but has been left here in order not to change the API."
        if not hasattr(self.transport, 'aservers'):
            raise NotImplementedError("Nextra transport %s doesn't know how to report manager IPs" % self.transport)
        return self.transport.get_manager_transport_parameters(manager)

    def _execute(self, cmd, **kwargs):
        cmdobj = command.XCLICommand(self.logger.name+'.cmd', cmd, kwargs, self.options)
        try:
            return self.transport.deliver(cmdobj)
        except exceptions.DeliveryFailedError, e:
            self.logger.warning('command %s failed: %s' % (cmdobj, e))
            raise

    def __str__(self):
        if self.logger.name == 'nx':
            return 'Unknown Nextra'
        return str(self.logger.name)

    def __repr__(self):
        return '<Nextra %s>' % self

    def _initialize(self):
        stored_options = self.options
        self.options['xiv-mode'] = 'yes'
        self._system_counters_get_on_startup()
        self._fetch_system_id()
        self._populate_commands()
        self.options = stored_options
        self._initialized = True
        self.logger.debug('nextra object initialized')

    def _initialization_command(self, command, **kwargs):
        try:
            return self._execute(command, **kwargs)
        except (exceptions.DeliveryFailedError, exceptions.ExecutionFailedError), e:
            raise exceptions.CantInstantiateError('%s failed initializing: %s' % (self, e))

    def _system_counters_get_on_startup(self):
        system_counters = self._initialization_command('system_counters_get').as_single_element
        pretty_status = system_counters.system_status.replace('_',' ')
        self.logger.debug('id %s, status %s' % (system_counters.machine_id, pretty_status))

    def _fetch_system_id(self):
        config = self._initialization_command('config_get').as_single_element
        self.system_id = config.machine_id
        # machine_id was deprecated in favor of system_id in P10, we only use system_id
        self.machine_id = config.machine_id
        self.logger = logging.getLogger(self.logger.name + '.' + self.system_id)

    @property
    def _system_id_high_octet(self):
        return system_id_string_to_high_octet(self.system_id)

    @property
    def _system_id_low_octet(self):
        return system_id_string_to_low_octet(self.system_id)

    @property
    def system_id_as_octets(self):
        return raw_system_id_string_to_dotted_octet_string(self.system_id)

    def _populate_commands(self):
        commands = self._initialization_command('help')
        namespaces = {}
        for command in commands:
            cmdobj = Command(self, command)
            if cmdobj.category not in namespaces:
                namespaces[cmdobj.category] = Namespace(self, cmdobj.category)
            namespaces[cmdobj.category]._add_command(cmdobj)
            setattr(self, cmdobj.name, cmdobj)
        setattr(self, 'categories', Namespace(self, 'categories'))
        for namespace_name, namespace in namespaces.items():
            setattr(self.categories, namespace_name, namespace)

class Namespace(object):
    def __init__(self, parent, name):
        super(Namespace, self).__init__()
        self._parent = weakref(parent)
        self._name = name

    def __repr__(self):
        return '<namespace %s.%s (%s items)>' % (self._parent, self._name,
            len(list((name, element) for (name, element) in self.__dict__.items() if not name.startswith('_'))))

    def _add_command(self, cmdobj):
        setattr(self, cmdobj.name, cmdobj)

class Command(object):
    def __init__(self, parent, command):
        super(Command, self).__init__()
        self.parent = weakref(parent)
        self.name = command.name
        self.description = command.description
        self.syntax = command.syntax
        self.category = command.category
        self.__doc__ = "%s: %s (syntax: %s, category: %s)" % (self.name, self.description, self.syntax, self.category)

    def _get_category(self):
        return self._category
    def _set_category(self, raw):
        self._category = raw.replace('-', '_')
        if self._category == '':
            self._category = 'unknown'
    category = property(_get_category, _set_category)

    def __str__(self):
        return self.name
    def __repr__(self):
        return "<command %s.%s>" % (self.parent, self.name)

    def __call__(self, *args, **kwargs):
        return self.parent._execute(self.name, **kwargs)
