import logging
import itertools
from elementtree.ElementTree import tostring, fromstring
from xml.parsers.expat import ExpatError
from elementtree import ElementTree
from xiv.dtypes import Bunch
from xiv.strutils import plural
from xiv_xsf.nextra.p9client import exceptions
import types

COMMAND_PATH = 'administrator/command'
RETURN_PATH = COMMAND_PATH+'/return'

class Response(object):
    """
    Container object for the response received from the Nextra after a command was executed. Since Nextra responses aren't always structured in the same manner, the response object can be accessed in several different ways. See the docstrings for 'as_single_element' and 'all' as well as the source for '__getitem__' for detailed usage instructions.

    Example usage (not for doctest use):
    # by default, we get all element types in a response
    >>> for volume in vol_list_response:
    ...     print volume.name
        ...
    # some commands should be force-parsed as a single element
    >>> print config_get_response.as_single_element.machine_id
        ...
    # some commands need to be subscripted for a specific tag
    >>> for event in event_list_response['event']:
    ...     if event.severity in ['Major', 'Critical']:
    ...         print event
    >>>
    """
    primary_element = None
    def __init__(self, xml, command, logger_name):
        # minor hack to improve logging prettiness
        if logger_name.endswith('.xp.rsp'):
            logger_name = logger_name.replace('.xp.rsp', '.rsp')
        self.logger = logging.getLogger(logger_name)
        self.logger.log(4, 'literal response: %s' % xml)
        self.command = command
        self.etree = self.parse_xml_to_etree(xml)
        self.verify_execution()

    def parse_xml_to_etree(self, xml):
        try:
            return fromstring(xml)
        except ExpatError, original_exception:
            exception = exceptions.DeliveryFailedError('corrupt response received for command %s' % self.command)
            exception.original_exception = original_exception
            exception.original_xml = xml
            raise exception

    def verify_execution(self):
        DELIVERY_FAILURE_EQUIVALENCE = [
            'MCL_UNSUPPORTED_TYPE',         # returned when xiv-mode command is sent to an inactive manager
            'MCL_TARGET_NODE_IS_HALTED',    # returned when xiv-mode command is sent to a halted manager
        ]
        # we import only here to avoid cyclic dependancy
        from xiv_xsf.nextra.p9client import exceptions
        if self.get_tag('aserver', 'status') != 'DELIVERY_SUCCESSFUL':
            raise exceptions.DeliveryFailedError('delivery status: ' + self.get_tag('aserver', 'status'))
        if self.get_tag(COMMAND_PATH+'/code') in DELIVERY_FAILURE_EQUIVALENCE:
            raise exceptions.DeliveryFailedError(self.get_tag(COMMAND_PATH+'/status_str'))
        if self.get_tag(COMMAND_PATH+'/status') != '0':
            raise exceptions.ExecutionFailedError(str(self.get_tag(COMMAND_PATH+'/status_str')),
                                                  original_etree=self.etree)

    @property
    def contained_element_types(self):
        return set(subelement.tag for subelement in self.as_return_etree.getchildren())

    @property
    def as_return_etree(self):
        return self.etree.find(RETURN_PATH)

    def all(self, element_type=None, response_path=None):
        """
        Generates Bunches, each representing a single subelement of the response. If an element_type is requested, only elements whose tag matches the element_type are returned. If the response has no subelements (for example, in a <return>-less command), yields None.
        """
        path = RETURN_PATH
        if response_path is not None:
            path += "/" + response_path
        response_element = self.etree.find(path)
        if response_element is None:
            return
        for subelement in self.etree.find(path).getchildren():
            if element_type is None or subelement.tag == element_type:
                yield populate_bunch_with_element(subelement)

    @property
    def as_single_element(self):
        """
        Processes the response as a single-element response, like config_get or system_counters_get.
        """
        if self.as_return_etree is None:
            return None
        if len(self.as_return_etree.getchildren()) == 1:
            return populate_bunch_with_element(self.as_return_etree.getchildren()[0])
        return populate_bunch_with_element(self.as_return_etree)

    @property
    def as_list(self, element_type=None, response_path=None):
        return list(self.all(element_type, response_path))

    def as_dict(self, key, element_type=None, response_path=None):
        result = {}
        for element in self.all(element_type, response_path):
            result[getattr(element, key)] = element
        return result
            
    def get_tag(self, path, tag='value'):
        return self.etree.find(path).get(tag)

    def __str__(self):
        return tostring(self.etree)

    def __repr__(self):
        from xiv_xsf.nextra.p9client.common import string_amount
        return '<XCLI response for %s; %s>' % (self.command, string_amount(list(self.all()), 'item'))

    def __iter__(self):
        return self.all()

    def __getitem__(self, item):
        if type(item) is types.StringType:
            return self.all(item)
        elif type(item) is types.IntType:
            # this is nextralib9's style of result parsing
            return list(self.all())[item]

    def __nonzero__(self):
        return bool(list(self.all()))


def populate_bunch_with_element(element):
    """
    Helper function to recursively populates a Bunch from an XML tree. Returns leaf XML elements as a simple value, branch elements are returned as Bunches containing their subelements as value or recursively generated Bunch members.
    """
    if 'value' in element.attrib:
        return element.get('value')
    current_bunch = Bunch()
    # Patch: See in xivpedia: Python_NextraLib#Inconsistent_CLI_XML_Output_Hack
    if element.get('id'):
        current_bunch['nextra_element_id'] = element.get('id')
    for subelement in element.getchildren():
        current_bunch[subelement.tag] = populate_bunch_with_element(subelement)
    return current_bunch
