import socket
import logging
import response
from xiv_xsf.nextra.p9client import exceptions
import configuration as conf
from scsy import RawSCSIInterface
import struct
from construct import Container
from xcli_over_scsi_constructs import XCLI_INPUT_BUFFER_SIZE, xcli_read_buffer, xcli_write_buffer

XCLI_OVER_SCSI_SEMAPHORE_NAME = 'xiv_xcli_over_scsi'

class XCLITransport(object):
    "Parent class for XCLI transports."
    def __init__(self, transport_object, logger_name):
        self.logger = logging.getLogger(logger_name)
        self.logger.log(4, 'initializing %s on %r', (self, transport_object))

    @classmethod
    def from_object(cls, obj, logger_name='xpt'):
        "This helper function will instantiate a Nextra transport from all supported object types. At the moment, it only supports the TCPTransport, which requires a dictionary of aservers' number:IP ({101:'1.1.1.1', ..., 103:'1.1.1.3'})."
        if isinstance(obj, cls):
            return obj
        if isinstance(obj, dict):
            return TCPTransport(obj, logger_name)
        if isinstance(obj, str):
            return SCSITransport(obj, logger_name)
        try:
            if obj[101]:
                return TCPTransport(obj, logger_name)
        except:
            pass
        raise exceptions.CantInstantiateError("unknown object %r passed as a transport initializer" % obj)

    def __str__(self):
        return self.__class__.__name__

class TCPTransport(XCLITransport):
    "Handles delivery of XML commands to/from the Nextra over TCP/IP."
    def __init__(self, aservers, logger_name, initial_timeout=conf.tcp_initial_timeout):
        "Receives a dictionary of aservers' number:IP ({101:'1.1.1.1', ..., 103:'1.1.1.3'}) and a logger name."
        super(TCPTransport, self).__init__(aservers, logger_name)
        self.aservers = self._validated_aservers(aservers)
        self.initial_timeout = initial_timeout
        self.timeout = initial_timeout
        self.set_active_aserver(aservers.keys())
        self.logger.debug('%s initialized' % self)

    def _validated_aservers(self, aservers):
        validated_aservers = {}
        for number, ip in aservers.items():
            validated_aservers[int(number)] = ip
        return validated_aservers

    def set_active_aserver(self, potential_numbers):
        self.logger.log(6, 'choosing aserver from potential %r' % potential_numbers)
        if hasattr(potential_numbers, '__iter__'):
            aserver = min(potential_numbers)
        else:
            aserver = int(potential_numbers)
        self.logger.log(6, 'active aserver %s' % aserver)
        self.active_aserver = aserver

    def deliver(self, command):
        rspobj, successful_aserver = self.iterate_aservers_until_delivery(command)
        if successful_aserver != self.active_aserver:
            self.logger.info('switching aserver from %s to %s' % (self.active_aserver, successful_aserver))
            self.set_active_aserver(successful_aserver)
        return rspobj

    def iterate_aservers_until_delivery(self, command):
        attempted_aservers = set()
        rounds_left = None
        while self.timeout < conf.tcp_timeout_error_threshold:
            for aserver in self.yield_aservers_with_active_first():
                try:
                    attempted_aservers.add(aserver)
                    self.logger.debug('sending %s to %s [%s]' % (command, aserver, self.aservers[aserver]))
                    raw_response = self.communicate_with_aserver(self.aservers[aserver], command.xml)
                    rspobj = response.Response(raw_response, command, self.logger.name+'.rsp')
                    return rspobj, aserver
                except socket.timeout:
                    self.log_timeout_exception(aserver)
                except (socket.error, exceptions.DeliveryFailedError), e:
                    if isinstance(e, exceptions.DeliveryFailedError):
                        # we reached one of the manafaces, but it was unable to process our request
                        #  this reduces the likelihood of a TCP timeout related issue, so we retry only
                        #  once more with a double timeout to make sure we give all managers time to answer
                        rounds_left = 1
                    level = logging.WARNING
                    # often enough passive manafaces respond faster than the active one, making erroneous
                    #  warnings ("Requested action is currently not supported by the recipient node"),
                    #  so we silence these
                    if str(e) == 'Requested action is currently not supported by the recipient node':
                        level = logging.DEBUG
                    self.logger.log(level, "aserver %s: %s" % (aserver, e))
            self.timeout *= 2
            if rounds_left is not None:
                rounds_left -= 1
                if rounds_left < 0:
                    break
        self.timeout = self.initial_timeout
        raise exceptions.DeliveryFailedError("delivery failed on all aservers, tried: %s" %
            ", ".join(tuple(str(attempted_aserver) for attempted_aserver in attempted_aservers)))

    def log_timeout_exception(self, aserver):
        if self.timeout > conf.tcp_timeout_error_threshold:
            severity = logging.WARNING
            future_action_string = ', giving up'
        elif self.timeout > conf.tcp_timeout_warning_threshold:
            severity = logging.WARNING
            future_action_string = ', retrying'
        else:
            severity = logging.DEBUG
            future_action_string = ', retrying'
        self.logger.log(severity, "aserver %s: timed out (%.1f)%s" % (aserver, self.timeout, future_action_string))

    def yield_aservers_with_active_first(self):
        """Yields the active aserver then all other aservers in undefined order"""
        # protect against the active aserver changing during iterations
        initial_active_aserver = self.active_aserver
        yield initial_active_aserver
        for aserver in self.aservers:
            if aserver == initial_active_aserver:
                continue
            yield aserver

    def communicate_with_aserver(self, ip, xml):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(self.timeout)
        s.connect((ip, 7777))
        s.settimeout(conf.tcp_post_connect_timeout)
        s.send(xml)
        s.shutdown(socket.SHUT_WR)
        return "".join(self.yield_raw_response_chunks(s))

    def yield_raw_response_chunks(self, s):
        while True:
            chunk = s.recv(1024)
            if not chunk:
                break
            yield chunk

    def get_manager_transport_parameters(self, manager):
        return self.aservers[int(manager)]

class SCSITransport(XCLITransport):
    def __init__(self, device, logger_name):
        "Receives a string representing the device to use for the XCLI CDBs"
        super(SCSITransport, self).__init__(device, logger_name)
        self.handle = self.open_device(device)
        self.logger.debug('%s initialized' % self)

    def open_device(self, device):
        try:
            RawSCSIInterface.verify_supported()
        except NotImplementedError, error:
            raise exceptions.CantInstantiateError(error)
        return RawSCSIInterface.get_lun_handle(device, syncronization_semaphore_name='/xiv_xcli_over_scsi')

    def get_manager_transport_parameters(self, manager):
        return self.handle

    def deliver(self, command):
        cdb = xcli_write_buffer.build(Container(command_length=len(command.xml)))
        self.logger.log(5, "write cdb: %s" % " ".join(hex(ord(ch)) for ch in cdb))
        self.handle.out_command(cdb=cdb, data_buffer=command.xml)
        cdb = xcli_read_buffer.build(Container(command_length=XCLI_INPUT_BUFFER_SIZE))
        self.logger.log(5, "read cdb: %s" % " ".join(hex(ord(ch)) for ch in cdb))
        raw_response = self.handle.in_command(cdb=cdb, data_buffer_length=XCLI_INPUT_BUFFER_SIZE)
        response_length = struct.unpack('>L', raw_response[0:4])[0] + 4
        chopped_response = raw_response[4:response_length]
        return response.Response(chopped_response, command, self.logger.name+'.rsp')
