from warnings import warn

from command import XCLICommand
from stubs import CommandStub
from response import XCLIResponse
from response.exceptions import XCLIException
from delivery import XCLIDelivery

from xiv_xsf.nextra.phase10.client.exceptions import CantInstantiateError, UnsupportedNextraVersion, \
                                                     IncorrectUsername, IncorrectPassword

class XCLIClient(object):
    def __init__(self, transport, options={}):
        super(XCLIClient, self).__init__()
        self.transport = transport
        self.options = options
        self.serial = None
        self.model = None
        self.type = None

    @classmethod
    def connect_ssl(cls, user, password, *endpoints):
        from transport.ssl import NextraClientSSLTransport
        if not endpoints:
            raise ValueError('must pass at least one endpoint')
        transport = NextraClientSSLTransport(*endpoints)
        options = dict(user=user, password=password)
        instance = cls(transport, options=options)
        instance.initialize()
        return instance

    @property
    def fully_qualified_serial(self):
        if self.type is None or self.model is None or self.serial is None:
            return None
        return "%s-%s-%s" % (self.type, self.model, self.serial)

    def initialize(self):
        self.transport.connect()
        try:
            self._get_version()
            self._get_configuration()
            self._populate_commands()
        except XCLIException, error:
            raise CantInstantiateError.choose_best_fit_instantiation_exception_class(error)(str(error))

    def _execute(self, command_to_execute, **arguments):
        try:
          remote_target = arguments['remote_target']  
          arguments.pop('remote_target')
        except KeyError, e:
          remote_target = None
        
        command = XCLICommand(command_to_execute, arguments, self.options, remote_target=remote_target)
        delivery = XCLIDelivery(transport=self.transport, command=command)
        returned_command_etree = delivery.execute()
        return XCLIResponse.create_valid_response_or_raise_proper_exception(command, returned_command_etree)

    def _get_version(self):
        self.system_version = self._execute('version_get').as_single_element.system_version

    def _get_configuration(self):
        if self.system_version.startswith('10.0'):
            configuration = self._execute('config_get').as_single_element
            self.serial = configuration.machine_serial_number
            self.model = configuration.machine_model
            self.type = configuration.machine_type
        elif self.system_version.startswith('10.1'):
            configuration = self._execute('config_get').as_dict('name')
            self.serial = configuration['machine_serial_number'].value
            self.model = configuration['machine_model'].value
            self.type = configuration['machine_type'].value
        else:
            raise UnsupportedNextraVersion('version %s never tested with this client' % (self.system_version,))

    def _populate_commands(self):
        for command in self._execute('help'):
            cmdobj = CommandStub(self, command)
            setattr(self, cmdobj.name, cmdobj)

    def __str__(self):
        if self.fully_qualified_serial is not None:
            return self.fully_qualified_serial
        return 'unknown XIV array'

    def __repr__(self):
        if self.transport.connected_endpoint:
            return '<XCLI client for %s via %s>' % (self, self.transport.connected_endpoint)
        return '<XCLI client for %s>' % self
