import logging
import itertools
import types
from xml.parsers.expat import ExpatError
try:
    from xml.etree.ElementTree import tostring
except ImportError:
    from elementtree.ElementTree import tostring

from xiv.strutils import plural
from xiv.dtypes import Bunch

from exceptions import ORDER_OF_CONCRETE_XCLI_EXCEPTIONS

RETURN_PATH = 'return'

def string_amount(container, element_name):
    "Utility prettyness function for formatting the amount of elements of a particular kind ('125 cats, one dog')"
    numeric_amount = len(container)
    if numeric_amount == 0:
        return 'empty'
    elif numeric_amount == 1:
        return 'one %s' % element_name
    else:
        return '%d %s' % (numeric_amount, plural(element_name))

class XCLIResponse(object):
    """Container object for the response received from the Nextra after a command was executed. Since Nextra responses aren't always structured in the same manner, the response object can be accessed in several different ways. See the docstrings for 'as_single_element' and 'all' as well as the source for '__getitem__' for detailed usage instructions.

    Example usage (not for doctest use):
    # by default, we get all element types in a response
    >>> for volume in vol_list_response:
    ...     print volume.name
        ...
    # some commands should be force-parsed as a single element
    >>> print config_get_response.as_single_element.machine_id
        ...
    # some commands need to be subscripted for a specific tag
    >>> for event in event_list_response['event']:
    ...     if event.severity in ['Major', 'Critical']:
    ...         print event
    >>>"""
    primary_element = None
    def __init__(self, command, response_etree):
        self.command = command
        self.response_etree = response_etree

    @classmethod
    def create_valid_response_or_raise_proper_exception(cls, command, response_etree):
        for potential_exception_type in ORDER_OF_CONCRETE_XCLI_EXCEPTIONS:
            potential_exception_type(command, response_etree).raise_if_necessary()
        return cls(command, response_etree)


    @property
    def contained_element_types(self):
        return set(subelement.tag for subelement in self.as_return_etree.getchildren())

    @property
    def as_return_etree(self):
        return self.response_etree.find(RETURN_PATH)

    def all(self, element_type=None, response_path=None):
        """
        Generates Bunches, each representing a single subelement of the response. If an element_type is requested, only elements whose tag matches the element_type are returned. If the response has no subelements (for example, in a <return>-less command), yields None.
        """
        path = RETURN_PATH
        if response_path is not None:
            path += "/" + response_path
        response_element = self.response_etree.find(path)
        if response_element is None:
            return
        for subelement in self.response_etree.find(path).getchildren():
            if element_type is None or subelement.tag == element_type:
                yield populate_bunch_with_element(subelement)

    @property
    def as_single_element(self):
        """
        Processes the response as a single-element response, like config_get or system_counters_get.
        """
        if self.as_return_etree is None:
            return None
        if len(self.as_return_etree.getchildren()) == 1:
            return populate_bunch_with_element(self.as_return_etree.getchildren()[0])
        return populate_bunch_with_element(self.as_return_etree)

    @property
    def as_list(self, element_type=None, response_path=None):
        return list(self.all(element_type, response_path))

    def as_dict(self, key, element_type=None, response_path=None):
        result = {}
        for element in self.all(element_type, response_path):
            result[getattr(element, key)] = element
        return result

    def xeval(self, expression, suffix='@value', prefix='command'):
        return self._xpath(self.response_etree.xeval, expression, prefix, suffix)
    def xpath(self, expression, suffix='@value', prefix='command'):
        return self._xpath(self.response_etree.xpath, expression, prefix, suffix)
    def _xpath(self, function, expression, prefix, suffix):
        if prefix is not None:
            expression = prefix + '/' + expression
        if suffix is not None:
            expression = expression + '/' + suffix
        return function(expression)
            
    def __str__(self):
        return tostring(self.response_etree)

    def __repr__(self):
        return '<XCLI response for %s; %s>' % (self.command, string_amount(list(self.all()), 'item'))

    def __iter__(self):
        return self.all()

    def __len__(self):
        return len(self.as_list)

    def __getitem__(self, item):
        if type(item) is types.StringType:
            return self.all(item)
        elif type(item) is types.IntType:
            # this is nextralib9's style of result parsing
            return list(self.all())[item]

    def __nonzero__(self):
        return bool(list(self.all()))


def populate_bunch_with_element(element):
    """
    Helper function to recursively populates a Bunch from an XML tree. Returns leaf XML elements as a simple value, branch elements are returned as Bunches containing their subelements as value or recursively generated Bunch members.
    """
    if 'value' in element.attrib:
        return element.get('value')
    current_bunch = Bunch()
    # Patch: See in xivpedia: Python_NextraLib#Inconsistent_CLI_XML_Output_Hack
    if element.get('id'):
        current_bunch['nextra_element_id'] = element.get('id')
    for subelement in element.getchildren():
        current_bunch[subelement.tag] = populate_bunch_with_element(subelement)
    return current_bunch
