import socket
import types
import errno
import os
from select import select
import time
from xml.parsers.expat import ExpatError

from xiv_xsf.nextra.phase10.client.xcli.transport import (NextraClientTransport, TransportSendError,
                                                          TransportReceiveError)
from xiv_xsf.nextra.phase10.client.xcli.transport.exceptions import UnableToConnect

from xiv.xmlutils import TerminationDetectingXMLParser

class EndpointInErrorState(Exception):
    pass

class SSLHandshakeFailed(Exception):
    pass

class AdministrationGatewaysConnectionErrors(Exception):
    def __init__(self, erroneous_endpoints):
        self.erroneous_endpoints = erroneous_endpoints
    @property
    def common_error_string(self):
        if len(set(endpoint.last_error_string for endpoint in self.erroneous_endpoints)) == 1:
            return self.erroneous_endpoints[0].last_errno
        return None
    def _common_error_format(self):
        assert self.common_error_string is not None, "error string isn't unique"
        return "encountered %r on all addresses" % self.erroneous_endpoints[0].last_error_string.lower()
    def simple_format(self):
        return "unable to connect to any of the addresses provided"
    def verbose_format(self):
        endpoints_and_errors_string = ", ".join("%s encountered %r" % (endpoint, endpoint.last_error_string.lower())
                                                                       for endpoint in self.erroneous_endpoints)
        return "%s: %s" % (self.simple_format(), endpoints_and_errors_string)
    def smart_format(self):
        if self.common_error_string is not None:
            return self._common_error_format()
        return self.simple_format()
    def __str__(self):
        return self.smart_format()

class AdministrationGatewayEndpoint(object):
    DEFAULT_ENDPOINT_PORT = 7778
    def __init__(self, host, port=DEFAULT_ENDPOINT_PORT):
        self.host = host
        self.port = port
        self.last_errno = None
        self.socket_handle = socket.socket()
        self.ssl_handle = None
        self.last_ssl_exception = None

    def __str__(self):
        if self.port == self.DEFAULT_ENDPOINT_PORT:
            return self.host
        return '%s:%s' % (self.host, self.port)

    def __repr__(self):
        if self.ssl_handle is not None:
            return '<%s [ssl established]>' % self
        if self.is_connected:
            return '<%s [connected]>' % self
        return '<%s>' % self

    @classmethod
    def from_anything(cls, obj):
        if isinstance(obj, cls):
            return obj
        # is it a string
        if isinstance(obj, basestring):
            if obj.count(':') == 0:
                return cls(obj, cls.DEFAULT_ENDPOINT_PORT)
            if obj.count(':') == 1:
                host, port = obj.split(':')
                port = int(port)
                return cls(host, port)
        # is it an iterable / tuple
        else:
            try:
                host, port = obj
                return cls(host, port)
            except ValueError:
                pass
        raise ValueError("%r isn't a valid endpoint specifier" % (obj,))

    def start_asynchronous_connection(self):
        self.socket_handle.setblocking(False)
        try:
            self.socket_handle.connect_ex((self.host, self.port))
        except socket.gaierror, error:
            raise UnableToConnect(message='address %r is not resolvable' % self.host, original_exception=error)

    def cancel_connection(self):
        self.socket_handle.close()
        self.socket_handle = socket.socket()
        self.ssl_handle = None
        self.last_ssl_exception = None

    def fileno(self):
        return self.socket_handle.fileno()

    def get_last_errno(self):
        self.last_errno = self.socket_handle.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
        return self.last_errno

    @property
    def last_error_string(self):
        if self.last_ssl_exception is not None:
            return self.last_ssl_exception_error_string()
        if self.last_errno == 0 or self.last_errno is None:
            return None
        return os.strerror(self.last_errno)

    def last_ssl_exception_error_string(self):
        # bloody braindamage in SSL exceptions, they have variable number of arguments
        for argument in self.last_ssl_exception.args:
            if isinstance(argument, types.StringType):
                return argument
        return str(self.last_ssl_exception.args[0])


    @property
    def is_ok(self):
        return self.last_errno == 0

    def switch_to_blocking_mode(self, timeout):
        self.socket_handle.setblocking(True)
        self.socket_handle.settimeout(timeout)

    @property
    def is_connected(self):
        try:
            self.socket_handle.getpeername()
            return True
        except socket.error:
            return False

    def perform_ssl_handshake(self):
        try:
            self.ssl_handle = socket.ssl(self.socket_handle)
            assert 'XIV' in self.ssl_handle.issuer(), "unexpected SSL issuer for endpoint %s" % self
        except socket.sslerror, error:
            self.last_ssl_exception = error
            raise SSLHandshakeFailed(self.last_error_string)

    def read(self, *args):
        assert self.ssl_handle is not None, "tried reading from null ssl handle"
        return self.ssl_handle.read(*args)

    def write(self, buf):
        assert self.ssl_handle is not None, "tried writing to null ssl handle"
        return self.ssl_handle.write(buf)

class NextraClientSSLTransport(NextraClientTransport):
    DEFAULT_SOCKET_TIMEOUT = 30
    ALLOWED_DELIVERY_ATTEMPTS = 2
    RECEPTION_ATTEMPT_TIMEOUT = 120
    def __init__(self, *endpoints, **kwargs):
        # HACK: I don't really want **kwargs, I just wanted the kwarg 'timeout'... implement that:
        self.timeout = kwargs.pop('timeout', self.DEFAULT_SOCKET_TIMEOUT)
        if kwargs:
           raise TypeError("__init__() got an unexpected keyword argument %r" % kwargs.keys()[0])

        self.remaining_timeout = None
        if not endpoints:
            raise ValueError('must pass at least one endpoint')
        self.endpoints = list(AdministrationGatewayEndpoint.from_anything(endpoint) for endpoint in endpoints)
        self.connected_endpoint = None

    def connect(self, allowed_endpoints=None):
        if allowed_endpoints is None:
            allowed_endpoints = self.endpoints
        self.remaining_timeout = self.timeout
        self._start_asynchronous_connection_to_endpoints(allowed_endpoints)
        try:
            self._choose_first_usable_endpoint(allowed_endpoints)
        except AdministrationGatewaysConnectionErrors, error:
            raise UnableToConnect(message=str(error), original_exception=error)
        self._clean_unused_endpoints()

    def _choose_first_usable_endpoint(self, endpoints):
        potential_endpoints = set(endpoints)
        while potential_endpoints:
            ready_endpoints = self._select_ready_endpoints_and_mark_errors(potential_endpoints)
            potential_endpoints.difference_update(set(ready_endpoints))
            for ready_endpoint in ready_endpoints:
                try:
                    self._set_connected_endpoint(ready_endpoint)
                    return
                except EndpointInErrorState, error:
                    pass
        raise AdministrationGatewaysConnectionErrors(endpoints)

    def _set_connected_endpoint(self, endpoint):
        if not endpoint.is_ok:
            raise EndpointInErrorState('%s is marked with error: %s' % (endpoint, endpoint.last_error_string))

        endpoint.switch_to_blocking_mode(self.timeout)
        try:
            endpoint.perform_ssl_handshake()
        except SSLHandshakeFailed, error:
            raise EndpointInErrorState('%s had an ssl error: %s' % (endpoint, error))

        self.connected_endpoint = endpoint

    def _start_asynchronous_connection_to_endpoints(self, endpoints):
        for endpoint in endpoints:
            endpoint.start_asynchronous_connection()

    def _clean_unused_endpoints(self):
        for endpoint in self.endpoints:
            if endpoint is self.connected_endpoint:
                continue
            endpoint.cancel_connection()

    def reconnect(self):
        for endpoint in self.endpoints:
            endpoint.cancel_connection()
        self.connect()

    def _select_ready_endpoints_and_mark_errors(self, potential_endpoints):
        ready_endpoints = self._perform_select_and_decrease_remaining_timeout(potential_endpoints)
        if ready_endpoints:
            return self._return_endpoints_with_retrieved_error_status(ready_endpoints)
        else:
            return self._return_endpoints_marked_as_timed_out(potential_endpoints)

    def _return_endpoints_marked_as_timed_out(self, endpoints):
        for endpoint in endpoints:
            endpoint.last_errno = errno.ETIMEDOUT
        return endpoints

    def _return_endpoints_with_retrieved_error_status(self, endpoints):
        for endpoint in endpoints:
            endpoint.get_last_errno()
        return endpoints

    def _perform_select_and_decrease_remaining_timeout(self, potential_endpoints):
        select_start = time.time()
        dummy_read, writable_endpoints, dummy_exception = select([], potential_endpoints, [], self.remaining_timeout)
        self.remaining_timeout -= time.time() - select_start
        if self.remaining_timeout < 0:
            self.remaining_timeout = 0
        return writable_endpoints

    def send_command_xml(self, data):
        try:
            self.connected_endpoint.write(data)
        except (socket.error, socket.sslerror), error:
            raise TransportSendError(message='socket/ssl', original_exception=error, endpoint=self.connected_endpoint)

    def receive_response_etree(self):
        reception_start = time.time()
        parser = TerminationDetectingXMLParser()
        while time.time() - reception_start < self.RECEPTION_ATTEMPT_TIMEOUT:
            try:
                parser.feed(self.connected_endpoint.read())
                if parser.top_element_closed:
                    return parser.close()
            except (socket.error, socket.sslerror), error:
                raise TransportReceiveError(endpoint=self.connected_endpoint, original_exception=error,
                                            message='socket/ssl')
            except ExpatError, error:
                raise TransportReceiveError(endpoint=self.connected_endpoint, original_exception=error,
                                            message='malformed XML')
