from xiv.network import command_runner
from xiv_xsf.nextra import p9constants
from xiv_xsf.nextra.p9client import system_id_utils
import os
import logging
import re
import shutil
import tempfile

SYSTEM_IMAGE_MOUNT_DIRECTORY = '/image_mnt'
SYSTEM_IMAGE_FILE_NAME = 'system_image.img'
SYSTEM_IMAGE_TRACE_IP_CONFIG_PATH = '/root_dir/trace/shm_setup.conf'
SYSTEM_IMAGE_HOSTS_PATH = '/etc/hosts'
TIMEZONE_PATH = '/etc/localtime'

class SystemImageConfigurationError(Exception):
    pass

class SystemImage(object):
    """
    Apply specific changes on existing system image in order to adjust it to a specific manaface
    """
    DEFAULT_LOGGER_NAME = "sysimage"
    DEFAULT_LOG_LEVEL = logging.DEBUG
    
    def __init__(self, compressed_system_image_path, nextra, manaface, logger_name = None):
        super(SystemImage,self).__init__()
        if logger_name is None:
            logger_name = self.DEFAULT_LOGGER_NAME
            self.logger = logging.getLogger(logger_name)
            self.logger.setLevel(self.DEFAULT_LOG_LEVEL)
        else:
            self.logger = logging.getLogger(logger_name)

        self.nextra = nextra
        self.logger.debug('checking manaface id')
        if manaface not in p9constants.MANAFACE_NUMBERS:
            raise SystemImageConfigurationError('manaface %s is not a valid manaface number' % manaface)
        self.manaface = manaface
        self.compressed_system_image_path = compressed_system_image_path
        self.uncompressed_system_image_path = re.compile('^(.*)\.gz$').match(compressed_system_image_path).groups()[0]

    def pre_configuration_initializations(self):
        try:
            command = command_runner.LocalCommandRunner("", command_runner.SimpleResultVerifier())
            command.executeStr(s=';'.join((
                'gunzip %s' % (self.compressed_system_image_path),
                'mkdir -p %s' % SYSTEM_IMAGE_MOUNT_DIRECTORY,
                'umount %s >/dev/null 2>&1' % SYSTEM_IMAGE_MOUNT_DIRECTORY,
                'mount -o loop %s %s' % (self.uncompressed_system_image_path, SYSTEM_IMAGE_MOUNT_DIRECTORY)
            )))
        except (command_runner.CommandFailureException, command_runner.CommandTimeout), e:
            raise SystemImageConfigurationError('preparing system image: %s' % e.stderr.strip())

    def post_configuration_image_closing(self):
        try:
            command = command_runner.LocalCommandRunner("", command_runner.SimpleResultVerifier())
            command.executeStr(s=';'.join((
                'umount %s' % SYSTEM_IMAGE_MOUNT_DIRECTORY,
                'gzip %s' % self.uncompressed_system_image_path
            )))
        except (command_runner.CommandFailureException, command_runner.CommandTimeout), e:
            raise SystemImageConfigurationError('closing system image: %s' % e.stderr.strip())

    def write_trace_ip_config_file(self):
        trace_file = file(os.path.join(SYSTEM_IMAGE_MOUNT_DIRECTORY, SYSTEM_IMAGE_TRACE_IP_CONFIG_PATH.lstrip('/')), 'w')
        trace_file.write("\n".join((
            'MODE=CLIENT',
            'IPS=' + ','.join(("14.%s.%s" % (self.nextra.system_id_as_octets, str(manaface_number)) for manaface_number in p9constants.MANAFACE_NUMBERS)),
            'USEC_SLEEP=100000',
            ''
        )))
        trace_file.close()

    def write_hosts_config_file(self):
        host_file = file(os.path.join(SYSTEM_IMAGE_MOUNT_DIRECTORY, SYSTEM_IMAGE_HOSTS_PATH.lstrip('/')), 'w')
        host_file.write("\n".join((
            '14.%s.%s manager' % (self.nextra.system_id_as_octets, self.manaface),
            ''
        )))
        host_file.close()

    def download_file_from_active_manager(self, remotepath, localpath):
        try:
            command = command_runner.RemoteCommandRunner('14.%s.%s' % (self.nextra.system_id_as_octets, self.nextra.active_manager))
            command.downloadFile(remotepath, localpath)
        except command_runner.CommandFailureException, e:
            raise SystemImageConfigurationError(
                    'Failed to copy file %s from ip %s to path %s: %s' % (
                        remotepath,
                        '14.%s.%s' % (self.nextra.system_id_as_octets, self.nextra.active_manager),
                        localpath,
                        e
                        )
                    )

    def insert_file_to_image(self, local_path, path_in_image):
        try:
            shutil.copyfile(local_path, os.path.join(SYSTEM_IMAGE_MOUNT_DIRECTORY, path_in_image.lstrip('/')))
        except shutil.Error, e:
            raise SystemImageConfigurationError("failed to copy file %s to system image path %s: %s" % (local_path, os.path.join(SYSTEM_IMAGE_MOUNT_DIRECTORY, path_in_image.lstrip('/')), e))

    def get_file_from_image(self, path_in_image, local_path):
        try:
            shutil.copyfile(os.path.join(SYSTEM_IMAGE_MOUNT_DIRECTORY, path_in_image.lstrip('/')), local_path)
        except shutil.Error, e:
            raise SystemImageConfigurationError("failed to copy file %s to system image path %s: %s" % (os.path.join(SYSTEM_IMAGE_MOUNT_DIRECTORY, path_in_image.lstrip('/')), local_path, e))

    def insert_timezone_file_from_active_manaface(self):
        TIMEZONE_TEMP_FILENAME = 'etc_timezone'
        temporary_timezone_path = os.path.join(tempfile.gettempdir(), TIMEZONE_TEMP_FILENAME)
        self.download_file_from_active_manager(TIMEZONE_PATH, temporary_timezone_path)
        self.insert_file_to_image(temporary_timezone_path, TIMEZONE_PATH)

    def configure(self):
        self.logger.debug('initializing')
        self.pre_configuration_initializations()
        self.logger.debug('writing trace ip configuration file in image')
        self.write_trace_ip_config_file()
        self.logger.debug('writing hosts configuration file in image')
        self.write_hosts_config_file()
        self.logger.debug('getting timezone file from active manager into system image')
        self.insert_timezone_file_from_active_manaface()
        self.logger.debug('packing image')
        self.post_configuration_image_closing()
        self.logger.debug('done')


