#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#------------------------------------------------------------------------------------------------- 
from construct import * 
from response import Response
from vpd import vpd_pages 
from xiv_xsf.scsy.commands.core import SCSIReadCommand


class Inquiry(SCSIReadCommand):
    """Options for SPC-4 INQUIRY command
       
    Optional Parameters:
    * allocation_length = [int16:36] size of buffer that will be created to
                          hold response
    * evpd              = [bool:False] enable vital product data. If enabled
                          will return data specified by the page_code  
    * page_code         = [int8:0] Type of data that will be returned if 
                          EVPD flag is enabled.
                          This value should be zero of evpd is not set.
    * control           = [int8:0] Not sure what this does
       
    
    Helper Functions:
    * getCorrectInquiryDataSize
        use to determine the correct buffer size based upon the contents of 
        an INQUIRY response 
    """

    DEFAULT_BUFFER_SIZE = 36

    def __init__(self):
        self._response = Response
        self._helperfunctions = ('evpd00','evpd80','evpd83')


    def _getCommandDefaults(self):
        return dict(operation_code=0x12,
                    evpd=0,
                    page_code=0,
                    control=0,
                    allocation_length = self.DEFAULT_BUFFER_SIZE)


    def _getCommandStruct(self):
        """ getCommandStruct documentation"""

        return Struct("INQUIRY_command",
        Const(UBInt8("operation_code"), 0x12),
        EmbeddedBitStruct(
            Padding(6), # reserved
            Padding(1), # obsolete
            Flag("evpd"),
        ),
        UBInt8("page_code"),
        UBInt16("allocation_length"),
        UBInt8("control"),
    )


    def evpd00(self,device, pages_only=True, raw=False):
        """Calls the EVPD page 0 command 'Supported VPD Pages'
        
        Use {response}.page_data.supported_vpd_pages or 
        pages_only = True
        
        Parameters:
        pages_only - [Bool:True] if true returns a list of supported
                     pages otherwise returns a vpd_pages constuct
        
        VPD Construct Sample: [note values in decimal]
            peripheral_qualifier = 0
            peripheral_device_type = 'direct_access_block'
            page_type_id = 0
            page_type = 'supported_pages'
            page_data = Container:
                page_type = 0
                page_length = 5
                supported_vpd_pages = [
                    0
                    128
                    131
                ]
        """
        command = self.getCommand(evpd = True, page_code = 0x00)
        
        if pages_only:
            return list(self.extended(device, command,auto_correct=True, 
                                    trim=True, raw=False
                                 ).page_data.supported_vpd_pages)
        else:
            return self.extended(device, command,True, True, raw)
    
    def evpd80(self,device, serial_only=True, raw=False):
        """Calls the EVPD page 0 command 'Unit Serial Number'
        
        Use {response}.page_data_product_serial_number or
        serial_only = True
        
        Parameters:
        serial_only - [Bool:True] return the serial as string or return
                      the whole VPD structure.
                      NOTE: if True, then the raw value will be ignored 
                      NOTE: if device does not support EPVPD 80 or does
                            return product_serial_number an empty string
                            shall be returned.
                      
        VPD Construct Sample:
            peripheral_qualifier = 0
            peripheral_device_type = 'direct_access_block'
            page_type_id = 128
            page_type = 'unit_serial_number'
            page_data = Container:
                page_type = 128
                page_length = 11
                product_serial_number = 'MN0001A0177'
        """
        command = self.getCommand(evpd = True, page_code = 0x80)
        if serial_only:
            data = self.extended(device, command,True, trim=True, raw=False)
            if hasattr(data, 'page_data') and hasattr(data.page_data, 'product_serial_number'):
                return data.page_data.product_serial_number
            else:
                #raise AttributeError("Device does not return product_serial_number")
                return ""
        else:
            return self.extended(device, command,True, True, raw)

    def evpd83(self,device, raw=False):
        """Calls the EVPD page 83 command 'Device Identification'
        """
        command = self.getCommand(evpd = True, page_code = 0x83, 
                                  allocation_length=252)
        return self.extended(device, command,True, True, raw)

    def extended(self,device, command,
                 auto_correct=True, trim=False, raw=False):
        """Executes the INQUIRY command
        
        Parameters:
        * command      - Inquiry CDB as returned by getCommand. See class
                         for option details
        * auto_correct - [Bool:True] if True and allocation_length is less then
                         required buffer size as specified in the response, 
                         will call the command again with the correct 
                         allocation_length (i.e. buffer size)
        * trim         - see SCSIReadCommand.extended
        * raw          - see SCSIReadCommand.extended
        
        """
        defaultInquirySize = self.getCorrectInquiryDataSize("")

        cmd_cdb    = command
        cmd_struct = self._getCommandStruct()
        cmd_bs     = cmd_struct.parse(command).allocation_length
        cmd_resp   = device.commandIn(cmd_cdb,cmd_bs)

        evpd              = cmd_struct.parse(cmd_cdb).evpd
        correctBufferSize = self.getCorrectInquiryDataSize(cmd_resp,evpd)

        if auto_correct and correctBufferSize > cmd_bs:
            # create new command and update the allocation_length
            new_cmd = cmd_struct.parse(command)
            new_cmd.allocation_length = correctBufferSize
            # rebuild the command and execute again
            cmd_cdb = cmd_struct.build(new_cmd)
            cmd_resp = device.commandIn(cmd_cdb,correctBufferSize)

        if trim:
            cmd_resp = cmd_resp[0:correctBufferSize]


        if raw:
           return cmd_resp
        elif evpd:
            return vpd_pages.parse(cmd_resp)          
        else:
            return Response.parse(cmd_resp)
               
                              

    def getCorrectInquiryDataSize(self,inquiry_buffer, evpd=False):
        """
        Determines the correct buffer size according to basic INQUIRY data 
        or VPD data.
        
        Returns:
        * If the return value > buffer_len then INQUIRY should be re-executed
          with the correct buffer size.
        * If the return value < buffer_len then only pass [0:return value] to 
          the INQUIRY construct for parsing so as to not pass any random data.
        
        
        Parameters:
        * evpd           - [Bool:False] True for response from EVPD inquiry
        * inquiry_buffer - string buffer containing at least the first 5 bytes 
                           of INQUIRY data. If the inquiry_buffer does not 
                           contain valid data, the default buffer size shall be
                           returned rather the raising an error.   
        """

        if None == inquiry_buffer or len(inquiry_buffer) < 5:
            return self.DEFAULT_BUFFER_SIZE

        if evpd:
            # byte 3 holds the additional length param = (n-3)
            # n is zero based to length is +1
            additional_length = ord(inquiry_buffer[3])
            total_length = additional_length + 3 + 1
        else:
            # byte 4 holds the additional length param = (n-4)
            # n is zero based to length is +1
            additional_length = ord(inquiry_buffer[4])
            total_length = additional_length + 4 + 1

        return total_length
