#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#------------------------------------------------------------------------------------------------- 


from construct import *
from data_types import *
from ..common import version_descriptor    

def _testValidData(num,subcon):
    """
        If used test that the following data is valid
        based upon the additional length field (byte4)
        in the Standard INQUIRY data format
        
        It is defined as total length(n) - 4 so we will 
        add the 4 so that the byte number in the table
        can be used rather then doing arithmetic
    """
    
    return If(lambda ctx: ctx.additional_length + 4 >= num  , subcon)

def _safeMetaRepeater(maxcount, size, offset, name, subcon):
    """ Determines how many repeaters of size 'size'
    can fit in remainder up to maxcount.
    """
    return  IfThenElse(name, lambda ctx: maxcount * size <= (ctx.additional_length + 4 - offset),
        MetaRepeater(maxcount, subcon),
        MetaRepeater(lambda ctx: (ctx.additional_length + 4 - offset)/size, subcon),
        )
    
def _safeMetaField(maxcount, size, offset, name):
    """ Determines how many repeaters of size 'size'
    can fit in remainder up to maxcount.
    """
    return  IfThenElse(name, lambda ctx: maxcount * size <= (ctx.additional_length + 4 - offset),
        MetaField(name, maxcount),
        MetaField(name, lambda ctx: (ctx.additional_length + 4 - offset)/size)
        )   

Response = Struct("INQUIRY_RESPONSE",
    EmbeddedBitStruct(
        peripheral_device_type, # peripheral device type is a 5 bit enum defined above
        Enum(Bits("peripheral_qualifier", 3),
            connected=0,
            not_connected=1,
            reserved=2,
            incapable=3,
            _default_=Pass, # 0x4-0x7 are vendor specific, so Pass the literal value as is
        ),
        Padding(7), # reserved
        Flag("removable_media"),
        Alias("rmb", "removable_media"),
    ),
    Enum(UBInt8("version"),
        no_conformance=0x0,
        # scsi=0x1, obsolete
        obsolete_2 = 0x2, # deviation: SPC4 says it's obsolete, I have a reasonably recent device returning it
        spc=0x3,
        spc_2=0x4,
        spc_3=0x5,
        spc_4=0x6,
        # unspecified=0x7, reserved
        # additional values > 0x7 are either obsolete or reserved and won't be copied here
    ),
    EmbeddedBitStruct(
        Bits("response_data_format", 4), # deviation: according to SPC4, this should be constantly 2
        Flag("hierarchial_support"),
        Alias("hisup", "hierarchial_support"),
        Flag("normal_aca_supported"),
        Alias("normaca", "normal_aca_supported"),
        Padding(2), # obsolete
    ),
    UBInt8("additional_length"),
    _testValidData(7,  EmbeddedBitStruct(
        Flag("protect"),
        Padding(2), # reserved
        Flag("third_party_copy"),
        Alias("flag_3pc", "third_party_copy"), # real name is just '3pc', didn't want to start with a digit
        Enum(Bits("target_port_group_support", 2),
            no_support=0,
            implicit=1,
            explicit=2,
            both=3,
        ),
        Flag("access_control_coordinator"),
        Alias("acc", "access_control_coordinator"),
        Flag("storage_array_controller_component_supported"),
        Alias("scc_supported", "storage_array_controller_component_supported"),
        Alias("sccs", "storage_array_controller_component_supported"),
        Flag("spi5_addr16"), # used only for SCSI Parallel Interface 5, reserved otherwise
        Padding(3), # obsolete
        Flag("multi_port"),
        Alias("multip", "multi_port"),
        Flag("vs_1"), # couldn't find an explanation for this in SPC-4
        Flag("enclosure_services"),
        Alias("encserv", "enclosure_services"),
        Padding(1), # obsolete
        Flag("vs_2"), # couldn't find an explanation for this in SPC-4
        Flag("command_queue"),
        Alias("cmdque", "command_queue"),
        Padding(2), # obsolete
        Flag("spi5_sync"), # used only for SCSI Parallel Interface 5, reserved otherwise
        Flag("spi5_wbus16"), # used only for SCSI Parallel Interface 5, reserved otherwise
        Padding(2), # obsolete
    )),
    _testValidData(15,
       String("vendor_identification", length=8, encoding=None, padchar=' ', paddir='right', trimdir='right'),
       ),
    _testValidData(31, 
       String("product_identification", length=16, encoding=None, padchar=' ', paddir='right', trimdir='right'),
       ),
    _testValidData(35, 
       String("product_revision_level", length=4, encoding=None, padchar=' ', paddir='right', trimdir='right'),
       ),
    _testValidData(36,
           # vendor specific field in a byte array starting at byte 36 up to 55 (20 bytes)
           # there is no requirement (at least we don't) to pas back all 20 bytes, so we'll
           # read up to the amount of data available  
           # adding 4 as additional_length = n-4    
        _safeMetaField(20,1,36, "vendor_specific"),
       ),
    _testValidData(56, 
       EmbeddedBitStruct(
            Flag("spi5_ius"), # used only for SCSI Parallel Interface 5, reserved otherwise
            Flag("spi5_qas"), # used only for SCSI Parallel Interface 5, reserved otherwise
            Bits("spi5_clocking", 2), # used only for SCSI Parallel Interface 5, reserved otherwise
            Padding(4), # reserved
            ),
        ),
    _testValidData(57, Padding(1), ),    
    _testValidData(58,
        # this section is made of up to 8 sets of 2 byte version descriptors
        _safeMetaRepeater(8,2,58,"version_descriptor", version_descriptor)
       ),
    _testValidData(96, Padding(22)), # reserved
    _testValidData(96,
        MetaField("vendor_specific_96", lambda ctx: ctx.additional_length + 5 - 96),
        ),
)


