#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#------------------------------------------------------------------------------------------------- 
# Vendor Product Pages
# 
# This will always return a composite struct
# The outer struct contains the following information
#  peripheral_qualifier   [int]
#  peripheral_device_type [string]
#  page_type_id           [int] - page type as int
#  page_type              [string] - page type as string
#  page_data              container for page specific data
#
#  Each page will contain a page_type_id parameter as an internal
#  assertion before processing the contents of the page
#
#  In the case of an unsupported page code, page_data shall contain
#  an array of the raw data, including the page_type field 
#
# NOTE: this only parses the output it does not validate
#

from construct import *
from data_types import  *


page_types = Enum(Byte("page_type"),
                       supported_pages = 0x00,
                       unit_serial_number = 0x80,
                       device_identification = 0x83,
                       _default_ = 'unsupported_page_code',
                  )

#===============================================================================
# General tables
#===============================================================================

#---------------------------------------------------------------------- Code Set
# See SPC-4 4.4.3

code_set = Enum(Nibble("code_set"),
    reserved = 0x0,
    binary   = 0x1,
    ascii    = 0x2,
    utf_8    = 0x3,
    _default_ = 'reserved',
    )

#----------------------------------------------------------- Protocol Identifier
# defined in SPC-4 7.5.1 Table 362
protocol_identifier = Enum(Nibble("protocol_identifier"),
    fibre_channel = 0x0,
    parallel_scsi = 0x1,
    ssa = 0x2,
    ieee_1394 = 0x3,
    scsi_remote_device_memory_access_protocol = 0x4,
    internet_scsi_iscsi = 0x5,
    sas_serial_scsi_protocol = 0x6,
    automation_drive_interface_transport_protocol = 0x7,
    at_attachement_interface = 0x8,
    no_specific_protocol = 0xF,
    _default_ = 'reserved',
    )

#===============================================================================
# Device Information
#===============================================================================
    
# SPC-4 7.7.3 table 459

designator_type = Enum(Nibble("designator_type"),
    vendor_specific = 0x0,
    t10_vendor_id_based = 0x1,
    eui_64_based = 0x2,
    naa = 0x3,
    relative_target_port_identifier = 0x4,
    target_port_group = 0x5,
    logical_unit_group = 0x6,
    md5_logical_unit_identifier = 0x7,
    scsi_name_string = 0x8,
    _default_ = 'reserved',
    )

#--------------------------------------------------------------- Vendor Specific
# SPC-4 table 460

vendor_specific_designator_field = Struct(
    "vendor_specific_designator_field",
    Bytes('vendor_specific_indentifier',
          lambda ctx: ctx['_']['designator_length'])
    )

#----------------------------------------------------------------- T10 Vendor ID
# SPC-4 Table 461

t10_vendor_id_based = Struct(
    "t10_vendor_id_based_field",
    String("t10_vendor_identifier",length=8),
    Bytes("vendor_specific_indentifier_string",
        length=lambda ctx: ctx['_']['designator_length']-8),
    )

#------------------------------------------------------------ IEEE EUI-64 Format
# The EUI-64 format is defined in SPC-4 section 7.7.3.5

#SPC-4 Table 463
eui_64_designator_8 = Struct("eui_64_designator_8",
    Bytes("ieee_company_id",3),
    Bytes("vendor_specific_extensiom_identifier",5),
    )

#SPC-4 Table 464
eui_64_designator_12 = Struct("eui_64_designator_12",
    Embedded(eui_64_designator_8),
    Bytes("directory_id",4),
    )

#SPC-4 Table 465
eui_64_designator_16 = Struct("eui_64_designator_16",
    UBInt64("identifier_extension"),
    Embedded(eui_64_designator_8),
    )

# SPC-4 Table 642 EUI-64 based designator format
# use designator_length to work out string size
# length or 8.12.16 determines different IEEE vendor format

eui_64_designator = Struct(
    "eui_64_designator",
    Peek(Bytes("eui_64_identifier_raw", 
               lambda ctx: ctx["_"]["designator_length"])),
    Switch("designator", lambda ctx: ctx["_"]["designator_length"],
       {
         8 : eui_64_designator_8,
        12 : eui_64_designator_12,
        16 : eui_64_designator_16,
       },
       ),
    )

#-------------------------------------------------------- Designation Descriptor
# SPC-4 Table 457
designation_descriptor = Struct(
    "designation_descriptor",
    EmbeddedBitStruct(
        Peek(Nibble("protocol_identifier_id")),
        protocol_identifier,
        Peek(Nibble("code_set_id")),
        code_set,
        Flag("piv"),
        Padding(1),
        Bits("association",2),
        Peek(Nibble("designator_type_id")),
        designator_type,
        ),
    Padding(1),
    UBInt8("designator_length"),
    Switch("designator", lambda ctx: ctx['designator_type'],
        {  
         'vendor_specific' : vendor_specific_designator_field,
         't10_vendor_id_based' : t10_vendor_id_based,
         'eui_64_based' : eui_64_designator,
        },
        default = MetaField('unsupported_designator_page', lambda ctx: ctx['designator_length']),
#        Range(0,lambda ctx: ctx['designator_length'],Byte('unsupported_designator_page')),
        
        )
    )


# SPC-4 Table 456
# defined in SPC-4 7.7.3
device_identification_vpd_page = Struct(
    "device_identification_vpd_page",
    Const(Byte("page_type"),0x83),
    UBInt16("page_length"),
    GreedyRepeater(designation_descriptor)
    )
#===============================================================================
# ASCII Information
#===============================================================================
# This is not yet supported by the enum, it is returned by pages 01h - 7Fh
# Best is to call the command and create this using raw data
# See SPC-4 7.7.2
#
# TODO: remove common header items once this is part of the common struct

ascii_info = Struct(
        "ascii_info",
        EmbeddedBitStruct(
           Bits("peripheral_qualifier",3),
           peripheral_device_type, 
           ),
        UBInt8("page_type_id"),
        Padding(1),
        UBInt8("page_length"),
        UBInt8("ascii_length"),
        String("ascii_information", lambda ctx: ctx.ascii_length),
#        Bytes("vendor_specific_information", lambda ctx:
#              ctx.page_length - ctx.ascii_length - 1),
        )
        
#===============================================================================
# Unit Serial Number
#===============================================================================
#defined in SPC-4 7.7.12 table 499
unit_serial_number = Struct(
        "unit_serial_number",
        Const(Byte("page_type"),0x80),
        Padding(1),
        UBInt8("page_length"),
        String("product_serial_number",lambda ctx: ctx["page_length"])
        )
    
#===============================================================================
# Supported VPD Pages
#===============================================================================
#defined in SPC-4 7.7.11
supported_vpd_pages = Struct(
        "supported_vpd_pages",
        Const(Byte("page_type"),0),
        Padding(1),
        UBInt8("page_length"),
        Repeater(lambda ctx: ctx["page_length"],Byte("supported_vpd_pages")),
        )

#===============================================================================
# Generic Struct
#===============================================================================
vpd_pages = Struct(
        "vpd_pages",
        EmbeddedBitStruct(
           Bits("peripheral_qualifier",3),
           peripheral_device_type, 
           ),
        Peek(UBInt8("page_type_id")),
        Peek(page_types),
        Switch("page_data", lambda ctx: ctx.page_type,
               {
                'supported_pages' : supported_vpd_pages,
                'device_identification' : device_identification_vpd_page,
                'unit_serial_number' : unit_serial_number
                },
                # if we don't support it, return as a byte array
                default = GreedyRepeater(Byte("raw_page_data") ),
                )
        )


#-------------------------------------------------------------------- Test Cases
if __name__ == "__main__":
    # case 1
    # supported VPD pages stating that it has 3 pages
    # and the page values are 0x00,0x80,0x83
    print vpd_pages.parse("\x00\x00\x00\x03\x00\x80\x83")
    
    # case 2
    # unsupported page type
    print vpd_pages.parse("\x00\x78\x00\x01\x02\x03\x04")
    
#===============================================================================
# Case 3
# VPD INQUIRY: Device Identification page
#  Designation descriptor number 1, descriptor length: 12
#    id_type: EUI-64 based,  code_set: Binary
#    associated with the addressed logical unit
#      EUI-64 based 8 byte identifier
#      IEEE Company_id: 0x1738
#      Vendor Specific Extension Identifier: 0x1a01d8
#      [0x00173800001a01d8]
#  Designation descriptor number 2, descriptor length: 40
#    id_type: T10 vendor identification,  code_set: ASCII
#    associated with the addressed logical unit
#      vendor id: IBM     
#      vendor specific: 2810XIV          MN0001A01D8
#  Designation descriptor number 3, descriptor length: 21
#    id_type: vendor specific [0x0],  code_set: ASCII
#    associated with the addressed logical unit
# 00     76 6f 6c 3d 64 67 2d 63  65 6e 74 6f 73 20 20 20    vol=dg-centos   
# 10     20                                                                  
#  Designation descriptor number 4, descriptor length: 22
#    id_type: vendor specific [0x0],  code_set: ASCII
#    associated with the addressed logical unit
# 00     68 6f 73 74 3d 64 67 2d  63 65 6e 74 6f 73 20 20    host=dg-centos  
# 10     20 20          
#===============================================================================
    print vpd_pages.parse("\x00\x83\x00\x5f\x01\x02\x00\x08\x00\x17\x38\x00" + \
                          "\x00\x1a\x01\xd8\x02\x01\x00\x24\x49\x42\x4d\x20" + \
                          "\x20\x20\x20\x20\x32\x38\x31\x30\x58\x49\x56\x20" + \
                          "\x20\x20\x20\x20\x20\x20\x20\x20\x20\x4d\x4e\x30" + \
                          "\x30\x30\x31\x41\x30\x31\x44\x38\x02\x00\x00\x11" + \
                          "\x76\x6f\x6c\x3d\x64\x67\x2d\x63\x65\x6e\x74\x6f" + \
                          "\x73\x20\x20\x20\x20\x02\x00\x00\x12\x68\x6f\x73" + \
                          "\x74\x3d\x64\x67\x2d\x63\x65\x6e\x74\x6f\x73\x20" + \
                          "\x20\x20\x20")
    
    
    # case 4
    print vpd_pages.parse("\x00\x80\x00\x0b\x4d\x4e\x30\x30\x30\x31\x41\x30" + \
                          "\x31\x37\x37")
