#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#------------------------------------------------------------------------------------------------- 
from construct import * 
from xiv_xsf.scsy.commands.core import SCSIReadCommand
from xiv_xsf.scsy.utils import ArbitrarySizeUnsignedInteger

Capacity10 = Struct("READ_CAPACITY_10_response",
    UNInt32("returned_logical_block_address"),
    UNInt32("logical_block_length_in_bytes"),
#    Value("total_size_bytes", lambda ctx: ctx.returned_logical_block_address * 
#                            ctx.logical_block_length_in_bytes),
    )

Capacity16 = Struct("READ_CAPACITY_16_response",
    UNInt64("returned_logical_block_address"),
    UNInt32("logical_block_length_in_bytes"),
    EmbeddedBitStruct(
        Padding(4), #reserved
        Bits("p_type",3),
        Flag("prot_en"),
        Padding(4), #reserved
        Nibble("logical_blocks_per_physical_block_exponent"),
        Padding(2), #reserved
        Bits("lowest_aligned_logical_block_address",14),
        ),
    Padding(16), #reserved
    )

class ReadCapacity10(SCSIReadCommand):
    #---------------------------------------------------------- Helper Functions
   
    def _getCommandDefaults(self):    
        return dict(control=0,
                    logical_block_address=0, 
                    operation_code=0x25,
                    pmi=0)
    
    def _getCommandStruct(self):
        return Struct("READ_CAPACITY_10_command",
            Const(UBInt8("operation_code"), 0x25),
            EmbeddedBitStruct(
                Padding(7), # reserved
                Padding(1), # obsolete
                ),
            UNInt32("logical_block_address"),
            Padding(2), # reserved
            EmbeddedBitStruct(
                Padding(7), # reserved
                Bit("pmi"), # obsolete
                ),
            
            UBInt8("control"),
            )
            
    #-------------------------------------------------------------- Core Methods
    def extended(self,device, command, trim=True, raw=True):
        """SCSI READ_CAPACITY_10 command

        Returns a struct with the following parameters
          returned_logical_block_address (int32)
          logical_block_length_in_bytes (int32)
        """
        
        cmd_resp =  device.commandIn(command,
                         8,
                         trim
                         )
        if raw:
            return cmd_resp
        else:
            return Capacity10.parse(cmd_resp)
        
        
class ReadCapacity16(SCSIReadCommand):
    #---------------------------------------------------------- Helper Functions
    default_payload = 2**18 # this is nexta centric
    
    def _getCommandDefaults(self):    
        return dict(control=0,
                    logical_block_address=0, 
                    allocation_length = 32,
                    operation_code=0x9E,
                    service_action=0x10,
                    pmi=0)
    
    def _getCommandStruct(self):
        return Struct("READ_CAPACITY_16_command",
            Const(UBInt8("operation_code"), 0x9E),
            EmbeddedBitStruct(
                Padding(3), # reserved
                Const(Bits("service_action",5), 0x10), # obsolete
                ),
            UNInt64("logical_block_address"),
            UNInt32("allocation_length"),
            EmbeddedBitStruct(
                Padding(7), # reserved
                Bit("pmi"), # obsolete
                ),
            
            UBInt8("control"),
            )
            
    #-------------------------------------------------------------- Core Methods
    def extended(self,device, command, trim=True, raw=True):
        """SCSI READ_CAPACITY_16 command

        Returns a struct with the following parameters
          returned_logical_block_address (int64)
          logical_block_length_in_bytes (int32)
          p_type (3 bits)
          prot_en (bit)
          logical_blocks_per_physical_block_exponent (4 bit)
          lowest_aligned_logical_block_address
        """
        
        cmd_resp =  device.commandIn(command,
                        self._getCommandStruct().parse(command).allocation_length,
                        trim
                        )
        if raw:
            return cmd_resp
        else:
            return Capacity16.parse(cmd_resp)