#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#------------------------------------------------------------------------------------------------- 
from construct import *
from array import *
from ..common.asc_codes import additional_sense_code

# response codes
# all values besides the ones defined below are RESERVED
#        ERROR TYPE      FORMAT
#  70h    Current         Fixed
#  71h    Deferred        Fixed
#  72h    Current         Descriptor
#  73h    Deferred        Descriptor
#  7Fh        Vendor Specific

# information for referrences
# SPC-4 - http://www.t10.org/ftp/t10/drafts/spc4/spc4r17.pdf
# OSD-2 - http://www.t10.org/ftp/t10/drafts/osd2/osd2r04c.pdf
# SBC-3 - http://www.t10.org/ftp/t10/drafts/sbc3/sbc3r17.pdf
# SSC-3 - http://www.t10.org/ftp/t10/drafts/ssc3/ssc3r04a.pdf

response_code = Enum(Bits("response_code", 7),
                      current_fixed = 0x70,
                      deferred_fixed = 0x71,
                      current_descriptor = 0x72,
                      deferred_descriptor = 0x73,
                      vendor_specific = 0x7f,
                      _default_ = 'reserved'
                      )

sense_key = Enum(Bits("sense_key", 4),
                     sense_no_sense = 0x00,
                     sense_recovered_error = 0x01,
                     sense_not_ready = 0x02,
                     sense_medium_error = 0x03,
                     sense_hardware_error = 0x04,
                     sense_illegal_request = 0x05,
                     sense_unit_attention = 0x06,
                     sense_data_protect = 0x07,
                     sense_blank_check = 0x08,
                     sense_vendor_specific = 0x09, 
                     sense_copy_aborted = 0x0a,
                     sense_aborted_command = 0x0b,
                     sense_obsolete = 0x0c,
                     sense_volume_overflow = 0x0d,
                     sense_miscompare = 0x0e,
                     sense_reserved =  0x0f,
                     )



             

vendor_sense_data = Struct("vendor_sense_data",
                           UBInt8("descriptor_type"),
                           UBInt8("additional_length"),
                           # TODO: confirm this is right, additional_lengthis defined as n-1
                           # where the first byte of vendor_specific is byte 2-n
                           MetaField("vendor_specific", lambda ctx: ctx['additional_length'] )
                           )
    # See SPC-4 4.5.2.2
descriptor_information = Struct("descriptor_information",
                                #Const(Byte("type"),0x00),
                                Const(Byte("additional_length"),0x0A),
                                EmbeddedBitStruct(
                                                  Const(Flag("valid"),True),
                                                  Padding(15),
                                                  ),
                                Bytes("information",8)            
                                )
    # See SPC-4 4.5.2.3
descriptor_command_specific = Struct("descriptor_command_specific",
                                #Const(Byte("type"),0x01),
                                Const(Byte("additional_length"),0x0A),
                                Padding(2),
                                Bytes("information",8)            
                                )

    # See SPC-4 4.5.2.5
descriptor_field_replacebale = Struct("descriptor_field_replacebale",
                                #Const(Byte("type"),0x03),
                                Const(Byte("additional_length"),0x02),
                                Padding(1),
                                UBInt8("field_replaceable_unit_code")            
                                )

    # See SSC-3 4.2.12(.2) Table 3 
descriptor_type_stream_commands = Struct("descriptor_type_stream_commands", 
                                #Const(Byte("type"),0x04),
                                Const(Byte("additional_length"),0x02),
                                Padding(1),
                                EmbeddedBitStruct(
                                    Flag("filemark"),
                                    Flag("eom"),
                                    Flag("ili"),
                                    Padding(5),
                                    )
                                )

  # see SBC-3 clause 4.14.3 Table 5
descriptor_type_block_commands = Struct("descriptor_type_block_commands", 
                                #Const(Byte("type"),0x05),
                                Const(Byte("additional_length"),0x02),
                                Padding(1),
                                EmbeddedBitStruct(
                                    Padding(2),
                                    Flag("ili"),
                                    Padding(5),
                                    )
                                )
 
    # See SPC-4 4.5.2.6
descriptor_progress_indicator = Struct("descriptor_progress_indicator",
                                #Const(Byte("type"),0x0A),
                                Const(Byte("additional_length"),0x06),
                                sense_key,
                                additional_sense_code, #2 bytes in my implementation
                                Padding(1),
                                UBInt16("progress_indication"),            
                                )

 # Sense Key Specific Data Descriptors
 
# see SPC-4 clause 4.5.2.4.2 Actual Retry Count
def parsed_sksd_field_pointer(ctx):
    if True == ctx['command_data']:
        type = "CDB"
    else:
        type = "Data-Out buffer"
         
    if True == ctx['bpv']:
        bits = "Bit %d" % ctx['bit_pointer']
    else:
        bits = ""
        
    return "Error in " + type +  " Byte %d " % ctx['field_pointer'] +  bits 
    
sksd_field_pointer =  Struct("sksd_field_pointer",
                               EmbeddedBitStruct(
                                   Flag("sense_key_specific_valid"),
                                   Flag("command_data"),
                                   Padding(2),
                                   Flag("bpv"),
                                   Bits("bit_pointer",3),  
                               ),
                               UBInt16("field_pointer"),
                               Value("parsed_sksd_field_pointer", 
                                     lambda ctx: parsed_sksd_field_pointer(ctx)),
                           )

# see SPC-4 clause 4.5.2.4.3 Actual Retry Count
sksd_actual_retry_count = Struct("sksd_actual_retry_count",
                          EmbeddedBitStruct(
                              Flag("sense_key_specific_valid"),
                              Padding(7),
                          ),
                          UBInt16("actual_retry_count"),
                      )

# see SPC-4 clause 4.5.2.4.4 Progress Indication
sksd_progress_indication = Struct("sksd_progress_indication",
                          EmbeddedBitStruct(
                              Flag("sense_key_specific_valid"),
                              Padding(7),
                          ),
                          UBInt16("progress_indication"),
                          Value("parsed_progress_indication", lambda ctx: "%d %%" % (ctx['progress_indication'] * 100 / 0x10000))
                      ) 

# see SPC-4 clause 4.5.2.4.5 Segment pointer
def parsed_sksd_segment_pointer(ctx):
    if True == ctx['sd']:
        offset_base = "segment descriptor"
    else:
        offset_base = "parameter list"
        
    if True == ctx['bpv']:
        bits = "Bit %d" % ctx['bit_pointer']
    else:
        bits = ""
       
    return "Relative to start of " + offset_base + \
           " Byte %d " % ctx['field_pointer'] + bits
     
sksd_segment_pointer =  Struct("sksd_segment_pointer",
                          EmbeddedBitStruct(
                              Flag("sense_key_specific_valid"),
                              Padding(1),
                              Flag("sd"),
                              Padding(1),
                              Flag("bpv"),
                              Bits("bit_pointer",3),
                          ),
                          UBInt16("field_pointer"),
                          Value("parsed_sksd_segment_pointer", 
                                lambda ctx: parsed_sksd_segment_pointer(ctx)
                                ),
                      ) 
 
# see SPC-4 clause 4.5.2.4.6 Unit attention condition
sksd_unit_attention_condition = Struct("sksd_unit_attention_condition",
                          EmbeddedBitStruct(
                              Flag("sense_key_specific_valid"),
                              Padding(6),
                              Flag("overflow"),
                          ),
                          Padding(2)
                      )

sksd_sense_data_generic = EmbeddedBitStruct(
                        Flag("sense_key_specific_valid"),
                        Bits("raw_data",23),
                        )

  # see SPC-4 clause 4.5.2.4.1 table 31
  # based upon the sense_key and the descriptor type
  # 0x02 sense key specific data descriptor, map the
  # sun descriptor type
def sense_key_specific_type(ctx):
    if 'sense_illegal_request' == ctx['sense_key']: return 'sksd_field_pointer'
    if 'sense_hardware_error'  == ctx['sense_key']: return 'sksd_actual_retry_count'
    if 'sense_medium_error'    == ctx['sense_key']: return 'sksd_actual_retry_count'
    if 'sense_no_sense'        == ctx['sense_key']: return 'sksd_progress_indication'
    if 'sense_not_ready'       == ctx['sense_key']: return 'sksd_progress_indication'
    if 'sense_copy_aborted'    == ctx['sense_key']: return 'sksd_segment_pointer'
    if 'sense_unit_attention'  == ctx['sense_key']: return 'sksd_unit_attention_condition'
    return 'sksd_sense_data_generic'

# split off from descriptor_sense_key_specific below because this 
  # part is specifically used in fixed format
sub_descriptor_sense_key_specific = Struct("sub_descriptor_sense_key_specific",
    Value('sense_key_specific_type', 
          lambda ctx: sense_key_specific_type(ctx['_'])),
    Switch("sense_key_specific", lambda ctx: ctx['sense_key_specific_type'],
        {
        'sksd_field_pointer' : sksd_field_pointer,
        'sksd_actual_retry_count'               : sksd_actual_retry_count, 
        'sksd_progress_indication'              : sksd_progress_indication,
        'sksd_segment_pointer'                  : sksd_segment_pointer,
        'sksd_unit_attention_condition'         : sksd_unit_attention_condition,                                    
        }, #default = sksd_sense_data_generic,
        ),
  )

# See SPC-4 clause  4.5.2.4.1
descriptor_sense_key_specific = Struct("descriptor_sense_key_specific",
    #Const(Byte("type"),0x02),
    Const(Byte("additional_length"),0x06),
    Padding(2),
    Embedded(sub_descriptor_sense_key_specific),
    Padding(1),            
    )

  
 
descriptor_type = Enum(Byte("descriptor_type"),
                      descriptor_type_information        = 0x00 ,
                      descriptor_type_command_specific   = 0x01 ,
                      descriptor_type_sense_key_specific = 0x02 ,
                      descriptor_type_field_replacebale  = 0x03 ,
                      descriptor_type_stream_commands    = 0x04 ,
                      descriptor_type_block_commands     = 0x05 ,
                      descriptor_type_progress_indicator = 0x0A ,
                      _default_ = Pass,
                )

descriptor_unknown_or_vendor_specific = Struct("descriptor_unknown_or_vendor_specific",
                 # swap the values and give it a nice (consistent) name
               Value("descriptor_type_value", lambda ctx: ctx['descriptor_type']),
               Value("descriptor_type", lambda ctx: "descriptor_unknown_or_vendor_specific" ),
#               OptionalGreedyRepeater(Byte("data")),
               UBInt8("additional_length"),
               MetaField("vendor_specific", lambda ctx: ctx['additional_length'] )
               )

sense_data_descriptor = Struct("sense_data_descriptor",
    descriptor_type,
    Embed(Switch("descriptor_data", lambda ctx: ctx['descriptor_type'],
      {
      'descriptor_type_information'        : descriptor_information,
      'descriptor_type_command_specific'   : descriptor_command_specific,
      'descriptor_type_sense_key_specific' : descriptor_sense_key_specific,
      'descriptor_type_field_replacebale'  : descriptor_field_replacebale,
      'descriptor_type_progress_indicator' : descriptor_progress_indicator,
      'descriptor_type_stream_commands'    : descriptor_type_stream_commands,
      'descriptor_type_block_commands'     : descriptor_type_block_commands,
       }, default = descriptor_unknown_or_vendor_specific
    )),
    )

    # see SPC-4 4.5.3                           
fixed_sense_data = Struct("fixed_sense_data",
      Padding(1), #obsolete
      EmbeddedBitStruct(
          Flag("filemark"),
          Flag("eom"),
          Flag("ili"),
          Padding(1),
          sense_key,
          ),
        Bytes("information",4),
        Byte("additional_sense_length"),
        Bytes("command_specific_information",4),
        additional_sense_code,
        UBInt8("field_replaceable_unit_code"), # see 4.5.2.5
        sub_descriptor_sense_key_specific,
        MetaField("vendor_specific", lambda ctx: 
                                      (ctx.additional_sense_length + 7) - 17),
        )
               
  
descriptor_sense_data = Struct("descriptor_sense_data",
                    EmbeddedBitStruct(
                            Padding(4),
                            sense_key,
                            ),
                    additional_sense_code,
                    Padding(3),
                    Byte("additional_sense_length"),
                    GreedyRepeater(sense_data_descriptor),
#                    sense_data_descriptor
                    )
    
sense_data = Struct("sense_data",
                   EmbeddedBitStruct(
                        Flag("valid"),
                        response_code,
                        ),
                        
                        Switch("data", lambda ctx: ctx['response_code'], 
                           {
                            'current_fixed' : fixed_sense_data,
                            'deferred_fixed' : fixed_sense_data,
                            'current_descriptor': descriptor_sense_data,
                            'deferred_descriptor': descriptor_sense_data,
                            'vendor_specific': vendor_sense_data,
                            },
                            default = vendor_sense_data
                            )
                    )

