#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#------------------------------------------------------------------------------------------------- 
from functools import partial, update_wrapper, wraps, WRAPPER_ASSIGNMENTS

import raw
import commands

from xiv_xsf.scsy.raw import SenseReceived, DeviceNotOpen

# TODO: why does the full import work but not the short form?
#from commands import core
from  xiv_xsf.scsy.commands import core
      
        
class Device(object):
    """Logical SCSI device that implements all SCSI commands
    
    Provides a high level abstraction to the scsy raw device and implements 
    all basic and extended SCSI commands defined in scsy.commands.
    
    Commands:
    The class provides the basic implementation of the scsy command by
    Device.command() e.g. Device.inquiry().
    If an extended version is defined, it may be called using the extended 
    keyword e.g. Device.inquiry.extended(options). The options are command
    specific and are specified in scsy.comands.{command}.options
    
    open({device_path}) must be called before SCSI commands may be called. 
    close() may be used to close the current device and the object may be
    reused by calling open() again.
    
    Exceptions:
    DeviceNotOpen - if any command is called before the device is opened.
    
    All other exceptions from scsy.raw and scsy.command are returned
    unprocessed. 
      
    """
    
    def __init__(self):
         
         self._cmdObjs    = dict()
         self._raw_device = raw.RawDevice()
         
         # scan the commands package for commands to add to device object
         # all commands must have a 'basic' method
         # if a command has an 'extended' method this shall also be included
         for command_name in dir(commands):   
             # set the basic and extended functions
             try:
                 if issubclass(eval("commands.%s" % command_name),
                               core.SCSICommand):
                     cmdObj = eval("commands.%s()" % command_name)
                     self._cmdObjs[command_name] = cmdObj
                     
                     # use partial to always pass in the device object as the
                     # first parameter
                     cmdname = command_name.lower()
                     setattr(self, cmdname, 
                             partial(cmdObj.basic, self._raw_device)
                             )
                     self_basic_cmd = getattr(self,cmdname)
                     update_wrapper(self_basic_cmd, cmdObj.basic)
                     
                     #set the extended command
                     setattr(self_basic_cmd, 'extended', 
                             partial(cmdObj.extended, self._raw_device))
                     update_wrapper(self_basic_cmd.extended, cmdObj.extended)
                     
                     #set the getCommand function
                     setattr(self_basic_cmd, 'getCommand', cmdObj.getCommand)
#                     update_wrapper(self_basic_cmd.getCommand,
#                                    cmdObj.getCommand,
#                                    WRAPPER_ASSIGNMENTS('__name__','__doc__'))
                            
                     # helper functions
                     # if the command as _helperfunctions then add each command
                     # under {command}.{function}
                     if hasattr(cmdObj,"_helperfunctions"):
                         for funcname in getattr(cmdObj,"_helperfunctions"):
                             #print "adding ", funcname
                             fn = getattr(cmdObj,funcname)
                             fnspec = partial(fn,self._raw_device)
                             update_wrapper(fnspec,fn)              
                             setattr(self_basic_cmd, funcname,fnspec) 
                                     
             except TypeError:
                # we expect non classes to raise a type error
                pass

    def open(self,device_path):
        """open the device and enable calling of SCSI commands"""
        # TODO: what happens if device is already open
        # TODO: test that the file is valid, probably best handled in device class
        self._raw_device.openHandle(device_path)
    
    def close(self):
        """close the device."""
        self._raw_device.closeHandle()
        
          
