#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#------------------------------------------------------------------------------------------------- 
from xiv_xsf.scsy.raw.raw_device import *
from array import *
import struct
from construct import *
from platform import processor
from struct import pack
import win32file


# Win32 SCSI_PASS_THROUGH structure
# taken from http://msdn.microsoft.com/en-us/library/ms810309.aspx
#
#
# typedef struct _SCSI_PASS_THROUGH {
#  USHORT  Length; - Contains the value of sizeof(SCSI_PASS_THROUGH) 44
#  UCHAR  ScsiStatus; - Reports the SCSI status that was returned by the 
#                       HBA or the target device. 
#  UCHAR  PathId; - Indicates the SCSI port or bus for the request.
#  UCHAR  TargetId; - Indicates the target controller or device on the bus. 
#  UCHAR  Lun; - Indicates the logical unit number of the device. 
#  UCHAR  CdbLength; - Indicates the size in bytes of the SCSI command 
#                      descriptor block.
#  UCHAR  SenseInfoLength; - Indicates the size in bytes of the 
#                            request-sense buffer.
#  UCHAR  DataIn; - Indicates whether the SCSI command will read or write data.
#                   This field must have one of three values:
#                   SCSI_IOCTL_DATA_IN     Read data from the device.
#                   SCSI_IOCTL_DATA_OUT     Write data to the device.
#                   SCSI_IOCTL_DATA_UNSPECIFIED     No data is transferred.
#  ULONG  DataTransferLength; - Indicates the size in bytes of the data 
#                   buffer. Many devices transfer chunks of data of 
#                   pre-defined length. The value in DataTransferLength 
#                   must be an integral multiple of this pre-defined, 
#                   minimum length that is specified by the device. 
#                   If an underrun occurs, the miniport driver must update
#                   this member to the number of bytes actually transferred.
#  ULONG  TimeOutValue; - Indicates the interval in seconds that the request
#                   can execute before the OS-specific port driver might 
#                   consider it timed out.
#  ULONG_PTR DataBuffer; - Pointer to the data buffer. 
#  ULONG  SenseInfoOffset; - Offset from the beginning of this structure to the
#                   request-sense buffer.
#  UCHAR  Cdb[16]; - Specifies the SCSI command descriptor block to be sent to the
#                   target device.
#}SCSI_PASS_THROUGH, *PSCSI_PASS_THROUGH;

windows_scsi_pass_through = Struct("SCSI_PASS_THROUGH_DIRECT",
    Const(ULInt16("length"),44),
    Byte("scsi_status"),
    Byte("path_id"),

    Byte("target_id"),
    Byte("lun"),
    Byte("cdb_length"),
    Byte("sense_info_length"),

    Byte("data_in"),
    Padding(3), # for alignment
    ULInt32("data_transfer_length"),

    ULInt32("time_out_value"),
    ULInt32("data_buffer"), #TODO: test x64
    Const(ULInt32("sense_info_offset"),48),

    Bytes("cdb",16),

	# On Win2k8 it only returns a SENSE BUFFER if there is
	# sense data returned. Conditionally set buffer
	# only if sense_info_length != 0
    If(lambda ctx: ctx.sense_info_length > 0,Padding(4)),
    If(lambda ctx: ctx.sense_info_length > 0,
       Bytes("sense_info_buffer",
             lambda ctx: ctx.sense_info_length),
       ),

    # helper value used by trim logic
    Value("total_bytes_xfer",
           lambda ctx: ctx.data_transfer_length)
    )

class Device(RawSCSIDevice):

    SCSI_PASS_THROUGH_DIRECT    = 0x4D014
    SCSI_IOCTL_DATA_IN          = 0
    SCSI_IOCTL_DATA_OUT         = 1
    SCSI_IOCTL_DATA_UNSPECIFIED = 2

    def openHandle(self, device_path):

        # TODO: get right flags for open
        # TODO: catch file open exceptions
        self.handle = win32file.CreateFile(device_path,
                    win32file.GENERIC_READ | win32file.GENERIC_WRITE,
                    win32file.FILE_SHARE_READ | win32file.FILE_SHARE_WRITE,
                    None,
                    win32file.OPEN_EXISTING,
                    0,0)

    def closeHandle(self):
        if None != self.handle:
            win32file.CloseHandle(self.handle)
            self.handle = None

    def sendRawCommand(self, direction_read,cdb,
                       data_ptr, data_len):

        # notes
        # we do not create a sense info buffer for windows
        # rather it returned with the result
        #
        # Todos
        # TODO: set the correct path, target and lun params
        # TODO: test struct alignment on 64 bit machines

        if None == self.handle:
            raise DeviceNotOpen

        if True == direction_read:
            flags = self.SCSI_IOCTL_DATA_OUT
        else:
            flags = self.SCSI_IOCTL_DATA_IN

        cdb_len = len(cdb)
        cdb = cdb + ((16-cdb_len) * "\00")

        data= windows_scsi_pass_through.build(Container(
                   length=44,
                   scsi_status=0, #SET ON RETURN
                   path_id = 0,
                   target_id = 0, # TODO: pass this as a param
                   lun = 0,
                   cdb_length = cdb_len,
                   sense_info_length = 252,
                   data_in = flags,
                   data_transfer_length = data_len,
                   time_out_value = SCSI_TIMEOUT,
                   data_buffer = data_ptr,
                   sense_info_offset = 48,
                   cdb = cdb,
                   sense_info_buffer = "\00" * 252,
                   total_bytes_xfer=0
                   ) )


        # TODO: add try catch
        # throws an IOError
        # max data length is 1024 bytes - longer will call a segmentation violation
        resp = win32file.DeviceIoControl(self.handle,
                                 self.SCSI_PASS_THROUGH_DIRECT,
                                 data,
                                 data_ptr,
                                 None)

        response = windows_scsi_pass_through.parse(resp)

        if 2 == response.scsi_status:
            raise SenseReceived(response.sense_info_buffer)

        return response


if __name__ == "__main__":
    from xiv_xsf.scsy.run_tests import *
    basic_test()
