#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#------------------------------------------------------------------------------------------------- 
from xiv_xsf.scsy.commands.sense import sense_data
from array import *
import struct 

"""
RawSCSIInterface

This the the common SCSI interface that is called regardless of the architecture.
It will attempt to determine the current architecture and instantiate the correct
api. If the architecture is not supported or recognized it will raise the
UnsupportedNativeArchitecture exception.

"""

SCSI_TIMEOUT = 15000

class RawSCSIException(Exception):
    pass

class UnsupportedNativeArchitecture(RawSCSIException):
    pass

class InvalidHandle(RawSCSIException):
    pass

class DeviceNotOpen(RawSCSIException):
    pass

class NotSCSIDevice(RawSCSIException):
    def __init__(self, device):
        self.device = device
    def __str__(self):
        return "The device '%s' is not a SCSI device" % self.device
    
class SenseReceived(RawSCSIException):
    def __init__(self, sense_buffer):
        self.sense_buffer = sense_buffer
    def __str__(self):   
        return "Sense Data Received:\n"+str(sense_data.parse(self.sense_buffer))
    def getSense(self):
        return sense_data.parse(self.sense_buffer)
    

class NonzeroStatus(RawSCSIException):
    def __init__(self, status):
        self.status = status
    def __str__(self):
        return "received nonzero status %d" % self.status
    
    
class RawSCSIDevice(object):
    DEFAULT_DATA_BUFFER_LENGTH = 2**18 # 256k
    
    def __init__(self):
        self.handle = None
    
    def openHandle(self):
        raise NotImplementedError("should be subclassed")
        
    def closeHandle(self):
        raise NotImplementedError("should be subclassed")
        
    def sendRawCommand(self, direction_read, cdb, data_ptr, data_len):
        raise NotImplementedError("should be subclassed")
        
    def commandIn(self,cdb, data_buffer_length=None, trim_buffer = False): 
        """
        Performs a SCSI IN(READ) command creating a buffer to hold the response data
        
        Parameters:
        * data_buffer_length: size of data_buffer to create, otherwise use default_size 
        * trim_buffer       : determines whether to return the whole buffer or only the buffer data returned by the SCSI command
            False [default] - return the whole buffer
            True            - only return valid data
        """ 
        
        if data_buffer_length is None:
            data_buffer_length = self.DEFAULT_DATA_BUFFER_LENGTH
 
        data_buffer = array('B',[0] * data_buffer_length)
        data_ptr, data_len = data_buffer.buffer_info()
    
        resp = self.sendRawCommand(True, cdb, data_ptr, data_len)
  
        # convert to a string buffer
        data_buffer = data_buffer.tostring()  
    
        if trim_buffer:
            bytes_xfer = resp.total_bytes_xfer
            return data_buffer[0:bytes_xfer]
        else:
            return data_buffer
        
            
    def commandOut(self, cdb, data_buffer, data_buffer_length = None):
        if data_buffer_length is None:
            data_buffer_length = len(data_buffer)

        data = array('B',data_buffer[0:data_buffer_length])
        data_ptr, data_len = data.buffer_info()
 
        resp = self.sendRawCommand(False, cdb, data_ptr, data_len)
        
        # TODO: parse the response to see that it is what we expected
        return True
    
    def __del__(self):
        # TODO: close semaphore
        self.closeHandle()
        
    def __repr__(self):
        # TODO: get proper values
        return '<%s to %s %s on %s>' % (self.__class__.__name__, "self.vendor_identification",
                                        "self.product_identification", "self.device")
        
        
