#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#------------------------------------------------------------------------------------------------- 

import subprocess
import os

class NoSuchFileError(IOError):
    """A simple, file doesn't exist error"""
    def __init__(self, filename):
        self.filename = filename

    def __str__(self):
        return "File '%s' doesn't exist" % self.filename

class ExternalCommandError(Exception):
    """Command has failed"""
    def __init__(self, cmdline, exitcode):
        self.cmdline = cmdline
        self.exitcode = exitcode

    def __str__(self):
        return "Command line '%s' has ended with exit code %s" % \
            (self.cmdline, str(self.exitcode) )

class SystemCommand(object):
    """A simple run-system-command class, wraps subprocess.Popen()
    a run() call populates the out, err, exitcode attributes"""

    out = None
    err = None
    exitcode = None

    def __init__(self, cmdline):
        """Constructor: gets the command line as an array
        i.e. [ "/bin/ls", "-l" ]
        """
        self.cmdline = cmdline
        self.sanity_checks()

    def sanity_checks(self):
        # Chk for file existance before, cuz subprocess doesn't throw an
        # indicative exception. (OSError)
        if not os.path.exists(self.cmdline[0]):
            raise NoSuchFileError(self.cmdline[0])

    def run(self, throw_on_bad_exit_code = True):
        """Runs the process and waits for the it to finish (blocking), 
        fetch stdout, stderr, exitcode.
        """
        proc = subprocess.Popen(self.cmdline, stdout=subprocess.PIPE, \
            stdin=subprocess.PIPE)
        out, err = proc.communicate()
        exitcode = proc.poll()

        if (throw_on_bad_exit_code and exitcode != 0):
            raise ExternalCommandError(" ".join(self.cmdline), exitcode)

        self.out = out
        self.err = err
        self.exitcode = exitcode

        return exitcode
